@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_DISABLER_H
 #define MUON_FUNCTIONS_DISABLER_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_disabler[];
+extern const struct func_impl impl_tbl_disabler[];
 #endif
