@@ -7,6 +7,8 @@
 #define MUON_FUNCTIONS_FILE_H
 #include "lang/func_lookup.h"
 
+bool file_is_dynamic_lib(struct workspace *wk, obj file);
+bool file_is_static_lib(struct workspace *wk, obj file);
 bool file_is_linkable(struct workspace *wk, obj file);
 
 extern const struct func_impl impl_tbl_file[];
