@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_MODULES_KEYVAL_H
 #define MUON_FUNCTIONS_MODULES_KEYVAL_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_module_keyval[];
+extern const struct func_impl impl_tbl_module_keyval[];
 #endif
