@@ -5,7 +5,7 @@
 
 #ifndef FUNCTIONS_SOURCE_SET_H
 #define FUNCTIONS_SOURCE_SET_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_source_set[];
+extern const struct func_impl impl_tbl_source_set[];
 #endif
