@@ -14,7 +14,8 @@ enum format_cb_result {
 	format_cb_skip,
 };
 
-bool version_compare(struct workspace *wk, uint32_t err_node, const struct str *ver, obj cmp_arr, bool *res);
+bool version_compare(const struct str *ver1, const struct str *_ver2);
+bool version_compare_list(struct workspace *wk, const struct str *ver, obj cmp_arr);
 
 typedef enum format_cb_result(
 	(*string_format_cb)(struct workspace *wk, uint32_t node, void *ctx, const struct str *key, uint32_t *elem));
@@ -22,4 +23,5 @@ typedef enum format_cb_result(
 bool string_format(struct workspace *wk, uint32_t err_node, obj str, obj *res, void *ctx, string_format_cb cb);
 
 extern const struct func_impl impl_tbl_string[];
+extern const struct func_impl impl_tbl_string_internal[];
 #endif
