@@ -18,11 +18,12 @@ enum vm_compile_mode {
 	vm_compile_mode_language_extended = 1 << 3,
 	vm_compile_mode_expr = 1 << 4,
 	vm_compile_mode_return_after_project = 1 << 5,
+	vm_compile_mode_relaxed_parse = 1 << 6,
 };
 
 void vm_compile_state_reset(struct workspace *wk);
 void vm_compile_initial_code_segment(struct workspace *wk);
 struct node;
 bool vm_compile_ast(struct workspace *wk, struct node *n, enum vm_compile_mode mode, uint32_t *entry);
-bool vm_compile(struct workspace *wk, struct source *src, enum vm_compile_mode mode, uint32_t *entry);
+bool vm_compile(struct workspace *wk, const struct source *src, enum vm_compile_mode mode, uint32_t *entry);
 #endif
