@@ -105,34 +105,36 @@ enum feature_opt_state {
 	feature_opt_disabled,
 };
 
-#define FOREACH_BUILTIN_MODULE(_)                     \
-	_(fs, "public", true)                         \
-	_(keyval, "public", true)                     \
-	_(pkgconfig, "public", true)                  \
-	_(python, "public", true)                     \
-	_(python3, "public", true)                    \
-	_(sourceset, "public", true)                  \
-	_(toolchain, "private", true)                 \
-	_(subprojects, "private", true)                 \
-	_(getopt, "private", true)                 \
-	_(cmake, "public", false)                     \
-	_(dlang, "public", false)                     \
-	_(gnome, "public", false)                     \
-	_(hotdoc, "public", false)                    \
-	_(i18n, "public", false)                      \
-	_(java, "public", false)                      \
-	_(modtest, "public", false)                   \
-	_(qt, "public", false)                        \
-	_(qt4, "public", false)                       \
-	_(qt5, "public", false)                       \
-	_(qt6, "public", false)                       \
-	_(unstable_cuda, "public", false)             \
-	_(unstable_external_project, "public", false) \
-	_(unstable_icestorm, "public", false)         \
-	_(unstable_rust, "public", false)             \
-	_(unstable_simd, "public", false)             \
-	_(unstable_wayland, "public", false)          \
-	_(windows, "public", false)
+#define FOREACH_BUILTIN_MODULE(_)                                               \
+	_(fs, "public/fs", true)                                                \
+	_(keyval, "public/keyval", true)                                        \
+	_(pkgconfig, "public/pkgconfig", true)                                  \
+	_(python, "public/python", true)                                        \
+	_(python3, "public/python3", true)                                      \
+	_(sourceset, "public/sourceset", true)                                  \
+	_(toolchain, "private/toolchain", true)                                 \
+	_(subprojects, "private/subprojects", true)                             \
+	_(getopt, "private/getopt", true)                                       \
+	_(curl, "private/curl", true)                                           \
+	_(json, "private/json", true)                                           \
+	_(cmake, "public/cmake", false)                                         \
+	_(dlang, "public/dlang", false)                                         \
+	_(gnome, "public/gnome", false)                                         \
+	_(hotdoc, "public/hotdoc", false)                                       \
+	_(i18n, "public/i18n", false)                                           \
+	_(java, "public/java", false)                                           \
+	_(modtest, "public/modtest", false)                                     \
+	_(qt, "public/qt", false)                                               \
+	_(qt4, "public/qt4", false)                                             \
+	_(qt5, "public/qt5", false)                                             \
+	_(qt6, "public/qt6", false)                                             \
+	_(unstable_cuda, "public/unstable-cuda", false)                         \
+	_(unstable_external_project, "public/unstable-external_project", false) \
+	_(unstable_icestorm, "public/unstable-icestorm", false)                 \
+	_(unstable_rust, "public/unstable-rust", false)                         \
+	_(unstable_simd, "public/unstable-simd", false)                         \
+	_(unstable_wayland, "public/unstable-wayland", false)                   \
+	_(windows, "public/windows", false)
 
 #define MODULE_ENUM(mod, path, implemented) module_##mod,
 enum module {
@@ -209,31 +211,51 @@ enum build_tgt_flags {
 	build_tgt_flag_pie = 1 << 6,
 };
 
+enum build_dep_flag {
+	build_dep_flag_recursive = 1 << 0,
+	build_dep_flag_both_libs_static = 1 << 1,
+	build_dep_flag_both_libs_shared = 1 << 2,
+	build_dep_flag_include_system = 1 << 3,
+	build_dep_flag_include_non_system = 1 << 4,
+	build_dep_flag_as_link_whole = 1 << 5,
+	build_dep_flag_partial = 1 << 6,
+	build_dep_flag_part_compile_args = 1 << 7,
+	build_dep_flag_part_includes = 1 << 8,
+	build_dep_flag_part_link_args = 1 << 9,
+	build_dep_flag_part_links = 1 << 10,
+	build_dep_flag_part_sources = 1 << 11,
+};
+
 struct build_dep {
 	enum compiler_language link_language;
+	obj frameworks; // not in raw
 
+	obj compile_args;
+	obj include_directories;
+	obj link_args;
 	obj link_whole;
-
 	obj link_with;
 	obj link_with_not_found;
-	obj frameworks;
-
-	obj link_args;
-	obj compile_args;
-
-	obj include_directories;
-
-	obj sources;
 	obj objects;
-
 	obj order_deps;
 	obj rpath;
+	obj sources;
 
-	struct {
-		obj deps;
-		obj order_deps;
-		obj link_with;
+	struct build_dep_raw {
+		enum build_dep_flag flags;
+
+		obj compile_args;
+		obj include_directories;
+		obj link_args;
 		obj link_whole;
+		obj link_with;
+		obj link_with_not_found;
+		obj objects;
+		obj order_deps;
+		obj rpath;
+		obj sources;
+
+		obj deps;
 	} raw;
 };
 
@@ -268,7 +290,14 @@ struct obj_build_target {
 	enum machine_kind machine;
 };
 
+enum default_both_libraries {
+	default_both_libraries_auto,
+	default_both_libraries_static,
+	default_both_libraries_shared,
+};
+
 struct obj_both_libs {
+	enum default_both_libraries default_both_libraries;
 	obj static_lib; // obj_build_target
 	obj dynamic_lib; // obj_build_target
 };
@@ -304,10 +333,19 @@ enum dependency_type {
 	dependency_type_pkgconf,
 	dependency_type_threads,
 	dependency_type_external_library,
-	dependency_type_appleframeworks,
+	dependency_type_system,
 	dependency_type_not_found,
 };
 
+enum dependency_public_type {
+	dependency_public_type_unset,
+	dependency_public_type_internal,
+	dependency_public_type_pkgconfig,
+	dependency_public_type_system,
+	dependency_public_type_library,
+	dependency_public_type_not_found,
+};
+
 enum dep_flags {
 	dep_flag_found = 1 << 0,
 };
@@ -327,6 +365,7 @@ struct obj_dependency {
 
 	enum dep_flags flags;
 	enum dependency_type type;
+	enum dependency_public_type public_type;
 	enum include_type include_type;
 	enum machine_kind machine;
 };
@@ -396,6 +435,7 @@ struct obj_compiler {
 	uint32_t type[toolchain_component_count];
 	obj ver;
 	obj libdirs;
+	obj fwdirs;
 	enum compiler_language lang;
 	enum machine_kind machine;
 };
@@ -425,6 +465,7 @@ struct obj_environment {
 struct obj_include_directory {
 	obj path;
 	bool is_system;
+	bool is_idirafter;
 };
 
 enum build_option_type {
@@ -434,6 +475,7 @@ enum build_option_type {
 	op_integer,
 	op_array,
 	op_feature,
+	op_shell_array,
 	build_option_type_count,
 };
 
@@ -532,7 +574,7 @@ struct obj_clear_mark {
 	struct bucket_arr_save obj_aos[obj_type_count - _obj_aos_start];
 };
 
-void make_obj(struct workspace *wk, obj *id, enum obj_type type);
+obj make_obj(struct workspace *wk, enum obj_type type);
 enum obj_type get_obj_type(struct workspace *wk, obj id);
 
 void make_default_objects(struct workspace *wk);
@@ -586,27 +628,27 @@ OBJ_GETTER(obj_iterator);
 
 #undef OBJ_GETTER
 
-struct sbuf;
+struct tstr;
 
 const char *obj_type_to_s(enum obj_type t);
 bool s_to_type_tag(const char *s, type_tag *t);
-void obj_to_s(struct workspace *wk, obj o, struct sbuf *sb);
-void obj_to_json(struct workspace *wk, obj o, struct sbuf *sb);
+void obj_to_s(struct workspace *wk, obj o, struct tstr *sb);
+bool obj_to_json(struct workspace *wk, obj o, struct tstr *sb);
 bool obj_equal(struct workspace *wk, obj left, obj right);
 bool obj_clone(struct workspace *wk_src, struct workspace *wk_dest, obj val, obj *ret);
 
 #define LO(...)                                                   \
-	if (log_should_print(log_debug)) {                        \
-		char buf[4096];                                   \
-		log_print_prefix(log_debug, buf, ARRAY_LEN(buf)); \
-		log_plain("%s", buf);                             \
-		obj_lprintf(wk, __VA_ARGS__);         \
+	{                        \
+		log_print(false, log_debug, "%s", ""); \
+		obj_lprintf(wk, log_debug, __VA_ARGS__);                     \
 	}
 #define LOBJ(object_id) LO("%s: %o\n", #object_id, object_id)
 
-bool obj_lprintf(struct workspace *wk, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 2, 3);
+bool obj_lprintf(struct workspace *wk, enum log_level lvl, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
 bool obj_fprintf(struct workspace *wk, FILE *f, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
 bool obj_printf(struct workspace *wk, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 2, 3);
+bool obj_vasprintf(struct workspace *wk, struct tstr *sb, const char *fmt, va_list ap);
+uint32_t obj_asprintf(struct workspace *wk, struct tstr *buf, const char *fmt, ...);
 uint32_t obj_vsnprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, va_list ap);
 uint32_t obj_snprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, ...)
 	MUON_ATTR_FORMAT(printf, 4, 5);
@@ -620,7 +662,7 @@ bool obj_array_foreach_flat(struct workspace *wk, obj arr, void *usr_ctx, obj_ar
 bool obj_array_in(struct workspace *wk, obj arr, obj val);
 bool obj_array_index_of(struct workspace *wk, obj arr, obj val, uint32_t *idx);
 obj *obj_array_index_pointer(struct workspace *wk, obj arr, int64_t i);
-void obj_array_index(struct workspace *wk, obj arr, int64_t i, obj *res);
+obj obj_array_index(struct workspace *wk, obj arr, int64_t i);
 void obj_array_extend(struct workspace *wk, obj arr, obj arr2);
 void obj_array_extend_nodup(struct workspace *wk, obj arr, obj arr2);
 void obj_array_dup(struct workspace *wk, obj arr, obj *res);
@@ -658,6 +700,10 @@ bool obj_dict_geti(struct workspace *wk, obj dict, uint32_t key, obj *val);
 void obj_dict_del(struct workspace *wk, obj dict, obj key);
 void obj_dict_del_str(struct workspace *wk, obj dict, const char *str);
 void obj_dict_del_strn(struct workspace *wk, obj dict, const char *str, uint32_t len);
+const struct str *obj_dict_index_as_str(struct workspace *wk, obj dict, const char *s);
+bool obj_dict_index_as_bool(struct workspace *wk, obj dict, const char *s);
+int64_t obj_dict_index_as_number(struct workspace *wk, obj dict, const char *s);
+obj obj_dict_index_as_obj(struct workspace *wk, obj dict, const char *s);
 
 bool obj_iterable_foreach(struct workspace *wk, obj dict_or_array, void *ctx, obj_dict_iterator cb);
 #endif
