@@ -49,6 +49,7 @@ enum log_level {
 	log_quiet,
 	log_error,
 	log_warn,
+	log_note,
 	log_info,
 	log_debug,
 	log_level_count,
@@ -61,10 +62,6 @@ enum toolchain_component {
 };
 #define toolchain_component_count 3 // Keep in sync with above
 
-struct complex_types {
-	type_tag options_dict_or_list;
-};
-
 union obj_dict_big_dict_value {
 	uint64_t u64;
 	struct {
@@ -72,4 +69,8 @@ union obj_dict_big_dict_value {
 	} val;
 };
 
+enum error_message_flag {
+	error_message_flag_no_source = 1 << 0,
+	error_message_flag_coalesce = 1 << 1,
+};
 #endif
