@@ -35,9 +35,9 @@
 #include "external/libarchive_null.c"
 #include "external/libcurl_null.c"
 #include "external/readline_builtin.c"
-#include "external/tinyjson_null.c"
 #include "formats/editorconfig.c"
 #include "formats/ini.c"
+#include "formats/json.c"
 #include "formats/lines.c"
 #include "formats/tap.c"
 #include "formats/xml.c"
@@ -67,8 +67,10 @@
 #include "functions/machine.c"
 #include "functions/meson.c"
 #include "functions/modules.c"
+#include "functions/modules/curl.c"
 #include "functions/modules/fs.c"
 #include "functions/modules/getopt.c"
+#include "functions/modules/json.c"
 #include "functions/modules/keyval.c"
 #include "functions/modules/pkgconfig.c"
 #include "functions/modules/python.c"
@@ -89,6 +91,7 @@
 #include "lang/fmt.c"
 #include "lang/func_lookup.c"
 #include "lang/lexer.c"
+#include "lang/lsp.c"
 #include "lang/object.c"
 #include "lang/object_iterators.c"
 #include "lang/parser.c"
@@ -144,11 +147,8 @@
 #include "platform/posix/uname.c"
 #endif
 
-#ifdef BOOTSTRAP_HAVE_LIBPKGCONF
-#include "external/libpkgconf.c"
-#else
-#include "external/libpkgconf_null.c"
-#endif
+#include "external/pkgconfig.c"
+#include "external/pkgconfig_exec.c"
 
 #ifdef BOOTSTRAP_NO_SAMU
 #include "external/samurai_null.c"
