@@ -11,6 +11,7 @@
 #include "args.h"
 #include "compilers.h"
 #include "error.h"
+#include "functions/both_libs.h"
 #include "functions/environment.h"
 #include "lang/workspace.h"
 #include "log.h"
@@ -35,14 +36,15 @@ push_args_null_terminated(struct workspace *wk, obj arr, char *const *argv)
 	}
 }
 
+// This function uses double quotes so that can be windows compatible
 void
-shell_escape_custom(struct workspace *wk, struct sbuf *sb, const char *str, const char *escape_inner, const char *need_escaping)
+shell_escape_custom(struct workspace *wk, struct tstr *sb, const char *str, const char *escape_inner, const char *need_escaping)
 {
 	const char *s;
 	bool do_esc = false;
 
 	if (!*str) {
-		sbuf_pushs(wk, sb, "\"\"");
+		tstr_pushs(wk, sb, "\"\"");
 		return;
 	}
 
@@ -54,74 +56,99 @@ shell_escape_custom(struct workspace *wk, struct sbuf *sb, const char *str, cons
 	}
 
 	if (!do_esc) {
-		sbuf_pushs(wk, sb, str);
+		tstr_pushs(wk, sb, str);
 		return;
 	}
 
-	sbuf_push(wk, sb, '"');
+	tstr_push(wk, sb, '"');
 
 	for (s = str; *s; ++s) {
 		if (strchr(escape_inner, *s)) {
-			sbuf_push(wk, sb, '\\');
+			tstr_push(wk, sb, '\\');
 		}
-		sbuf_push(wk, sb, *s);
+		tstr_push(wk, sb, *s);
 	}
 
-	sbuf_push(wk, sb, '"');
+	tstr_push(wk, sb, '"');
+}
+
+// POSIX shell quoting rules
+//
+// $ is intentionally omitted from these defines
+//
+// 2.2 Quoting
+// The application shall quote the following characters if they are to
+// represent themselves:
+//
+// |  &  ;  <  >  (  )  $  `  \  "  '  <space>  <tab>  <newline>
+// and the following might need to be quoted under certain circumstances. That
+// is, these characters are sometimes special depending on conditions described
+// elsewhere in this volume of POSIX.1-2024:
+//
+// *  ?  [  ]  ^  -  !  #  ~  =  %  {  ,  }
+// (We also omit some of these)
+//
+// 2.2.3 Double-Quotes
+// Enclosing characters in double-quotes ("") shall preserve the literal value
+// of all characters within the double-quotes, with the exception of the
+// characters backquote, <dollar-sign>, and <backslash>
+//
+// We also add " here.
+#define SH_DOUBLE_QUOTE_CHARS "\\\"`$"
+#define SH_QUOTE_CHARS SH_DOUBLE_QUOTE_CHARS "|&;<>()' \t\n" "*?[]!#"
+
+void
+shell_escape(struct workspace *wk, struct tstr *sb, const char *str)
+{
+	shell_escape_custom(wk, sb, str, SH_DOUBLE_QUOTE_CHARS, SH_QUOTE_CHARS);
 }
 
 void
-shell_escape(struct workspace *wk, struct sbuf *sb, const char *str)
+shell_escape_cmd(struct workspace *wk, struct tstr *sb, const char *str)
 {
-	shell_escape_custom(wk, sb, str, "\"$\\", "\"'$ \\><&#()\n");
-}
-
-void
-shell_escape_no_dollar(struct workspace *wk, struct sbuf *sb, const char *str)
-{
-	shell_escape_custom(wk, sb, str, "\"\\", "\"' \\><&#()\n");
+	shell_escape_custom(wk, sb, str, "\\\"", "\\\" \t\r\n");
 }
 
 static void
-simple_escape(struct workspace *wk, struct sbuf *sb, const char *str, const char *need_escaping, char esc_char)
+simple_escape(struct workspace *wk, struct tstr *sb, const char *str, const char *need_escaping, char esc_char)
 {
 	const char *s = str;
 
 	for (; *s; ++s) {
 		if (strchr(need_escaping, *s)) {
-			sbuf_push(wk, sb, esc_char);
+			tstr_push(wk, sb, esc_char);
 		} else if (*s == '\n') {
 			assert(false && "newlines cannot be escaped");
 		}
 
-		sbuf_push(wk, sb, *s);
+		tstr_push(wk, sb, *s);
 	}
 }
 
 void
-ninja_escape(struct workspace *wk, struct sbuf *sb, const char *str)
+ninja_escape(struct workspace *wk, struct tstr *sb, const char *str)
 {
 	simple_escape(wk, sb, str, " :$", '$');
 }
 
 static void
-shell_ninja_escape(struct workspace *wk, struct sbuf *sb, const char *str)
+shell_ninja_escape(struct workspace *wk, struct tstr *sb, const char *str)
 {
-	SBUF_manual(tmp);
+	TSTR_manual(tmp);
 
 	shell_escape(wk, &tmp, str);
 	simple_escape(wk, sb, tmp.buf, "$\n", '$');
 
-	sbuf_destroy(&tmp);
+	tstr_destroy(&tmp);
 }
 
 void
-pkgconf_escape(struct workspace *wk, struct sbuf *sb, const char *str)
+pkgconf_escape(struct workspace *wk, struct tstr *sb, const char *str)
 {
 	simple_escape(wk, sb, str, " ", '\\');
 }
 
-typedef void((*escape_func)(struct workspace *wk, struct sbuf *sb, const char *str));
+typedef void((*escape_func)(struct workspace *wk, struct tstr *sb, const char *str));
 
 struct join_args_iter_ctx {
 	uint32_t i, len;
@@ -136,7 +163,7 @@ join_args_iter(struct workspace *wk, void *_ctx, obj val)
 
 	const char *s = get_cstr(wk, val);
 
-	SBUF(esc);
+	TSTR(esc);
 	if (ctx->escape) {
 		ctx->escape(wk, &esc, s);
 
@@ -177,12 +204,6 @@ join_args_shell(struct workspace *wk, obj arr)
 	return join_args(wk, arr, shell_escape);
 }
 
-obj
-join_args_shell_no_dollar(struct workspace *wk, obj arr)
-{
-	return join_args(wk, arr, shell_escape_no_dollar);
-}
-
 obj
 join_args_ninja(struct workspace *wk, obj arr)
 {
@@ -218,9 +239,9 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 	case obj_string: str = src; break;
 	case obj_file:
 		if (ctx->mode & arr_to_args_relativize_paths) {
-			SBUF(rel);
+			TSTR(rel);
 			path_relative_to(wk, &rel, wk->build_root, get_file_path(wk, src));
-			str = sbuf_into_str(wk, &rel);
+			str = tstr_into_str(wk, &rel);
 			break;
 		}
 		str = *get_obj_file(wk, src);
@@ -231,7 +252,7 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 		}
 		str = get_obj_alias_target(wk, src)->name;
 		break;
-	case obj_both_libs: src = get_obj_both_libs(wk, src)->dynamic_lib;
+	case obj_both_libs: src = decay_both_libs(wk, src);
 	/* fallthrough */
 	case obj_build_target: {
 		if (!(ctx->mode & arr_to_args_build_target)) {
@@ -240,10 +261,10 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 
 		struct obj_build_target *tgt = get_obj_build_target(wk, src);
 
-		SBUF(rel);
+		TSTR(rel);
 		if (ctx->mode & arr_to_args_relativize_paths) {
 			path_relative_to(wk, &rel, wk->build_root, get_cstr(wk, tgt->build_path));
-			str = sbuf_into_str(wk, &rel);
+			str = tstr_into_str(wk, &rel);
 		} else {
 			str = tgt->build_path;
 		}
@@ -294,7 +315,7 @@ type_err:
 bool
 arr_to_args(struct workspace *wk, enum arr_to_args_flags mode, obj arr, obj *res)
 {
-	make_obj(wk, res, obj_array);
+	*res = make_obj(wk, obj_array);
 
 	struct arr_to_args_ctx ctx = { .mode = mode, .res = *res };
 
@@ -364,3 +385,11 @@ env_to_envstr(struct workspace *wk, const char **res, uint32_t *envc, obj val)
 	*res = get_str(wk, ctx.str)->s;
 	*envc = get_obj_dict(wk, dict)->len;
 }
+
+obj
+make_shell_escaped_str(struct workspace *wk, const char *s)
+{
+	TSTR(buf);
+	shell_escape(wk, &buf, s);
+	return tstr_into_str(wk, &buf);
+}
