@@ -18,8 +18,8 @@
 #include "backend/output.h"
 #include "error.h"
 #include "external/samurai.h"
+#include "functions/kernel.h"
 #include "log.h"
-#include "options.h"
 #include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
@@ -72,8 +72,6 @@ write_tgt_iter(struct workspace *wk, void *_ctx, obj tgt_id)
 
 	switch (t) {
 	case obj_alias_target: ret = ninja_write_alias_tgt(wk, tgt_id, ctx); break;
-	case obj_both_libs: tgt_id = get_obj_both_libs(wk, tgt_id)->dynamic_lib;
-	/* fallthrough */
 	case obj_build_target: ret = ninja_write_build_tgt(wk, tgt_id, ctx); break;
 	case obj_custom_target: ret = ninja_write_custom_tgt(wk, tgt_id, ctx); break;
 	default: UNREACHABLE;
@@ -88,6 +86,7 @@ write_tgt_iter(struct workspace *wk, void *_ctx, obj tgt_id)
 	obj_clear(wk, &mk);
 
 	TracyCZoneAutoE;
+	log_progress_inc(wk);
 	return ret;
 }
 
@@ -140,6 +139,16 @@ ninja_write_build(struct workspace *wk, void *_ctx, FILE *out)
 		ninja_coverage_write_targets(wk, out);
 	}
 
+	{ // Add install target for compatibility with meson
+		fprintf(out,
+			"build install: phony muon-internal__install\n"
+			"build muon-internal__install: CUSTOM_COMMAND\n"
+			" desc = Installing$ files\n"
+			" COMMAND = %s install\n"
+			" pool = console\n\n",
+			wk->argv0);
+	}
+
 	if (!wrote_default) {
 		fprintf(out,
 			"build muon_do_nothing: phony\n"
@@ -153,7 +162,7 @@ bool
 ninja_write_all(struct workspace *wk)
 {
 	struct write_build_ctx ctx = { 0 };
-	make_obj(wk, &ctx.compiler_rule_arr, obj_array);
+	ctx.compiler_rule_arr = make_obj(wk, obj_array);
 
 	obj_array_push(wk, wk->backend_output_stack, make_str(wk, "ninja_write_all"));
 
@@ -163,17 +172,36 @@ ninja_write_all(struct workspace *wk)
 
 	obj_array_pop(wk, wk->backend_output_stack);
 
+	{
+		TSTR(ninja_log);
+		path_join(wk, &ninja_log, wk->build_root, output_path.private_dir);
+		path_push(wk, &ninja_log, ".ninja_log");
+
+		if (fs_file_exists(ninja_log.buf)) {
+			obj args = make_obj(wk, obj_array);
+			obj_array_push(wk, args, make_str(wk, "-t"));
+			obj_array_push(wk, args, make_str(wk, "restat"));
+			obj_array_push(wk, args, make_str(wk, "build.ninja"));
+
+			ninja_run(wk,
+				args,
+				wk->build_root,
+				0,
+				ninja_run_flag_ignore_errors | ninja_run_flag_require_ninja);
+		}
+	}
+
 	{ /* compile_commands.json */
 		TracyCZoneN(tctx_compdb, "output compile_commands.json", true);
 
 		obj compdb_args;
-		make_obj(wk, &compdb_args, obj_array);
+		compdb_args = make_obj(wk, obj_array);
 		obj_array_push(wk, compdb_args, make_str(wk, "-C"));
 		obj_array_push(wk, compdb_args, make_str(wk, wk->build_root));
 		obj_array_push(wk, compdb_args, make_str(wk, "-t"));
 		obj_array_push(wk, compdb_args, make_str(wk, "compdb"));
 		obj_array_extend_nodup(wk, compdb_args, ctx.compiler_rule_arr);
-		if (!ninja_run(wk, compdb_args, wk->build_root, "compile_commands.json")) {
+		if (!ninja_run(wk, compdb_args, wk->build_root, "compile_commands.json", 0)) {
 			LOG_E("error writing compile_commands.json");
 		}
 
@@ -184,15 +212,44 @@ ninja_write_all(struct workspace *wk)
 }
 
 bool
-ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture)
+ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture, enum ninja_run_flag flags)
 {
+	bool use_internal_samurai = false;
+	{
+		obj found_ninja;
+		obj version = make_obj(wk, obj_array);
+		obj_array_push(wk, version, make_str(wk, ">=1.7.1"));
+		struct find_program_ctx fp_ctx = {
+			.res = &found_ninja,
+			.version = version,
+			.requirement = requirement_required,
+			.machine = machine_kind_build,
+		};
+		if (!find_program(wk, &fp_ctx, make_str(wk, "ninja"))) {
+			return false;
+		} else if (!fp_ctx.found) {
+			return false;
+		}
+
+		struct obj_external_program *ep = get_obj_external_program(wk, found_ninja);
+
+		if (str_eql(get_str(wk, obj_array_index(wk, ep->cmd_array, 0)), &STRL(wk->argv0))) {
+			use_internal_samurai = true;
+		} else {
+			obj cmd_array_dup;
+			obj_array_dup(wk, ep->cmd_array, &cmd_array_dup);
+			obj_array_extend_nodup(wk, cmd_array_dup, args);
+			args = cmd_array_dup;
+		}
+	}
+
 	const char *argstr;
 	uint32_t argstr_argc;
 
 	bool ret = false;
 	char *const *argv = NULL;
 	uint32_t argc;
-	SBUF_manual(cwd);
+	TSTR_manual(cwd);
 
 	if (chdir) {
 		path_copy_cwd(NULL, &cwd);
@@ -202,7 +259,13 @@ ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture
 		}
 	}
 
-	if (have_samurai) {
+	if (use_internal_samurai) {
+		if (flags & ninja_run_flag_require_ninja) {
+			if (!(flags & ninja_run_flag_ignore_errors)) {
+				obj_lprintf(wk, log_error, "ninja is required for the command `ninja %o`, but only internal samurai was found\n", args);
+			}
+			goto ret;
+		}
 		join_args_argstr(wk, &argstr, &argstr_argc, args);
 		argc = argstr_to_argv(argstr, argstr_argc, "samu", &argv);
 
@@ -223,41 +286,20 @@ ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture
 	} else {
 		struct run_cmd_ctx cmd_ctx = { 0 };
 
-		SBUF_manual(cmd);
+		TSTR_manual(cmd);
 		const char *prepend = NULL;
 
-		obj ninja_opt_id;
-		if (!get_option(wk, NULL, &WKSTR("env.NINJA"), &ninja_opt_id)) {
-			UNREACHABLE;
-		}
-		const struct obj_option *ninja_opt = get_obj_option(wk, ninja_opt_id);
-
-		if (ninja_opt->source == option_value_source_default && have_samurai) {
-			obj cmd_arr;
-			make_obj(wk, &cmd_arr, obj_array);
-			obj_array_push(wk, cmd_arr, make_str(wk, wk->argv0));
-			obj_array_push(wk, cmd_arr, make_str(wk, "samu"));
-			obj_array_extend_nodup(wk, cmd_arr, args);
-			args = cmd_arr;
-		} else if (ninja_opt->source == option_value_source_default
-			   && (fs_find_cmd(NULL, &cmd, "samu") || fs_find_cmd(NULL, &cmd, "ninja"))) {
-			prepend = cmd.buf;
-		} else {
-			obj cmd_arr;
-			obj_array_dup(wk, ninja_opt->val, &cmd_arr);
-			obj_array_extend_nodup(wk, cmd_arr, args);
-			args = cmd_arr;
-		}
-
 		join_args_argstr(wk, &argstr, &argstr_argc, args);
 		argc = argstr_to_argv(argstr, argstr_argc, prepend, &argv);
 
-		if (!capture) {
+		if (!capture && !(flags & ninja_run_flag_ignore_errors)) {
 			cmd_ctx.flags |= run_cmd_ctx_flag_dont_capture;
 		}
 
 		if (!run_cmd_argv(&cmd_ctx, argv, NULL, 0)) {
-			LOG_E("%s", cmd_ctx.err_msg);
+			if (!(flags & ninja_run_flag_ignore_errors)) {
+				LOG_E("%s", cmd_ctx.err_msg);
+			}
 			goto run_cmd_done;
 		}
 
@@ -269,7 +311,7 @@ ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture
 
 		ret = cmd_ctx.status == 0;
 run_cmd_done:
-		sbuf_destroy(&cmd);
+		tstr_destroy(&cmd);
 		run_cmd_ctx_destroy(&cmd_ctx);
 	}
 
@@ -282,6 +324,6 @@ ret:
 		path_chdir(cwd.buf);
 	}
 
-	sbuf_destroy(&cwd);
+	tstr_destroy(&cwd);
 	return ret;
 }
