@@ -39,10 +39,8 @@ enum write_tgt_src_flag {
 static obj
 write_tgt_source(struct workspace *wk, struct write_tgt_source_ctx *ctx, enum compiler_language lang, obj val, enum write_tgt_src_flag flags)
 {
-	const char *src = get_file_path(wk, val);
-
 	/* build paths */
-	SBUF(dest_path);
+	TSTR(dest_path);
 	if ((flags & write_tgt_src_flag_pch)) {
 		if (!tgt_src_to_pch_path(wk, ctx->tgt, lang, val, &dest_path)) {
 			return 0;
@@ -53,13 +51,18 @@ write_tgt_source(struct workspace *wk, struct write_tgt_source_ctx *ctx, enum co
 		}
 	}
 
-	obj dest = sbuf_into_str(wk, &dest_path);
+	obj dest = tstr_into_str(wk, &dest_path);
 
-	if (!(flags & write_tgt_src_flag_pch)) {
+	if ((flags & write_tgt_src_flag_pch)) {
+		if (get_obj_type(wk, val) == obj_build_target) {
+			return dest;
+		}
+	} else {
 		obj_array_push(wk, ctx->object_names, dest);
 	}
 
-	SBUF(src_path);
+	TSTR(src_path);
+	const char *src = get_file_path(wk, val);
 	path_relative_to(wk, &src_path, wk->build_root, src);
 
 	/* build rules and args */
@@ -76,12 +79,12 @@ write_tgt_source(struct workspace *wk, struct write_tgt_source_ctx *ctx, enum co
 			UNREACHABLE;
 		}
 
-		obj_array_index(wk, rule_name_arr, 0, &rule_name);
-		obj_array_index(wk, rule_name_arr, 1, &specialized_rule);
+		rule_name = obj_array_index(wk, rule_name_arr, 0);
+		specialized_rule = obj_array_index(wk, rule_name_arr, 1);
 	}
 
-	SBUF(esc_dest_path);
-	SBUF(esc_path);
+	TSTR(esc_dest_path);
+	TSTR(esc_path);
 
 	ninja_escape(wk, &esc_dest_path, dest_path.buf);
 	ninja_escape(wk, &esc_path, src_path.buf);
@@ -128,8 +131,8 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 	struct obj_build_target *tgt = get_obj_build_target(wk, tgt_id);
 	L("writing rules for target '%s'", get_cstr(wk, tgt->build_name));
 
-	SBUF(esc_path);
-	SBUF(rel_build_path);
+	TSTR(esc_path);
+	TSTR(rel_build_path);
 	{
 		path_relative_to(wk, &rel_build_path, wk->build_root, get_cstr(wk, tgt->build_path));
 		ninja_escape(wk, &esc_path, rel_build_path.buf);
@@ -141,7 +144,7 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		.out = wctx->out,
 	};
 
-	make_obj(wk, &ctx.object_names, obj_array);
+	ctx.object_names = make_obj(wk, obj_array);
 
 	ctx.args = tgt->dep_internal;
 
@@ -160,8 +163,8 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 					esc_path.buf,
 					get_cstr(wk, order_deps));
 				ctx.have_order_deps = false;
-				make_obj(wk, &implicit_deps, obj_array);
-				obj_array_push(wk, implicit_deps, make_strf(wk, "%s-order_deps", esc_path.buf));
+				implicit_deps = make_obj(wk, obj_array);
+				obj_array_push(wk, implicit_deps, make_strf(wk, "%s-order_deps", rel_build_path.buf));
 				have_custom_order_deps = true;
 			} else {
 				ctx.order_deps = order_deps;
@@ -170,6 +173,9 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 	}
 
 	if (implicit_deps) {
+		// We need to go ahead and join implicit deps here so that they can be
+		// used by pch generation if necessary.  The pch will then be added to
+		// the list of implicit deps for the rest of the compiled sources.
 		ctx.implicit_deps = join_args_ninja(wk, implicit_deps);
 	}
 
@@ -179,9 +185,13 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 			enum compiler_language lang = k;
 
 			if (!implicit_deps) {
-				make_obj(wk, &implicit_deps, obj_array);
+				implicit_deps = make_obj(wk, obj_array);
 			}
-			obj_array_push(wk, implicit_deps, write_tgt_source(wk, &ctx, lang, v, write_tgt_src_flag_pch));
+			obj src = write_tgt_source(wk, &ctx, lang, v, write_tgt_src_flag_pch);
+			if (!src) {
+				UNREACHABLE;
+			}
+			obj_array_push(wk, implicit_deps, src);
 		}
 
 		ctx.implicit_deps = join_args_ninja(wk, implicit_deps);
@@ -193,9 +203,9 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		{
 			const char *src = get_file_path(wk, v);
 
-			SBUF(path);
+			TSTR(path);
 			path_relative_to(wk, &path, wk->build_root, src);
-			obj_array_push(wk, ctx.object_names, sbuf_into_str(wk, &path));
+			obj_array_push(wk, ctx.object_names, tstr_into_str(wk, &path));
 		}
 
 		obj_array_for(wk, tgt->src, v) {
@@ -211,7 +221,7 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 	}
 
 	obj implicit_link_deps;
-	make_obj(wk, &implicit_link_deps, obj_array);
+	implicit_link_deps = make_obj(wk, obj_array);
 	if (have_custom_order_deps) {
 		obj_array_push(wk, implicit_link_deps, make_strf(wk, "%s-order_deps", rel_build_path.buf));
 	}
@@ -244,10 +254,13 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		link_args = get_cstr(wk, join_args_shell_ninja(wk, ctx.args.link_args));
 		break;
 	case tgt_static_library:
+		if (!get_obj_array(wk, ctx.object_names)->len) {
+			goto done;
+		}
 		linker_type = "static";
 		link_args = 0;
 		break;
-	default: assert(false); return false;
+	default: UNREACHABLE_RETURN;
 	}
 
 	fprintf(wctx->out, "build %s", esc_path.buf);
@@ -265,6 +278,7 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		linker_type);
 
 	fputs(get_cstr(wk, join_args_ninja(wk, ctx.object_names)), wctx->out);
+
 	if (get_obj_array(wk, implicit_link_deps)->len) {
 		implicit_link_deps = join_args_ninja(wk, implicit_link_deps);
 		fputs(" | ", wctx->out);
@@ -282,6 +296,8 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		wctx->wrote_default = true;
 		fprintf(wctx->out, "\ndefault %s\n", esc_path.buf);
 	}
+
+done:
 	fprintf(wctx->out, "\n");
 	return true;
 }
