@@ -65,12 +65,12 @@ xc_bool(struct xc_ctx *ctx, bool b)
 static obj
 xc_quoted_str(struct xc_ctx *ctx, const char *s)
 {
-	SBUF(tmp);
+	TSTR(tmp);
 	shell_escape(ctx->wk, &tmp, s);
 	if (tmp.buf[0] != '"') {
 		return make_strf(ctx->wk, "\"%s\"", s);
 	} else {
-		return sbuf_into_str(ctx->wk, &tmp);
+		return tstr_into_str(ctx->wk, &tmp);
 	}
 }
 
@@ -101,8 +101,7 @@ struct pbx_item {
 static enum obj_type
 xc_pbx_type(struct xc_ctx *ctx, obj pbx)
 {
-	obj o;
-	obj_array_index(ctx->wk, pbx, 0, &o);
+	obj o = obj_array_index(ctx->wk, pbx, 0);
 	return (enum obj_type)o;
 }
 
@@ -148,7 +147,7 @@ static obj
 xc_pbx_new(struct xc_ctx *ctx, enum obj_type t)
 {
 	obj pbx;
-	make_obj(ctx->wk, &pbx, obj_array);
+	pbx = make_obj(ctx->wk, obj_array);
 	obj_array_push(ctx->wk, pbx, (obj)t);
 	return pbx;
 }
@@ -311,10 +310,10 @@ xc_project_target_sources(struct xc_ctx *ctx, struct project *proj, obj tgt, obj
 	obj sorted;
 	obj_array_sort(ctx->wk, ctx, sources, xc_source_sort, &sorted);
 
-	SBUF(rel);
-	SBUF(prev_dirname);
-	SBUF(dirname);
-	SBUF(basename);
+	TSTR(rel);
+	TSTR(prev_dirname);
+	TSTR(dirname);
+	TSTR(basename);
 	const char *base = get_cstr(ctx->wk, proj->source_root);
 
 	uint32_t stack_base = ctx->stack.len;
@@ -446,8 +445,6 @@ xc_project_target_group(struct xc_ctx *ctx, struct project *proj, obj tgt)
 		// Not sure how to handle these?
 		break;
 	}
-	case obj_both_libs: tgt = get_obj_both_libs(ctx->wk, tgt)->dynamic_lib;
-	/* fallthrough */
 	case obj_build_target: return xc_project_target_build_tgt(ctx, proj, tgt); break;
 	case obj_custom_target: return xc_project_target_custom_tgt(ctx, proj, tgt); break;
 	default: UNREACHABLE;
@@ -464,7 +461,7 @@ static obj
 xc_project_meson_group(struct xc_ctx *ctx, struct project *proj)
 {
 	obj files, f;
-	make_obj(ctx->wk, &files, obj_array);
+	files = make_obj(ctx->wk, obj_array);
 
 	const char *path, *proj_source_root = get_cstr(ctx->wk, proj->source_root);
 
@@ -472,7 +469,7 @@ xc_project_meson_group(struct xc_ctx *ctx, struct project *proj)
 		path = get_cstr(ctx->wk, f);
 		if (path_is_subpath(proj_source_root, path)) {
 			obj new;
-			make_obj(ctx->wk, &new, obj_file);
+			new = make_obj(ctx->wk, obj_file);
 			*get_obj_file(ctx->wk, new) = f;
 			obj_array_push(ctx->wk, files, new);
 		}
@@ -530,7 +527,7 @@ xc_build_configuration_list(struct xc_ctx *ctx, struct project *proj, struct obj
 			(void)_lang;
 			obj_array_for(ctx->wk, args, v) {
 				const struct str *s = get_str(ctx->wk, v);
-				if (str_startswith(s, &WKSTR("-D")) && s->len > 2) {
+				if (str_startswith(s, &STR("-D")) && s->len > 2) {
 					xc_pbx_push_v(ctx, pbx_gcc_preprocessor_definitions, xc_quoted(ctx, make_strn(ctx->wk, s->s + 2, s->len - 2)));
 				}
 			}
@@ -660,7 +657,7 @@ xc_project_ninja_build(struct xc_ctx *ctx, struct project *proj)
 
 	ctx->legacy_target_uuid = xc_pbx_push_root_object(ctx, pbx);
 
-	SBUF(scheme_path);
+	TSTR(scheme_path);
 	path_join(ctx->wk, &scheme_path, ctx->proj_path, "xcshareddata");
 	path_push(ctx->wk, &scheme_path, "xcschemes");
 	if (!fs_mkdir_p(scheme_path.buf)) {
@@ -705,7 +702,7 @@ xc_project_main(struct workspace *_wk, void *_ctx, FILE *_out)
 	ctx->out = _out;
 	struct project *proj = ctx->proj;
 
-	make_obj(ctx->wk, &ctx->tgt_build_files, obj_dict);
+	ctx->tgt_build_files = make_obj(ctx->wk, obj_dict);
 
 	obj pbx_main = xc_pbx_new(ctx, obj_dict);
 	xc_pbx_push_kv(ctx, pbx_main, "archiveVersion", make_number(ctx->wk, 1));
@@ -770,9 +767,9 @@ xc_main(struct workspace *_wk, void *_ctx, FILE *workspace_out)
 		struct project *proj = ctx->proj = arr_get(&ctx->wk->projects, i);
 		ctx->master_project = i == 0;
 
-		SBUF(proj_name);
-		sbuf_pushf(ctx->wk, &proj_name, "%s.xcodeproj", get_cstr(ctx->wk, proj->cfg.name));
-		SBUF(proj_path);
+		TSTR(proj_name);
+		tstr_pushf(ctx->wk, &proj_name, "%s.xcodeproj", get_cstr(ctx->wk, proj->cfg.name));
+		TSTR(proj_path);
 		path_push(ctx->wk, &proj_path, ctx->wk->build_root);
 		path_push(ctx->wk, &proj_path, "xcode-projects");
 		path_push(ctx->wk, &proj_path, proj_name.buf);
@@ -786,7 +783,7 @@ xc_main(struct workspace *_wk, void *_ctx, FILE *workspace_out)
 		xml_node_push_attr(w, xcworkspace_file_ref, "location", make_strf(ctx->wk, "container:%s", proj_path.buf));
 		xml_node_push_child(w, xcworkspace, xcworkspace_file_ref);
 
-		SBUF(pbx_path);
+		TSTR(pbx_path);
 		path_join(ctx->wk, &pbx_path, proj_path.buf, "project.pbxproj");
 
 		ctx->pbx_path = pbx_path.buf;
@@ -803,7 +800,7 @@ xc_main(struct workspace *_wk, void *_ctx, FILE *workspace_out)
 bool
 xcode_write_all(struct workspace *wk)
 {
-	SBUF(xcworkspace_path);
+	TSTR(xcworkspace_path);
 	path_join(wk, &xcworkspace_path, wk->build_root, "main.xcworkspace");
 	if (!fs_mkdir_p(xcworkspace_path.buf)) {
 		return false;
