@@ -35,15 +35,15 @@ static enum iteration_result
 copy_subdir_iter(void *_ctx, const char *path)
 {
 	struct copy_subdir_ctx *ctx = _ctx;
-	SBUF(src);
-	SBUF(dest);
+	TSTR(src);
+	TSTR(dest);
 
 	path_join(ctx->wk, &src, ctx->src_base, path);
 	path_join(ctx->wk, &dest, ctx->dest_base, path);
 
-	SBUF(rel);
+	TSTR(rel);
 	path_relative_to(ctx->wk, &rel, ctx->src_root, src.buf);
-	obj rel_str = sbuf_into_str(ctx->wk, &rel);
+	obj rel_str = tstr_into_str(ctx->wk, &rel);
 
 	if (fs_dir_exists(src.buf)) {
 		if (ctx->exclude_directories && obj_array_in(ctx->wk, ctx->exclude_directories, rel_str)) {
@@ -107,12 +107,12 @@ install_iter(struct workspace *wk, void *_ctx, obj v_id)
 	struct install_ctx *ctx = _ctx;
 	struct obj_install_target *in = get_obj_install_target(wk, v_id);
 
-	SBUF(dest_dirname);
+	TSTR(dest_dirname);
 	const char *dest = get_cstr(wk, in->dest), *src = get_cstr(wk, in->src);
 
 	assert(in->type == install_target_symlink || in->type == install_target_emptydir || path_is_absolute(src));
 
-	SBUF(full_dest_dir);
+	TSTR(full_dest_dir);
 	if (ctx->destdir) {
 		path_join_absolute(wk, &full_dest_dir, get_cstr(wk, ctx->destdir), dest);
 		dest = full_dest_dir.buf;
@@ -207,15 +207,15 @@ install_scripts_iter(struct workspace *wk, void *_ctx, obj install_script)
 	struct install_ctx *ctx = _ctx;
 
 	obj install_script_skip_if_destdir, install_script_dry_run, install_script_cmdline;
-	obj_array_index(wk, install_script, 0, &install_script_skip_if_destdir);
-	obj_array_index(wk, install_script, 1, &install_script_dry_run);
-	obj_array_index(wk, install_script, 2, &install_script_cmdline);
+	install_script_skip_if_destdir = obj_array_index(wk, install_script, 0);
+	install_script_dry_run = obj_array_index(wk, install_script, 1);
+	install_script_cmdline = obj_array_index(wk, install_script, 2);
 
 	bool script_skip_if_destdir = get_obj_bool(wk, install_script_skip_if_destdir);
 	bool script_can_dry_run = get_obj_bool(wk, install_script_dry_run);
 
 	obj env;
-	make_obj(wk, &env, obj_dict);
+	env = make_obj(wk, obj_dict);
 	if (ctx->destdir) {
 		obj_dict_set(wk, env, make_str(wk, "DESTDIR"), ctx->destdir);
 	}
@@ -266,12 +266,12 @@ bool
 install_run(struct install_options *opts)
 {
 	bool ret = true;
-	SBUF_manual(install_src);
+	TSTR_manual(install_src);
 	path_join(NULL, &install_src, output_path.private_dir, output_path.install);
 
 	FILE *f;
 	f = fs_fopen(install_src.buf, "rb");
-	sbuf_destroy(&install_src);
+	tstr_destroy(&install_src);
 
 	if (!f) {
 		return false;
@@ -293,24 +293,24 @@ install_run(struct install_options *opts)
 	};
 
 	obj install_targets, install_scripts, source_root;
-	obj_array_index(&wk, install, 0, &install_targets);
-	obj_array_index(&wk, install, 1, &install_scripts);
-	obj_array_index(&wk, install, 2, &source_root);
-	obj_array_index(&wk, install, 3, &ctx.prefix);
+	install_targets = obj_array_index(&wk, install, 0);
+	install_scripts = obj_array_index(&wk, install, 1);
+	source_root = obj_array_index(&wk, install, 2);
+	ctx.prefix = obj_array_index(&wk, install, 3);
 
-	SBUF(build_root);
+	TSTR(build_root);
 	path_copy_cwd(&wk, &build_root);
-	wk.build_root = get_cstr(&wk, sbuf_into_str(&wk, &build_root));
+	wk.build_root = get_cstr(&wk, tstr_into_str(&wk, &build_root));
 	wk.source_root = get_cstr(&wk, source_root);
 
 	if ((opts->destdir)) {
-		SBUF(full_prefix);
-		SBUF(abs_destdir);
+		TSTR(full_prefix);
+		TSTR(abs_destdir);
 		path_make_absolute(&wk, &abs_destdir, opts->destdir);
 		path_join_absolute(&wk, &full_prefix, abs_destdir.buf, get_cstr(&wk, ctx.prefix));
 
-		ctx.full_prefix = sbuf_into_str(&wk, &full_prefix);
-		ctx.destdir = sbuf_into_str(&wk, &abs_destdir);
+		ctx.full_prefix = tstr_into_str(&wk, &full_prefix);
+		ctx.destdir = tstr_into_str(&wk, &abs_destdir);
 	} else {
 		ctx.full_prefix = ctx.prefix;
 	}
