@@ -16,6 +16,7 @@
 #include "buf_size.h"
 #include "compilers.h"
 #include "error.h"
+#include "functions/string.h"
 #include "guess.h"
 #include "lang/object_iterators.h"
 #include "lang/typecheck.h"
@@ -83,9 +84,9 @@ compiler_check_cache_get(struct workspace *wk, obj key, struct compiler_check_ca
 	obj arr;
 	if (obj_dict_index(wk, wk->compiler_check_cache, key, &arr)) {
 		obj cache_res;
-		obj_array_index(wk, arr, 0, &cache_res);
+		cache_res = obj_array_index(wk, arr, 0);
 		val->success = get_obj_bool(wk, cache_res);
-		obj_array_index(wk, arr, 1, &val->value);
+		val->value = obj_array_index(wk, arr, 1);
 		return true;
 	} else {
 		return false;
@@ -104,7 +105,7 @@ compiler_check_cache_set(struct workspace *wk, obj key, const struct compiler_ch
 		obj_array_set(wk, arr, 0, make_obj_bool(wk, val->success));
 		obj_array_set(wk, arr, 1, val->value);
 	} else {
-		make_obj(wk, &arr, obj_array);
+		arr = make_obj(wk, obj_array);
 
 		obj_array_push(wk, arr, make_obj_bool(wk, val->success));
 		obj_array_push(wk, arr, val->value);
@@ -187,12 +188,27 @@ toolchain_component_from_s(const char *name, uint32_t *res)
 	return toolchain_id_lookup(name, toolchain_component_name, ARRAY_LEN(toolchain_component_name), res);
 }
 
+const struct toolchain_id *
+toolchain_component_type_to_s(enum toolchain_component comp, uint32_t val)
+{
+	const struct toolchain_id *ids = 0;
+	switch (comp) {
+	case toolchain_component_compiler: ids = compiler_type_name; break;
+	case toolchain_component_linker: ids = linker_type_name; break;
+	case toolchain_component_static_linker: ids = static_linker_type_name; break;
+	}
+
+	return &ids[val];
+}
+
 static const char *compiler_language_names[compiler_language_count] = {
 	[compiler_language_null] = "null",
 	[compiler_language_c] = "c",
 	[compiler_language_c_hdr] = "c_hdr",
 	[compiler_language_cpp] = "cpp",
 	[compiler_language_cpp_hdr] = "cpp_hdr",
+	[compiler_language_objc_hdr] = "objc_hdr",
+	[compiler_language_objcpp_hdr] = "objcpp_hdr",
 	[compiler_language_c_obj] = "c_obj",
 	[compiler_language_objc] = "objc",
 	[compiler_language_objcpp] = "objcpp",
@@ -201,6 +217,17 @@ static const char *compiler_language_names[compiler_language_count] = {
 	[compiler_language_nasm] = "nasm",
 };
 
+static const char *compiler_language_gcc_names[compiler_language_count] = {
+	[compiler_language_c] = "c",
+	[compiler_language_c_hdr] = "c-header",
+	[compiler_language_cpp] = "c++",
+	[compiler_language_cpp_hdr] = "c++-header",
+	[compiler_language_objc] = "objective-c",
+	[compiler_language_objc_hdr] = "objective-c-header",
+	[compiler_language_objcpp] = "objective-c++",
+	[compiler_language_objcpp_hdr] = "objective-c++-header",
+};
+
 const char *
 compiler_language_to_s(enum compiler_language l)
 {
@@ -211,10 +238,10 @@ compiler_language_to_s(enum compiler_language l)
 bool
 s_to_compiler_language(const char *_s, enum compiler_language *l)
 {
-	const struct str s = WKSTR(_s);
+	const struct str s = STRL(_s);
 	uint32_t i;
 	for (i = 0; i < compiler_language_count; ++i) {
-		if (str_eqli(&s, &WKSTR(compiler_language_names[i]))) {
+		if (str_eqli(&s, &STRL(compiler_language_names[i]))) {
 			*l = i;
 			return true;
 		}
@@ -223,6 +250,18 @@ s_to_compiler_language(const char *_s, enum compiler_language *l)
 	return false;
 }
 
+enum compiler_language
+compiler_language_to_hdr(enum compiler_language lang)
+{
+	switch (lang) {
+	case compiler_language_c: return compiler_language_c_hdr;
+	case compiler_language_cpp: return compiler_language_cpp_hdr;
+	case compiler_language_objc: return compiler_language_objc_hdr;
+	case compiler_language_objcpp: return compiler_language_objcpp_hdr;
+	default: UNREACHABLE_RETURN;
+	}
+}
+
 static const char *compiler_language_exts[compiler_language_count][10] = {
 	[compiler_language_c] = { "c" },
 	[compiler_language_c_hdr] = { "h" },
@@ -272,6 +311,8 @@ coalesce_link_languages(enum compiler_language cur, enum compiler_language new)
 	case compiler_language_null:
 	case compiler_language_c_hdr:
 	case compiler_language_cpp_hdr:
+	case compiler_language_objc_hdr:
+	case compiler_language_objcpp_hdr:
 	case compiler_language_llvm_ir: break;
 	case compiler_language_assembly:
 		if (!cur) {
@@ -322,27 +363,27 @@ run_cmd_arr(struct workspace *wk, struct run_cmd_ctx *cmd_ctx, obj cmd_arr, cons
 		}
 
 		obj status, err, out;
-		obj_array_index(wk, cache_val.value, 0, &status);
-		obj_array_index(wk, cache_val.value, 1, &out);
-		obj_array_index(wk, cache_val.value, 2, &err);
+		status = obj_array_index(wk, cache_val.value, 0);
+		out = obj_array_index(wk, cache_val.value, 1);
+		err = obj_array_index(wk, cache_val.value, 2);
 
 		cmd_ctx->status = get_obj_number(wk, status);
 
 		const struct str *err_str = get_str(wk, err), *out_str = get_str(wk, out);
 
-		cmd_ctx->out = (struct sbuf){
+		cmd_ctx->out = (struct tstr){
 			.buf = (char *)out_str->s,
 			.len = out_str->len,
 			.cap = out_str->len,
-			.flags = sbuf_flag_overflown,
+			.flags = tstr_flag_overflown,
 			.s = out,
 		};
 
-		cmd_ctx->err = (struct sbuf){
+		cmd_ctx->err = (struct tstr){
 			.buf = (char *)err_str->s,
 			.len = err_str->len,
 			.cap = err_str->len,
-			.flags = sbuf_flag_overflown,
+			.flags = tstr_flag_overflown,
 			.s = err,
 		};
 
@@ -352,15 +393,18 @@ run_cmd_arr(struct workspace *wk, struct run_cmd_ctx *cmd_ctx, obj cmd_arr, cons
 	bool success = true;
 
 	if (!run_cmd(cmd_ctx, argstr, argc, NULL, 0)) {
+		L("failed to run command %s", argstr);
+		run_cmd_print_error(cmd_ctx, log_debug);
+
 		run_cmd_ctx_destroy(cmd_ctx);
-		success = false;
+		return false;
 	}
 
 	cache_val.success = success;
-	make_obj(wk, &cache_val.value, obj_array);
+	cache_val.value = make_obj(wk, obj_array);
 
 	obj status;
-	make_obj(wk, &status, obj_number);
+	status = make_obj(wk, obj_number);
 	set_obj_number(wk, status, cmd_ctx->status);
 	obj_array_push(wk, cache_val.value, status);
 	obj_array_push(wk, cache_val.value, make_strn(wk, cmd_ctx->out.buf, cmd_ctx->out.len));
@@ -400,8 +444,8 @@ compiler_detect_c_or_cpp(struct workspace *wk, obj cmd_arr, obj comp_id)
 		goto detection_over;
 	}
 
-	if (str_containsi(&SBUF_WKSTR(&cmd_ctx.out), &WKSTR("clang"))) {
-		if (str_contains(&SBUF_WKSTR(&cmd_ctx.out), &WKSTR("Apple"))) {
+	if (str_containsi(&TSTR_STR(&cmd_ctx.out), &STR("clang"))) {
+		if (str_contains(&TSTR_STR(&cmd_ctx.out), &STR("Apple"))) {
 			type = compiler_apple_clang;
 		} else if (strstr(cmd_ctx.out.buf, "CL.EXE COMPATIBILITY")) {
 			type = compiler_clang_cl;
@@ -471,17 +515,17 @@ compiler_detect_nasm(struct workspace *wk, obj cmd_arr, obj comp_id)
 		uint32_t addr_bits = host_machine.address_bits;
 
 		const char *plat;
-		SBUF(define);
+		TSTR(define);
 
 		if (host_machine.is_windows) {
 			plat = "win";
-			sbuf_pushf(wk, &define, "WIN%d", addr_bits);
+			tstr_pushf(wk, &define, "WIN%d", addr_bits);
 		} else if (host_machine.sys == machine_system_darwin) {
 			plat = "macho";
-			sbuf_pushs(wk, &define, "MACHO");
+			tstr_pushs(wk, &define, "MACHO");
 		} else {
 			plat = "elf";
-			sbuf_pushs(wk, &define, "ELF");
+			tstr_pushs(wk, &define, "ELF");
 		}
 
 		obj_array_push(wk, new_cmd, make_strf(wk, "-f%s%d", plat, addr_bits));
@@ -527,7 +571,7 @@ compiler_get_libdirs(struct workspace *wk, struct obj_compiler *comp)
 				.len = e ? (uint32_t)(e - s) : strlen(s),
 			};
 
-			comp->libdirs = str_split(wk, &str, &WKSTR(ENV_PATH_SEP_STR));
+			comp->libdirs = str_split(wk, &str, &STR(ENV_PATH_SEP_STR));
 			goto done;
 		}
 
@@ -540,7 +584,7 @@ done:
 	if (!comp->libdirs) {
 		const char *libdirs[] = { "/usr/lib", "/usr/local/lib", "/lib", NULL };
 
-		make_obj(wk, &comp->libdirs, obj_array);
+		comp->libdirs = make_obj(wk, obj_array);
 
 		uint32_t i;
 		for (i = 0; libdirs[i]; ++i) {
@@ -567,9 +611,7 @@ compiler_refine_host_machine(struct workspace *wk, struct obj_compiler *comp)
 static bool
 compiler_detect_cmd_arr(struct workspace *wk, obj comp, enum compiler_language lang, obj cmd_arr)
 {
-	if (log_should_print(log_debug)) {
-		obj_lprintf(wk, "checking compiler %o\n", cmd_arr);
-	}
+	obj_lprintf(wk, log_debug, "checking compiler %o\n", cmd_arr);
 
 	switch (lang) {
 	case compiler_language_c:
@@ -603,6 +645,8 @@ static_linker_detect(struct workspace *wk, obj comp, enum compiler_language lang
 {
 	struct obj_compiler *compiler = get_obj_compiler(wk, comp);
 
+	obj_lprintf(wk, log_debug, "checking static linker %o\n", cmd_arr);
+
 	struct run_cmd_ctx cmd_ctx = { 0 };
 	if (!run_cmd_arr(wk,
 		    &cmd_ctx,
@@ -646,6 +690,8 @@ linker_detect(struct workspace *wk, obj comp, enum compiler_language lang, obj c
 		}
 	}
 
+	obj_lprintf(wk, log_debug, "checking linker %o\n", cmd_arr);
+
 	struct run_cmd_ctx cmd_ctx = { 0 };
 	if (!run_cmd_arr(wk, &cmd_ctx, cmd_arr, guess_version_arg(wk, msvc_like))) {
 		run_cmd_ctx_destroy(&cmd_ctx);
@@ -676,7 +722,7 @@ toolchain_exe_detect(struct workspace *wk,
 	}
 
 	obj cmd_arr_opt;
-	get_option(wk, NULL, &WKSTR(toolchain_exe_option_name), &cmd_arr_opt);
+	get_option(wk, NULL, &STRL(toolchain_exe_option_name), &cmd_arr_opt);
 	struct obj_option *cmd_arr = get_obj_option(wk, cmd_arr_opt);
 
 	if (cmd_arr->source > option_value_source_default) {
@@ -686,7 +732,7 @@ toolchain_exe_detect(struct workspace *wk,
 	uint32_t i;
 	for (i = 0; exe_list[i]; ++i) {
 		obj cmd_arr;
-		make_obj(wk, &cmd_arr, obj_array);
+		cmd_arr = make_obj(wk, obj_array);
 		obj_array_push(wk, cmd_arr, make_str(wk, exe_list[i]));
 
 		if (cb(wk, comp, lang, cmd_arr)) {
@@ -700,18 +746,29 @@ toolchain_exe_detect(struct workspace *wk,
 static bool
 toolchain_linker_detect(struct workspace *wk, obj comp, enum compiler_language lang)
 {
+	const char **exe_list = NULL;
+
 	struct obj_compiler *compiler = get_obj_compiler(wk, comp);
 
-	enum linker_type t = compilers[compiler->type[toolchain_component_compiler]].default_linker;
 	if (host_machine.sys == machine_system_windows) {
 		if (compiler->type[toolchain_component_compiler] == compiler_clang) {
-			t = linker_lld_link;
+			static const char *clang_list[] = { "lld-link", NULL };
+			exe_list = clang_list;
+		} else {
+			static const char *msvc_list[] = { "link", NULL };
+			exe_list = msvc_list;
 		}
+	} else {
+		static const char *default_list[] = { "lld", "ld", NULL };
+		exe_list = default_list;
 	}
 
-	const char *exe_list[] = { linker_type_to_s(t), "ld", NULL };
-
-	return toolchain_exe_detect(wk, "env.LD", exe_list, comp, lang, linker_detect);
+	return toolchain_exe_detect(wk,
+		toolchain_component_option_name[lang][toolchain_component_linker],
+		exe_list,
+		comp,
+		lang,
+		linker_detect);
 }
 
 static bool
@@ -735,14 +792,6 @@ toolchain_static_linker_detect(struct workspace *wk, obj comp, enum compiler_lan
 static bool
 toolchain_compiler_detect(struct workspace *wk, obj comp, enum compiler_language lang)
 {
-	static const char *compiler_option[compiler_language_count] = {
-		[compiler_language_c] = "env.CC",
-		[compiler_language_cpp] = "env.CXX",
-		[compiler_language_objc] = "env.OBJC",
-		[compiler_language_objcpp] = "env.OBJCPP",
-		[compiler_language_nasm] = "env.NASM",
-	};
-
 	const char **exe_list = NULL;
 
 	if (host_machine.sys == machine_system_windows) {
@@ -767,7 +816,12 @@ toolchain_compiler_detect(struct workspace *wk, obj comp, enum compiler_language
 		exe_list = default_executables[lang];
 	}
 
-	return toolchain_exe_detect(wk, compiler_option[lang], exe_list, comp, lang, compiler_detect_cmd_arr);
+	return toolchain_exe_detect(wk,
+		toolchain_component_option_name[lang][toolchain_component_compiler],
+		exe_list,
+		comp,
+		lang,
+		compiler_detect_cmd_arr);
 }
 
 bool
@@ -777,20 +831,20 @@ toolchain_detect(struct workspace *wk, obj *comp, enum machine_kind machine, enu
 		return true;
 	}
 
-	make_obj(wk, comp, obj_compiler);
+	*comp = make_obj(wk, obj_compiler);
 
 	if (!toolchain_compiler_detect(wk, *comp, lang)) {
-		LOG_E("failed to detect compiler");
+		LOG_W("failed to detect compiler for %s", compiler_language_to_s(lang));
 		return false;
 	}
 
 	if (!toolchain_linker_detect(wk, *comp, lang)) {
-		LOG_E("failed to detect linker");
+		LOG_W("failed to detect linker for %s", compiler_language_to_s(lang));
 		return false;
 	}
 
 	if (!toolchain_static_linker_detect(wk, *comp, lang)) {
-		LOG_E("failed to detect static linker");
+		LOG_W("failed to detect static linker for %s", compiler_language_to_s(lang));
 		return false;
 	}
 
@@ -798,11 +852,11 @@ toolchain_detect(struct workspace *wk, obj *comp, enum machine_kind machine, enu
 
 	struct obj_compiler *compiler = get_obj_compiler(wk, *comp);
 
-	LLOG_I("detected %s compiler for %s: %s ",
-		machine_kind_to_s(machine),
-		compiler_language_to_s(lang),
+	LLOG_I("%s: detected %s ",
+		compiler_log_prefix(lang, machine),
 		compiler_type_to_s(compiler->type[toolchain_component_compiler]));
 	obj_lprintf(wk,
+		log_info,
 		"%o (%o), linker: %s (%o), static_linker: %s (%o)\n",
 		compiler->ver,
 		compiler->cmd_arr[toolchain_component_compiler],
@@ -1007,6 +1061,15 @@ TOOLCHAIN_PROTO_1s(compiler_gcc_args_include_system)
 	return &args;
 }
 
+TOOLCHAIN_PROTO_1s(compiler_gcc_args_include_dirafter)
+{
+	TOOLCHAIN_ARGS({ "-idirafter", NULL });
+
+	argv[1] = s1;
+
+	return &args;
+}
+
 TOOLCHAIN_PROTO_2s(compiler_gcc_args_deps)
 {
 	TOOLCHAIN_ARGS({ "-MD", "-MQ", NULL, "-MF", NULL });
@@ -1042,7 +1105,7 @@ TOOLCHAIN_PROTO_1i(compiler_gcc_args_warning_lvl)
 
 	switch ((enum compiler_warning_lvl)i1) {
 	case compiler_warning_lvl_everything: UNREACHABLE; break;
-	case compiler_warning_lvl_3: argv[args.len] = "-Wpedantic"; ++args.len;
+	case compiler_warning_lvl_3: argv[args.len] = "-pedantic"; ++args.len;
 	/* fallthrough */
 	case compiler_warning_lvl_2: argv[args.len] = "-Wextra"; ++args.len;
 	/* fallthrough */
@@ -1071,7 +1134,7 @@ TOOLCHAIN_PROTO_0(compiler_gcc_args_warn_everything)
 	TOOLCHAIN_ARGS({ "-Wall",
 		"-Winvalid-pch",
 		"-Wextra",
-		"-Wpedantic",
+		"-pedantic",
 		"-Wcast-qual",
 		"-Wconversion",
 		"-Wfloat-equal",
@@ -1139,6 +1202,20 @@ TOOLCHAIN_PROTO_0(compiler_gcc_args_warn_everything)
 	return &args;
 }
 
+TOOLCHAIN_PROTO_1i(compiler_gcc_args_force_language)
+{
+	TOOLCHAIN_ARGS({ "-x", 0 });
+
+	if (compiler_language_gcc_names[i1]) {
+		args.args[1] = compiler_language_gcc_names[i1];
+		args.len = 2;
+	} else {
+		args.len = 0;
+	}
+
+	return &args;
+}
+
 TOOLCHAIN_PROTO_0(compiler_clang_args_warn_everything)
 {
 	TOOLCHAIN_ARGS({ "-Weverything" });
@@ -1152,6 +1229,19 @@ TOOLCHAIN_PROTO_1s(compiler_clang_args_include_pch)
 	return &args;
 }
 
+TOOLCHAIN_PROTO_0(compiler_clang_args_emit_pch)
+{
+	TOOLCHAIN_ARGS({ "-Xclang", "-emit-pch" });
+	return &args;
+}
+
+TOOLCHAIN_PROTO_0(compiler_clang_args_pch_extension)
+{
+	TOOLCHAIN_ARGS({ ".pch" })
+
+	return &args;
+}
+
 TOOLCHAIN_PROTO_1s(compiler_gcc_args_set_std)
 {
 	static char buf[BUF_SIZE_S];
@@ -1244,6 +1334,12 @@ TOOLCHAIN_PROTO_1s(compiler_gcc_args_color_output)
 	static char buf[BUF_SIZE_S];
 	TOOLCHAIN_ARGS({ buf });
 
+	if (comp->type[toolchain_component_compiler] == compiler_gcc
+		&& (!comp->ver || version_compare(get_str(wk, comp->ver), &STR("<4.9.0")))) {
+		args.len = 0;
+		return &args;
+	}
+
 	snprintf(buf, BUF_SIZE_S, "-fdiagnostics-color=%s", s1);
 
 	return &args;
@@ -1275,7 +1371,7 @@ TOOLCHAIN_PROTO_1s(compiler_gcc_args_include_pch)
 	static char a[BUF_SIZE_S], b[BUF_SIZE_S];
 	TOOLCHAIN_ARGS({ "-I", a, "-include", b })
 
-	SBUF(buf);
+	TSTR(buf);
 	path_dirname(wk, &buf, s1);
 	assert(buf.len + 1 < sizeof(a));
 	memcpy(a, buf.buf, buf.len + 1);
@@ -1292,12 +1388,9 @@ TOOLCHAIN_PROTO_0(compiler_gcc_args_pch_extension)
 {
 	TOOLCHAIN_ARGS({ ".gch" })
 
-	/* s1; */
-
 	return &args;
 }
 
-
 /* cl compilers
  * see mesonbuild/compilers/mixins/visualstudio.py for reference
  */
@@ -1360,7 +1453,7 @@ TOOLCHAIN_PROTO_1s(compiler_cl_args_output)
 	static char buf[BUF_SIZE_S];
 	TOOLCHAIN_ARGS({ buf });
 
-	if (str_endswithi(&WKSTR(s1), &WKSTR(".exe"))) {
+	if (str_endswithi(&STRL(s1), &STR(".exe"))) {
 		snprintf(buf, BUF_SIZE_S, "/Fe%s", s1);
 	} else {
 		snprintf(buf, BUF_SIZE_S, "/Fo%s", s1);
@@ -1802,6 +1895,7 @@ build_compilers(void)
 	gcc.args.winvalid_pch = compiler_gcc_args_winvalid_pch;
 	gcc.args.set_std = compiler_gcc_args_set_std;
 	gcc.args.include_system = compiler_gcc_args_include_system;
+	gcc.args.include_dirafter = compiler_gcc_args_include_dirafter;
 	gcc.args.pgo = compiler_gcc_args_pgo;
 	gcc.args.pic = compiler_gcc_args_pic;
 	gcc.args.pie = compiler_gcc_args_pie;
@@ -1815,12 +1909,15 @@ build_compilers(void)
 	gcc.args.permissive = compiler_gcc_args_permissive;
 	gcc.args.include_pch = compiler_gcc_args_include_pch;
 	gcc.args.pch_ext = compiler_gcc_args_pch_extension;
+	gcc.args.force_language = compiler_gcc_args_force_language;
 	gcc.default_linker = linker_ld;
 	gcc.default_static_linker = static_linker_ar_gcc;
 
 	struct compiler clang = gcc;
 	clang.args.warn_everything = compiler_clang_args_warn_everything;
 	clang.args.include_pch = compiler_clang_args_include_pch;
+	clang.args.emit_pch = compiler_clang_args_emit_pch;
+	clang.args.pch_ext = compiler_clang_args_pch_extension;
 	clang.default_linker = linker_clang;
 
 	struct compiler apple_clang = clang;
@@ -2216,3 +2313,21 @@ toolchain_dump(struct workspace *wk, struct obj_compiler *comp, struct toolchain
 #undef TOOLCHAIN_ARG_MEMBER
 #undef TOOLCHAIN_ARG_MEMBER_
 }
+
+const char *
+compiler_log_prefix(enum compiler_language lang, enum machine_kind machine)
+{
+	static char buf[256];
+
+	if (machine == machine_kind_build) {
+		snprintf(buf,
+			sizeof(buf),
+			"%s %s machine compiler",
+			compiler_language_to_s(lang),
+			machine_kind_to_s(machine));
+	} else {
+		snprintf(buf, sizeof(buf), "%s compiler", compiler_language_to_s(lang));
+	}
+
+	return buf;
+}
