@@ -7,7 +7,6 @@
 #include "compat.h"
 
 #include <ctype.h>
-#include <errno.h>
 #include <inttypes.h>
 #include <string.h>
 
@@ -497,9 +496,9 @@ err:
 }
 
 static struct samu_nodearray *
-samu_depsparse_msvc(struct samu_ctx *ctx, struct sbuf *out, struct samu_string *deps_prefix)
+samu_depsparse_msvc(struct samu_ctx *ctx, struct tstr *out, struct samu_string *deps_prefix)
 {
-	const struct str prefix = deps_prefix ? WKSTR(deps_prefix->s) : WKSTR("Note: including file: ");
+	const struct str prefix = deps_prefix ? STRL(deps_prefix->s) : STR("Note: including file: ");
 	const char *whitespace = " \r\t";
 	struct str line = { .s = out->buf, .len = 0 };
 	int32_t i;
@@ -525,9 +524,9 @@ samu_depsparse_msvc(struct samu_ctx *ctx, struct sbuf *out, struct samu_string *
 		}
 
 		if (!seen_prefix
-			&& (str_endswithi(&line, &WKSTR(".c")) || str_endswithi(&line, &WKSTR(".cc"))
-				|| str_endswithi(&line, &WKSTR(".cxx")) || str_endswithi(&line, &WKSTR(".cpp"))
-				|| str_endswithi(&line, &WKSTR(".c++")))) {
+			&& (str_endswithi(&line, &STR(".c")) || str_endswithi(&line, &STR(".cc"))
+				|| str_endswithi(&line, &STR(".cxx")) || str_endswithi(&line, &STR(".cpp"))
+				|| str_endswithi(&line, &STR(".c++")))) {
 			// cl.exe likes to output the name of the compiled
 			// file.  ninja filters this here so we do the same.
 			goto cont;
@@ -547,14 +546,14 @@ samu_depsparse_msvc(struct samu_ctx *ctx, struct sbuf *out, struct samu_string *
 
 			str_to_lower(&line);
 
-			SBUF_manual(buf);
-			SBUF_manual(path);
-			sbuf_pushn(0, &buf, line.s, line.len);
+			TSTR_manual(buf);
+			TSTR_manual(path);
+			tstr_pushn(0, &buf, line.s, line.len);
 			path_make_absolute(0, &path, buf.buf);
 
 			// Skip system headers
-			if (str_contains(&line, &WKSTR("program files"))
-				|| str_contains(&line, &WKSTR("microsoft visual studio"))) {
+			if (str_contains(&line, &STR("program files"))
+				|| str_contains(&line, &STR("microsoft visual studio"))) {
 				goto cont;
 			}
 
@@ -565,8 +564,8 @@ samu_depsparse_msvc(struct samu_ctx *ctx, struct sbuf *out, struct samu_string *
 					path.len,
 				});
 
-			sbuf_destroy(&buf);
-			sbuf_destroy(&path);
+			tstr_destroy(&buf);
+			tstr_destroy(&path);
 		} else {
 			for (i = 0; (uint32_t)i < line.len; ++i) {
 				samu_bufadd(&ctx->arena, &ctx->deps.buf, line.s[i]);
@@ -627,7 +626,7 @@ samu_depsload(struct samu_ctx *ctx, struct samu_edge *e)
 }
 
 void
-samu_depsrecord(struct samu_ctx *ctx, struct sbuf *output, const char **filtered_output, struct samu_edge *e)
+samu_depsrecord(struct samu_ctx *ctx, struct tstr *output, const char **filtered_output, struct samu_edge *e)
 {
 	struct samu_string *deptype_str, *depfile;
 	struct samu_nodearray *deps;
