@@ -29,7 +29,10 @@
 static void
 samu_usage(struct samu_ctx *ctx)
 {
-	fprintf(stderr, "usage: %s [-C dir] [-f buildfile] [-j maxjobs] [-k maxfail] [-n]\n", ctx->argv0);
+	fprintf(stderr, "usage: %s [-C dir] [-d debugflag] [-f buildfile] [-j maxjobs]"
+		" [-k maxfail] [-n] [-t tool] [-v] [-w warnflag] [target...]\n"
+		"       %s -h | --help\n",
+		ctx->argv0, ctx->argv0);
 	exit(2);
 }
 
@@ -77,9 +80,9 @@ samu_jobsflag(struct samu_ctx *ctx, const char *flag)
 	char *end;
 
 	num = strtol(flag, &end, 10);
-	if (*end || num < 0)
+	if (*end || num <= 0)
 		samu_fatal("invalid -j parameter");
-	ctx->buildopts.maxjobs = num > 0 ? num : -1;
+	ctx->buildopts.maxjobs = num;
 }
 
 static void
@@ -151,7 +154,7 @@ samu_main(int argc, char *argv[], struct samu_opts *opts)
 	case '-':
 		arg = SAMU_EARGF(samu_usage(ctx));
 		if (strcmp(arg, "version") == 0) {
-			samu_printf(ctx, "%d.%d.0\n", ninjamajor, ninjaminor);
+			samu_printf(ctx, "%d.%d.0\n", samu_ninjamajor, samu_ninjaminor);
 			return true;
 		} else if (strcmp(arg, "verbose") == 0) {
 			ctx->buildopts.verbose = true;
@@ -184,7 +187,7 @@ samu_main(int argc, char *argv[], struct samu_opts *opts)
 		ctx->buildopts.dryrun = true;
 		break;
 	case 't':
-		tool = samu_toolget(SAMU_EARGF(samu_usage(ctx)));
+		tool = samu_toolget(SAMU_EARGF(samu_toollist(ctx)));
 		goto argdone;
 	case 'v':
 		ctx->buildopts.verbose = true;
