@@ -17,7 +17,7 @@ func_array_length(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_number);
+	*res = make_obj(wk, obj_number);
 	set_obj_number(wk, *res, get_obj_array(wk, self)->len);
 	return true;
 }
@@ -40,7 +40,7 @@ func_array_get(struct workspace *wk, obj self, obj *res)
 			return false;
 		}
 	} else {
-		obj_array_index(wk, self, i, res);
+		*res = obj_array_index(wk, self, i);
 	}
 
 	return true;
@@ -125,7 +125,6 @@ func_array_slice(struct workspace *wk, obj self, obj *res)
 	bounds_adjust(a->len, &start);
 	bounds_adjust(a->len, &end);
 
-	start = MIN(start, a->len - 1);
 	end = MIN(end, a->len);
 
 	*res = obj_array_slice(wk, self, start, end);
@@ -143,6 +142,17 @@ func_array_clear(struct workspace *wk, obj self, obj *res)
 	return true;
 }
 
+static bool
+func_array_dedup(struct workspace *wk, obj self, obj *res)
+{
+	if (!pop_args(wk, 0, 0)) {
+		return false;
+	}
+
+	obj_array_dedup(wk, self, res);
+	return true;
+}
+
 const struct func_impl impl_tbl_array[] = {
 	{ "length", func_array_length, tc_number, true },
 	{ "get", func_array_get, tc_any, true },
@@ -157,5 +167,6 @@ const struct func_impl impl_tbl_array_internal[] = {
 	{ "delete", func_array_delete },
 	{ "slice", func_array_slice, tc_array, true },
 	{ "clear", func_array_clear },
+	{ "dedup", func_array_dedup, tc_array, true },
 	{ NULL, NULL },
 };
