@@ -36,6 +36,7 @@
 #include "platform/filesystem.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
+#include "version.h"
 #include "wrap.h"
 
 static bool
@@ -70,7 +71,7 @@ project_add_language(struct workspace *wk,
 	}
 
 	if (compiler) {
-		make_obj(wk, &comp_id, obj_compiler);
+		comp_id = make_obj(wk, obj_compiler);
 		struct obj_compiler *c = get_obj_compiler(wk, comp_id);
 		*c = *get_obj_compiler(wk, compiler);
 		c->lang = l;
@@ -108,7 +109,7 @@ project_add_language(struct workspace *wk,
 		const enum compiler_type type = comp->type[toolchain_component_compiler];
 		if (type == compiler_clang || type == compiler_apple_clang) {
 			obj llvm_ir_compiler;
-			make_obj(wk, &llvm_ir_compiler, obj_compiler);
+			llvm_ir_compiler = make_obj(wk, obj_compiler);
 			struct obj_compiler *c = get_obj_compiler(wk, llvm_ir_compiler);
 			*c = *comp;
 			c->type[toolchain_component_compiler] = compiler_clang_llvm_ir;
@@ -166,7 +167,7 @@ func_project(struct workspace *wk, obj _, obj *res)
 		kw_version,
 	};
 	struct args_kw akw[] = {
-		[kw_default_options] = { "default_options", wk->complex_types.options_dict_or_list },
+		[kw_default_options] = { "default_options", COMPLEX_TYPE_PRESET(tc_cx_options_dict_or_list) },
 		[kw_license] = { "license", TYPE_TAG_LISTIFY | obj_string },
 		[kw_license_files] = { "license_files", TYPE_TAG_LISTIFY | obj_string },
 		[kw_meson_version] = { "meson_version", obj_string },
@@ -203,7 +204,7 @@ func_project(struct workspace *wk, obj _, obj *res)
 	}
 
 #ifndef MUON_BOOTSTRAPPED
-	if (wk->cur_project == 0 && !str_eql(get_str(wk, an[0].val), &WKSTR("muon"))) {
+	if (wk->cur_project == 0 && !str_eql(get_str(wk, an[0].val), &STR("muon"))) {
 		vm_error_at(wk,
 			an[0].node,
 			"This muon has not been fully bootstrapped. It can only be used to setup muon itself.");
@@ -211,6 +212,19 @@ func_project(struct workspace *wk, obj _, obj *res)
 	}
 #endif
 
+#ifdef MUON_BOOTSTRAPPED
+	if (akw[kw_meson_version].set) {
+		if (!version_compare(&STRL(muon_version.meson_compat), get_str(wk, akw[kw_meson_version].val))) {
+			vm_error_at(wk,
+				akw[kw_meson_version].node,
+				"meson compatibility version %s does not meet requirement: %o",
+				muon_version.meson_compat,
+				akw[kw_meson_version].val);
+			return false;
+		}
+	}
+#endif
+
 	obj val;
 	obj_array_for(wk, an[1].val, val) {
 		bool _found;
@@ -272,7 +286,7 @@ func_project(struct workspace *wk, obj _, obj *res)
 	}
 
 	{ // subprojects
-		SBUF(subprojects_path);
+		TSTR(subprojects_path);
 		path_join(wk,
 			&subprojects_path,
 			get_cstr(wk, current_project(wk)->source_root),
@@ -284,9 +298,9 @@ func_project(struct workspace *wk, obj _, obj *res)
 		}
 	}
 
-	LOG_I("configuring '%s', version: %s",
-		get_cstr(wk, current_project(wk)->cfg.name),
-		get_cstr(wk, current_project(wk)->cfg.version));
+	LLOG_I(CLR(c_bold, c_magenta) "%s" CLR(0), get_cstr(wk, current_project(wk)->cfg.name));
+	log_plain_version_string(log_info, get_cstr(wk, current_project(wk)->cfg.version));
+	log_plain(log_info, "\n");
 
 	current_project(wk)->initialized = true;
 	return true;
@@ -297,7 +311,7 @@ get_project_argument_array(struct workspace *wk, obj dict, enum compiler_languag
 {
 	obj arg_arr;
 	if (!obj_dict_geti(wk, dict, l, &arg_arr)) {
-		make_obj(wk, &arg_arr, obj_array);
+		arg_arr = make_obj(wk, obj_array);
 		obj_dict_seti(wk, dict, l, arg_arr);
 	}
 
@@ -547,23 +561,12 @@ func_files(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	return coerce_files(wk, an[0].node, an[0].val, res);
+	return coerce_files(wk, 0, an[0].val, res);
 }
 
-struct find_program_iter_ctx {
-	bool found;
-	uint32_t node, version_node;
-	obj version, version_argument;
-	obj dirs;
-	obj *res;
-	enum requirement_type requirement;
-	enum machine_kind machine;
-	struct args_kw *default_options;
-};
-
 struct find_program_custom_dir_ctx {
 	const char *prog;
-	struct sbuf *buf;
+	struct tstr *buf;
 	bool found;
 };
 
@@ -582,19 +585,19 @@ find_program_custom_dir_iter(struct workspace *wk, void *_ctx, obj val)
 	return ir_cont;
 }
 
-static bool
-find_program_check_override(struct workspace *wk, struct find_program_iter_ctx *ctx, obj prog)
+bool
+find_program_check_override(struct workspace *wk, struct find_program_ctx *ctx, obj prog)
 {
 	obj override;
 	if (!obj_dict_index(wk, wk->find_program_overrides[ctx->machine], prog, &override)) {
 		return true;
 	}
 
-	obj over = 0, op;
+	obj override_version = 0, op;
 	switch (get_obj_type(wk, override)) {
 	case obj_array:
-		obj_array_index(wk, override, 0, &op);
-		obj_array_index(wk, override, 1, &over);
+		op = obj_array_index(wk, override, 0);
+		override_version = obj_array_index(wk, override, 1);
 		break;
 	case obj_python_installation:
 	case obj_external_program:
@@ -606,27 +609,24 @@ find_program_check_override(struct workspace *wk, struct find_program_iter_ctx *
 		}
 
 		if (ctx->version) {
-			find_program_guess_version(wk, ep->cmd_array, ctx->version_argument, &over);
+			find_program_guess_version(wk, ep->cmd_array, ctx->version_argument, &override_version);
 		}
 		break;
 	default: UNREACHABLE;
 	}
 
-	if (ctx->version && over) {
-		bool comparison_result;
-		if (!version_compare(wk, ctx->version_node, get_str(wk, over), ctx->version, &comparison_result)) {
-			return false;
-		} else if (!comparison_result) {
+	if (ctx->version && override_version) {
+		if (!version_compare_list(wk, get_str(wk, override_version), ctx->version)) {
 			return true;
 		}
 	}
 
 	if (get_obj_type(wk, op) == obj_file) {
 		obj newres;
-		make_obj(wk, &newres, obj_external_program);
+		newres = make_obj(wk, obj_external_program);
 		struct obj_external_program *ep = get_obj_external_program(wk, newres);
 		ep->found = true;
-		make_obj(wk, &ep->cmd_array, obj_array);
+		ep->cmd_array = make_obj(wk, obj_array);
 		obj_array_push(wk, ep->cmd_array, *get_obj_file(wk, op));
 		op = newres;
 	}
@@ -637,15 +637,15 @@ find_program_check_override(struct workspace *wk, struct find_program_iter_ctx *
 }
 
 static bool
-find_program_check_fallback(struct workspace *wk, struct find_program_iter_ctx *ctx, obj prog)
+find_program_check_fallback(struct workspace *wk, struct find_program_ctx *ctx, obj prog)
 {
 	obj fallback_arr, subproj_name;
 	if (obj_dict_index(wk, current_project(wk)->wrap_provides_exes, prog, &fallback_arr)) {
 		obj_array_flatten_one(wk, fallback_arr, &subproj_name);
 
 		obj subproj;
-		if (!subproject(wk, subproj_name, requirement_auto, ctx->default_options, NULL, &subproj)
-			&& get_obj_subproject(wk, subproj)->found) {
+		if (!(subproject(wk, subproj_name, requirement_auto, ctx->default_options, NULL, &subproj)
+			    && get_obj_subproject(wk, subproj)->found)) {
 			return true;
 		}
 
@@ -656,9 +656,10 @@ find_program_check_fallback(struct workspace *wk, struct find_program_iter_ctx *
 			if (!obj_dict_index(wk, wk->find_program_overrides[ctx->machine], prog, &_)) {
 				vm_warning_at(wk,
 					0,
-					"subproject %o claims to provide %o, but did not override it",
+					"subproject %o claims to provide %o for the %s machine, but did not override it",
 					subproj_name,
-					prog);
+					prog,
+					machine_kind_to_s(ctx->machine));
 			}
 		}
 	}
@@ -666,12 +667,32 @@ find_program_check_fallback(struct workspace *wk, struct find_program_iter_ctx *
 	return true;
 }
 
-static bool
-find_program(struct workspace *wk, struct find_program_iter_ctx *ctx, obj prog)
+bool
+find_program_check_version(struct workspace *wk, struct find_program_ctx *ctx, obj ver)
+{
+	if (!ctx->version) {
+		return true;
+	}
+
+	if (!ver) {
+		return true; // no version to check against
+	}
+
+	if (!version_compare_list(wk, get_str(wk, ver), ctx->version)) {
+		LO("version %o does not meet requirement: %o\n", ver, ctx->version);
+		return false;
+	}
+
+	return true;
+}
+
+bool
+find_program(struct workspace *wk, struct find_program_ctx *ctx, obj prog)
 {
 	const char *str;
 	obj ver = 0;
 	bool guessed_ver = false;
+	obj cmd_array = 0;
 
 	type_tag tc_allowed = tc_file | tc_string | tc_external_program | tc_python_installation;
 	if (!typecheck(wk, ctx->node, prog, tc_allowed)) {
@@ -693,9 +714,9 @@ find_program(struct workspace *wk, struct find_program_iter_ctx *ctx, obj prog)
 	default: UNREACHABLE_RETURN;
 	}
 
-	const char *path;
+	const char *path = 0;
 
-	SBUF(buf);
+	TSTR(buf);
 	struct find_program_custom_dir_ctx dir_ctx = {
 		.buf = &buf,
 		.prog = str,
@@ -707,24 +728,35 @@ find_program(struct workspace *wk, struct find_program_iter_ctx *ctx, obj prog)
 		const bool is_meson = strcmp(str, "meson") == 0;
 		const bool is_muon = !is_meson && strcmp(str, "muon") == 0;
 		if (is_meson || is_muon) {
-			make_obj(wk, ctx->res, obj_external_program);
-			struct obj_external_program *ep = get_obj_external_program(wk, *ctx->res);
-			ep->found = true;
-			make_obj(wk, &ep->cmd_array, obj_array);
-
 			const char *argv0_resolved;
-			SBUF(argv0);
+			TSTR(argv0);
 			if (fs_find_cmd(wk, &argv0, wk->argv0)) {
 				argv0_resolved = argv0.buf;
 			} else {
 				argv0_resolved = wk->argv0;
 			}
-			obj_array_push(wk, ep->cmd_array, make_str(wk, argv0_resolved));
+
+			obj ver = 0;
+			cmd_array = make_obj(wk, obj_array);
+			obj_array_push(wk, cmd_array, make_str(wk, argv0_resolved));
 
 			if (is_meson) {
-				obj_array_push(wk, ep->cmd_array, make_str(wk, "meson"));
+				obj_array_push(wk, cmd_array, make_str(wk, "meson"));
+				ver = make_str(wk, muon_version.meson_compat);
+			} else {
+				ver = make_str(wk, muon_version.version);
 			}
 
+			if (!find_program_check_version(wk, ctx, ver)) {
+				return true;
+			}
+
+			*ctx->res = make_obj(wk, obj_external_program);
+			struct obj_external_program *ep = get_obj_external_program(wk, *ctx->res);
+			ep->found = true;
+			ep->cmd_array = cmd_array;
+			ep->ver = ver;
+			ep->guessed_ver = true;
 			ctx->found = true;
 			return true;
 		}
@@ -777,7 +809,24 @@ find_program_step_4:
 		goto found;
 	}
 
-	/* 6. PATH environment variable */
+	/* 6. PATH environment variable, and program specific environment variables */
+	{
+		// TODO: avoid duplicating this
+		// TODO: should this apply to compiler envvars, e.g. env.CC?
+		const char *program_specific_envvar = 0;
+		if (strcmp(str, "ninja") == 0) {
+			program_specific_envvar = "env.NINJA";
+		} else if (strcmp(str, "ar") == 0) {
+			program_specific_envvar = "env.AR";
+		} else if (strcmp(str, "pkg-config") == 0 || strcmp(str, "pkgconf") == 0) {
+			program_specific_envvar = "env.PKG_CONFIG";
+		}
+
+		if (program_specific_envvar) {
+			get_option_value(wk, NULL, program_specific_envvar, &cmd_array);
+		}
+	}
+
 	if (fs_find_cmd(wk, &buf, str)) {
 		path = buf.buf;
 		goto found;
@@ -802,13 +851,19 @@ find_program_step_8:
 	/* 8. Special cases, only if the binary was not found by regular means */
 	if (t == obj_string) {
 		if (have_samurai && (strcmp(str, "ninja") == 0 || strcmp(str, "samu") == 0)) {
-			make_obj(wk, ctx->res, obj_external_program);
+			obj ver = make_strf(wk, "%d.%d.0", samu_ninjamajor, samu_ninjaminor);
+			if (!find_program_check_version(wk, ctx, ver)) {
+				return true;
+			}
+
+			*ctx->res = make_obj(wk, obj_external_program);
 			struct obj_external_program *ep = get_obj_external_program(wk, *ctx->res);
 			ep->found = true;
-			make_obj(wk, &ep->cmd_array, obj_array);
+			ep->cmd_array = make_obj(wk, obj_array);
 			obj_array_push(wk, ep->cmd_array, make_str(wk, wk->argv0));
 			obj_array_push(wk, ep->cmd_array, make_str(wk, "samu"));
-
+			ep->guessed_ver = true;
+			ep->ver = ver;
 			ctx->found = true;
 			return true;
 		}
@@ -816,27 +871,21 @@ find_program_step_8:
 
 	return true;
 found: {
-	obj cmd_array;
-	make_obj(wk, &cmd_array, obj_array);
-	obj_array_push(wk, cmd_array, make_str(wk, path));
+	if (!cmd_array) {
+		cmd_array = make_obj(wk, obj_array);
+		obj_array_push(wk, cmd_array, make_str(wk, path));
+	}
 
 	if (ctx->version) {
 		find_program_guess_version(wk, cmd_array, ctx->version_argument, &ver);
 		guessed_ver = true;
 
-		if (!ver) {
-			return true; // no version to check against
-		}
-
-		bool comparison_result;
-		if (!version_compare(wk, ctx->version_node, get_str(wk, ver), ctx->version, &comparison_result)) {
-			return false;
-		} else if (!comparison_result) {
+		if (!find_program_check_version(wk, ctx, ver)) {
 			return true;
 		}
 	}
 
-	make_obj(wk, ctx->res, obj_external_program);
+	*ctx->res = make_obj(wk, obj_external_program);
 	struct obj_external_program *ep = get_obj_external_program(wk, *ctx->res);
 	ep->found = true;
 	ep->cmd_array = cmd_array;
@@ -868,7 +917,7 @@ func_find_program(struct workspace *wk, obj _, obj *res)
 		[kw_dirs] = { "dirs", TYPE_TAG_LISTIFY | obj_string },
 		[kw_version] = { "version", TYPE_TAG_LISTIFY | obj_string },
 		[kw_version_argument] = { "version_argument", obj_string },
-		[kw_default_options] = { "default_options", wk->complex_types.options_dict_or_list },
+		[kw_default_options] = { "default_options", COMPLEX_TYPE_PRESET(tc_cx_options_dict_or_list) },
 		0,
 	};
 	if (!pop_args(wk, an, akw)) {
@@ -884,13 +933,13 @@ func_find_program(struct workspace *wk, obj _, obj *res)
 		if (akw[kw_disabler].set && get_obj_bool(wk, akw[kw_disabler].val)) {
 			*res = obj_disabler;
 		} else {
-			make_obj(wk, res, obj_external_program);
+			*res = make_obj(wk, obj_external_program);
 			get_obj_external_program(wk, *res)->found = false;
 		}
 		return true;
 	}
 
-	struct find_program_iter_ctx ctx = {
+	struct find_program_ctx ctx = {
 		.node = an[0].node,
 		.version = akw[kw_version].val,
 		.version_argument = akw[kw_version_argument].val,
@@ -922,7 +971,7 @@ func_find_program(struct workspace *wk, obj _, obj *res)
 		if (akw[kw_disabler].set && get_obj_bool(wk, akw[kw_disabler].val)) {
 			*res = obj_disabler;
 		} else {
-			make_obj(wk, res, obj_external_program);
+			*res = make_obj(wk, obj_external_program);
 			get_obj_external_program(wk, *res)->found = false;
 		}
 	}
@@ -974,11 +1023,11 @@ func_generator(struct workspace *wk, obj _, obj *res)
 	}
 
 	obj command;
-	make_obj(wk, &command, obj_array);
+	command = make_obj(wk, obj_array);
 	obj_array_push(wk, command, an[0].val);
 	obj_array_extend(wk, command, akw[kw_arguments].val);
 
-	make_obj(wk, res, obj_generator);
+	*res = make_obj(wk, obj_generator);
 	struct obj_generator *gen = get_obj_generator(wk, *res);
 
 	gen->output = akw[kw_output].val;
@@ -1011,6 +1060,8 @@ func_assert(struct workspace *wk, obj _, obj *res)
 	if (!get_obj_bool(wk, an[0].val)) {
 		if (an[1].set) {
 			LOG_E("%s", get_cstr(wk, an[1].val));
+		} else {
+			vm_error(wk, "assertion failed");
 		}
 		return false;
 	}
@@ -1019,7 +1070,7 @@ func_assert(struct workspace *wk, obj _, obj *res)
 }
 
 static bool
-func_debug(struct workspace *wk, obj _, obj *res)
+func_log_common(struct workspace *wk, enum log_level lvl)
 {
 	struct args_norm an[] = { { tc_message }, ARG_TYPE_NULL };
 
@@ -1027,57 +1078,52 @@ func_debug(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	log_plain(log_clr() ? "\033[35mdebug\033[0m: " : "debug: ");
+	if (lvl == log_error) {
+		struct source *src = 0;
+		struct source_location loc = { 0 };
+		vm_lookup_inst_location(&wk->vm, wk->vm.ip - 1, &loc, &src);
+
+		struct detailed_source_location dloc = { 0 };
+		get_detailed_source_location(src, loc, &dloc, get_detailed_source_location_flag_multiline);
+
+		log_plain(lvl, "%s:%d:%d: ", src->label, dloc.line, dloc.col);
+	}
+
+	log_print(false, lvl, "%s", "");
 	obj val;
 	obj_array_for(wk, an[0].val, val) {
-		obj_lprintf(wk, "%#o ", val);
+		obj_lprintf(wk, lvl, "%#o ", val);
 	}
-	log_plain("\n");
-	*res = 0;
+	log_plain(lvl, "\n");
 
 	return true;
 }
 
+static bool
+func_debug(struct workspace *wk, obj _, obj *res)
+{
+	return func_log_common(wk, log_debug);
+}
+
 static bool
 func_message(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { tc_message }, ARG_TYPE_NULL };
-
-	if (!pop_args(wk, an, NULL)) {
-		return false;
-	}
-
-	log_plain(log_clr() ? "\033[34mmessage\033[0m " : "message: ");
-	obj val;
-	obj_array_for(wk, an[0].val, val) {
-		obj_lprintf(wk, "%#o ", val);
-	}
-	log_plain("\n");
-	*res = 0;
-
-	return true;
+	return func_log_common(wk, log_note);
 }
 
 static bool
 func_error(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { tc_message }, ARG_TYPE_NULL };
-
-	if (!pop_args(wk, an, NULL)) {
-		return false;
-	}
-
-	log_plain(log_clr() ? "\033[31merror\033[0m " : "error: ");
-	obj val;
-	obj_array_for(wk, an[0].val, val) {
-		obj_lprintf(wk, "%#o ", val);
-	}
-	log_plain("\n");
-	*res = 0;
-
+	func_log_common(wk, log_error);
 	return false;
 }
 
+static bool
+func_warning(struct workspace *wk, obj _, obj *res)
+{
+	return func_log_common(wk, log_warn);
+}
+
 static bool
 func_print(struct workspace *wk, obj _, obj *res)
 {
@@ -1087,27 +1133,7 @@ func_print(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	log_plain("%s", get_cstr(wk, an[0].val));
-	*res = 0;
-
-	return true;
-}
-
-static bool
-func_warning(struct workspace *wk, obj _, obj *res)
-{
-	struct args_norm an[] = { { tc_message }, ARG_TYPE_NULL };
-
-	if (!pop_args(wk, an, NULL)) {
-		return false;
-	}
-
-	log_plain(log_clr() ? "\033[33mwarn\033[0m " : "warn: ");
-	obj val;
-	obj_array_for(wk, an[0].val, val) {
-		obj_lprintf(wk, "%#o ", val);
-	}
-	log_plain("\n");
+	log_plain(log_info, "%s", get_cstr(wk, an[0].val));
 	*res = 0;
 
 	return true;
@@ -1139,7 +1165,7 @@ func_run_command(struct workspace *wk, obj _, obj *res)
 	{
 		obj arg0;
 		obj cmd_file = 0;
-		struct find_program_iter_ctx find_program_ctx = {
+		struct find_program_ctx find_program_ctx = {
 			.node = an[0].node,
 			.res = &cmd_file,
 			.requirement = requirement_auto,
@@ -1151,7 +1177,7 @@ func_run_command(struct workspace *wk, obj _, obj *res)
 			return false;
 		}
 
-		obj_array_index(wk, an[0].val, 0, &arg0);
+		arg0 = obj_array_index(wk, an[0].val, 0);
 
 		if (get_obj_type(wk, arg0) == obj_compiler) {
 			obj cmd_arr = get_obj_compiler(wk, arg0)->cmd_arr[toolchain_component_compiler];
@@ -1198,10 +1224,10 @@ func_run_command(struct workspace *wk, obj _, obj *res)
 	if (!run_cmd(&cmd_ctx, argstr, argc, envstr, envc)) {
 		vm_error(wk, "%s", cmd_ctx.err_msg);
 		if (cmd_ctx.out.len) {
-			log_plain("stdout:\n%s", cmd_ctx.out.buf);
+			log_plain(log_info, "stdout:\n%s", cmd_ctx.out.buf);
 		}
 		if (cmd_ctx.err.len) {
-			log_plain("stderr:\n%s", cmd_ctx.err.buf);
+			log_plain(log_info, "stderr:\n%s", cmd_ctx.err.buf);
 		}
 
 		goto ret;
@@ -1210,24 +1236,24 @@ func_run_command(struct workspace *wk, obj _, obj *res)
 	if (akw[kw_check].set && get_obj_bool(wk, akw[kw_check].val) && cmd_ctx.status != 0) {
 		vm_error(wk, "command failed");
 		if (cmd_ctx.out.len) {
-			log_plain("stdout:\n%s", cmd_ctx.out.buf);
+			log_plain(log_info, "stdout:\n%s", cmd_ctx.out.buf);
 		}
 		if (cmd_ctx.err.len) {
-			log_plain("stderr:\n%s", cmd_ctx.err.buf);
+			log_plain(log_info, "stderr:\n%s", cmd_ctx.err.buf);
 		}
 
 		goto ret;
 	}
 
-	make_obj(wk, res, obj_run_result);
+	*res = make_obj(wk, obj_run_result);
 	struct obj_run_result *run_result = get_obj_run_result(wk, *res);
 	run_result->status = cmd_ctx.status;
 	if (akw[kw_capture].set && !get_obj_bool(wk, akw[kw_capture].val)) {
 		run_result->out = make_str(wk, "");
 		run_result->err = make_str(wk, "");
 	} else {
-		run_result->out = sbuf_into_str(wk, &cmd_ctx.out);
-		run_result->err = sbuf_into_str(wk, &cmd_ctx.err);
+		run_result->out = tstr_into_str(wk, &cmd_ctx.out);
+		run_result->err = tstr_into_str(wk, &cmd_ctx.err);
 	}
 
 	ret = true;
@@ -1306,8 +1332,7 @@ func_subdir(struct workspace *wk, obj _, obj *res)
 	enum kwargs {
 		kw_if_found,
 	};
-	type_tag if_found_type = wk->vm.in_analyzer ? tc_any : TYPE_TAG_LISTIFY | tc_dependency;
-	struct args_kw akw[] = { [kw_if_found] = { "if_found", if_found_type }, 0 };
+	struct args_kw akw[] = { [kw_if_found] = { "if_found", TYPE_TAG_LISTIFY | tc_dependency }, 0 };
 	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
@@ -1321,17 +1346,17 @@ func_subdir(struct workspace *wk, obj _, obj *res)
 		}
 	}
 
-	SBUF(build_dir);
+	TSTR(build_dir);
 
 	obj old_cwd = current_project(wk)->cwd;
 	obj old_build_dir = current_project(wk)->build_dir;
 
-	SBUF(new_cwd);
+	TSTR(new_cwd);
 	path_join(wk, &new_cwd, get_cstr(wk, old_cwd), get_cstr(wk, an[0].val));
-	current_project(wk)->cwd = make_str(wk, new_cwd.buf);
+	current_project(wk)->cwd = tstr_into_str(wk, &new_cwd);
 
 	path_join(wk, &build_dir, get_cstr(wk, old_build_dir), get_cstr(wk, an[0].val));
-	current_project(wk)->build_dir = sbuf_into_str(wk, &build_dir);
+	current_project(wk)->build_dir = tstr_into_str(wk, &build_dir);
 
 	bool ret = false;
 	if (!wk->vm.in_analyzer) {
@@ -1344,11 +1369,11 @@ func_subdir(struct workspace *wk, obj _, obj *res)
 
 	{
 		enum build_language lang;
-		const char *build_file = determine_build_file(wk, new_cwd.buf, &lang);
+		const char *build_file = determine_build_file(wk, new_cwd.buf, &lang, false);
 		if (!build_file) {
 			goto ret;
 		}
-		ret = wk->vm.behavior.eval_project_file(wk, build_file, lang, 0);
+		ret = wk->vm.behavior.eval_project_file(wk, build_file, lang, 0, 0);
 	}
 
 ret:
@@ -1366,13 +1391,13 @@ func_configuration_data(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	make_obj(wk, res, obj_configuration_data);
+	*res = make_obj(wk, obj_configuration_data);
 
 	if (an[0].set) {
 		get_obj_configuration_data(wk, *res)->dict = an[0].val;
 	} else {
 		obj dict;
-		make_obj(wk, &dict, obj_dict);
+		dict = make_obj(wk, obj_dict);
 		get_obj_configuration_data(wk, *res)->dict = dict;
 	}
 
@@ -1406,7 +1431,7 @@ func_add_test_setup(struct workspace *wk, obj _, obj *ret)
 	}
 
 	obj test_setup;
-	make_obj(wk, &test_setup, obj_array);
+	test_setup = make_obj(wk, obj_array);
 
 	obj env = 0;
 	if (akw[kw_env].set && !coerce_environment_from_kwarg(wk, &akw[kw_env], false, &env)) {
@@ -1432,7 +1457,7 @@ func_add_test_setup(struct workspace *wk, obj _, obj *ret)
 	obj_array_push(wk, test_setup, akw[kw_timeout_multiplier].val);
 
 	if (!current_project(wk)->test_setups) {
-		make_obj(wk, &current_project(wk)->test_setups, obj_array);
+		current_project(wk)->test_setups = make_obj(wk, obj_array);
 	}
 
 	obj_array_push(wk, current_project(wk)->test_setups, test_setup);
@@ -1447,7 +1472,7 @@ struct add_test_depends_ctx {
 static enum iteration_result
 add_test_depends_iter(struct workspace *wk, void *_ctx, obj val)
 {
-	SBUF(rel);
+	TSTR(rel);
 	struct add_test_depends_ctx *ctx = _ctx;
 
 	switch (get_obj_type(wk, val)) {
@@ -1461,16 +1486,19 @@ add_test_depends_iter(struct workspace *wk, void *_ctx, obj val)
 		}
 
 		path_relative_to(wk, &rel, wk->build_root, get_file_path(wk, val));
-		obj_array_push(wk, ctx->t->depends, sbuf_into_str(wk, &rel));
+		obj_array_push(wk, ctx->t->depends, tstr_into_str(wk, &rel));
 		break;
 
-	case obj_both_libs: val = get_obj_both_libs(wk, val)->dynamic_lib;
-	/* fallthrough */
+	case obj_both_libs: {
+		add_test_depends_iter(wk, ctx, get_obj_both_libs(wk, val)->dynamic_lib);
+		add_test_depends_iter(wk, ctx, get_obj_both_libs(wk, val)->static_lib);
+		break;
+	}
 	case obj_build_target: {
 		struct obj_build_target *tgt = get_obj_build_target(wk, val);
 
 		path_relative_to(wk, &rel, wk->build_root, get_cstr(wk, tgt->build_path));
-		obj_array_push(wk, ctx->t->depends, sbuf_into_str(wk, &rel));
+		obj_array_push(wk, ctx->t->depends, tstr_into_str(wk, &rel));
 		break;
 	}
 	case obj_custom_target:
@@ -1541,7 +1569,7 @@ add_test_common(struct workspace *wk, enum test_category cat)
 		};
 
 		for (protocol = 0; (uint32_t)protocol < ARRAY_LEN(protocol_names); ++protocol) {
-			if (str_eql(get_str(wk, akw[kw_protocol].val), &WKSTR(protocol_names[protocol]))) {
+			if (str_eql(get_str(wk, akw[kw_protocol].val), &STRL(protocol_names[protocol]))) {
 				break;
 			}
 		}
@@ -1567,7 +1595,10 @@ add_test_common(struct workspace *wk, enum test_category cat)
 
 	obj args = exe_args;
 	if (akw[kw_args].set) {
-		if (!arr_to_args(wk, arr_to_args_build_target | arr_to_args_custom_target, akw[kw_args].val, &args)) {
+		if (!arr_to_args(wk,
+			    arr_to_args_build_target | arr_to_args_custom_target | arr_to_args_external_program,
+			    akw[kw_args].val,
+			    &args)) {
 			return false;
 		}
 
@@ -1578,7 +1609,7 @@ add_test_common(struct workspace *wk, enum test_category cat)
 	}
 
 	obj test;
-	make_obj(wk, &test, obj_test);
+	test = make_obj(wk, obj_test);
 	struct obj_test *t = get_obj_test(wk, test);
 
 	if (!coerce_environment_from_kwarg(wk, &akw[kw_env], false, &t->env)) {
@@ -1602,7 +1633,7 @@ add_test_common(struct workspace *wk, enum test_category cat)
 	}
 
 	struct add_test_depends_ctx deps_ctx = { .t = t };
-	make_obj(wk, &t->depends, obj_array);
+	t->depends = make_obj(wk, obj_array);
 	add_test_depends_iter(wk, &deps_ctx, an[1].val);
 	if (akw[kw_depends].set) {
 		obj_array_foreach(wk, akw[kw_depends].val, &deps_ctx, add_test_depends_iter);
@@ -1628,7 +1659,7 @@ func_benchmark(struct workspace *wk, obj _, obj *ret)
 }
 
 struct join_paths_ctx {
-	struct sbuf *buf;
+	struct tstr *buf;
 };
 
 static enum iteration_result
@@ -1653,7 +1684,7 @@ func_join_paths(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	SBUF(join_paths_buf);
+	TSTR(join_paths_buf);
 	struct join_paths_ctx ctx = {
 		.buf = &join_paths_buf,
 	};
@@ -1662,7 +1693,7 @@ func_join_paths(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	*res = sbuf_into_str(wk, ctx.buf);
+	*res = tstr_into_str(wk, ctx.buf);
 	return true;
 }
 
@@ -1694,9 +1725,9 @@ func_environment(struct workspace *wk, obj _, obj *res)
 	enum environment_set_mode mode = environment_set_mode_set;
 	if (akw[kw_method].set) {
 		const struct str modes[] = {
-			[environment_set_mode_set] = WKSTR("set"),
-			[environment_set_mode_append] = WKSTR("append"),
-			[environment_set_mode_prepend] = WKSTR("prepend"),
+			[environment_set_mode_set] = STR("set"),
+			[environment_set_mode_append] = STR("append"),
+			[environment_set_mode_prepend] = STR("prepend"),
 		}, *method = get_str(wk, akw[kw_method].val);
 
 		uint32_t i;
@@ -1714,9 +1745,9 @@ func_environment(struct workspace *wk, obj _, obj *res)
 		mode = i;
 	}
 
-	make_obj(wk, res, obj_environment);
+	*res = make_obj(wk, obj_environment);
 	struct obj_environment *d = get_obj_environment(wk, *res);
-	make_obj(wk, &d->actions, obj_array);
+	d->actions = make_obj(wk, obj_array);
 
 	if (an[0].set) {
 		obj dict;
@@ -1764,7 +1795,7 @@ func_import(struct workspace *wk, obj _, obj *res)
 	bool found = false;
 
 	if (requirement == requirement_skip) {
-		make_obj(wk, res, obj_module);
+		*res = make_obj(wk, obj_module);
 	} else {
 		if (module_import(wk, get_cstr(wk, an[0].val), true, res)) {
 			found = true;
@@ -1880,14 +1911,33 @@ func_subdir_done(struct workspace *wk, obj _, obj *res)
 	UNREACHABLE_RETURN;
 }
 
+static void
+summary_push_kv(struct workspace *wk, obj dest, obj k, obj v, obj attr)
+{
+	obj wrapped_v = make_obj(wk, obj_array);
+	obj_array_push(wk, wrapped_v, attr);
+	obj_array_push(wk, wrapped_v, v);
+	obj_dict_set(wk, dest, k, wrapped_v);
+}
+
 static bool
 func_summary(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { tc_any }, { tc_any, .optional = true }, ARG_TYPE_NULL };
+	type_tag value_base_type = tc_number | tc_bool | tc_string | tc_external_program | tc_dependency
+				   | tc_feature_opt;
+	type_tag value_type = make_complex_type(wk,
+		complex_type_or,
+		value_base_type,
+		make_complex_type(wk, complex_type_nested, tc_array, value_base_type));
+	type_tag dict_type = make_complex_type(wk, complex_type_nested, tc_dict, value_type);
+
+	struct args_norm an[] = { { make_complex_type(wk, complex_type_or, tc_string, dict_type) },
+		{ value_type, .optional = true },
+		ARG_TYPE_NULL };
 	enum kwargs {
 		kw_section,
-		kw_bool_yn, // ignored
-		kw_list_sep, // ignored
+		kw_bool_yn,
+		kw_list_sep,
 	};
 	struct args_kw akw[] = {
 		[kw_section] = { "section", obj_string, },
@@ -1899,32 +1949,45 @@ func_summary(struct workspace *wk, obj _, obj *res)
 		return false;
 	}
 
-	obj sec = akw[kw_section].set ? akw[kw_section].val : make_str(wk, "");
-	obj dict;
+	obj attr = 0;
+	if (akw[kw_list_sep].val) {
+		if (!attr) {
+			attr = make_obj(wk, obj_dict);
+		}
+		obj_dict_set(wk, attr, make_str(wk, "list_sep"), akw[kw_list_sep].val);
+	}
+	if (akw[kw_bool_yn].val) {
+		if (!attr) {
+			attr = make_obj(wk, obj_dict);
+		}
+		obj_dict_set(wk, attr, make_str(wk, "bool_yn"), akw[kw_bool_yn].val);
+	}
+
+	obj section = akw[kw_section].set ? akw[kw_section].val : make_str(wk, "");
+
+	obj dest;
+	if (!obj_dict_index(wk, current_project(wk)->summary, section, &dest)) {
+		dest = make_obj(wk, obj_dict);
+		obj_dict_set(wk, current_project(wk)->summary, section, dest);
+	}
 
 	if (an[1].set) {
-		if (!typecheck(wk, an[0].node, an[0].val, obj_string)) {
+		if (!typecheck(wk, an[0].node, an[0].val, tc_string)) {
 			return false;
 		}
 
-		make_obj(wk, &dict, obj_dict);
-		obj_dict_set(wk, dict, an[0].val, an[1].val);
+		summary_push_kv(wk, dest, an[0].val, an[1].val, attr);
 	} else {
-		if (!typecheck(wk, an[0].node, an[0].val, obj_dict)) {
+		if (!typecheck(wk, an[0].node, an[0].val, dict_type)) {
 			return false;
 		}
 
-		dict = an[0].val;
+		obj k, v;
+		obj_dict_for(wk, an[0].val, k, v) {
+			summary_push_kv(wk, dest, k, v, attr);
+		}
 	}
 
-	obj prev;
-	if (obj_dict_index(wk, current_project(wk)->summary, sec, &prev)) {
-		obj ndict;
-		obj_dict_merge(wk, prev, dict, &ndict);
-		dict = ndict;
-	}
-
-	obj_dict_set(wk, current_project(wk)->summary, sec, dict);
 	return true;
 }
 
@@ -1935,7 +1998,7 @@ make_alias_target(struct workspace *wk, obj name, obj deps)
 	assert(get_obj_type(wk, deps) == obj_array && "Alias target list must be an array.");
 
 	obj id;
-	make_obj(wk, &id, obj_alias_target);
+	id = make_obj(wk, obj_alias_target);
 	struct obj_alias_target *alias_tgt = get_obj_alias_target(wk, id);
 
 	alias_tgt->name = name;
@@ -1954,8 +2017,11 @@ push_alias_target_deps_iter(struct workspace *wk, void *_ctx, obj val)
 	struct alias_target_iter_ctx *ctx = _ctx;
 	enum obj_type t = get_obj_type(wk, val);
 	switch (t) {
-	case obj_both_libs: val = get_obj_both_libs(wk, val)->dynamic_lib;
-	/* fallthrough */
+	case obj_both_libs: {
+		push_alias_target_deps_iter(wk, ctx, get_obj_both_libs(wk, val)->dynamic_lib);
+		push_alias_target_deps_iter(wk, ctx, get_obj_both_libs(wk, val)->static_lib);
+		break;
+	}
 	case obj_alias_target:
 	case obj_build_target:
 	case obj_custom_target: obj_array_push(wk, ctx->deps, val); break;
@@ -1978,7 +2044,7 @@ func_alias_target(struct workspace *wk, obj _, obj *res)
 	L("adding alias target '%s'", get_cstr(wk, an[0].val));
 
 	obj deps_id;
-	make_obj(wk, &deps_id, obj_array);
+	deps_id = make_obj(wk, obj_array);
 
 	struct alias_target_iter_ctx iter_ctx = {
 		.deps = deps_id,
@@ -2032,7 +2098,7 @@ func_range(struct workspace *wk, obj _, obj *res)
 		params.step = 1;
 	}
 
-	make_obj(wk, res, obj_iterator);
+	*res = make_obj(wk, obj_iterator);
 	struct obj_iterator *iter = get_obj_iterator(wk, *res);
 	iter->type = obj_iterator_type_range;
 	iter->data.range = params;
@@ -2065,7 +2131,7 @@ func_p(struct workspace *wk, obj _, obj *res)
 		obj_inspect(wk, an[0].val);
 	} else {
 		const char *fmt = akw[kw_pretty].set && get_obj_bool(wk, akw[kw_pretty].val) ? "%#o\n" : "%o\n";
-		obj_lprintf(wk, fmt, an[0].val);
+		obj_lprintf(wk, log_info, fmt, an[0].val);
 	}
 
 	*res = an[0].val;
@@ -2084,7 +2150,7 @@ func_serial_load(struct workspace *wk, obj _, obj *res)
 	coerce_string(wk, an[0].node, an[0].val, &str);
 
 	FILE *f;
-	if (str_eql(get_str(wk, str), &WKSTR("-"))) {
+	if (str_eql(get_str(wk, str), &STR("-"))) {
 		f = stdin;
 	} else if (!(f = fs_fopen(get_cstr(wk, str), "rb"))) {
 		return false;
@@ -2177,6 +2243,30 @@ func_exit(struct workspace *wk, obj _, obj *res)
 	return true;
 }
 
+static bool
+func_create_enum(struct workspace *wk, obj _, obj *res)
+{
+	struct args_norm an[] = {
+		{ tc_string, .desc = "The value for this enum" },
+		{ TYPE_TAG_LISTIFY | tc_string, "The list of possible values for this enum" },
+		ARG_TYPE_NULL,
+	};
+	if (!pop_args(wk, an, NULL)) {
+		return false;
+	}
+
+	const struct str *s = get_str(wk, an[0].val);
+
+	if (!obj_array_in(wk, an[1].val, an[0].val)) {
+		vm_error_at(wk, an[0].node, "value %o not in list of values", an[0].val);
+		return false;
+	}
+
+	*res = make_strn_enum(wk, s->s, s->len, an[1].val);
+
+	return true;
+}
+
 // clang-format off
 const struct func_impl impl_tbl_kernel[] =
 {
@@ -2188,7 +2278,7 @@ const struct func_impl impl_tbl_kernel[] =
 	{ "add_project_link_arguments", func_add_project_link_arguments },
 	{ "add_test_setup", func_add_test_setup },
 	{ "alias_target", func_alias_target, tc_alias_target },
-	{ "assert", func_assert },
+	{ "assert", func_assert, .flags = func_impl_flag_throws_error },
 	{ "benchmark", func_benchmark },
 	{ "both_libraries", func_both_libraries, tc_both_libs },
 	{ "build_target", func_build_target, tc_build_target | tc_both_libs },
@@ -2200,12 +2290,12 @@ const struct func_impl impl_tbl_kernel[] =
 	{ "dependency", func_dependency, tc_dependency, true },
 	{ "disabler", func_disabler, tc_disabler },
 	{ "environment", func_environment, tc_environment },
-	{ "error", func_error },
+	{ "error", func_error, .flags = func_impl_flag_throws_error },
 	{ "executable", func_executable, tc_build_target },
 	{ "files", func_files, tc_array },
 	{ "find_program", func_find_program, tc_external_program },
 	{ "generator", func_generator, tc_generator },
-	{ "get_option", func_get_option, tc_string | tc_number | tc_bool | tc_feature_opt | tc_array, },
+	{ "get_option", func_get_option, tc_string | tc_number | tc_bool | tc_feature_opt | tc_array, true, },
 	{ "get_variable", func_get_variable, tc_any, true },
 	{ "import", func_import, tc_module, true },
 	{ "include_directories", func_include_directories, tc_array },
@@ -2237,16 +2327,17 @@ const struct func_impl impl_tbl_kernel[] =
 	{ "vcs_tag", func_vcs_tag, tc_custom_target },
 	{ "warning", func_warning },
 	// non-standard muon extensions
-	{ "p", func_p, tc_any, true, .extension = true },
+	{ "p", func_p, tc_any, true, .flags = func_impl_flag_extension },
 	{ NULL, NULL },
 };
 
 const struct func_impl impl_tbl_kernel_internal[] = {
-	{ "assert", func_assert },
+	{ "assert", func_assert, .flags = func_impl_flag_throws_error },
+	{ "configure_file", func_configure_file, tc_file },
 	{ "configuration_data", func_configuration_data, tc_configuration_data },
 	{ "disabler", func_disabler, tc_disabler },
 	{ "environment", func_environment, tc_environment },
-	{ "error", func_error },
+	{ "error", func_error, .flags = func_impl_flag_throws_error },
 	{ "files", func_files, tc_array },
 	{ "find_program", func_find_program, tc_external_program },
 	{ "get_variable", func_get_variable, tc_any, true },
@@ -2256,7 +2347,7 @@ const struct func_impl impl_tbl_kernel_internal[] = {
 	{ "join_paths", func_join_paths, tc_string, true },
 	{ "message", func_message },
 	{ "range", func_range, tc_array, true },
-	{ "run_command", func_run_command, tc_run_result, .fuzz_unsafe = true },
+	{ "run_command", func_run_command, tc_run_result, .flags = func_impl_flag_sandbox_disable },
 	{ "set_variable", func_set_variable, 0, true },
 	{ "unset_variable", func_unset_variable, 0, true },
 	{ "warning", func_warning },
@@ -2264,10 +2355,11 @@ const struct func_impl impl_tbl_kernel_internal[] = {
 	{ "p", func_p, tc_any, true },
 	{ "print", func_print, tc_any },
 	{ "serial_load", func_serial_load, tc_any },
-	{ "serial_dump", func_serial_dump, .fuzz_unsafe = true },
+	{ "serial_dump", func_serial_dump, .flags = func_impl_flag_sandbox_disable },
 	{ "is_null", func_is_null, tc_bool, true },
 	{ "typeof", func_typeof, tc_string, true },
 	{ "exit", func_exit },
+	{ "create_enum", func_create_enum, tc_string, true, .desc = "Create a string enum.  The resulting string will warn if it is compared against a value that it can never contain." },
 	{ NULL, NULL },
 };
 
