@@ -103,7 +103,7 @@ determine_linker_from_objects_iter(struct workspace *wk, void *_ctx, obj val)
 	 */
 
 	const struct str *o = get_str(wk, *get_obj_file(wk, val));
-	SBUF(path);
+	TSTR(path);
 	path_basename(wk, &path, o->s);
 	const char *first_dot = strrchr(path.buf, '.');
 	path.len = first_dot - path.buf;
@@ -162,6 +162,7 @@ struct process_build_tgt_sources_ctx {
 	uint32_t err_node;
 	obj tgt_id;
 	obj res;
+	obj prepend_include_directories;
 	bool implicit_include_directories;
 };
 
@@ -174,15 +175,15 @@ process_source_include(struct workspace *wk, struct process_build_tgt_sources_ct
 		return true;
 	}
 
-	SBUF(dir);
-	SBUF(path);
+	TSTR(dir);
+	TSTR(path);
 	path_relative_to(wk, &path, wk->build_root, src);
 
 	struct obj_build_target *tgt = get_obj_build_target(wk, ctx->tgt_id);
-	obj_array_push(wk, tgt->dep_internal.order_deps, sbuf_into_str(wk, &path));
+	obj_array_push(wk, tgt->dep_internal.order_deps, tstr_into_str(wk, &path));
 
 	if (!tgt->dep_internal.raw.order_deps) {
-		make_obj(wk, &tgt->dep_internal.raw.order_deps, obj_array);
+		tgt->dep_internal.raw.order_deps = make_obj(wk, obj_array);
 	}
 	obj_array_push(wk, tgt->dep_internal.raw.order_deps, val);
 
@@ -193,14 +194,31 @@ process_source_include(struct workspace *wk, struct process_build_tgt_sources_ct
 	path_dirname(wk, &dir, src);
 
 	obj inc;
-	make_obj(wk, &inc, obj_include_directory);
+	inc = make_obj(wk, obj_include_directory);
 	struct obj_include_directory *d = get_obj_include_directory(wk, inc);
-	d->path = sbuf_into_str(wk, &dir);
-	obj_array_push(wk, tgt->dep_internal.include_directories, inc);
+	d->path = tstr_into_str(wk, &dir);
+	obj_array_push(wk, ctx->prepend_include_directories, inc);
+
+	// mkdir so that the include dir doesn't get pruned later on
+	if (!fs_mkdir_p(dir.buf)) {
+		return false;
+	}
 
 	return true;
 }
 
+static void
+build_tgt_inc_required_compiler(struct workspace *wk, struct obj_build_target *tgt, enum compiler_language lang)
+{
+	obj n;
+	if (obj_dict_geti(wk, tgt->required_compilers, lang, &n)) {
+		obj_dict_seti(wk, tgt->required_compilers, lang, n + 1);
+	} else {
+		obj_dict_seti(wk, tgt->required_compilers, lang, 1);
+	}
+}
+
+
 static enum iteration_result
 build_tgt_push_source_files_iter(struct workspace *wk, void *_ctx, obj val)
 {
@@ -227,12 +245,7 @@ build_tgt_push_source_files_iter(struct workspace *wk, void *_ctx, obj val)
 		return ir_cont;
 	}
 
-	obj n;
-	if (obj_dict_geti(wk, tgt->required_compilers, lang, &n)) {
-		obj_dict_seti(wk, tgt->required_compilers, lang, n + 1);
-	} else {
-		obj_dict_seti(wk, tgt->required_compilers, lang, 1);
-	}
+	build_tgt_inc_required_compiler(wk, tgt, lang);
 
 	obj_array_push(wk, ctx->res, val);
 	return ir_cont;
@@ -355,11 +368,11 @@ setup_dllname(struct workspace *wk, struct obj_build_target *tgt, const char *pl
 	}
 
 	if (tgt->type == tgt_dynamic_library) {
-		SBUF(implib);
-		sbuf_pushf(wk, &implib, "%s-implib.lib", plain_name);
-		SBUF(path);
+		TSTR(implib);
+		tstr_pushf(wk, &implib, "%s-implib.lib", plain_name);
+		TSTR(path);
 		path_join(wk, &path, get_cstr(wk, tgt->build_dir), implib.buf);
-		tgt->implib = sbuf_into_str(wk, &path);
+		tgt->implib = tstr_into_str(wk, &path);
 	}
 }
 
@@ -370,8 +383,8 @@ setup_shared_object_symlinks(struct workspace *wk,
 	obj *plain_name_install,
 	obj *soname_install)
 {
-	SBUF(soname_symlink);
-	SBUF(plain_name_symlink);
+	TSTR(soname_symlink);
+	TSTR(plain_name_symlink);
 
 	if (!fs_mkdir_p(get_cstr(wk, tgt->build_dir))) {
 		return false;
@@ -384,18 +397,18 @@ setup_shared_object_symlinks(struct workspace *wk,
 			return false;
 		}
 
-		*soname_install = sbuf_into_str(wk, &soname_symlink);
+		*soname_install = tstr_into_str(wk, &soname_symlink);
 	}
 
-	if (!str_eql(&WKSTR(plain_name), get_str(wk, tgt->soname))
-		&& !str_eql(&WKSTR(plain_name), get_str(wk, tgt->build_name))) {
+	if (!str_eql(&STRL(plain_name), get_str(wk, tgt->soname))
+		&& !str_eql(&STRL(plain_name), get_str(wk, tgt->build_name))) {
 		path_join(wk, &plain_name_symlink, get_cstr(wk, tgt->build_dir), plain_name);
 
 		if (!fs_make_symlink(get_cstr(wk, tgt->soname), plain_name_symlink.buf, true)) {
 			return false;
 		}
 
-		*plain_name_install = sbuf_into_str(wk, &plain_name_symlink);
+		*plain_name_install = tstr_into_str(wk, &plain_name_symlink);
 	}
 
 	return true;
@@ -509,7 +522,7 @@ create_target(struct workspace *wk,
 	obj *res)
 {
 	char plain_name[BUF_SIZE_2k + 1] = { 0 };
-	make_obj(wk, res, obj_build_target);
+	*res = make_obj(wk, obj_build_target);
 	struct obj_build_target *tgt = get_obj_build_target(wk, *res);
 	tgt->type = type;
 	tgt->name = an[0].val;
@@ -517,30 +530,42 @@ create_target(struct workspace *wk,
 	tgt->build_dir = current_project(wk)->build_dir;
 	tgt->machine = coerce_machine_kind(wk, &akw[bt_kw_native]);
 	tgt->callstack = vm_callstack(wk);
-	make_obj(wk, &tgt->args, obj_dict);
-	make_obj(wk, &tgt->src, obj_array);
-	make_obj(wk, &tgt->required_compilers, obj_dict);
-	make_obj(wk, &tgt->extra_files, obj_array);
-	build_dep_init(wk, &tgt->dep_internal);
+	tgt->args = make_obj(wk, obj_dict);
+	tgt->src = make_obj(wk, obj_array);
+	tgt->required_compilers = make_obj(wk, obj_dict);
+	tgt->extra_files = make_obj(wk, obj_array);
 
-	{ // linker args (process before dependencies so link_with libs come first on link line
-		if (akw[bt_kw_link_with].set) {
-			if (!dep_process_link_with(
-				    wk, akw[bt_kw_link_with].node, akw[bt_kw_link_with].val, &tgt->dep_internal)) {
-				return false;
+	{ // dep internal setup
+		enum build_dep_flag flags = 0;
+		if (get_option_default_both_libraries(wk, 0, 0) == default_both_libraries_auto) {
+			if (tgt->type & tgt_static_library) {
+				flags |= build_dep_flag_both_libs_static;
+				flags |= build_dep_flag_recursive;
+			} else if (tgt->type & tgt_dynamic_library) {
+				flags |= build_dep_flag_both_libs_shared;
+				flags |= build_dep_flag_recursive;
 			}
 		}
 
-		if (akw[bt_kw_link_whole].set) {
-			if (!dep_process_link_whole(
-				    wk, akw[bt_kw_link_whole].node, akw[bt_kw_link_whole].val, &tgt->dep_internal)) {
-				return false;
-			}
+		obj rpath = make_obj(wk, obj_array);
+		if (akw[bt_kw_build_rpath].set) {
+			obj_array_push(wk, rpath, akw[bt_kw_build_rpath].val);
 		}
-	}
 
-	if (akw[bt_kw_dependencies].set) {
-		dep_process_deps(wk, akw[bt_kw_dependencies].val, &tgt->dep_internal);
+		if (akw[bt_kw_install_rpath].set) {
+			obj_array_push(wk, rpath, akw[bt_kw_install_rpath].val);
+		}
+
+		struct build_dep_raw raw = {
+			.link_with = akw[bt_kw_link_with].val,
+			.link_whole = akw[bt_kw_link_whole].val,
+			.deps = akw[bt_kw_dependencies].val,
+			.rpath = rpath,
+		};
+
+		if (!dependency_create(wk, &raw, &tgt->dep_internal, flags)) {
+			return false;
+		}
 	}
 
 	if (!deps_check_machine_matches(wk,
@@ -552,7 +577,7 @@ create_target(struct workspace *wk,
 		return false;
 	}
 
-	if (akw[bt_kw_override_options].set) { // override options
+	if (akw[bt_kw_override_options].set) {
 		if (!parse_and_set_override_options(wk,
 			    akw[bt_kw_override_options].node,
 			    akw[bt_kw_override_options].val,
@@ -591,7 +616,7 @@ create_target(struct workspace *wk,
 			bool pie = false;
 
 			if (akw[bt_kw_pie].set) {
-				pie = get_obj_bool(wk, akw[bt_kw_pie].set);
+				pie = get_obj_bool(wk, akw[bt_kw_pie].val);
 
 				if (pie && (tgt->type & tgt_executable) != tgt_executable) {
 					vm_error_at(wk, akw[bt_kw_pie].node, "pie cannot be set for non-executables");
@@ -617,15 +642,15 @@ create_target(struct workspace *wk,
 		struct args_kw *vis = &akw[bt_kw_gnu_symbol_visibility];
 		if (vis->set && get_str(wk, vis->val)->len) {
 			const struct str *str = get_str(wk, vis->val);
-			if (str_eql(str, &WKSTR("default"))) {
+			if (str_eql(str, &STR("default"))) {
 				tgt->visibility = compiler_visibility_default;
-			} else if (str_eql(str, &WKSTR("hidden"))) {
+			} else if (str_eql(str, &STR("hidden"))) {
 				tgt->visibility = compiler_visibility_hidden;
-			} else if (str_eql(str, &WKSTR("internal"))) {
+			} else if (str_eql(str, &STR("internal"))) {
 				tgt->visibility = compiler_visibility_internal;
-			} else if (str_eql(str, &WKSTR("protected"))) {
+			} else if (str_eql(str, &STR("protected"))) {
 				tgt->visibility = compiler_visibility_protected;
-			} else if (str_eql(str, &WKSTR("inlineshidden"))) {
+			} else if (str_eql(str, &STR("inlineshidden"))) {
 				tgt->visibility = compiler_visibility_inlineshidden;
 			} else {
 				vm_error_at(wk, vis->node, "unknown visibility '%s'", get_cstr(wk, vis->val));
@@ -654,19 +679,20 @@ create_target(struct workspace *wk,
 	}
 
 	{ /* tgt_build_path */
-		SBUF(path);
+		TSTR(path);
 		path_join(wk, &path, get_cstr(wk, tgt->build_dir), get_cstr(wk, tgt->build_name));
 
 		tgt->build_path = make_str(wk, path.buf);
 
-		sbuf_pushs(wk, &path, ".p");
+		tstr_pushs(wk, &path, ".p");
 
-		tgt->private_path = sbuf_into_str(wk, &path);
+		tgt->private_path = tstr_into_str(wk, &path);
 	}
 
 	bool implicit_include_directories = akw[bt_kw_implicit_include_directories].set ?
 						    get_obj_bool(wk, akw[bt_kw_implicit_include_directories].val) :
 						    true;
+	obj prepend_include_directories = make_obj(wk, obj_array);
 
 	{ // sources
 		if (akw[bt_kw_objects].set) {
@@ -678,7 +704,7 @@ create_target(struct workspace *wk,
 			obj_array_dedup(wk, tgt->objects, &deduped);
 			tgt->objects = deduped;
 		} else {
-			make_obj(wk, &tgt->objects, obj_array);
+			tgt->objects = make_obj(wk, obj_array);
 		}
 
 		obj_array_extend(wk, tgt->objects, tgt->dep_internal.objects);
@@ -697,6 +723,7 @@ create_target(struct workspace *wk,
 				.err_node = an[1].node,
 				.res = tgt->src,
 				.tgt_id = *res,
+				.prepend_include_directories = prepend_include_directories,
 				.implicit_include_directories = implicit_include_directories,
 			};
 
@@ -723,22 +750,23 @@ create_target(struct workspace *wk,
 	}
 
 	{ // include directories
-		obj inc_dirs;
-		make_obj(wk, &inc_dirs, obj_array);
 		uint32_t node = an[0].node; // TODO: not a very informative error node
+		obj include_directories = make_obj(wk, obj_array);
 
 		if (implicit_include_directories) {
-			obj_array_push(wk, inc_dirs, current_project(wk)->cwd);
+			obj_array_push(wk, include_directories, current_project(wk)->cwd);
 		}
 
 		if (akw[bt_kw_include_directories].set) {
 			node = akw[bt_kw_include_directories].node;
 
-			obj_array_extend(wk, inc_dirs, akw[bt_kw_include_directories].val);
+			obj_array_extend(wk, include_directories, akw[bt_kw_include_directories].val);
 		}
 
+		obj_array_extend_nodup(wk, include_directories, prepend_include_directories);
+
 		obj coerced;
-		if (!coerce_include_dirs(wk, node, inc_dirs, false, &coerced)) {
+		if (!coerce_include_dirs(wk, node, include_directories, false, &coerced)) {
 			return false;
 		}
 
@@ -788,8 +816,7 @@ create_target(struct workspace *wk,
 			enum compiler_language l;
 		} pch_args[] = {
 #define E(lang, s) { &akw[bt_kw_##lang##s], compiler_language_##lang },
-#define TOOLCHAIN_ENUM(lang) \
-	E(lang, _pch)
+#define TOOLCHAIN_ENUM(lang) E(lang, _pch)
 			FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
 #undef TOOLCHAIN_ENUM
 #undef E
@@ -801,15 +828,39 @@ create_target(struct workspace *wk,
 				continue;
 			}
 
-			obj pch_file;
-			if (!coerce_file(wk, pch_args[i].kw->node, pch_args[i].kw->val, &pch_file)) {
-				return false;
+			obj pch;
+
+			if (get_obj_type(wk, pch_args[i].kw->val) == obj_build_target) {
+				struct obj_build_target *pch_tgt = get_obj_build_target(wk, pch_args[i].kw->val);
+				if ((pch_tgt->type & tgt_static_library) != tgt_static_library ) {
+					vm_error_at(wk, pch_args[i].kw->node, "this build target must be a static library to be used as a pch");
+					return false;
+				}
+
+				obj pch_tgt_pch;
+				if (!obj_dict_geti(wk, pch_tgt->pch, pch_args[i].l, &pch_tgt_pch)) {
+					vm_error_at(wk, pch_args[i].kw->node, "this build target does not define a pch for %s", compiler_language_to_s(pch_args[i].l));
+					return false;
+				}
+
+				if (get_obj_type(wk, pch_tgt_pch) == obj_build_target) {
+					vm_error_at(wk, pch_args[i].kw->node, "cannot use a build target for pch which itself uses a build target for pch");
+					return false;
+				}
+
+				pch = pch_args[i].kw->val;
+			} else {
+				if (!coerce_file(wk, pch_args[i].kw->node, pch_args[i].kw->val, &pch)) {
+					return false;
+				}
 			}
 
 			if (!tgt->pch) {
-				make_obj(wk, &tgt->pch, obj_dict);
+				tgt->pch = make_obj(wk, obj_dict);
 			}
-			obj_dict_seti(wk, tgt->pch, pch_args[i].l, pch_file);
+			obj_dict_seti(wk, tgt->pch, pch_args[i].l, pch);
+
+			build_tgt_inc_required_compiler(wk, tgt, pch_args[i].l);
 		}
 	}
 
@@ -866,16 +917,16 @@ create_target(struct workspace *wk,
 			}
 		}
 
-		SBUF(install_src);
+		TSTR(install_src);
 		path_join(wk, &install_src, get_cstr(wk, tgt->build_dir), get_cstr(wk, tgt->build_name));
 
-		SBUF(install_dest);
+		TSTR(install_dest);
 		path_join(wk, &install_dest, get_cstr(wk, install_dir), get_cstr(wk, tgt->build_name));
 
 		struct obj_install_target *install_tgt;
 		if (!(install_tgt = push_install_target(wk,
-			      sbuf_into_str(wk, &install_src),
-			      sbuf_into_str(wk, &install_dest),
+			      tstr_into_str(wk, &install_src),
+			      tstr_into_str(wk, &install_dest),
 			      akw[bt_kw_install_mode].val))) {
 			return false;
 		}
@@ -896,16 +947,6 @@ create_target(struct workspace *wk,
 		return false;
 	}
 
-	{ // rpaths
-		if (akw[bt_kw_build_rpath].set) {
-			obj_array_push(wk, tgt->dep_internal.rpath, akw[bt_kw_build_rpath].val);
-		}
-
-		if (akw[bt_kw_install_rpath].set) {
-			obj_array_push(wk, tgt->dep_internal.rpath, akw[bt_kw_install_rpath].val);
-		}
-	}
-
 	tgt->dep = (struct build_dep){
 		.link_language = tgt->dep_internal.link_language,
 		.include_directories = tgt->dep_internal.include_directories,
@@ -926,6 +967,15 @@ create_target(struct workspace *wk,
 		obj_array_extend(wk, tgt->dep_internal.link_args, akw[bt_kw_link_args].val);
 	}
 
+	if (tgt->flags & build_tgt_generated_include) {
+		const char *private_path = get_cstr(wk, tgt->private_path);
+
+		// mkdir so that the include dir doesn't get pruned later on
+		if (!fs_mkdir_p(private_path)) {
+			return false;
+		}
+	}
+
 	L("adding build target %s", get_cstr(wk, tgt->build_name));
 	obj_array_push(wk, current_project(wk)->targets, *res);
 	return true;
@@ -951,6 +1001,27 @@ typecheck_string_or_empty_array(struct workspace *wk, struct args_kw *kw)
 	}
 }
 
+static bool
+both_libs_can_reuse_objects(struct workspace *wk, struct args_kw *akw)
+{
+	bool lib_specific_args[] = {
+#define E(lang, s) akw[bt_kw_##lang##s].set
+#define TOOLCHAIN_ENUM(lang) E(lang, _static_args), E(lang, _shared_args),
+		FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+#undef E
+	};
+
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(lib_specific_args); ++i) {
+		if (lib_specific_args[i]) {
+			return false;
+		}
+	}
+
+	return true;
+}
+
 static bool
 tgt_common(struct workspace *wk, obj *res, enum tgt_type type, enum tgt_type argtype, bool tgt_type_from_kw)
 {
@@ -991,12 +1062,12 @@ tgt_common(struct workspace *wk, obj *res, enum tgt_type type, enum tgt_type arg
 		[bt_kw_gui_app] = { "gui_app", obj_bool },
 		[bt_kw_link_language] = { "link_language", obj_string },
 		[bt_kw_win_subsystem] = { "win_subsystem", obj_string },
-		[bt_kw_override_options] = { "override_options", wk->complex_types.options_dict_or_list },
+		[bt_kw_override_options] = { "override_options", COMPLEX_TYPE_PRESET(tc_cx_options_dict_or_list) },
 		[bt_kw_link_args] = { "link_args", TYPE_TAG_LISTIFY | obj_string },
 #define E(lang, s, t) [bt_kw_##lang##s] = { #lang #s, t }
 #define TOOLCHAIN_ENUM(lang)                                                                                 \
 	E(lang, _args, TYPE_TAG_LISTIFY | obj_string), E(lang, _static_args, TYPE_TAG_LISTIFY | obj_string), \
-		E(lang, _shared_args, TYPE_TAG_LISTIFY | obj_string), E(lang, _pch, tc_string | tc_file),
+		E(lang, _shared_args, TYPE_TAG_LISTIFY | obj_string), E(lang, _pch, tc_string | tc_file | tc_build_target),
 		FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
 #undef TOOLCHAIN_ENUM
 #undef E
@@ -1056,30 +1127,36 @@ tgt_common(struct workspace *wk, obj *res, enum tgt_type type, enum tgt_type arg
 			continue;
 		}
 
+		bool ignore_sources = false;
+
 		if (tgt && !multi_target) {
 			multi_target = true;
-			make_obj(wk, res, obj_array);
+			*res = make_obj(wk, obj_array);
 			obj_array_push(wk, *res, tgt);
 
-			// If this target is a multi-target (both_libraries),
-			// set the objects argument with objects from the
-			// previous target
+			if (both_libs_can_reuse_objects(wk, akw)) {
+				// If this target is a multi-target (both_libraries),
+				// set the objects argument with objects from the
+				// previous target
 
-			obj objects;
-			if (!build_target_extract_all_objects(wk, an[0].node, tgt, &objects, true)) {
-				return false;
-			}
+				obj objects;
+				if (!build_target_extract_all_objects(wk, an[0].node, tgt, &objects, true)) {
+					return false;
+				}
 
-			if (akw[bt_kw_objects].set) {
-				obj_array_extend(wk, akw[bt_kw_objects].val, objects);
-			} else {
-				akw[bt_kw_objects].set = true;
-				akw[bt_kw_objects].val = objects;
-				akw[bt_kw_objects].node = an[0].node;
+				if (akw[bt_kw_objects].set) {
+					obj_array_extend(wk, akw[bt_kw_objects].val, objects);
+				} else {
+					akw[bt_kw_objects].set = true;
+					akw[bt_kw_objects].val = objects;
+					akw[bt_kw_objects].node = an[0].node;
+				}
+
+				ignore_sources = true;
 			}
 		}
 
-		if (!create_target(wk, an, akw, t, multi_target, &tgt)) {
+		if (!create_target(wk, an, akw, t, ignore_sources, &tgt)) {
 			return false;
 		}
 
@@ -1092,10 +1169,10 @@ tgt_common(struct workspace *wk, obj *res, enum tgt_type type, enum tgt_type arg
 
 	if (multi_target) {
 		obj val;
-		make_obj(wk, &val, obj_both_libs);
+		val = make_obj(wk, obj_both_libs);
 		struct obj_both_libs *both = get_obj_both_libs(wk, val);
-		obj_array_index(wk, *res, 0, &both->static_lib);
-		obj_array_index(wk, *res, 1, &both->dynamic_lib);
+		both->static_lib = obj_array_index(wk, *res, 0);
+		both->dynamic_lib = obj_array_index(wk, *res, 1);
 		*res = val;
 
 		assert(get_obj_build_target(wk, both->static_lib)->type == tgt_static_library);
