@@ -48,14 +48,6 @@ file_exists_with_content(struct workspace *wk, const char *dest, const char *out
 	return false;
 }
 
-static void
-configure_file_skip_whitespace(const struct source *src, uint32_t *i)
-{
-	while (src->src[*i] && is_whitespace_except_newline(src->src[*i])) {
-		++(*i);
-	}
-}
-
 static uint32_t
 configure_var_len(const char *p)
 {
@@ -74,154 +66,107 @@ configure_var_len(const char *p)
 enum configure_file_syntax {
 	configure_file_syntax_mesondefine = 0 << 0,
 	configure_file_syntax_cmakedefine = 1 << 0,
-	configure_file_syntax_mesonvar = 0 << 1,
-	configure_file_syntax_cmakevar = 1 << 1,
+
+	configure_file_syntax_mesonvar = 1 << 1,
+	configure_file_syntax_cmakevar = 1 << 2,
 };
 
-static bool
-substitute_config(struct workspace *wk,
-	uint32_t dict,
-	uint32_t in_node,
-	const char *in,
-	obj out,
-	enum configure_file_syntax syntax)
+struct configure_file_context {
+	const char *name;
+	obj dict;
+	const struct tstr *in;
+	enum configure_file_syntax syntax;
+};
+
+struct configure_file_var_patterns {
+	struct {
+		struct str start, end;
+		enum configure_file_syntax type;
+	} pats[2];
+	uint32_t len;
+};
+
+static void
+configure_file_var_pattern_push(struct configure_file_var_patterns *var_patterns,
+	enum configure_file_syntax syntax,
+	enum configure_file_syntax enable_flag,
+	const struct str *start,
+	const struct str *end)
 {
-	const char *define;
-	if (syntax & configure_file_syntax_cmakedefine) {
-		define = "#cmakedefine ";
-	} else {
-		define = "#mesondefine ";
+	if (!(syntax & enable_flag)) {
+		return;
 	}
 
-	const char *varstart;
-	char varend;
-	if (syntax & configure_file_syntax_cmakevar) {
-		varstart = "${";
-		varend = '}';
-	} else {
-		varstart = "@";
-		varend = '@';
+	assert(var_patterns->len < ARRAY_LEN(var_patterns->pats));
+
+	var_patterns->pats[var_patterns->len].start = *start;
+	var_patterns->pats[var_patterns->len].end = *end;
+	var_patterns->pats[var_patterns->len].type = enable_flag;
+
+	++var_patterns->len;
+}
+
+MUON_ATTR_FORMAT(printf, 4, 5)
+static void
+configure_file_log(struct configure_file_context *ctx,
+	struct source_location loc,
+	enum log_level lvl,
+	const char *fmt,
+	...)
+{
+	va_list args;
+	va_start(args, fmt);
+
+	struct source src = {
+		.src = ctx->in->buf,
+		.len = ctx->in->len,
+		.label = ctx->name,
+	};
+
+	error_messagev(&src, loc, lvl, fmt, args);
+
+	va_end(args);
+}
+
+static bool
+configure_file_var_patterns_match(struct configure_file_var_patterns *var_patterns,
+	const struct tstr *in,
+	uint32_t off,
+	uint32_t *match_idx)
+{
+	const struct str s = { in->buf + off, in->len - off };
+
+	for (uint32_t i = 0; i < var_patterns->len; ++i) {
+		if (str_startswith(&s, &var_patterns->pats[i].start)) {
+			*match_idx = i;
+			return true;
+		}
 	}
-	const uint32_t define_len = strlen(define), varstart_len = strlen(varstart);
+	return false;
+}
 
-	bool ret = true;
-	struct source src;
+static bool
+substitute_config_variables(struct workspace *wk, struct configure_file_context *ctx, struct tstr *out)
+{
+	const struct tstr *in = ctx->in;
 
-	if (!fs_read_entire_file(in, &src)) {
-		ret = false;
-		goto cleanup;
-	}
-
-	SBUF_manual(out_buf);
+	struct configure_file_var_patterns var_patterns = { 0 };
+	configure_file_var_pattern_push(
+		&var_patterns, ctx->syntax, configure_file_syntax_cmakevar, &STR("${"), &STR("}"));
+	configure_file_var_pattern_push(
+		&var_patterns, ctx->syntax, configure_file_syntax_mesonvar, &STR("@"), &STR("@"));
 
 	struct source_location location = { 0, 1 }, id_location;
-	uint32_t i, id_start, id_len, col = 0, orig_i;
+	uint32_t i, id_start;
+	uint32_t match_idx = 0;
 	obj elem;
-	char tmp_buf[BUF_SIZE_1k] = { 0 };
 
-	for (i = 0; i < src.len; ++i) {
+	for (i = 0; i < in->len; ++i) {
 		location.off = i;
-		if (src.src[i] == '\n') {
-			col = i + 1;
-		}
 
-		if (i == col && strncmp(&src.src[i], define, define_len) == 0) {
-			i += define_len;
+		struct str src_str = { in->buf + i, in->len - i };
 
-			configure_file_skip_whitespace(&src, &i);
-
-			id_start = i;
-			id_location = location;
-			++id_location.off;
-			id_len = configure_var_len(&src.src[id_start]);
-			i += id_len;
-
-			const char *sub = NULL, *deftype = "#define";
-
-			configure_file_skip_whitespace(&src, &i);
-
-			if (!(src.src[i] == '\n' || src.src[i] == 0)) {
-				if (syntax & configure_file_syntax_cmakedefine) {
-					if (src.src[i] == '@'
-						&& strncmp(&src.src[i + 1], &src.src[id_start], id_len) == 0
-						&& src.src[i + 1 + id_len] == '@') {
-						i += 2 + id_len;
-						configure_file_skip_whitespace(&src, &i);
-
-						if (!(src.src[i] == '\n' || src.src[i] == 0)) {
-							goto extraneous_cmake_chars;
-						}
-					} else {
-extraneous_cmake_chars:
-						orig_i = i;
-
-						while (src.src[i] && src.src[i] != '\n') {
-							++i;
-						}
-
-						/* id_location.col = orig_i - location.col + 1; */
-						error_messagef(&src,
-							id_location,
-							log_warn,
-							"ignoring trailing characters (%.*s) in cmakedefine",
-							i - orig_i,
-							&src.src[orig_i]);
-					}
-				} else {
-					/* id_location.col = i - location.col + 1; */
-					error_messagef(&src,
-						id_location,
-						log_error,
-						"expected exactly one token on mesondefine line");
-					return false;
-				}
-			}
-
-			if (i == id_start) {
-				error_messagef(&src, id_location, log_error, "key of zero length not supported");
-				return false;
-			} else if (!obj_dict_index_strn(wk, dict, &src.src[id_start], id_len, &elem)) {
-				deftype = "/* undef";
-				sub = "*/";
-				goto write_mesondefine;
-			}
-
-			switch (get_obj_type(wk, elem)) {
-			case obj_bool: {
-				if (!get_obj_bool(wk, elem)) {
-					deftype = "#undef";
-				}
-				break;
-			}
-			case obj_string: {
-				sub = get_cstr(wk, elem);
-				break;
-			}
-			case obj_number:
-				snprintf(tmp_buf, BUF_SIZE_1k, "%" PRId64, get_obj_number(wk, elem));
-				sub = tmp_buf;
-				break;
-			default:
-				error_messagef(&src,
-					id_location,
-					log_error,
-					"invalid type for %s: '%s'",
-					define,
-					obj_type_to_s(get_obj_type(wk, elem)));
-				return false;
-			}
-
-write_mesondefine:
-			sbuf_pushn(wk, &out_buf, deftype, strlen(deftype));
-			sbuf_pushn(wk, &out_buf, " ", 1);
-			sbuf_pushn(wk, &out_buf, &src.src[id_start], id_len);
-			if (sub) {
-				sbuf_pushn(wk, &out_buf, " ", 1);
-				sbuf_pushn(wk, &out_buf, sub, strlen(sub));
-			}
-
-			i -= 1; // so we catch the newline
-		} else if (src.src[i] == '\\') {
+		if (in->buf[i] == '\\') {
 			/* cope with weird config file escaping rules :(
 			 *
 			 * - Backslashes not directly preceeding a format character are not
@@ -236,13 +181,13 @@ write_mesondefine:
 			uint32_t j, output_backslashes;
 			bool output_format_char = false;
 
-			for (j = 1; src.src[i + j] && src.src[i + j] == '\\'; ++j) {
+			for (j = 1; in->buf[i + j] == '\\'; ++j) {
 			}
 
-			if (strncmp(&src.src[i + j], varstart, varstart_len) == 0) {
+			if (configure_file_var_patterns_match(&var_patterns, in, i + j, &match_idx)) {
 				output_backslashes = j / 2;
 
-				if (*varstart == '@') {
+				if (var_patterns.pats[match_idx].type == configure_file_syntax_mesonvar) {
 					output_format_char = true;
 					i += j;
 				} else {
@@ -260,64 +205,236 @@ write_mesondefine:
 			}
 
 			for (j = 0; j < output_backslashes; ++j) {
-				sbuf_pushn(wk, &out_buf, "\\", 1);
+				tstr_pushn(wk, out, "\\", 1);
 			}
 
 			if (output_format_char) {
-				sbuf_pushn(wk, &out_buf, varstart, varstart_len);
-				i += varstart_len - 1;
+				tstr_pushs(wk, out, var_patterns.pats[match_idx].start.s);
+				i += var_patterns.pats[match_idx].start.len - 1;
 			}
-		} else if (strncmp(&src.src[i], varstart, varstart_len) == 0) {
-			i += varstart_len;
+		} else if (configure_file_var_patterns_match(&var_patterns, in, i, &match_idx)) {
+			i += var_patterns.pats[match_idx].start.len;
 			id_start = i;
 			id_location = location;
-			i += configure_var_len(&src.src[id_start]);
+			i += configure_var_len(&in->buf[id_start]);
 
-			if (src.src[i] != varend) {
+			src_str = (struct str){ in->buf + i, in->len - i };
+			if (!str_startswith(&src_str, &var_patterns.pats[match_idx].end)) {
 				i = id_start - 1;
-				sbuf_pushn(wk, &out_buf, varstart, varstart_len);
+				tstr_pushs(wk, out, var_patterns.pats[match_idx].start.s);
 				continue;
 			}
 
 			if (i <= id_start) {
 				// This means we got a key of length zero
-				sbuf_pushs(wk, &out_buf, "@@");
+				tstr_pushs(wk, out, "@@");
 				continue;
-			} else if (!obj_dict_index_strn(wk, dict, &src.src[id_start], i - id_start, &elem)) {
-				error_messagef(&src, id_location, log_error, "key not found in configuration data");
+			} else if (!obj_dict_index_strn(wk, ctx->dict, &in->buf[id_start], i - id_start, &elem)) {
+				configure_file_log(ctx,
+					id_location,
+					log_error,
+					"key %.*s not found in configuration data",
+					i - id_start,
+					&in->buf[id_start]);
 				return false;
 			}
 
 			obj sub;
-			if (!coerce_string(wk, in_node, elem, &sub)) {
-				error_messagef(&src, id_location, log_error, "unable to substitute value");
+			if (!coerce_string(wk, 0, elem, &sub)) {
+				configure_file_log(ctx, id_location, log_error, "unable to substitute value");
 				return false;
 			}
 
 			const struct str *ss = get_str(wk, sub);
-			sbuf_pushn(wk, &out_buf, ss->s, ss->len);
+			tstr_pushn(wk, out, ss->s, ss->len);
 		} else {
-			sbuf_pushn(wk, &out_buf, &src.src[i], 1);
+			tstr_pushn(wk, out, &in->buf[i], 1);
 		}
 	}
 
-	if (file_exists_with_content(wk, get_cstr(wk, out), out_buf.buf, out_buf.len)) {
-		goto cleanup;
+	return true;
+}
+
+static bool
+substitute_config_defines(struct workspace *wk, struct configure_file_context *ctx, struct tstr *out)
+{
+	struct str define = { 0 };
+	if (ctx->syntax & configure_file_syntax_cmakedefine) {
+		define = STR("#cmakedefine");
+	} else {
+		define = STR("#mesondefine");
 	}
 
-	if (!fs_write(get_cstr(wk, out), (uint8_t *)out_buf.buf, out_buf.len)) {
+	struct source_location location = { 0, 1 };
+	obj elem;
+
+	const char *e, *s = ctx->in->buf, *eof = s + ctx->in->len;
+
+	while (s < eof) {
+		if (!(e = strchr(s, '\n'))) {
+			e = eof;
+		}
+
+		location.off = s - ctx->in->buf;
+
+		struct str line = { s, e - s };
+
+		if (str_startswith(&line, &define)) {
+			obj arr = str_split(wk, &line, 0);
+
+			if (ctx->syntax & configure_file_syntax_cmakedefine) {
+				bool cmake_bool = str_startswith(&line, &STR("#cmakedefine01"));
+
+				if (get_obj_array(wk, arr)->len < 2) {
+					configure_file_log(
+						ctx, location, log_error, "#cmakedefine does not contain >= 2 tokens");
+					return false;
+				}
+
+				obj varname = obj_array_index(wk, arr, 1);
+				if (!obj_dict_index(wk, ctx->dict, varname, &elem)) {
+					if (cmake_bool) {
+						tstr_pushf(wk, out, "#define %s 0\n", get_str(wk, varname)->s);
+					} else {
+						tstr_pushf(wk, out, "/* #undef %s */\n", get_str(wk, varname)->s);
+					}
+				} else {
+					if (!cmake_bool && !coerce_truthiness(wk, elem)) {
+						tstr_pushf(wk, out, "/* #undef %s */\n", get_str(wk, varname)->s);
+					} else {
+						tstr_pushf(wk, out, "#define %s", get_str(wk, varname)->s);
+
+						obj v;
+						obj_array_for_(wk, arr, v, iter)
+						{
+							if (iter.i < 2) {
+								continue;
+							}
+
+							if (obj_dict_index(wk, ctx->dict, v, &elem)) {
+								obj str;
+								if (!coerce_string(wk, 0, v, &str)) {
+									return false;
+								}
+								tstr_pushf(wk, out, " %s", get_cstr(wk, str));
+							} else {
+								tstr_pushf(wk, out, " %s", get_str(wk, v)->s);
+							}
+						}
+
+						tstr_push(wk, out, '\n');
+					}
+				}
+			} else {
+				if (get_obj_array(wk, arr)->len != 2) {
+					configure_file_log(ctx,
+						location,
+						log_error,
+						"#mesondefine does not contain exactly two tokens");
+					return false;
+				}
+
+				obj varname = obj_array_index(wk, arr, 1);
+				if (!obj_dict_index(wk, ctx->dict, varname, &elem)) {
+					tstr_pushf(wk, out, "/* #undef %s */\n", get_str(wk, varname)->s);
+				} else {
+					if (!typecheck(wk, 0, elem, tc_string | tc_bool | tc_number)) {
+						return false;
+					}
+
+					switch (get_obj_type(wk, elem)) {
+					case obj_string: {
+						tstr_pushf(wk,
+							out,
+							"#define %s %s\n",
+							get_str(wk, varname)->s,
+							get_cstr(wk, elem));
+						break;
+					}
+					case obj_bool: {
+						if (get_obj_bool(wk, elem)) {
+							tstr_pushf(wk, out, "#define %s\n", get_str(wk, varname)->s);
+						} else {
+							tstr_pushf(wk, out, "#undef %s\n", get_str(wk, varname)->s);
+						}
+						break;
+					}
+					case obj_number: {
+						tstr_pushf(wk,
+							out,
+							"#define %s %" PRId64 "\n",
+							get_str(wk, varname)->s,
+							get_obj_number(wk, elem));
+						break;
+					}
+					default: UNREACHABLE;
+					}
+				}
+			}
+		} else {
+			tstr_pushn(wk, out, line.s, line.len);
+			tstr_push(wk, out, '\n');
+		}
+
+		s = e + 1;
+	}
+
+	return true;
+}
+
+static bool
+substitute_config(struct workspace *wk,
+	obj dict,
+	const char *input_path,
+	const char *output_path,
+	enum configure_file_syntax syntax)
+{
+	bool ret = true;
+	struct source src = { 0 };
+
+	if (!fs_read_entire_file(input_path, &src)) {
 		ret = false;
 		goto cleanup;
 	}
 
-	if (!fs_copy_metadata(in, get_cstr(wk, out))) {
+	const struct tstr src_tstr = { (char *)src.src, src.len };
+
+	struct configure_file_context ctx = {
+		.name = input_path,
+		.dict = dict,
+		.in = &src_tstr,
+		.syntax = syntax,
+	};
+
+	TSTR(out1);
+	if (!substitute_config_defines(wk, &ctx, &out1)) {
+		ret = false;
+		goto cleanup;
+	}
+
+	ctx.in = &out1;
+	TSTR(out2);
+	if (!substitute_config_variables(wk, &ctx, &out2)) {
+		ret = false;
+		goto cleanup;
+	}
+
+	if (file_exists_with_content(wk, output_path, out2.buf, out2.len)) {
+		goto cleanup;
+	}
+
+	if (!fs_write(output_path, (uint8_t *)out2.buf, out2.len)) {
+		ret = false;
+		goto cleanup;
+	}
+
+	if (!fs_copy_metadata(input_path, output_path)) {
 		ret = false;
 		goto cleanup;
 	}
 
 cleanup:
 	fs_source_destroy(&src);
-	sbuf_destroy(&out_buf);
 	return ret;
 }
 
@@ -329,13 +446,13 @@ generate_config(struct workspace *wk,
 	uint32_t node,
 	obj out_path)
 {
-	SBUF_manual(out_buf);
+	TSTR_manual(out_buf);
 
 	if (macro_name) {
-		sbuf_pushf(
+		tstr_pushf(
 			wk, &out_buf, "#ifndef %s\n#define %s\n", get_cstr(wk, macro_name), get_cstr(wk, macro_name));
 	} else if (format == configure_file_output_format_json) {
-		sbuf_push(wk, &out_buf, '{');
+		tstr_push(wk, &out_buf, '{');
 	}
 
 	bool ret = false, first = true;
@@ -352,12 +469,12 @@ generate_config(struct workspace *wk,
 
 		if (format == configure_file_output_format_json) {
 			if (!first) {
-				sbuf_push(wk, &out_buf, ',');
+				tstr_push(wk, &out_buf, ',');
 			}
 			first = false;
 
-			sbuf_push_json_escaped_quoted(wk, &out_buf, get_str(wk, key));
-			sbuf_push(wk, &out_buf, ':');
+			tstr_push_json_escaped_quoted(wk, &out_buf, get_str(wk, key));
+			tstr_push(wk, &out_buf, ':');
 		}
 
 		switch (t) {
@@ -367,7 +484,7 @@ generate_config(struct workspace *wk,
 			switch (format) {
 			case configure_file_output_format_c:
 			case configure_file_output_format_nasm:
-				sbuf_pushf(wk,
+				tstr_pushf(wk,
 					&out_buf,
 					"%cdefine %s %s\n",
 					define_prefix,
@@ -375,7 +492,7 @@ generate_config(struct workspace *wk,
 					get_cstr(wk, val));
 				break;
 			case configure_file_output_format_json: {
-				sbuf_push_json_escaped_quoted(wk, &out_buf, get_str(wk, val));
+				tstr_push_json_escaped_quoted(wk, &out_buf, get_str(wk, val));
 				break;
 			}
 			}
@@ -387,13 +504,13 @@ generate_config(struct workspace *wk,
 			case configure_file_output_format_c:
 			case configure_file_output_format_nasm:
 				if (get_obj_bool(wk, val)) {
-					sbuf_pushf(wk, &out_buf, "%cdefine %s\n", define_prefix, get_cstr(wk, key));
+					tstr_pushf(wk, &out_buf, "%cdefine %s\n", define_prefix, get_cstr(wk, key));
 				} else {
-					sbuf_pushf(wk, &out_buf, "%cundef %s\n", define_prefix, get_cstr(wk, key));
+					tstr_pushf(wk, &out_buf, "%cundef %s\n", define_prefix, get_cstr(wk, key));
 				}
 				break;
 			case configure_file_output_format_json:
-				sbuf_pushs(wk, &out_buf, get_obj_bool(wk, val) ? "true" : "false");
+				tstr_pushs(wk, &out_buf, get_obj_bool(wk, val) ? "true" : "false");
 				break;
 			}
 			break;
@@ -403,7 +520,7 @@ generate_config(struct workspace *wk,
 			switch (format) {
 			case configure_file_output_format_c:
 			case configure_file_output_format_nasm:
-				sbuf_pushf(wk,
+				tstr_pushf(wk,
 					&out_buf,
 					"%cdefine %s %" PRId64 "\n",
 					define_prefix,
@@ -413,7 +530,7 @@ generate_config(struct workspace *wk,
 			case configure_file_output_format_json: {
 				char buf[32] = { 0 };
 				snprintf(buf, sizeof(buf), "%" PRId64, get_obj_number(wk, val));
-				sbuf_pushs(wk, &out_buf, buf);
+				tstr_pushs(wk, &out_buf, buf);
 				break;
 			}
 			}
@@ -423,9 +540,9 @@ generate_config(struct workspace *wk,
 	}
 
 	if (macro_name) {
-		sbuf_pushf(wk, &out_buf, "#endif\n");
+		tstr_pushf(wk, &out_buf, "#endif\n");
 	} else if (format == configure_file_output_format_json) {
-		sbuf_pushs(wk, &out_buf, "}\n");
+		tstr_pushs(wk, &out_buf, "}\n");
 	}
 
 	if (!file_exists_with_content(wk, get_cstr(wk, out_path), out_buf.buf, out_buf.len)) {
@@ -436,7 +553,7 @@ generate_config(struct workspace *wk,
 
 	ret = true;
 ret:
-	sbuf_destroy(&out_buf);
+	tstr_destroy(&out_buf);
 	return ret;
 }
 
@@ -453,9 +570,9 @@ configure_file_with_command(struct workspace *wk,
 
 	{
 		obj f;
-		make_obj(wk, &f, obj_file);
+		f = make_obj(wk, obj_file);
 		*get_obj_file(wk, f) = out_path;
-		make_obj(wk, &output_arr, obj_array);
+		output_arr = make_obj(wk, obj_array);
 		obj_array_push(wk, output_arr, f);
 	}
 
@@ -463,7 +580,7 @@ configure_file_with_command(struct workspace *wk,
 		// XXX: depfile for configure_file is not supported, this is
 		// only here to make the types align
 		obj f;
-		make_obj(wk, &f, obj_file);
+		f = make_obj(wk, obj_file);
 		*get_obj_file(wk, f) = depfile;
 		depfile = f;
 	}
@@ -474,7 +591,7 @@ configure_file_with_command(struct workspace *wk,
 		.output = output_arr,
 		.depfile = depfile,
 	};
-	make_obj(wk, &opts.depends, obj_array);
+	opts.depends = make_obj(wk, obj_array);
 
 	if (!process_custom_target_commandline(wk, &opts, command, &args)) {
 		return false;
@@ -485,12 +602,12 @@ configure_file_with_command(struct workspace *wk,
 	const char *argstr, *envstr;
 	uint32_t argc, envc;
 
-	if (!path_chdir(get_cstr(wk, current_project(wk)->build_dir))) {
+	if (!path_chdir(workspace_build_dir(wk))) {
 		return false;
 	}
 
 	obj env;
-	make_obj(wk, &env, obj_dict);
+	env = make_obj(wk, obj_dict);
 	set_default_environment_vars(wk, env, true);
 
 	join_args_argstr(wk, &argstr, &argc, args);
@@ -535,7 +652,7 @@ array_to_elem_or_err(struct workspace *wk, uint32_t node, uint32_t arr, uint32_t
 		return false;
 	}
 
-	obj_array_index(wk, arr, 0, res);
+	*res = obj_array_index(wk, arr, 0);
 	return true;
 }
 
@@ -560,7 +677,7 @@ perform_output_string_substitutions(struct workspace *wk, uint32_t node, uint32_
 			}
 			assert(e);
 
-			SBUF(buf);
+			TSTR(buf);
 			char *c;
 			path_basename(wk, &buf, get_file_path(wk, e));
 
@@ -576,7 +693,7 @@ perform_output_string_substitutions(struct workspace *wk, uint32_t node, uint32_
 				return false;
 			}
 
-			SBUF(buf);
+			TSTR(buf);
 			path_basename(wk, &buf, get_file_path(wk, e));
 			str_app(wk, &str, buf.buf);
 			s += len - 1;
@@ -654,11 +771,11 @@ func_configure_file(struct workspace *wk, obj _, obj *res)
 	enum configure_file_output_format output_format = configure_file_output_format_c;
 	if (akw[kw_output_format].set) {
 		const struct str *output_format_str = get_str(wk, akw[kw_output_format].val);
-		if (str_eql(output_format_str, &WKSTR("c"))) {
+		if (str_eql(output_format_str, &STR("c"))) {
 			output_format = configure_file_output_format_c;
-		} else if (str_eql(output_format_str, &WKSTR("nasm"))) {
+		} else if (str_eql(output_format_str, &STR("nasm"))) {
 			output_format = configure_file_output_format_nasm;
-		} else if (str_eql(output_format_str, &WKSTR("json"))) {
+		} else if (str_eql(output_format_str, &STR("json"))) {
 			output_format = configure_file_output_format_json;
 		} else {
 			vm_error_at(
@@ -681,22 +798,33 @@ func_configure_file(struct workspace *wk, obj _, obj *res)
 		}
 
 		const char *out = get_cstr(wk, subd);
-		SBUF(out_path);
+		TSTR(out_path);
 
 		if (!path_is_basename(out)) {
-			vm_error_at(wk, akw[kw_output].node, "config file output '%s' contains path separator", out);
+			if (wk->vm.lang_mode == language_external) {
+				vm_error_at(wk,
+					akw[kw_output].node,
+					"config file output '%s' contains path separator",
+					out);
+				return false;
+			}
+
+			TSTR(dir);
+			path_dirname(wk, &dir, out);
+			if (!fs_mkdir_p(dir.buf)) {
+				return false;
+			}
+		}
+
+		if (!fs_mkdir_p(workspace_build_dir(wk))) {
 			return false;
 		}
 
-		if (!fs_mkdir_p(get_cstr(wk, current_project(wk)->build_dir))) {
-			return false;
-		}
-
-		path_join(wk, &out_path, get_cstr(wk, current_project(wk)->build_dir), out);
+		path_join(wk, &out_path, workspace_build_dir(wk), out);
 
 		LOG_I("configuring '%s'", out_path.buf);
-		output_str = sbuf_into_str(wk, &out_path);
-		make_obj(wk, res, obj_file);
+		output_str = tstr_into_str(wk, &out_path);
+		*res = make_obj(wk, obj_file);
 		*get_obj_file(wk, *res) = output_str;
 	}
 
@@ -782,11 +910,12 @@ copy_err:
 
 			if (akw[kw_format].set) {
 				const struct str *fmt = get_str(wk, akw[kw_format].val);
-				if (str_eql(fmt, &WKSTR("meson"))) {
+				if (str_eql(fmt, &STR("meson"))) {
 					syntax = configure_file_syntax_mesondefine | configure_file_syntax_mesonvar;
-				} else if (str_eql(fmt, &WKSTR("cmake"))) {
-					syntax = configure_file_syntax_cmakedefine | configure_file_syntax_cmakevar;
-				} else if (str_eql(fmt, &WKSTR("cmake@"))) {
+				} else if (str_eql(fmt, &STR("cmake"))) {
+					syntax = configure_file_syntax_cmakedefine | configure_file_syntax_cmakevar
+						 | configure_file_syntax_mesonvar;
+				} else if (str_eql(fmt, &STR("cmake@"))) {
 					syntax = configure_file_syntax_cmakedefine | configure_file_syntax_mesonvar;
 				} else {
 					vm_error_at(
@@ -795,7 +924,7 @@ copy_err:
 				}
 			}
 
-			if (!substitute_config(wk, dict, akw[kw_input].node, path, output_str, syntax)) {
+			if (!substitute_config(wk, dict, path, get_cstr(wk, output_str), syntax)) {
 				return false;
 			}
 		} else {
