@@ -15,6 +15,7 @@
 #include "coerce.h"
 #include "error.h"
 #include "functions/generator.h"
+#include "functions/kernel.h"
 #include "functions/kernel/custom_target.h"
 #include "functions/string.h"
 #include "install.h"
@@ -35,7 +36,7 @@ static bool
 prefix_plus_index(const struct str *ss, const char *prefix, int64_t *index)
 {
 	uint32_t len = strlen(prefix);
-	if (str_startswith(ss, &WKSTR(prefix))) {
+	if (str_startswith(ss, &STRL(prefix))) {
 		return str_to_i(&(struct str){ .s = &ss->s[len], .len = ss->len - len }, index, false);
 	}
 
@@ -45,7 +46,7 @@ prefix_plus_index(const struct str *ss, const char *prefix, int64_t *index)
 static void
 str_relative_to_build_root(struct workspace *wk, struct custom_target_cmd_fmt_ctx *ctx, const char *path_orig, obj *res)
 {
-	SBUF(rel);
+	TSTR(rel);
 	const char *path = path_orig;
 
 	if (!ctx->opts->relativize) {
@@ -67,11 +68,11 @@ str_relative_to_build_root(struct workspace *wk, struct custom_target_cmd_fmt_ct
 		// path elements, and will be assumed to be on PATH, which
 		// either results in the wrong executable being run, or a
 		// command not found error.
-		SBUF(exe);
+		TSTR(exe);
 		path_executable(wk, &exe, rel.buf);
-		*res = sbuf_into_str(wk, &exe);
+		*res = tstr_into_str(wk, &exe);
 	} else {
-		*res = sbuf_into_str(wk, &rel);
+		*res = tstr_into_str(wk, &rel);
 	}
 }
 
@@ -115,7 +116,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 
 	enum cmd_arg_fmt_key key;
 	for (key = 0; key < cmd_arg_fmt_key_count; ++key) {
-		if (!str_eql(strkey, &WKSTR(key_names[key].key))) {
+		if (!str_eql(strkey, &STRL(key_names[key].key))) {
 			continue;
 		}
 
@@ -137,7 +138,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		if (!boundscheck(wk, ctx->opts->err_node, get_obj_array(wk, arr)->len, &index)) {
 			return format_cb_error;
 		}
-		obj_array_index(wk, arr, 0, &e);
+		e = obj_array_index(wk, arr, 0);
 
 		str_relative_to_build_root(wk, ctx, get_file_path(wk, e), elem);
 		return format_cb_found;
@@ -157,9 +158,9 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 	case key_private_dir: {
 		/* @PRIVATE_DIR@ (since 0.50.1): path to a directory where the
 		 * custom target must store all its intermediate files. */
-		SBUF(path);
+		TSTR(path);
 		path_join(wk, &path, get_cstr(wk, current_project(wk)->build_dir), get_cstr(wk, ctx->opts->name));
-		sbuf_pushs(wk, &path, ".p");
+		tstr_pushs(wk, &path, ".p");
 
 		if (!fs_mkdir_p(path.buf)) {
 			return format_cb_error;
@@ -202,15 +203,14 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 			return format_cb_error;
 		}
 
-		obj in0;
-		obj_array_index(wk, ctx->opts->input, 0, &in0);
+		obj in0 = obj_array_index(wk, ctx->opts->input, 0);
 		const struct str *orig_str = get_str(wk, *get_obj_file(wk, in0));
 
-		SBUF(plainname);
+		TSTR(plainname);
 		path_basename(wk, &plainname, orig_str->s);
 
 		if (key == key_basename) {
-			SBUF(basename);
+			TSTR(basename);
 			path_without_ext(wk, &basename, plainname.buf);
 
 			str_relative_to_build_root(wk, ctx, basename.buf, elem);
@@ -244,7 +244,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		return format_cb_error;
 	}
 
-	obj_array_index(wk, arr, index, &e);
+	e = obj_array_index(wk, arr, index);
 
 	str_relative_to_build_root(wk, ctx, get_file_path(wk, e), elem);
 	return format_cb_found;
@@ -257,6 +257,7 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 
 	obj ss;
 	enum obj_type t = get_obj_type(wk, val);
+	obj extra_args = 0;
 
 	switch (t) {
 	case obj_both_libs:
@@ -264,8 +265,8 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 	case obj_external_program:
 	case obj_python_installation:
 	case obj_file: {
-		obj str, args;
-		if (!coerce_executable(wk, ctx->opts->err_node, val, &str, &args)) {
+		obj str;
+		if (!coerce_executable(wk, ctx->opts->err_node, val, &str, &extra_args)) {
 			return ir_err;
 		}
 
@@ -274,31 +275,24 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 		if (!ctx->skip_depends) {
 			obj_array_push(wk, ctx->opts->depends, ss);
 		}
-
-		if (args) {
-			obj_array_push(wk, *ctx->res, ss);
-			obj_array_extend_nodup(wk, *ctx->res, args);
-			return ir_cont;
-		}
-
 		break;
 	}
 	case obj_string: {
-		if (ctx->opts->input && str_eql(get_str(wk, val), &WKSTR("@INPUT@"))) {
+		if (ctx->opts->input && str_eql(get_str(wk, val), &STR("@INPUT@"))) {
 			ctx->skip_depends = true;
 			if (!obj_array_foreach(wk, ctx->opts->input, ctx, custom_target_cmd_fmt_iter)) {
 				return ir_err;
 			}
 			ctx->skip_depends = false;
 			return ir_cont;
-		} else if (ctx->opts->output && str_eql(get_str(wk, val), &WKSTR("@OUTPUT@"))) {
+		} else if (ctx->opts->output && str_eql(get_str(wk, val), &STR("@OUTPUT@"))) {
 			ctx->skip_depends = true;
 			if (!obj_array_foreach(wk, ctx->opts->output, ctx, custom_target_cmd_fmt_iter)) {
 				return ir_err;
 			}
 			ctx->skip_depends = false;
 			goto cont;
-		} else if (ctx->opts->extra_args_valid && str_eql(get_str(wk, val), &WKSTR("@EXTRA_ARGS@"))) {
+		} else if (ctx->opts->extra_args_valid && str_eql(get_str(wk, val), &STR("@EXTRA_ARGS@"))) {
 			if (ctx->opts->extra_args) {
 				obj_array_extend(wk, *ctx->res, ctx->opts->extra_args);
 				ctx->opts->extra_args_used = true;
@@ -311,6 +305,25 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 			return ir_err;
 		}
 		ss = s;
+
+		if (ctx->i == 0 && !path_is_absolute(get_str(wk, ss)->s)) {
+			obj argv0 = 0;
+			struct find_program_ctx find_program_ctx = {
+				.res = &argv0,
+				.requirement = requirement_required,
+				.machine = coerce_machine_kind(wk, 0),
+			};
+			if (!find_program(wk, &find_program_ctx, ss)) {
+				return false;
+			}
+
+			if (!find_program_ctx.found) {
+				vm_error_at(wk, ctx->opts->err_node, "program %o not found", ss);
+				return false;
+			}
+
+			return custom_target_cmd_fmt_iter(wk, ctx, argv0);
+		}
 		break;
 	}
 	case obj_custom_target: {
@@ -335,6 +348,10 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 	assert(get_obj_type(wk, ss) == obj_string);
 
 	obj_array_push(wk, *ctx->res, ss);
+
+	if (extra_args) {
+		obj_array_extend_nodup(wk, *ctx->res, extra_args);
+	}
 cont:
 	++ctx->i;
 	return ir_cont;
@@ -346,7 +363,7 @@ process_custom_target_commandline(struct workspace *wk,
 	obj arr,
 	obj *res)
 {
-	make_obj(wk, res, obj_array);
+	*res = make_obj(wk, obj_array);
 
 	struct custom_target_cmd_fmt_ctx ctx = {
 		.opts = opts,
@@ -361,6 +378,7 @@ process_custom_target_commandline(struct workspace *wk,
 		vm_error_at(wk, opts->err_node, "cmd cannot be empty");
 		return false;
 	}
+
 	return true;
 }
 
@@ -378,7 +396,7 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 
 	enum cmd_output_fmt_key key;
 	for (key = 0; key < cmd_output_fmt_key_count; ++key) {
-		if (str_eql(strkey, &WKSTR(key_names[key]))) {
+		if (str_eql(strkey, &STRL(key_names[key]))) {
 			break;
 		}
 	}
@@ -397,14 +415,14 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 	}
 
 	obj in0;
-	obj_array_index(wk, ctx->opts->input, 0, &in0);
+	in0 = obj_array_index(wk, ctx->opts->input, 0);
 	const struct str *ss = get_str(wk, *get_obj_file(wk, in0));
-	SBUF(buf);
+	TSTR(buf);
 
 	switch (key) {
 	case key_plainname: path_basename(wk, &buf, ss->s); break;
 	case key_basename: {
-		SBUF(basename);
+		TSTR(basename);
 		path_basename(wk, &basename, ss->s);
 		path_without_ext(wk, &buf, basename.buf);
 		break;
@@ -412,7 +430,7 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 	default: assert(false && "unreachable"); return format_cb_error;
 	}
 
-	*elem = sbuf_into_str(wk, &buf);
+	*elem = tstr_into_str(wk, &buf);
 	return format_cb_found;
 }
 
@@ -429,7 +447,7 @@ custom_command_output_format_iter(struct workspace *wk, void *_ctx, obj v)
 	}
 
 	obj f;
-	make_obj(wk, &f, obj_file);
+	f = make_obj(wk, obj_file);
 	*get_obj_file(wk, f) = s;
 
 	obj_array_push(wk, ctx->opts->output, f);
@@ -472,21 +490,21 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 {
 	obj input, raw_output, output, args;
 
-	make_obj(wk, res, obj_custom_target);
+	*res = make_obj(wk, obj_custom_target);
 	struct obj_custom_target *tgt = get_obj_custom_target(wk, *res);
 	tgt->name = opts->name;
 	tgt->callstack = vm_callstack(wk);
 
 	// A custom_target won't have a name if it is from a generator
 	if (opts->name) { /* private path */
-		SBUF(path);
+		TSTR(path);
 		path_join(wk, &path, get_cstr(wk, current_project(wk)->build_dir), get_cstr(wk, opts->name));
-		sbuf_pushs(wk, &path, ".p");
-		tgt->private_path = sbuf_into_str(wk, &path);
+		tstr_pushs(wk, &path, ".p");
+		tgt->private_path = tstr_into_str(wk, &path);
 	}
 
 	if (opts->input_orig) {
-		make_obj(wk, &input, obj_array);
+		input = make_obj(wk, obj_array);
 
 		struct process_custom_tgt_sources_ctx ctx = {
 			.err_node = opts->input_node,
@@ -496,7 +514,7 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 
 		if (get_obj_type(wk, opts->input_orig) != obj_array) {
 			obj arr_input;
-			make_obj(wk, &arr_input, obj_array);
+			arr_input = make_obj(wk, obj_array);
 			obj_array_push(wk, arr_input, opts->input_orig);
 			opts->input_orig = arr_input;
 		}
@@ -516,7 +534,7 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 			return false;
 		}
 
-		make_obj(wk, &output, obj_array);
+		output = make_obj(wk, obj_array);
 		struct custom_target_cmd_fmt_ctx ctx = {
 			.opts = &(struct process_custom_target_commandline_opts) {
 				.err_node = opts->output_node,
@@ -568,7 +586,7 @@ make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, o
 		.extra_args = opts->extra_args,
 		.extra_args_valid = opts->extra_args_valid,
 	};
-	make_obj(wk, &cmdline_opts.depends, obj_array);
+	cmdline_opts.depends = make_obj(wk, obj_array);
 	if (!process_custom_target_commandline(wk, &cmdline_opts, opts->command_orig, &args)) {
 		return false;
 	}
@@ -619,7 +637,7 @@ install_custom_target(struct workspace *wk,
 
 	if (get_obj_array(wk, install_dir)->len == 1) {
 		obj i0;
-		obj_array_index(wk, install_dir, 0, &i0);
+		i0 = obj_array_index(wk, install_dir, 0);
 		install_dir = i0;
 	}
 
@@ -684,7 +702,7 @@ func_custom_target(struct workspace *wk, obj _, obj *res)
 		}
 
 		obj v;
-		obj_array_index(wk, akw[kw_output].val, 0, &v);
+		v = obj_array_index(wk, akw[kw_output].val, 0);
 		name = v;
 	}
 
@@ -775,12 +793,20 @@ func_vcs_tag(struct workspace *wk, obj _, obj *res)
 		kw_command,
 		kw_fallback,
 		kw_replace_string,
+		kw_install,
+		kw_install_dir,
+		kw_install_mode,
+		kw_install_tag,
 	};
 	struct args_kw akw[] = { [kw_input] = { "input", TYPE_TAG_LISTIFY | tc_coercible_files, .required = true },
 		[kw_output] = { "output", obj_string, .required = true },
 		[kw_command] = { "command", tc_command_array | tc_both_libs },
 		[kw_fallback] = { "fallback", obj_string },
 		[kw_replace_string] = { "replace_string", obj_string },
+		[kw_install] = { "install", obj_bool },
+		[kw_install_dir] = { "install_dir", TYPE_TAG_LISTIFY | tc_string | tc_bool },
+		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
+		[kw_install_tag] = { "install_tag", tc_string }, // TODO
 		0 };
 
 	if (!pop_args(wk, NULL, akw)) {
@@ -797,7 +823,7 @@ func_vcs_tag(struct workspace *wk, obj _, obj *res)
 	}
 
 	obj command;
-	make_obj(wk, &command, obj_array);
+	command = make_obj(wk, obj_array);
 
 	push_args_null_terminated(wk,
 		command,
@@ -856,6 +882,10 @@ func_vcs_tag(struct workspace *wk, obj _, obj *res)
 	struct obj_custom_target *tgt = get_obj_custom_target(wk, *res);
 	tgt->flags |= custom_target_build_always_stale;
 
+	if (!install_custom_target(wk, tgt, &akw[kw_install], 0, akw[kw_install_dir].val, akw[kw_install_mode].val)) {
+		return false;
+	}
+
 	obj_array_push(wk, current_project(wk)->targets, *res);
 	return true;
 }
