@@ -9,6 +9,7 @@
 
 #include "error.h"
 #include "functions/kernel/options.h"
+#include "lang/analyze.h"
 #include "lang/typecheck.h"
 #include "options.h"
 #include "platform/assert.h"
@@ -23,10 +24,15 @@ build_option_type_from_s(struct workspace *wk, uint32_t node, uint32_t name, enu
 		[op_integer] = "integer",
 		[op_array] = "array",
 		[op_feature] = "feature",
+		[op_shell_array] = "shell_array",
 	};
 
 	enum build_option_type type;
 	for (type = 0; type < build_option_type_count; ++type) {
+		if (type == op_shell_array && !initializing_builtin_options) {
+			continue;
+		}
+
 		if (strcmp(build_option_type_name[type], get_cstr(wk, name)) == 0) {
 			*res = type;
 			return true;
@@ -72,16 +78,6 @@ func_option(struct workspace *wk, obj self, obj *res)
 		kw_kind = kwargs_count,
 	};
 
-	// TODO: this winds up creating 4 typeinfo objects every time you call
-	// option, it'd be nice to have some caching
-	type_tag deprecated_type = make_complex_type(wk,
-		complex_type_or,
-		tc_string | tc_bool,
-		make_complex_type(wk,
-			complex_type_or,
-			make_complex_type(wk, complex_type_nested, tc_dict, tc_string),
-			make_complex_type(wk, complex_type_nested, tc_array, tc_string)));
-
 	struct args_kw akw[] = { [kw_type] = { "type", obj_string },
 		[kw_value] = { "value", tc_any },
 		[kw_description] = { "description", obj_string },
@@ -89,7 +85,7 @@ func_option(struct workspace *wk, obj self, obj *res)
 		[kw_max] = { "max", obj_number },
 		[kw_min] = { "min", obj_number },
 		[kw_yield] = { "yield", obj_bool },
-		[kw_deprecated] = { "deprecated", deprecated_type },
+		[kw_deprecated] = { "deprecated", COMPLEX_TYPE_PRESET(tc_cx_options_deprecated_kw) },
 		[kw_kind] = { 0 },
 		0 };
 
@@ -153,26 +149,25 @@ func_option(struct workspace *wk, obj self, obj *res)
 	} else {
 		switch (type) {
 		case op_string: val = make_str(wk, ""); break;
-		case op_boolean:
-			val = make_obj_bool(wk, true);
-			break;
+		case op_boolean: val = make_obj_bool(wk, true); break;
 		case op_combo:
 			if (!get_obj_array(wk, akw[kw_choices].val)->len) {
 				vm_error_at(wk, akw[kw_choices].node, "combo option with no choices");
 				return false;
 			}
 
-			obj_array_index(wk, akw[kw_choices].val, 0, &val);
+			val = obj_array_index(wk, akw[kw_choices].val, 0);
 			break;
 		case op_array:
+		case op_shell_array:
 			if (akw[kw_choices].set) {
 				val = akw[kw_choices].val;
 			} else {
-				make_obj(wk, &val, obj_array);
+				val = make_obj(wk, obj_array);
 			}
 			break;
 		case op_feature:
-			make_obj(wk, &val, obj_feature_opt);
+			val = make_obj(wk, obj_feature_opt);
 			set_obj_feature_opt(wk, val, feature_opt_auto);
 			break;
 		default: UNREACHABLE_RETURN;
@@ -180,7 +175,7 @@ func_option(struct workspace *wk, obj self, obj *res)
 	}
 
 	obj opt;
-	make_obj(wk, &opt, obj_option);
+	opt = make_obj(wk, obj_option);
 	struct obj_option *o = get_obj_option(wk, opt);
 	o->name = an[0].val;
 	o->type = type;
@@ -193,9 +188,9 @@ func_option(struct workspace *wk, obj self, obj *res)
 	o->ip = wk->vm.ip - 1;
 
 	if (akw[kw_kind].set) {
-		if (str_eql(&WKSTR("default"), get_str(wk, akw[kw_kind].val))) {
+		if (str_eql(&STR("default"), get_str(wk, akw[kw_kind].val))) {
 			o->kind = build_option_kind_default;
-		} else if (str_eql(&WKSTR("prefixed_dir"), get_str(wk, akw[kw_kind].val))) {
+		} else if (str_eql(&STR("prefixed_dir"), get_str(wk, akw[kw_kind].val))) {
 			o->kind = build_option_kind_prefixed_dir;
 		} else {
 			vm_error_at(wk, akw[kw_kind].node, "invalid kind: %o", akw[kw_kind].val);
@@ -237,6 +232,23 @@ func_get_option(struct workspace *wk, obj self, obj *res)
 	}
 
 	struct obj_option *o = get_obj_option(wk, opt);
-	*res = o->val;
+
+	if (wk->vm.in_analyzer) {
+		type_tag t = 0;
+		switch (o->type) {
+		case op_string: t =  tc_string; break;
+		case op_boolean: t =  tc_bool; break;
+		case op_combo: t = COMPLEX_TYPE(o->choices, complex_type_enum); break;
+		case op_integer: t = tc_number; break;
+		case op_shell_array: t = tc_array; break;
+		case op_array: t = tc_array; break;
+		case op_feature: t = tc_feature_opt; break;
+		default: UNREACHABLE;
+		}
+
+		*res = make_typeinfo(wk, t);
+	} else {
+		*res = o->val;
+	}
 	return true;
 }
