@@ -18,37 +18,37 @@
 
 static bool
 subproject_prepare(struct workspace *wk,
-	struct sbuf *cwd_buf,
+	struct tstr *cwd_buf,
 	const char **cwd,
-	struct sbuf *build_dir_buf,
+	struct tstr *build_dir_buf,
 	const char **build_dir,
 	bool required,
 	bool *found)
 {
-	SBUF(wrap_path);
-	sbuf_pushf(wk, &wrap_path, "%s.wrap", *cwd);
+	TSTR(wrap_path);
+	tstr_pushf(wk, &wrap_path, "%s.wrap", *cwd);
 
 	if (!wk->vm.in_analyzer && fs_file_exists(wrap_path.buf)) {
 		bool wrap_ok = false;
 
-		SBUF(base_path);
+		TSTR(base_path);
 
 		path_dirname(wk, &base_path, *cwd);
 
-		struct wrap wrap = { 0 };
-		struct wrap_opts wrap_opts = {
-			.allow_download = get_option_wrap_mode(wk) != wrap_mode_nodownload,
-			.subprojects = base_path.buf,
-		};
-		if (!wrap_handle(wrap_path.buf, &wrap, &wrap_opts)) {
+		struct wrap_handle_ctx wrap_ctx
+			= { .opts = {
+				    .allow_download = get_option_wrap_mode(wk) != wrap_mode_nodownload,
+				    .subprojects = base_path.buf,
+			    } };
+		if (!wrap_handle(wk, wrap_path.buf, &wrap_ctx)) {
 			goto wrap_cleanup;
 		}
 
-		if (wrap.fields[wf_directory]) {
-			path_join(wk, cwd_buf, base_path.buf, wrap.fields[wf_directory]);
+		if (wrap_ctx.wrap.fields[wf_directory]) {
+			path_join(wk, cwd_buf, base_path.buf, wrap_ctx.wrap.fields[wf_directory]);
 
 			path_dirname(wk, &base_path, *build_dir);
-			path_join(wk, build_dir_buf, base_path.buf, wrap.fields[wf_directory]);
+			path_join(wk, build_dir_buf, base_path.buf, wrap_ctx.wrap.fields[wf_directory]);
 
 			*cwd = cwd_buf->buf;
 			*build_dir = build_dir_buf->buf;
@@ -57,7 +57,7 @@ subproject_prepare(struct workspace *wk,
 		wrap_ok = true;
 
 wrap_cleanup:
-		wrap_destroy(&wrap);
+		wrap_destroy(&wrap_ctx.wrap);
 
 		if (!wrap_ok) {
 			if (required) {
@@ -70,7 +70,7 @@ wrap_cleanup:
 		}
 	}
 
-	SBUF(src);
+	TSTR(src);
 	path_join(wk, &src, *cwd, "meson.build");
 
 	if (!fs_file_exists(src.buf)) {
@@ -100,14 +100,14 @@ subproject(struct workspace *wk,
 		return true;
 	}
 
-	make_obj(wk, res, obj_subproject);
+	*res = make_obj(wk, obj_subproject);
 	if (req == requirement_skip) {
 		return true;
 	}
 
 	const char *subproj_name = get_cstr(wk, name);
-	SBUF(cwd);
-	SBUF(build_dir);
+	TSTR(cwd);
+	TSTR(build_dir);
 
 	path_join(wk,
 		&cwd,
@@ -125,8 +125,8 @@ subproject(struct workspace *wk,
 	bool found;
 
 	const char *sp_cwd = cwd.buf, *sp_build_dir = build_dir.buf;
-	SBUF(sp_cwd_buf);
-	SBUF(sp_build_dir_buf);
+	TSTR(sp_cwd_buf);
+	TSTR(sp_build_dir_buf);
 
 	if (!subproject_prepare(
 		    wk, &sp_cwd_buf, &sp_cwd, &sp_build_dir_buf, &sp_build_dir, req == requirement_required, &found)) {
@@ -150,13 +150,7 @@ subproject(struct workspace *wk,
 	if (versions && versions->set) {
 		struct project *subp = arr_get(&wk->projects, subproject_id);
 
-		bool compare_result;
-		if (!version_compare(
-			    wk, versions->node, get_str(wk, subp->cfg.version), versions->val, &compare_result)) {
-			goto not_found;
-		}
-
-		if (!compare_result) {
+		if (!version_compare_list(wk, get_str(wk, subp->cfg.version), versions->val)) {
 			if (req == requirement_required) {
 				vm_error_at(wk,
 					versions->node,
@@ -168,7 +162,7 @@ subproject(struct workspace *wk,
 		}
 	}
 
-	make_obj(wk, res, obj_subproject);
+	*res = make_obj(wk, obj_subproject);
 	struct obj_subproject *sub = get_obj_subproject(wk, *res);
 	sub->id = subproject_id;
 	sub->found = true;
@@ -181,8 +175,8 @@ subproject(struct workspace *wk,
 		}
 
 		obj sub_name, var_name, dep;
-		obj_array_index(wk, v, 0, &sub_name);
-		obj_array_index(wk, v, 1, &var_name);
+		sub_name = obj_array_index(wk, v, 0);
+		var_name = obj_array_index(wk, v, 1);
 
 		if (!obj_equal(wk, sub_name, name)) {
 			continue;
@@ -238,7 +232,7 @@ func_subproject(struct workspace *wk, obj _, obj *res)
 
 	};
 	struct args_kw akw[] = {
-		[kw_default_options] = { "default_options", wk->complex_types.options_dict_or_list },
+		[kw_default_options] = { "default_options", COMPLEX_TYPE_PRESET(tc_cx_options_dict_or_list) },
 		[kw_required] = { "required", tc_required_kw },
 		[kw_version] = { "version", TYPE_TAG_LISTIFY | obj_string },
 		0,
