@@ -9,10 +9,11 @@
 
 #include "buf_size.h"
 #include "embedded.h"
-#include "error.h"
 #include "functions/modules.h"
+#include "functions/modules/curl.h"
 #include "functions/modules/fs.h"
 #include "functions/modules/getopt.h"
+#include "functions/modules/json.h"
 #include "functions/modules/keyval.h"
 #include "functions/modules/pkgconfig.h"
 #include "functions/modules/python.h"
@@ -26,8 +27,8 @@
 #include "platform/filesystem.h"
 #include "platform/path.h"
 
-#define MODULE_INFO(mod, path_prefix, _implemented) \
-	{ .name = #mod, .path = path_prefix "/" #mod, .implemented = _implemented },
+#define MODULE_INFO(mod, _path, _implemented) \
+	{ .name = #mod, .path = _path, .implemented = _implemented },
 const struct module_info module_info[module_count] = { FOREACH_BUILTIN_MODULE(MODULE_INFO) };
 #undef MODULE_INFO
 
@@ -53,39 +54,30 @@ struct module_lookup_script_opts {
 
 static bool
 module_lookup_script(struct workspace *wk,
-	struct sbuf *path,
+	struct tstr *path,
 	struct obj_module *m,
 	const struct module_lookup_script_opts *opts)
 {
-	struct source src;
-
-	if (opts->embedded) {
-		if (!(embedded_get(path->buf, &src))) {
-			return false;
-		}
-	} else {
-		if (!fs_file_exists(path->buf)) {
-			return false;
-		}
-
-		if (!fs_read_entire_file(path->buf, &src)) {
-			UNREACHABLE;
-		}
-	}
-
-	src.label = get_cstr(wk, sbuf_into_str(wk, path));
-	src.len = strlen(src.src);
-
 	bool ret = false;
-	enum language_mode old_language_mode = wk->vm.lang_mode;
-	wk->vm.lang_mode = language_extended;
-
 	bool stack_popped = false;
+	stack_push(&wk->stack, wk->vm.lang_mode, language_extended);
 	stack_push(&wk->stack, wk->vm.scope_stack, wk->vm.behavior.scope_stack_dup(wk, wk->vm.default_scope_stack));
 
 	obj res;
-	if (!eval(wk, &src, build_language_meson, 0, &res)) {
-		goto ret;
+	if (opts->embedded) {
+		struct source src;
+		if (!(embedded_get(path->buf, &src))) {
+			goto ret;
+		}
+		src.label = get_cstr(wk, tstr_into_str(wk, path));
+		src.len = strlen(src.src);
+		if (!eval(wk, &src, build_language_meson, 0, &res)) {
+			goto ret;
+		}
+	} else {
+		if (!wk->vm.behavior.eval_project_file(wk, path->buf, build_language_meson, 0, &res)) {
+			goto ret;
+		}
 	}
 
 	if (!typecheck_custom(wk,
@@ -102,6 +94,7 @@ module_lookup_script(struct workspace *wk,
 		m->exports = res;
 	} else {
 		stack_pop(&wk->stack, wk->vm.scope_stack);
+		stack_pop(&wk->stack, wk->vm.lang_mode);
 		stack_popped = true;
 		obj k, v;
 		obj_dict_for(wk, res, k, v) {
@@ -113,8 +106,8 @@ module_lookup_script(struct workspace *wk,
 ret:
 	if (!stack_popped) {
 		stack_pop(&wk->stack, wk->vm.scope_stack);
+		stack_pop(&wk->stack, wk->vm.lang_mode);
 	}
-	wk->vm.lang_mode = old_language_mode;
 	return ret;
 }
 
@@ -145,22 +138,22 @@ module_import(struct workspace *wk, const char *name, bool encapsulate, obj *res
 
 		bool loop = true;
 		struct str path;
-		SBUF(path_interpolated);
-		SBUF(module_path);
+		TSTR(path_interpolated);
+		TSTR(module_path);
 		const char *p, *sep;
 
 		{
 			struct project *proj;
 			if (wk->vm.lang_mode == language_external && (proj = current_project(wk)) && proj->module_dir) {
-				sbuf_pushs(wk, &module_path, module_paths[wk->vm.lang_mode]);
+				tstr_pushs(wk, &module_path, module_paths[wk->vm.lang_mode]);
 
-				sbuf_pushs(wk, &module_path, ";file:");
+				tstr_pushs(wk, &module_path, ";file:");
 
-				SBUF(new_module_path);
+				TSTR(new_module_path);
 				path_push(wk, &new_module_path, get_cstr(wk, proj->source_root));
 				path_push(wk, &new_module_path, get_cstr(wk, proj->module_dir));
 				path_push(wk, &new_module_path, "%.meson");
-				sbuf_pushn(wk, &module_path, new_module_path.buf, new_module_path.len);
+				tstr_pushn(wk, &module_path, new_module_path.buf, new_module_path.len);
 				p = module_path.buf;
 			} else {
 				p = module_paths[wk->vm.lang_mode];
@@ -182,7 +175,7 @@ module_import(struct workspace *wk, const char *name, bool encapsulate, obj *res
 					const struct str schema_str = { path.s, sep - path.s };
 					for (schema = 0; (uint32_t)schema < ARRAY_LEN(schema_type_str); ++schema) {
 						if (schema_type_str[schema]
-							&& str_eql(&WKSTR(schema_type_str[schema]), &schema_str)) {
+							&& str_eql(&STRL(schema_type_str[schema]), &schema_str)) {
 							break;
 						}
 					}
@@ -203,14 +196,14 @@ missing_schema:
 			}
 
 			{ // Interpolate path
-				sbuf_clear(&path_interpolated);
+				tstr_clear(&path_interpolated);
 
 				uint32_t i;
 				for (i = 0; i < path.len; ++i) {
 					if (path.s[i] == '%') {
-						sbuf_pushs(wk, &path_interpolated, name);
+						tstr_pushs(wk, &path_interpolated, name);
 					} else {
-						sbuf_push(wk, &path_interpolated, path.s[i]);
+						tstr_push(wk, &path_interpolated, path.s[i]);
 					}
 				}
 			}
@@ -224,7 +217,7 @@ missing_schema:
 				};
 
 				if (encapsulate) {
-					make_obj(wk, res, obj_module);
+					*res = make_obj(wk, obj_module);
 					m = get_obj_module(wk, *res);
 
 					if (obj_dict_index_strn(wk,
@@ -240,13 +233,12 @@ missing_schema:
 					if (encapsulate) {
 						obj_dict_set(wk,
 							wk->vm.modules,
-							sbuf_into_str(wk, &path_interpolated),
+							tstr_into_str(wk, &path_interpolated),
 							*res);
 					}
 
 					if (schema == schema_type_file) {
-						obj_array_push(
-							wk, wk->regenerate_deps, sbuf_into_str(wk, &path_interpolated));
+						workspace_add_regenerate_dep(wk, tstr_into_str(wk, &path_interpolated));
 					}
 
 					if (wk->vm.error) {
@@ -266,7 +258,7 @@ missing_schema:
 						return false;
 					}
 
-					make_obj(wk, res, obj_module);
+					*res = make_obj(wk, obj_module);
 					m = get_obj_module(wk, *res);
 					m->module = mod_type;
 					m->found = has_impl;
@@ -304,6 +296,8 @@ struct func_impl_group module_func_impl_groups[module_count][language_mode_count
 	[module_toolchain]   = { { 0 },                         { impl_tbl_module_toolchain }          },
 	[module_subprojects] = { { 0 },                         { impl_tbl_module_subprojects }        },
 	[module_getopt]      = { { 0 },                         { impl_tbl_module_getopt }             },
+	[module_curl]        = { { 0 },                         { impl_tbl_module_curl }             },
+	[module_json]        = { { 0 },                         { impl_tbl_module_json }             },
 };
 
 const struct func_impl impl_tbl_module[] = {
