@@ -11,6 +11,7 @@
 #include "args.h"
 #include "error.h"
 #include "formats/editorconfig.h"
+#include "functions/both_libs.h"
 #include "functions/kernel/custom_target.h"
 #include "functions/modules/fs.h"
 #include "lang/func_lookup.h"
@@ -35,7 +36,7 @@ fs_coerce_file_path(struct workspace *wk, uint32_t node, obj o, bool abs_build_t
 	switch (t) {
 	case obj_string: ss = get_str(wk, o); break;
 	case obj_file: ss = get_str(wk, *get_obj_file(wk, o)); break;
-	case obj_both_libs: o = get_obj_both_libs(wk, o)->dynamic_lib;
+	case obj_both_libs: o = decay_both_libs(wk, o);
 	/* fallthrough */
 	case obj_build_target: {
 		struct obj_build_target *tgt = get_obj_build_target(wk, o);
@@ -44,9 +45,9 @@ fs_coerce_file_path(struct workspace *wk, uint32_t node, obj o, bool abs_build_t
 			return name;
 		}
 
-		SBUF(joined);
+		TSTR(joined);
 		path_join(wk, &joined, get_cstr(wk, tgt->build_dir), name);
-		return get_cstr(wk, sbuf_into_str(wk, &joined));
+		return get_cstr(wk, tstr_into_str(wk, &joined));
 	}
 	case obj_custom_target: {
 		o = get_obj_custom_target(wk, o)->output;
@@ -72,7 +73,7 @@ fs_coerce_file_path(struct workspace *wk, uint32_t node, obj o, bool abs_build_t
 }
 
 static bool
-fix_file_path(struct workspace *wk, uint32_t err_node, obj path, enum fix_file_path_opts opts, struct sbuf *buf)
+fix_file_path(struct workspace *wk, uint32_t err_node, obj path, enum fix_file_path_opts opts, struct tstr *buf)
 {
 	const char *s = fs_coerce_file_path(wk, err_node, path, false);
 	if (!s) {
@@ -89,7 +90,11 @@ fix_file_path(struct workspace *wk, uint32_t err_node, obj path, enum fix_file_p
 				return false;
 			}
 
-			path_join(wk, buf, home, &s[1]);
+			if (s[1] == '/' && s[2]) {
+				path_join(wk, buf, home, &s[2]);
+			} else {
+				path_copy(wk, buf, home);
+			}
 		} else if (opts & fix_file_path_noabs) {
 			path_copy(wk, buf, s);
 		} else {
@@ -120,7 +125,7 @@ func_module_fs_lookup_common(struct workspace *wk,
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, opts, &path)) {
 		return false;
 	}
@@ -161,14 +166,14 @@ func_module_fs_parent(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noabs, &path)) {
 		return false;
 	}
 
-	SBUF(buf);
+	TSTR(buf);
 	path_dirname(wk, &buf, path.buf);
-	*res = sbuf_into_str(wk, &buf);
+	*res = tstr_into_str(wk, &buf);
 	return true;
 }
 
@@ -188,12 +193,12 @@ func_module_fs_read(struct workspace *wk, obj self, obj *res)
 	}
 
 	if (akw[kw_encoding].set) {
-		if (!str_eql(get_str(wk, akw[kw_encoding].val), &WKSTR("utf-8"))) {
+		if (!str_eql(get_str(wk, akw[kw_encoding].val), &STR("utf-8"))) {
 			vm_error_at(wk, akw[kw_encoding].node, "only 'utf-8' supported");
 		}
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -230,12 +235,12 @@ func_module_fs_expanduser(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
 
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -247,14 +252,14 @@ func_module_fs_name(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noexpanduser, &path)) {
 		return false;
 	}
 
-	SBUF(basename);
+	TSTR(basename);
 	path_basename(wk, &basename, path.buf);
-	*res = sbuf_into_str(wk, &basename);
+	*res = tstr_into_str(wk, &basename);
 	return true;
 }
 
@@ -266,12 +271,12 @@ func_module_fs_stem(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noexpanduser, &path)) {
 		return false;
 	}
 
-	SBUF(basename);
+	TSTR(basename);
 	path_basename(wk, &basename, path.buf);
 
 	char *dot;
@@ -280,7 +285,7 @@ func_module_fs_stem(struct workspace *wk, obj self, obj *res)
 		basename.len = strlen(basename.buf);
 	}
 
-	*res = sbuf_into_str(wk, &basename);
+	*res = tstr_into_str(wk, &basename);
 	return true;
 }
 
@@ -294,19 +299,19 @@ func_module_fs_as_posix(struct workspace *wk, obj self, obj *res)
 
 	const char *path = get_cstr(wk, an[0].val), *p;
 
-	SBUF(buf);
+	TSTR(buf);
 	for (p = path; *p; ++p) {
 		if (*p == '\\') {
-			sbuf_push(wk, &buf, '/');
+			tstr_push(wk, &buf, '/');
 			if (*(p + 1) == '\\') {
 				++p;
 			}
 		} else {
-			sbuf_push(wk, &buf, *p);
+			tstr_push(wk, &buf, *p);
 		}
 	}
 
-	*res = sbuf_into_str(wk, &buf);
+	*res = tstr_into_str(wk, &buf);
 	return true;
 }
 
@@ -318,7 +323,7 @@ func_module_fs_replace_suffix(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noabs, &path)) {
 		return false;
 	}
@@ -330,8 +335,8 @@ func_module_fs_replace_suffix(struct workspace *wk, obj self, obj *res)
 		path.len = strlen(path.buf);
 	}
 
-	sbuf_pushs(wk, &path, get_cstr(wk, an[1].val));
-	*res = sbuf_into_str(wk, &path);
+	tstr_pushs(wk, &path, get_cstr(wk, an[1].val));
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -343,12 +348,12 @@ func_module_fs_hash(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	if (!str_eql(get_str(wk, an[1].val), &WKSTR("sha256"))) {
+	if (!str_eql(get_str(wk, an[1].val), &STR("sha256"))) {
 		vm_error_at(wk, an[1].node, "only sha256 is supported");
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -362,11 +367,7 @@ func_module_fs_hash(struct workspace *wk, obj self, obj *res)
 	calc_sha_256(hash, src.src, src.len); // TODO: other hash algos
 
 	char buf[65] = { 0 };
-	uint32_t i, bufi = 0;
-	for (i = 0; i < 32; ++i) {
-		snprintf(&buf[bufi], 3, "%x", hash[i]);
-		bufi += 2;
-	}
+	sha256_to_str(hash, buf);
 
 	*res = make_str(wk, buf);
 
@@ -382,7 +383,7 @@ func_module_fs_size(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -398,7 +399,7 @@ func_module_fs_size(struct workspace *wk, obj self, obj *res)
 	}
 
 	assert(size < INT64_MAX);
-	make_obj(wk, res, obj_number);
+	*res = make_obj(wk, obj_number);
 	set_obj_number(wk, *res, size);
 	return true;
 }
@@ -411,12 +412,12 @@ func_module_fs_is_samepath(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path1);
+	TSTR(path1);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path1)) {
 		return false;
 	}
 
-	SBUF(path2);
+	TSTR(path2);
 	if (!fix_file_path(wk, an[1].node, an[1].val, 0, &path2)) {
 		return false;
 	}
@@ -449,7 +450,7 @@ func_module_fs_copyfile(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -458,13 +459,13 @@ func_module_fs_copyfile(struct workspace *wk, obj self, obj *res)
 	if (an[1].set) {
 		output = an[1].val;
 	} else {
-		SBUF(dest);
+		TSTR(dest);
 		path_basename(wk, &dest, path.buf);
-		output = sbuf_into_str(wk, &dest);
+		output = tstr_into_str(wk, &dest);
 	}
 
 	obj command;
-	make_obj(wk, &command, obj_array);
+	command = make_obj(wk, obj_array);
 	push_args_null_terminated(wk,
 		command,
 		(char *const[]){
@@ -519,8 +520,8 @@ func_module_fs_relative_to(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(b1);
-	SBUF(b2);
+	TSTR(b1);
+	TSTR(b2);
 
 	if (path_is_absolute(p1)) {
 		path_copy(wk, &b1, p1);
@@ -534,9 +535,9 @@ func_module_fs_relative_to(struct workspace *wk, obj self, obj *res)
 		path_join(wk, &b2, workspace_cwd(wk), p2);
 	}
 
-	SBUF(path);
+	TSTR(path);
 	path_relative_to(wk, &path, b2.buf, b1.buf);
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -569,7 +570,7 @@ func_module_fs_write(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -599,7 +600,7 @@ func_module_fs_copy(struct workspace *wk, obj self, obj *res)
 
 	bool force = akw[kw_force].set ? get_obj_bool(wk, akw[kw_force].val) : false;
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, 0, &path)) {
 		return false;
 	}
@@ -617,9 +618,9 @@ func_module_fs_cwd(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(cwd);
+	TSTR(cwd);
 	path_copy_cwd(wk, &cwd);
-	*res = sbuf_into_str(wk, &cwd);
+	*res = tstr_into_str(wk, &cwd);
 	return true;
 }
 
@@ -631,9 +632,9 @@ func_module_fs_make_absolute(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	path_make_absolute(wk, &path, get_cstr(wk, an[0].val));
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -705,9 +706,9 @@ func_module_fs_without_ext(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	path_without_ext(wk, &path, get_cstr(wk, an[0].val));
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -731,10 +732,10 @@ func_module_fs_add_suffix(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	path_copy(wk, &path, get_cstr(wk, an[0].val));
-	sbuf_pushs(wk, &path, get_cstr(wk, an[1].val));
-	*res = sbuf_into_str(wk, &path);
+	tstr_pushs(wk, &path, get_cstr(wk, an[1].val));
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -746,9 +747,9 @@ func_module_fs_executable(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	path_executable(wk, &path, get_cstr(wk, an[0].val));
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
@@ -783,20 +784,20 @@ func_module_fs_glob_cb(void *_ctx, const char *name)
 
 	/* L("path: %s, pat: %.*s", name, pat.len, pat.s); */
 
-	if (str_eql(&pat, &WKSTR("**"))) {
+	if (str_eql(&pat, &STR("**"))) {
 		subctx.mode = fs_glob_mode_recurse;
 		subctx.matchprefix = "/";
 		goto recurse_all_mode;
 	}
 
-	if (str_eql_glob(&pat, &WKSTR(name))) {
-		SBUF(path);
+	if (str_eql_glob(&pat, &STRL(name))) {
+		TSTR(path);
 		path_join(wk, &path, ctx->prefix, name);
 		subctx.prefix = path.buf;
 
 		subctx.pat += pat.len;
 		if (!*subctx.pat) {
-			obj_array_push(wk, ctx->res, sbuf_into_str(wk, &path));
+			obj_array_push(wk, ctx->res, tstr_into_str(wk, &path));
 		} else {
 			++subctx.pat;
 			if (fs_dir_exists(path.buf)) {
@@ -809,9 +810,9 @@ func_module_fs_glob_cb(void *_ctx, const char *name)
 
 	return ir_cont;
 recurse_all_mode: {
-	SBUF(path);
+	TSTR(path);
 	path_join(wk, &path, subctx.matchprefix, name);
-	SBUF(full_path);
+	TSTR(full_path);
 	path_join(wk, &full_path, subctx.prefix, name);
 
 	subctx.prefix = full_path.buf;
@@ -820,7 +821,7 @@ recurse_all_mode: {
 	/* L("recurse: path: %s, pat: %s", path.buf, ctx->pat); */
 
 	if (editorconfig_pattern_match(ctx->pat, path.buf)) {
-		obj_array_push(wk, ctx->res, sbuf_into_str(wk, &full_path));
+		obj_array_push(wk, ctx->res, tstr_into_str(wk, &full_path));
 	}
 
 	if (fs_dir_exists(full_path.buf)) {
@@ -840,7 +841,7 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(pat);
+	TSTR(pat);
 	{
 		const struct str *_pat = get_str(wk, an[0].val);
 		if (str_has_null(_pat)) {
@@ -851,7 +852,7 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 		path_join(wk, &pat, workspace_cwd(wk), _pat->s);
 	}
 
-	SBUF(prefix);
+	TSTR(prefix);
 	{
 		uint32_t i, prefix_len = 0;
 		for (i = 0; pat.buf[i]; ++i) {
@@ -867,7 +868,7 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 			}
 		}
 
-		sbuf_pushn(wk, &prefix, pat.buf, prefix_len);
+		tstr_pushn(wk, &prefix, pat.buf, prefix_len);
 	}
 
 	if (prefix.len) {
@@ -879,7 +880,7 @@ func_module_fs_glob(struct workspace *wk, obj self, obj *res)
 
 	/* L("prefix: %.*s, pat: %s", prefix.len, prefix.buf, pat.buf); */
 
-	make_obj(wk, res, obj_array);
+	*res = make_obj(wk, obj_array);
 
 	if (!fs_dir_exists(prefix.buf)) {
 		vm_error(wk, "Path \"%s\" does not exist", prefix.buf);
@@ -908,7 +909,7 @@ delete_suffix_recursive(void *_ctx, const char *path)
 	struct delete_suffixes_ctx *ctx = _ctx;
 	struct stat sb;
 
-	SBUF(name);
+	TSTR(name);
 
 	path_join(NULL, &name, ctx->base_dir, path);
 
@@ -917,7 +918,7 @@ delete_suffix_recursive(void *_ctx, const char *path)
 	// machinery found this file. In this case, the file no longer exists, so
 	// just continue.
 	if (!fs_exists(name.buf)) {
-		sbuf_destroy(&name);
+		tstr_destroy(&name);
 		return ir_cont;
 	}
 
@@ -946,7 +947,7 @@ delete_suffix_recursive(void *_ctx, const char *path)
 	ir_res = ir_cont;
 
 ret:
-	sbuf_destroy(&name);
+	tstr_destroy(&name);
 	return ir_res;
 }
 
@@ -975,18 +976,17 @@ func_module_fs_canonicalize(struct workspace *wk, obj self, obj *res)
 		return false;
 	}
 
-	SBUF(path);
+	TSTR(path);
 	if (!fix_file_path(wk, an[0].node, an[0].val, fix_file_path_noexpanduser, &path)) {
 		return false;
 	}
 
-	*res = sbuf_into_str(wk, &path);
+	*res = tstr_into_str(wk, &path);
 	return true;
 }
 
 const struct func_impl impl_tbl_module_fs_internal[] = {
 	{ "as_posix", func_module_fs_as_posix, tc_string, true },
-	{ "copyfile", func_module_fs_copyfile },
 	{ "exists", func_module_fs_exists, tc_bool },
 	{ "expanduser", func_module_fs_expanduser, tc_string },
 	{ "hash", func_module_fs_hash, tc_string },
@@ -1004,17 +1004,17 @@ const struct func_impl impl_tbl_module_fs_internal[] = {
 	{ "stem", func_module_fs_stem, tc_string, true },
 	// non-standard muon extensions
 	{ "add_suffix", func_module_fs_add_suffix, tc_string, true, .desc = "Append a suffix to a path" },
-	{ "copy", func_module_fs_copy, .fuzz_unsafe = true, .desc = "Copy a file to a destination at configure time" },
+	{ "copy", func_module_fs_copy, .flags = func_impl_flag_sandbox_disable, .desc = "Copy a file to a destination at configure time" },
 	{ "cwd", func_module_fs_cwd, tc_string, .desc = "Return the muon's current working directory" },
 	{ "executable", func_module_fs_executable, tc_string, true, .desc = "Prepend ./ to a path if it is a basename" },
 	{ "glob", func_module_fs_glob, tc_array, .desc = "Gather a list of files matching a glob expression" },
 	{ "is_basename", func_module_fs_is_basename, tc_bool, true, .desc = "Check if a path is a basename" },
 	{ "is_subpath", func_module_fs_is_subpath, tc_bool, true, .desc = "Check if a path is a subpath of another path" },
 	{ "make_absolute", func_module_fs_make_absolute, tc_string, .desc = "Make a path absolute by prepending the current working directory" },
-	{ "mkdir", func_module_fs_mkdir, .fuzz_unsafe = true, .desc = "Create a directory" },
-	{ "rmdir", func_module_fs_rmdir, .fuzz_unsafe = true, .desc = "Remove a directory" },
+	{ "mkdir", func_module_fs_mkdir, .flags = func_impl_flag_sandbox_disable, .desc = "Create a directory" },
+	{ "rmdir", func_module_fs_rmdir, .flags = func_impl_flag_sandbox_disable, .desc = "Remove a directory" },
 	{ "without_ext", func_module_fs_without_ext, tc_string, true, .desc = "Get a path with its extension removed" },
-	{ "write", func_module_fs_write, .fuzz_unsafe = true, .desc = "Write a file" },
+	{ "write", func_module_fs_write, .flags = func_impl_flag_sandbox_disable, .desc = "Write a file" },
 	{ "delete_with_suffix", func_module_fs_delete_with_suffix, .desc = "Recursively delete all files with a matching suffix" },
 	{ "canonicalize", func_module_fs_canonicalize, tc_string, true, .desc = "Make a path absolute and replace .."  },
 	{ NULL, NULL },
