@@ -10,12 +10,13 @@
 #include "buf_size.h"
 #include "coerce.h"
 #include "embedded.h"
-#include "external/tinyjson.h"
+#include "formats/json.h"
 #include "functions/external_program.h"
 #include "functions/modules/python.h"
 #include "install.h"
 #include "lang/object.h"
 #include "lang/typecheck.h"
+#include "log.h"
 #include "options.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
@@ -31,39 +32,44 @@ introspect_python_interpreter(struct workspace *wk, const char *path, struct obj
 
 	struct run_cmd_ctx cmd_ctx = { 0 };
 	char *const var_args[] = { (char *)path, "-c", (char *)src.src, 0 };
-	if (!run_cmd_argv(&cmd_ctx, var_args, NULL, 0) || cmd_ctx.status != 0) {
+	if (!run_cmd_argv_checked(&cmd_ctx, var_args, NULL, 0)) {
 		return false;
 	}
 
 	obj res_introspect;
-	bool success = muon_json_to_dict(wk, cmd_ctx.out.buf, &res_introspect);
+	{
+		bool success = muon_json_to_obj(wk, &TSTR_STR(&cmd_ctx.out), &res_introspect);
+		run_cmd_ctx_destroy(&cmd_ctx);
 
-	if (!success) {
-		goto end;
+		if (!success) {
+			return false;
+		}
 	}
 
-	if (!obj_dict_index_str(wk, res_introspect, "version", &python->language_version)) {
-		success = false;
-		goto end;
+	if (get_obj_type(wk, res_introspect) != obj_dict) {
+		LOG_E("introspection object is not a dictionary");
+		return false;
 	}
 
-	if (!obj_dict_index_str(wk, res_introspect, "sysconfig_paths", &python->sysconfig_paths)) {
-		success = false;
-		goto end;
+	struct {
+		const char *key;
+		obj *dest;
+	} key_to_dest[] = {
+		{ "version", &python->language_version },
+		{ "sysconfig_paths", &python->sysconfig_paths },
+		{ "variables", &python->sysconfig_vars },
+		{ "install_paths", &python->install_paths },
+	};
+
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(key_to_dest); ++i) {
+		if (!obj_dict_index_str(wk, res_introspect, key_to_dest[i].key, key_to_dest[i].dest)) {
+			LOG_E("introspection object missing key '%s'", key_to_dest[i].key);
+			return false;
+		}
 	}
 
-	if (!obj_dict_index_str(wk, res_introspect, "variables", &python->sysconfig_vars)) {
-		success = false;
-		goto end;
-	}
-
-	if (!obj_dict_index_str(wk, res_introspect, "install_paths", &python->install_paths)) {
-		success = false;
-		goto end;
-	}
-end:
-	run_cmd_ctx_destroy(&cmd_ctx);
-	return success;
+	return true;
 }
 
 static bool
@@ -71,8 +77,8 @@ python_module_present(struct workspace *wk, const char *pythonpath, const char *
 {
 	struct run_cmd_ctx cmd_ctx = { 0 };
 
-	SBUF(importstr);
-	sbuf_pushf(wk, &importstr, "import %s", mod);
+	TSTR(importstr);
+	tstr_pushf(wk, &importstr, "import %s", mod);
 
 	char *const *args = (char *const[]){ (char *)pythonpath, "-c", importstr.buf, 0 };
 
@@ -107,16 +113,16 @@ iterate_required_module_list(struct workspace *wk, void *ctx, obj val)
 }
 
 static bool
-build_python_installation(struct workspace *wk, obj self, obj *res, struct sbuf cmd_path, bool found, bool pure)
+build_python_installation(struct workspace *wk, obj self, obj *res, struct tstr cmd_path, bool found, bool pure)
 {
-	make_obj(wk, res, obj_python_installation);
+	*res = make_obj(wk, obj_python_installation);
 	struct obj_python_installation *python = get_obj_python_installation(wk, *res);
 	python->pure = pure;
-	make_obj(wk, &python->prog, obj_external_program);
+	python->prog = make_obj(wk, obj_external_program);
 	struct obj_external_program *ep = get_obj_external_program(wk, python->prog);
 	ep->found = found;
-	make_obj(wk, &ep->cmd_array, obj_array);
-	obj_array_push(wk, ep->cmd_array, sbuf_into_str(wk, &cmd_path));
+	ep->cmd_array = make_obj(wk, obj_array);
+	obj_array_push(wk, ep->cmd_array, tstr_into_str(wk, &cmd_path));
 
 	if (found && !introspect_python_interpreter(wk, cmd_path.buf, python)) {
 		vm_error(wk, "failed to introspect python");
@@ -165,7 +171,7 @@ func_module_python_find_installation(struct workspace *wk, obj self, obj *res)
 		}
 	}
 
-	SBUF(cmd_path);
+	TSTR(cmd_path);
 	bool found = fs_find_cmd(wk, &cmd_path, cmd);
 	if (!found && (requirement == requirement_required)) {
 		vm_error(wk, "%s not found", cmd);
@@ -231,17 +237,17 @@ func_module_python3_find_python(struct workspace *wk, obj self, obj *res)
 		cmd = get_cstr(wk, an[0].val);
 	}
 
-	SBUF(cmd_path);
+	TSTR(cmd_path);
 	if (!fs_find_cmd(wk, &cmd_path, cmd)) {
 		vm_error(wk, "python3 not found");
 		return false;
 	}
 
-	make_obj(wk, res, obj_external_program);
+	*res = make_obj(wk, obj_external_program);
 	struct obj_external_program *ep = get_obj_external_program(wk, *res);
 	ep->found = true;
-	make_obj(wk, &ep->cmd_array, obj_array);
-	obj_array_push(wk, ep->cmd_array, sbuf_into_str(wk, &cmd_path));
+	ep->cmd_array = make_obj(wk, obj_array);
+	obj_array_push(wk, ep->cmd_array, tstr_into_str(wk, &cmd_path));
 
 	return true;
 }
@@ -325,7 +331,7 @@ func_python_installation_has_var(struct workspace *wk, obj self, obj *res)
 static bool
 get_install_dir(struct workspace *wk, obj self, bool pure, const char *subdir, obj *res)
 {
-	SBUF(installdir);
+	TSTR(installdir);
 
 	obj prefix;
 	get_option_value(wk, current_project(wk), "prefix", &prefix);
@@ -335,7 +341,7 @@ get_install_dir(struct workspace *wk, obj self, bool pure, const char *subdir, o
 	if (pure) {
 		obj puredir;
 		get_option_value(wk, current_project(wk), "python.purelibdir", &puredir);
-		if (!str_eql(get_str(wk, puredir), &WKSTR(""))) {
+		if (!str_eql(get_str(wk, puredir), &STR(""))) {
 			path_push(wk, &installdir, get_cstr(wk, puredir));
 		} else {
 			if (!obj_dict_index_str(wk, py->install_paths, "purelib", &puredir)) {
@@ -346,7 +352,7 @@ get_install_dir(struct workspace *wk, obj self, bool pure, const char *subdir, o
 	} else {
 		obj platdir;
 		get_option_value(wk, current_project(wk), "python.platlibdir", &platdir);
-		if (!str_eql(get_str(wk, platdir), &WKSTR(""))) {
+		if (!str_eql(get_str(wk, platdir), &STR(""))) {
 			path_push(wk, &installdir, get_cstr(wk, platdir));
 		} else {
 			if (!obj_dict_index_str(wk, py->install_paths, "platlib", &platdir)) {
@@ -360,7 +366,7 @@ get_install_dir(struct workspace *wk, obj self, bool pure, const char *subdir, o
 		path_push(wk, &installdir, subdir);
 	}
 
-	*res = sbuf_into_str(wk, &installdir);
+	*res = tstr_into_str(wk, &installdir);
 
 	return true;
 }
@@ -413,12 +419,12 @@ py_install_data_rename_iter(struct workspace *wk, void *_ctx, obj val)
 	obj dest;
 
 	obj rename;
-	obj_array_index(wk, ctx->rename, ctx->i, &rename);
+	rename = obj_array_index(wk, ctx->rename, ctx->i);
 
-	SBUF(d);
+	TSTR(d);
 	path_join(wk, &d, get_cstr(wk, ctx->dest), get_cstr(wk, rename));
 
-	dest = sbuf_into_str(wk, &d);
+	dest = tstr_into_str(wk, &d);
 
 	push_install_target(wk, src, dest, ctx->mode);
 
@@ -533,10 +539,53 @@ func_python_installation_interpreter_path(struct workspace *wk, obj self, obj *r
 		return false;
 	}
 
-	obj_array_index(wk, get_obj_external_program(wk, self)->cmd_array, 0, res);
+	*res = obj_array_index(wk, get_obj_external_program(wk, self)->cmd_array, 0);
 	return true;
 }
 
+static bool
+func_python_installation_dependency(struct workspace *wk, obj self, obj *res)
+{
+	struct arr kwargs;
+	func_kwargs_lookup(wk, 0, "dependency", &kwargs);
+	kwargs_arr_push(wk, &kwargs, &(struct args_kw){ "embed", obj_bool });
+
+	if (!pop_args(wk, 0, (struct args_kw *)kwargs.e)) {
+		kwargs_arr_destroy(wk, &kwargs);
+		return false;
+	}
+	kwargs_arr_destroy(wk, &kwargs);
+
+	vm_error(wk, "unimplemented");
+	return false;
+}
+
+static bool
+func_python_installation_extension_module(struct workspace *wk, obj self, obj *res)
+{
+	struct args_norm an[] = {
+		{ obj_string },
+		{ TYPE_TAG_GLOB | tc_coercible_files | tc_generated_list },
+		ARG_TYPE_NULL,
+	};
+
+	struct arr kwargs;
+	func_kwargs_lookup(wk, 0, "shared_module", &kwargs);
+	kwargs_arr_del(wk, &kwargs, "name_suffix");
+	kwargs_arr_del(wk, &kwargs, "name_prefix");
+	kwargs_arr_push(wk, &kwargs, &(struct args_kw){ "subdir", obj_string });
+	kwargs_arr_push(wk, &kwargs, &(struct args_kw){ "limited_api", obj_string });
+
+	if (!pop_args(wk, an, (struct args_kw *)kwargs.e)) {
+		kwargs_arr_destroy(wk, &kwargs);
+		return false;
+	}
+	kwargs_arr_destroy(wk, &kwargs);
+
+	vm_error(wk, "unimplemented");
+	return false;
+}
+
 static obj
 python_self_transform(struct workspace *wk, obj self)
 {
@@ -573,5 +622,7 @@ struct func_impl impl_tbl_python_installation[] = {
 	{ "install_sources", func_python_installation_install_sources },
 	{ "language_version", func_python_installation_language_version, tc_string },
 	{ "path", func_python_installation_interpreter_path, tc_string },
+	{ "dependency", func_python_installation_dependency, tc_dependency },
+	{ "extension_module", func_python_installation_extension_module, tc_build_target },
 	{ NULL, NULL },
 };
