@@ -2,7 +2,6 @@
  * SPDX-FileCopyrightText: Stone Tickle <lattis@mochiro.moe>
  * SPDX-License-Identifier: GPL-3.0-only
  */
-
 #include "compat.h"
 
 #include <string.h>
@@ -16,6 +15,8 @@
 #include "log.h"
 #include "options.h"
 #include "platform/assert.h"
+#include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 #include "tracy.h"
 
@@ -28,6 +29,7 @@ static struct {
 	struct vm_ops unpatched_ops;
 	struct obj_typeinfo az_injected_native_func_return;
 	struct hash branch_map;
+	struct hash dict_locations;
 	struct arr visited_ops;
 } analyzer;
 
@@ -40,17 +42,6 @@ struct az_file_entrypoint {
 
 static struct arr az_entrypoint_stack, az_entrypoint_stacks;
 
-struct assignment {
-	const char *name;
-	obj o;
-	bool accessed, default_var;
-	struct source_location location;
-	uint32_t src_idx;
-
-	uint32_t ep_stacks_i;
-	uint32_t ep_stack_len;
-};
-
 enum branch_map_type {
 	branch_map_type_normal,
 	branch_map_type_ternary,
@@ -72,7 +63,7 @@ az_set_error(void)
 }
 
 /******************************************************************************
- * typeinfo utilities
+ * utilities
  ******************************************************************************/
 
 struct obj_tainted_by_typeinfo_opts {
@@ -167,7 +158,7 @@ obj
 make_typeinfo(struct workspace *wk, type_tag t)
 {
 	obj res;
-	make_obj(wk, &res, obj_typeinfo);
+	res = make_obj(wk, obj_typeinfo);
 	struct obj_typeinfo *type = get_obj_typeinfo(wk, res);
 	type->type = t;
 	return res;
@@ -200,19 +191,22 @@ coerce_type_tag(struct workspace *wk, obj r)
 static void
 merge_types(struct workspace *wk, struct obj_typeinfo *a, obj r)
 {
-	a->type |= coerce_type_tag(wk, r);
+	a->type = flatten_type(wk, a->type) | flatten_type(wk, coerce_type_tag(wk, r));
 }
 
-struct az_ctx {
-	type_tag expected;
-	type_tag found;
-	const struct func_impl *found_func;
-	obj found_func_obj, found_func_module;
-	struct obj_typeinfo ti;
-	obj l;
-};
+uint32_t
+az_dict_member_location_lookup_str(struct workspace *wk, obj dict, const char *key)
+{
+	uint64_t *hv;
+	if ((hv = hash_get(&analyzer.dict_locations, &dict))) {
+		obj ip;
+		if (obj_dict_index_str(wk, *hv, key, &ip)) {
+			return ip;
+		}
+	}
 
-typedef void((az_for_each_type_cb)(struct workspace *wk, struct az_ctx *ctx, uint32_t n_id, type_tag t, obj *res));
+	return 0;
+}
 
 /******************************************************************************
  * scope handling
@@ -247,12 +241,12 @@ print_scope_stack(struct workspace *wk)
 			if (i == 0) {
 				L("    root scope:");
 				obj_dict_for(wk, scope_group, k, v) {
-					struct assignment *assign = bucket_arr_get(&assignments, v);
+					struct az_assignment *assign = bucket_arr_get(&assignments, v);
 					LO("      %o: %s %o\n", k, assign->accessed ? "a" : "_", assign->o);
 				}
 			} else {
 				obj_array_for(wk, scope_group, scope) {
-					struct assignment *assign = bucket_arr_get(&assignments, v);
+					struct az_assignment *assign = bucket_arr_get(&assignments, v);
 					L("    scope group:");
 					obj_dict_for(wk, scope, k, v) {
 						LO("      %o: %s %o\n", k, assign->accessed ? "a" : "_", assign->o);
@@ -275,8 +269,7 @@ assign_lookup_with_scope(struct workspace *wk, const char *name, obj *scope, obj
 	bool found = false;
 	obj local_scope;
 	obj_array_for(wk, wk->vm.scope_stack, local_scope) {
-		obj base;
-		obj_array_index(wk, local_scope, 0, &base);
+		obj base = obj_array_index(wk, local_scope, 0);
 
 		uint32_t local_scope_len = get_obj_array(wk, local_scope)->len;
 		if (local_scope_len > 1) {
@@ -286,7 +279,7 @@ assign_lookup_with_scope(struct workspace *wk, const char *name, obj *scope, obj
 			// example: [{c: 4}]          -- take last
 			for (i = local_scope_len - 1; i >= 1; --i) {
 				obj scope_group;
-				obj_array_index(wk, local_scope, i, &scope_group);
+				scope_group = obj_array_index(wk, local_scope, i);
 				*scope = obj_array_get_tail(wk, scope_group);
 
 				if (obj_dict_index_str(wk, *scope, name, res)) {
@@ -308,8 +301,8 @@ assign_lookup_with_scope(struct workspace *wk, const char *name, obj *scope, obj
 	return found;
 }
 
-static struct assignment *
-assign_lookup(struct workspace *wk, const char *name)
+struct az_assignment *
+az_assign_lookup(struct workspace *wk, const char *name)
 {
 	obj res, _;
 	if (assign_lookup_with_scope(wk, name, &_, &res)) {
@@ -341,9 +334,9 @@ az_unassign(struct workspace *wk, const char *name)
 
 static void
 check_reassign_to_different_type(struct workspace *wk,
-	struct assignment *a,
+	struct az_assignment *a,
 	obj new_val,
-	struct assignment *new_a,
+	struct az_assignment *new_a,
 	uint32_t n_id)
 {
 	if (!az_diagnostic_enabled(az_diagnostic_reassign_to_conflicting_type)) {
@@ -384,9 +377,10 @@ push_assignment(struct workspace *wk, const char *name, obj o, uint32_t ip)
 	// an obj (for storage in the scope dict)
 	obj v = assignments.len;
 	bucket_arr_push(&assignments,
-		&(struct assignment){
+		&(struct az_assignment){
 			.name = name,
 			.o = o,
+			.ip = ip,
 			.location = loc,
 			.src_idx = src_idx,
 			.ep_stacks_i = ep_stacks_i,
@@ -395,8 +389,8 @@ push_assignment(struct workspace *wk, const char *name, obj o, uint32_t ip)
 	return v;
 }
 
-static struct assignment *
-scope_assign(struct workspace *wk, const char *name, obj o, uint32_t n_id, enum variable_assignment_mode mode)
+static struct az_assignment *
+scope_assign(struct workspace *wk, const char *name, obj o, uint32_t ip, enum variable_assignment_mode mode)
 {
 	TracyCZoneAutoS;
 	obj scope = 0;
@@ -419,8 +413,8 @@ scope_assign(struct workspace *wk, const char *name, obj o, uint32_t n_id, enum
 
 	assert(scope);
 
-	struct assignment *a = 0;
-	if (analyzer.impure_loop_depth && (a = assign_lookup(wk, name))) {
+	struct az_assignment *a = 0;
+	if (analyzer.impure_loop_depth && (a = az_assign_lookup(wk, name))) {
 		// When overwriting a variable in an impure loop turn it into a
 		// typeinfo so that it gets marked as impure.
 		enum obj_type new_type = get_obj_type(wk, o);
@@ -433,17 +427,18 @@ scope_assign(struct workspace *wk, const char *name, obj o, uint32_t n_id, enum
 	if (obj_dict_index(wk, scope, make_str(wk, name), &aid)) {
 		// The assignment was found so just reassign it here
 		a = bucket_arr_get(&assignments, aid);
-		check_reassign_to_different_type(wk, a, o, NULL, n_id);
+		check_reassign_to_different_type(wk, a, o, NULL, ip);
 
 		a->o = o;
+		a->ip = ip;
 		TracyCZoneAutoE;
 		return a;
 	}
 
-	aid = push_assignment(wk, name, o, n_id);
+	aid = push_assignment(wk, name, o, ip);
 	obj_dict_set(wk, scope, make_str(wk, name), aid);
 
-	struct assignment *assign = bucket_arr_get(&assignments, aid);
+	struct az_assignment *assign = bucket_arr_get(&assignments, aid);
 	assign->accessed = accessed;
 
 	TracyCZoneAutoE;
@@ -455,7 +450,7 @@ push_scope_group(struct workspace *wk)
 {
 	obj local_scope = obj_array_get_tail(wk, wk->vm.scope_stack);
 	obj scope_group;
-	make_obj(wk, &scope_group, obj_array);
+	scope_group = make_obj(wk, obj_array);
 	obj_array_push(wk, local_scope, scope_group);
 }
 
@@ -466,12 +461,12 @@ push_scope_group_scope(struct workspace *wk)
 	obj scope_group = obj_array_get_tail(wk, local_scope);
 	obj scope;
 
-	make_obj(wk, &scope, obj_dict);
+	scope = make_obj(wk, obj_dict);
 	obj_array_push(wk, scope_group, scope);
 }
 
 static void
-merge_objects(struct workspace *wk, struct assignment *dest, struct assignment *src)
+merge_objects(struct workspace *wk, struct az_assignment *dest, struct az_assignment *src)
 {
 	type_tag dest_type = get_obj_type(wk, dest->o);
 	type_tag src_type = get_obj_type(wk, src->o);
@@ -505,15 +500,7 @@ pop_scope_group(struct workspace *wk)
 
 	obj scope_group = obj_array_pop(wk, local_scope);
 
-	obj merged, base;
-	obj_array_index(wk, scope_group, 0, &merged);
-
-	if (get_obj_array(wk, local_scope)->len == 1) {
-		obj_array_index(wk, local_scope, 0, &base);
-	} else {
-		obj prev_scope_group = obj_array_get_tail(wk, local_scope);
-		base = obj_array_get_tail(wk, prev_scope_group);
-	}
+	obj merged = obj_array_index(wk, scope_group, 0);
 
 	{ // First, merge all scopes other than the root scope into `merged`
 		bool first = true;
@@ -527,7 +514,7 @@ pop_scope_group(struct workspace *wk)
 			obj k, aid;
 			obj_dict_for(wk, scope, k, aid) {
 				obj bid;
-				struct assignment *b, *a = bucket_arr_get(&assignments, aid);
+				struct az_assignment *b, *a = bucket_arr_get(&assignments, aid);
 				if (obj_dict_index(wk, merged, k, &bid)) {
 					b = bucket_arr_get(&assignments, bid);
 					merge_objects(wk, b, a);
@@ -542,8 +529,8 @@ pop_scope_group(struct workspace *wk)
 		obj k, aid;
 		obj_dict_for(wk, merged, k, aid) {
 			(void)k;
-			struct assignment *a = bucket_arr_get(&assignments, aid), *b;
-			if ((b = assign_lookup(wk, a->name))) {
+			struct az_assignment *a = bucket_arr_get(&assignments, aid), *b;
+			if ((b = az_assign_lookup(wk, a->name))) {
 				merge_objects(wk, b, a);
 			} else {
 				if (false) {
@@ -562,7 +549,7 @@ pop_scope_group(struct workspace *wk)
 					}
 				}
 
-				b = scope_assign(wk, a->name, a->o, 0, assign_local);
+				b = scope_assign(wk, a->name, a->o, a->ip, assign_local);
 				b->accessed = a->accessed;
 				b->location = a->location;
 				b->src_idx = a->src_idx;
@@ -614,12 +601,12 @@ az_op_az_branch(struct workspace *wk)
 
 	push_scope_group(wk);
 
-	/* L("---> branching, merging @ %03x", cur_branch_group.merge_point); */
+	// L("---> branching, merging @ %03x", cur_branch_group.merge_point);
 
 	bool pure = true;
 	obj branch, expr_result = 0;
 	obj_array_for(wk, branches, branch) {
-		/* L("--> branch"); */
+		// L("--> branch");
 		cur_branch_group.branch = (union az_branch_element){ .i64 = get_obj_number(wk, branch) }.data;
 		cur_branch_group.result = (struct branch_map_data){ 0 };
 		wk->vm.ip = cur_branch_group.branch.ip;
@@ -647,7 +634,7 @@ az_op_az_branch(struct workspace *wk)
 		}
 	}
 
-	/* L("<--- all branches merged %03x <---", cur_branch_group.merge_point); */
+	// L("<--- all branches merged %03x <---", cur_branch_group.merge_point);
 
 	pop_scope_group(wk);
 
@@ -754,31 +741,6 @@ az_op_return_end(struct workspace *wk)
 	analyzer.unpatched_ops.ops[op_return_end](wk);
 }
 
-#if 0
-static void
-az_op_add_store(struct workspace *wk)
-{
-	obj a, b;
-
-	b = object_stack_pop(&wk->vm.stack);
-	obj a_id = vm_get_constant(wk->vm.code.e, &wk->vm.ip);
-
-	const struct str *id = get_str(wk, a_id);
-	if (!wk->vm.behavior.get_variable(wk, id->s, &a)) {
-		vm_error(wk, "undefined object %s", get_cstr(wk, a_id));
-		object_stack_push(wk, make_typeinfo(wk, tc_any));
-		return;
-	}
-
-	object_stack_push(wk, a);
-	object_stack_push(wk, b);
-	analyzer.unpatched_ops.ops[op_add](wk);
-
-	obj res = object_stack_peek(&wk->vm.stack, 1);
-	wk->vm.behavior.assign_variable(wk, id->s, res, 0, assign_reassign);
-}
-#endif
-
 /******************************************************************************
  * analyzer behaviors
  ******************************************************************************/
@@ -792,7 +754,7 @@ az_assign_wrapper(struct workspace *wk, const char *name, obj o, uint32_t n_id,
 static bool
 az_lookup_wrapper(struct workspace *wk, const char *name, obj *res)
 {
-	struct assignment *a = assign_lookup(wk, name);
+	struct az_assignment *a = az_assign_lookup(wk, name);
 	if (a) {
 		a->accessed = true;
 		*res = a->o;
@@ -806,9 +768,9 @@ static void
 az_push_local_scope(struct workspace *wk)
 {
 	obj scope_group;
-	make_obj(wk, &scope_group, obj_array);
+	scope_group = make_obj(wk, obj_array);
 	obj scope;
-	make_obj(wk, &scope, obj_dict);
+	scope = make_obj(wk, obj_dict);
 	obj_array_push(wk, scope_group, scope);
 	obj_array_push(wk, wk->vm.scope_stack, scope_group);
 }
@@ -825,10 +787,10 @@ az_scope_stack_dup(struct workspace *wk, obj scope_stack)
 {
 	obj dup, local_scope, scope_group, scope;
 	obj local_scope_dup, scope_group_dup, scope_dup;
-	make_obj(wk, &dup, obj_array);
+	dup = make_obj(wk, obj_array);
 
 	obj_array_for(wk, scope_stack, local_scope) {
-		make_obj(wk, &local_scope_dup, obj_array);
+		local_scope_dup = make_obj(wk, obj_array);
 
 		uint32_t i = 0;
 		obj_array_for(wk, local_scope, scope_group) {
@@ -836,7 +798,7 @@ az_scope_stack_dup(struct workspace *wk, obj scope_stack)
 				obj_dict_dup(wk, scope_group, &scope_dup);
 				obj_array_push(wk, local_scope_dup, scope_dup);
 			} else {
-				make_obj(wk, &scope_group_dup, obj_array);
+				scope_group_dup = make_obj(wk, obj_array);
 
 				obj_array_for(wk, scope_group, scope) {
 					obj_dict_dup(wk, scope, &scope_dup);
@@ -859,38 +821,36 @@ static bool
 az_eval_project_file(struct workspace *wk,
 	const char *path,
 	enum build_language lang,
-	enum eval_project_file_flags flags)
+	enum eval_project_file_flags flags,
+	obj *res)
 {
-	const char *newpath = path;
-	if (analyzer.opts->file_override && strcmp(analyzer.opts->file_override, path) == 0) {
-		bool ret = false;
-		struct source src = { 0 };
-		if (!fs_read_entire_file("-", &src)) {
-			return false;
-		}
-		src.label = get_cstr(wk, make_str(wk, path));
-
-		obj res;
-
+	obj override;
+	if (analyzer.opts->file_override && obj_dict_index_str(wk, analyzer.opts->file_override, path, &override)) {
 		enum eval_mode eval_mode = 0;
 		if (flags & eval_project_file_flag_first) {
 			eval_mode |= eval_mode_first;
 		}
-
-		if (!eval(wk, &src, lang, eval_mode, &res)) {
-			goto ret;
+		if (analyzer.opts->relaxed_parse) {
+			eval_mode |= eval_mode_relaxed_parse;
 		}
 
-		ret = true;
-ret:
-		return ret;
-	}
+		const struct source *src = arr_get(&analyzer.opts->file_override_src, override);
 
-	if (analyzer.opts->analyze_project_call_only) {
-		flags |= eval_project_file_flag_return_after_project;
-	}
+		struct source weak_src = *src;
+		weak_src.is_weak_reference = true;
 
-	return eval_project_file(wk, newpath, lang, flags);
+		return eval(wk, &weak_src, lang, eval_mode, res);
+	} else {
+		if (analyzer.opts->analyze_project_call_only) {
+			flags |= eval_project_file_flag_return_after_project;
+		}
+
+		if (analyzer.opts->relaxed_parse) {
+			flags |= eval_project_file_flag_relaxed_parse;
+		}
+
+		return eval_project_file(wk, path, lang, flags, res);
+	}
 }
 
 static void
@@ -900,10 +860,8 @@ az_execute_loop(struct workspace *wk)
 
 	uint32_t cip;
 	while (wk->vm.run) {
-		if (log_should_print(log_debug)) {
-			/* LL("%-50s", vm_dis_inst(wk, wk->vm.code.e, wk->vm.ip)); */
-			/* object_stack_print(wk, &wk->vm.stack); */
-		}
+		// LL("%-50s", vm_dis_inst(wk, wk->vm.code.e, wk->vm.ip));
+		// object_stack_print(wk, &wk->vm.stack);
 
 		cip = wk->vm.ip;
 		++wk->vm.ip;
@@ -959,9 +917,10 @@ az_func_lookup(struct workspace *wk, obj self, const char *name, uint32_t *idx,
 		uint32_t idx;
 		uint32_t matches;
 	} lookup_res = { 0 };
-	type_tag t = ti->type;
 	uint32_t i;
 
+	type_tag t = flatten_type(wk, ti->type);
+
 	// Strip disabler from type list
 	if ((t & tc_disabler) == tc_disabler) {
 		t &= ~tc_disabler;
@@ -1097,9 +1056,10 @@ az_native_func_dispatch(struct workspace *wk, uint32_t func_idx, obj self, obj *
 		pure = false;
 	}
 
+	bool is_subdir = false;
 	if (!self) {
-		if (strcmp(native_funcs[func_idx].name, "subdir") == 0
-			|| strcmp(native_funcs[func_idx].name, "subproject") == 0
+		is_subdir = strcmp(native_funcs[func_idx].name, "subdir") == 0;
+		if (is_subdir || strcmp(native_funcs[func_idx].name, "subproject") == 0
 			|| strcmp(native_funcs[func_idx].name, "dependency") == 0) {
 			pop_args_ctx.allow_impure_args_except_first = true;
 		} else if (strcmp(native_funcs[func_idx].name, "p") == 0) {
@@ -1125,6 +1085,10 @@ az_native_func_dispatch(struct workspace *wk, uint32_t func_idx, obj self, obj *
 		return func_ok;
 	}
 
+	if (is_subdir) {
+		vm_warning(wk, "Unable to analyze this subdir call.  Diagnostics will be incomplete");
+	}
+
 	if (func_idx == az_func_impl_group.off) {
 		// This means the native function we called was
 		// az_injected_native_func and we should respect the return
@@ -1133,10 +1097,6 @@ az_native_func_dispatch(struct workspace *wk, uint32_t func_idx, obj self, obj *
 		*res = make_typeinfo(wk, native_funcs[func_idx].return_type);
 	}
 
-	if (!args_ok) {
-		// Add error here if func was subdir
-	}
-
 	return args_ok;
 }
 
@@ -1213,7 +1173,7 @@ az_op_call(struct workspace *wk)
 			obj _k, aid;
 			obj_dict_for(wk, root_scope, _k, aid) {
 				(void)_k;
-				struct assignment *assign = bucket_arr_get(&assignments, aid);
+				struct az_assignment *assign = bucket_arr_get(&assignments, aid);
 				assign->accessed = true;
 			}
 		}
@@ -1223,6 +1183,112 @@ az_op_call(struct workspace *wk)
 	}
 }
 
+static void
+az_op_constant_dict(struct workspace *wk)
+{
+	obj b;
+	uint32_t ip = wk->vm.ip;
+	uint32_t i, len = vm_get_constant(wk->vm.code.e, &ip);
+	b = make_obj(wk, obj_dict);
+	for (i = 0; i < len; ++i) {
+		obj key = object_stack_peek(&wk->vm.stack, (len - i) * 2 - 1);
+		if (wk->vm.in_analyzer && get_obj_type(wk, key) == obj_typeinfo) {
+			continue;
+		}
+
+		obj_dict_set(wk, b, key, object_stack_peek_entry(&wk->vm.stack, (len - i) * 2)->ip);
+	}
+
+	analyzer.unpatched_ops.ops[op_constant_dict](wk);
+
+	obj dict = object_stack_peek(&wk->vm.stack, 1);
+	hash_set(&analyzer.dict_locations, &dict, b);
+}
+
+static void
+az_dict_locations_merge(struct workspace *wk, obj a, obj b, obj tgt)
+{
+	uint64_t *la, *lb;
+	if (!(la = hash_get(&analyzer.dict_locations, &a))) {
+		UNREACHABLE;
+	} else if (!(lb = hash_get(&analyzer.dict_locations, &b))) {
+		UNREACHABLE;
+	}
+
+	obj merged;
+	obj_dict_merge(wk, *la, *lb, &merged);
+
+	hash_set(&analyzer.dict_locations, &tgt, merged);
+}
+
+static void
+az_op_add(struct workspace *wk)
+{
+	obj b = object_stack_peek(&wk->vm.stack, 1);
+	obj a = object_stack_peek(&wk->vm.stack, 2);
+	analyzer.unpatched_ops.ops[op_add](wk);
+	obj c = object_stack_peek(&wk->vm.stack, 1);
+
+	if (get_obj_type(wk, c) == obj_dict && get_obj_type(wk, b) == obj_dict && get_obj_type(wk, a) == obj_dict) {
+		az_dict_locations_merge(wk, a, b, c);
+	}
+}
+
+static void
+az_op_store(struct workspace *wk)
+{
+	obj dup = 0;
+
+	enum op_store_flags flags;
+	{
+		uint32_t ip = wk->vm.ip;
+		flags = vm_get_constant(wk->vm.code.e, &ip);
+	}
+
+	if (flags & op_store_flag_member) {
+		obj tgt, key;
+		tgt = object_stack_peek(&wk->vm.stack, 2);
+		key = object_stack_peek(&wk->vm.stack, 3);
+
+		if (get_obj_type(wk, tgt) == obj_dict && get_obj_type(wk, key) == obj_string) {
+			const struct obj_stack_entry *e;
+			e = object_stack_peek_entry(&wk->vm.stack, 1);
+
+			uint64_t *hv;
+			if (!(hv = hash_get(&analyzer.dict_locations, &tgt))) {
+				hash_set(&analyzer.dict_locations, &tgt, make_obj(wk, obj_dict));
+				hv = hash_get(&analyzer.dict_locations, &tgt);
+			}
+
+			obj_dict_set(wk, *hv, key, e->ip);
+		}
+	} else {
+		obj tgt = object_stack_peek(&wk->vm.stack, 2);
+		if (get_obj_type(wk, tgt) == obj_dict) {
+			if (flags & op_store_flag_add_store) {
+				obj val = object_stack_peek(&wk->vm.stack, 1);
+				if (get_obj_type(wk, val) == obj_dict) {
+					az_dict_locations_merge(wk, tgt, val, tgt);
+				}
+			} else {
+				uint64_t *hv;
+				if (!(hv = hash_get(&analyzer.dict_locations, &tgt))) {
+					UNREACHABLE;
+				}
+
+				obj_dict_dup_light(wk, *hv, &dup);
+			}
+		}
+	}
+
+	analyzer.unpatched_ops.ops[op_store](wk);
+
+	if (dup) {
+		obj tgt = object_stack_peek(&wk->vm.stack, 1);
+		hash_set(&analyzer.dict_locations, &tgt, dup);
+	}
+}
+
 /******************************************************************************
  * eval trace helpers
  ******************************************************************************/
@@ -1282,9 +1348,9 @@ eval_trace_print_level(struct workspace *wk, struct eval_trace_print_ctx *ctx, o
 			}
 		}
 
-		SBUF(rel);
+		TSTR(rel);
 		if (path_is_absolute(get_cstr(wk, v))) {
-			SBUF(cwd);
+			TSTR(cwd);
 			path_copy_cwd(wk, &cwd);
 			path_relative_to(wk, &rel, cwd.buf, get_cstr(wk, v));
 			printf("%s\n", rel.buf);
@@ -1322,7 +1388,6 @@ static const struct {
 	{ "unused-variable", az_diagnostic_unused_variable },
 	{ "reassign-to-conflicting-type", az_diagnostic_reassign_to_conflicting_type },
 	{ "dead-code", az_diagnostic_dead_code },
-	{ "redirect-script-error", az_diagnostic_redirect_script_error },
 };
 
 bool
@@ -1371,7 +1436,86 @@ az_warn_dead_code(struct workspace *wk,
 		= { .off = start_loc->off, .len = (end_loc->off - start_loc->off) + end_loc->len };
 	struct source *src = arr_get(&wk->vm.src, src_idx);
 
-	error_message(src, merged, log_warn, "dead code");
+	error_message(src, merged, log_warn, 0, "dead code");
+}
+
+void
+analyze_opts_init(struct workspace *wk, struct az_opts *opts)
+{
+	*opts = (struct az_opts){ 0 };
+	opts->file_override = make_obj(wk, obj_dict);
+	arr_init(&opts->file_override_src, 8, sizeof(struct source));
+}
+
+void
+analyze_opts_destroy(struct workspace *wk, struct az_opts *opts)
+{
+	TracyCZoneAutoS;
+	uint32_t i;
+	for (i = 0; i < opts->file_override_src.len; ++i) {
+		struct source *src = arr_get(&opts->file_override_src, i);
+		fs_source_destroy(src);
+	}
+
+	arr_destroy(&opts->file_override_src);
+	TracyCZoneAutoE;
+}
+
+bool
+analyze_opts_push_override(struct workspace *wk,
+	struct az_opts *opts,
+	const char *override,
+	const char *content_path,
+	const struct str *content)
+{
+	TracyCZoneAutoS;
+	uint32_t idx;
+	struct source *src = 0;
+
+	TSTR(abs);
+	path_make_absolute(wk, &abs, override);
+	obj path = tstr_into_str(wk, &abs);
+
+	if (obj_dict_index(wk, opts->file_override, path, &idx)) {
+		src = arr_get(&opts->file_override_src, idx);
+		fs_source_destroy(src);
+	}
+
+	if (!content && !content_path) {
+		obj_dict_del(wk, opts->file_override, path);
+		TracyCZoneAutoE;
+		return true;
+	}
+
+	if (!src) {
+		idx = opts->file_override_src.len;
+		arr_push(&opts->file_override_src, &(struct source){ 0 });
+		src = arr_peek(&opts->file_override_src, 1);
+		obj_dict_set(wk, opts->file_override, path, idx);
+	}
+
+	assert(!src->src);
+	assert(!src->len);
+
+	if (content) {
+		void *buf = z_calloc(content->len + 1, 1);
+		memcpy(buf, content->s, content->len);
+		*src = (struct source){
+			.type = source_type_file,
+			.len = content->len,
+			.src = buf,
+		};
+	} else {
+		if (!fs_read_entire_file(content_path, src)) {
+			TracyCZoneAutoE;
+			return false;
+		}
+	}
+
+	src->label = get_cstr(wk, path);
+
+	TracyCZoneAutoE;
+	return true;
 }
 
 /******************************************************************************
@@ -1379,29 +1523,31 @@ az_warn_dead_code(struct workspace *wk,
  ******************************************************************************/
 
 bool
-do_analyze_internal(struct workspace *wk, struct az_opts *opts)
+do_analyze(struct workspace *wk, struct az_opts *opts)
 {
+	TracyCZoneAutoS;
+
 	bool res = false;
 	analyzer.opts = opts;
-	workspace_init_bare(wk);
 
-	bucket_arr_init(&assignments, 512, sizeof(struct assignment));
+	bucket_arr_init(&assignments, 512, sizeof(struct az_assignment));
 	hash_init(&analyzer.branch_map, 1024, sizeof(uint32_t));
+	hash_init(&analyzer.dict_locations, 1024, sizeof(obj));
 	arr_init_flags(&analyzer.visited_ops, 1024, 1, arr_flag_zero_memory);
 
 	{ /* re-initialize the default scope */
 		obj original_scope, scope_group, scope;
-		obj_array_index(wk, wk->vm.default_scope_stack, 0, &original_scope);
-		make_obj(wk, &wk->vm.default_scope_stack, obj_array);
-		make_obj(wk, &scope_group, obj_array);
-		make_obj(wk, &scope, obj_dict);
+		original_scope = obj_array_index(wk, wk->vm.default_scope_stack, 0);
+		wk->vm.default_scope_stack = make_obj(wk, obj_array);
+		scope_group = make_obj(wk, obj_array);
+		scope = make_obj(wk, obj_dict);
 		obj_array_push(wk, scope_group, scope);
 		obj_array_push(wk, wk->vm.default_scope_stack, scope_group);
 		obj k, v;
 		obj_dict_for(wk, original_scope, k, v) {
 			obj aid = push_assignment(wk, get_cstr(wk, k), v, 0);
 
-			struct assignment *a = bucket_arr_get(&assignments, aid);
+			struct az_assignment *a = bucket_arr_get(&assignments, aid);
 			a->default_var = true;
 
 			obj_dict_set(wk, scope, k, aid);
@@ -1432,7 +1578,9 @@ do_analyze_internal(struct workspace *wk, struct az_opts *opts)
 	wk->vm.ops.ops[op_return] = az_op_return;
 	wk->vm.ops.ops[op_return_end] = az_op_return_end;
 	wk->vm.ops.ops[op_call] = az_op_call;
-	/* wk->vm.ops.ops[op_add_store] = az_op_add_store; */
+	wk->vm.ops.ops[op_store] = az_op_store;
+	wk->vm.ops.ops[op_constant_dict] = az_op_constant_dict;
+	wk->vm.ops.ops[op_add] = az_op_add;
 
 	error_diagnostic_store_init(wk);
 
@@ -1440,54 +1588,68 @@ do_analyze_internal(struct workspace *wk, struct az_opts *opts)
 	arr_init(&az_entrypoint_stacks, 32, sizeof(struct az_file_entrypoint));
 
 	if (analyzer.opts->eval_trace) {
-		make_obj(wk, &wk->vm.dbg_state.eval_trace, obj_array);
+		wk->vm.dbg_state.eval_trace = make_obj(wk, obj_array);
 	}
 
-	if (analyzer.opts->file_override) {
-		const char *root = determine_project_root(wk, analyzer.opts->file_override);
-		if (root) {
-			SBUF(cwd);
-			path_copy_cwd(wk, &cwd);
+	if (analyzer.opts->auto_chdir_root) {
+		obj first_override = 0;
+		if (analyzer.opts->file_override) {
+			obj _v;
+			obj_dict_for(wk, analyzer.opts->file_override, first_override, _v) {
+				(void)_v;
+				break;
+			}
+		}
 
-			if (strcmp(cwd.buf, root) != 0) {
-				path_chdir(root);
-				wk->source_root = root;
+		if (first_override) {
+			const char *root = determine_project_root(wk, get_cstr(wk, first_override));
+			if (root) {
+				TSTR(cwd);
+				path_copy_cwd(wk, &cwd);
+
+				if (strcmp(cwd.buf, root) != 0) {
+					path_chdir(root);
+					wk->source_root = root;
+				}
 			}
 		}
 	}
 
-	if (opts->internal_file) {
-		struct assignment *a = scope_assign(wk, "argv", make_typeinfo(wk, tc_array), 0, assign_local);
+	wk->vm.lang_mode = opts->lang_mode;
+
+	if (wk->vm.lang_mode == language_internal) {
+		struct az_assignment *a = scope_assign(wk, "argv", make_typeinfo(wk, tc_array), 0, assign_local);
 		a->default_var = true;
-
-		wk->vm.lang_mode = language_extended;
-
-		struct source src;
-		if (!fs_read_entire_file(opts->internal_file, &src)) {
-			res = false;
-		} else {
-			obj _v;
-			res = eval(wk, &src, build_language_meson, 0, &_v);
-		}
 	} else {
-		uint32_t project_id;
 		workspace_init_runtime(wk);
 		workspace_init_startup_files(wk);
 
 		{
 			obj wrap_mode;
-			get_option(wk, 0, &WKSTR("wrap_mode"), &wrap_mode);
+			get_option(wk, 0, &STR("wrap_mode"), &wrap_mode);
 			set_option(
 				wk, wrap_mode, make_str(wk, "forcefallback"), option_value_source_commandline, false);
 		}
+	}
 
+	if (opts->single_file) {
+		struct source src;
+		if (!fs_read_entire_file(opts->single_file, &src)) {
+			res = false;
+		} else {
+			obj _v;
+			res = eval(wk, &src, build_language_meson, 0, &_v);
+		}
+
+	} else {
+		uint32_t project_id;
 		res = eval_project(wk, NULL, wk->source_root, wk->build_root, &project_id);
 	}
 
 	if (az_diagnostic_enabled(az_diagnostic_unused_variable)) {
 		uint32_t i;
 		for (i = 0; i < assignments.len; ++i) {
-			struct assignment *a = bucket_arr_get(&assignments, i);
+			struct az_assignment *a = bucket_arr_get(&assignments, i);
 			if (!a->default_var && !a->accessed && *a->name != '_') {
 				const char *msg = get_cstr(wk, make_strf(wk, "unused variable %s", a->name));
 				enum log_level lvl = log_warn;
@@ -1569,33 +1731,18 @@ do_analyze_internal(struct workspace *wk, struct az_opts *opts)
 
 			for (j = 0; j < len; ++j) {
 				error_diagnostic_store_push(
-					ep_stack[j].src_idx, ep_stack[j].location, lvl, "in subdir");
+					ep_stack[j].src_idx, ep_stack[j].location, lvl, "errors in subdir");
 			}
 		}
 
 		i += len;
 	}
 
-	bool saw_error;
 	if (analyzer.opts->eval_trace) {
 		eval_trace_print(wk, wk->vm.dbg_state.eval_trace);
-	} else if (analyzer.opts->get_definition_for) {
-		bool found = false;
-		uint32_t i;
-		for (i = 0; i < assignments.len; ++i) {
-			struct assignment *a = bucket_arr_get(&assignments, i);
-			if (strcmp(a->name, analyzer.opts->get_definition_for) == 0) {
-				struct source *src = arr_get(&wk->vm.src, a->src_idx);
-				list_line_range(src, a->location, 1);
-				found = true;
-			}
-		}
-
-		if (!found) {
-			LOG_W("couldn't find definition for %s", analyzer.opts->get_definition_for);
-		}
 	} else {
-		error_diagnostic_store_replay(analyzer.opts->replay_opts, &saw_error);
+		bool saw_error;
+		saw_error = error_diagnostic_store_replay(wk, analyzer.opts->replay_opts);
 
 		if (saw_error || analyzer.error) {
 			res = false;
@@ -1606,15 +1753,9 @@ do_analyze_internal(struct workspace *wk, struct az_opts *opts)
 	arr_destroy(&az_entrypoint_stack);
 	arr_destroy(&az_entrypoint_stacks);
 	arr_destroy(&analyzer.visited_ops);
-	return res;
-}
-
-bool
-do_analyze(struct az_opts *opts)
-{
-	struct workspace wk;
-	bool res = do_analyze_internal(&wk, opts);
-	workspace_destroy(&wk);
+	hash_destroy(&analyzer.branch_map);
+	hash_destroy(&analyzer.dict_locations);
+	TracyCZoneAutoE;
 	return res;
 }
 
@@ -1626,5 +1767,7 @@ analyze_project_call(struct workspace *wk)
 		.analyze_project_call_only = true,
 	};
 
-	return do_analyze_internal(wk, &opts);
+	workspace_init_bare(wk);
+
+	return do_analyze(wk, &opts);
 }
