@@ -8,9 +8,11 @@
 #include <stdarg.h>
 #include <string.h>
 
+#include "buf_size.h"
 #include "lang/analyze.h"
 #include "lang/compiler.h"
 #include "lang/func_lookup.h"
+#include "lang/object_iterators.h"
 #include "lang/parser.h"
 #include "lang/typecheck.h"
 #include "lang/vm.h"
@@ -120,26 +122,36 @@ enum vm_compile_block_flags {
 static void vm_compile_block(struct workspace *wk, struct node *n, enum vm_compile_block_flags flags);
 static void vm_compile_expr(struct workspace *wk, struct node *n);
 
-static void
-vm_comp_node(struct workspace *wk, struct node *n)
+static bool
+vm_comp_node_skip(enum node_type t)
 {
-	assert(n->type != node_type_stmt);
-
-	/* L("compiling %s", node_to_s(wk, n)); */
-
-	switch (n->type) {
-	case node_type_group:
-	case node_type_stmt: UNREACHABLE;
-
+	switch (t) {
 	case node_type_id_lit:
 	case node_type_args:
 	case node_type_def_args:
 	case node_type_kw:
 	case node_type_list:
 	case node_type_foreach_args:
-		// Skipped
-		break;
+	case node_type_stmt:
+		return true;
+	default: return false;
+	}
+}
 
+static void
+vm_comp_node(struct workspace *wk, struct node *n)
+{
+	if (vm_comp_node_skip(n->type)) {
+		return;
+	}
+
+	/* L("compiling %s", node_to_s(wk, n)); */
+
+	if (n->flags & node_flag_breakpoint) {
+		push_code(wk, op_dbg_break);
+	}
+
+	switch (n->type) {
 	case node_type_stringify: push_code(wk, op_stringify); break;
 	case node_type_index: push_code(wk, op_index); break;
 	case node_type_negate: push_code(wk, op_negate); break;
@@ -183,7 +195,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 	case node_type_number:
 		push_code(wk, op_constant);
 		obj o;
-		make_obj(wk, &o, obj_number);
+		o = make_obj(wk, obj_number);
 		set_obj_number(wk, o, n->data.num);
 		push_constant(wk, o);
 		break;
@@ -191,6 +203,10 @@ vm_comp_node(struct workspace *wk, struct node *n)
 		push_code(wk, op_constant);
 		push_constant(wk, n->data.num ? obj_bool_true : obj_bool_false);
 		break;
+	case node_type_null:
+		push_code(wk, op_constant);
+		push_constant(wk, 0);
+		break;
 	case node_type_string:
 		push_code(wk, op_constant);
 		push_constant(wk, n->data.str);
@@ -228,7 +244,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 		if (n->r->type == node_type_id_lit) {
 			name = get_str(wk, n->r->data.str);
 
-			if (str_eql(name, &WKSTR("subdir_done"))) {
+			if (str_eql(name, &STR("subdir_done"))) {
 				push_location(wk, n);
 
 				vm_comp_assert_inline_func_args(wk, n, n->l, 0, 0, 0);
@@ -238,7 +254,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 
 				vm_comp_op_return(wk);
 				break;
-			} else if (str_eql(name, &WKSTR("set_variable"))) {
+			} else if (str_eql(name, &STR("set_variable"))) {
 				push_location(wk, n);
 
 				vm_comp_assert_inline_func_args(wk, n, n->l, 2, 2, 0);
@@ -246,7 +262,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 				push_code(wk, op_swap);
 				push_op_store(wk, 0);
 				break;
-			} else if (str_eql(name, &WKSTR("get_variable"))) {
+			} else if (str_eql(name, &STR("get_variable"))) {
 				push_location(wk, n);
 
 				vm_comp_assert_inline_func_args(wk, n, n->l, 1, 2, 0);
@@ -257,7 +273,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 					push_code(wk, op_try_load);
 				}
 				break;
-			} else if (str_eql(name, &WKSTR("disabler"))) {
+			} else if (str_eql(name, &STR("disabler"))) {
 				push_location(wk, n);
 
 				vm_comp_assert_inline_func_args(wk, n, n->l, 0, 0, 0);
@@ -265,7 +281,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 				push_code(wk, op_constant);
 				push_constant(wk, obj_disabler);
 				break;
-			} else if (str_eql(name, &WKSTR("is_disabler"))) {
+			} else if (str_eql(name, &STR("is_disabler"))) {
 				/* jmp_if_disabler >-,
 				 * pop               |
 				 * const false       |
@@ -321,7 +337,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 		}
 
 		if (known && (wk->vm.compiler_state.mode & vm_compile_mode_return_after_project)
-			&& str_eql(name, &WKSTR("project"))) {
+			&& str_eql(name, &STR("project"))) {
 			push_code(wk, op_return_end);
 		}
 
@@ -474,7 +490,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 			++patch_tgts;
 			push_constant(wk, 0);
 
-			make_obj(wk, &az_branches, obj_array);
+			az_branches = make_obj(wk, obj_array);
 			push_constant(wk, az_branches);
 		}
 
@@ -539,7 +555,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 			end_jmp[2] = wk->vm.code.len;
 			push_constant(wk, 0);
 
-			make_obj(wk, &az_branches, obj_array);
+			az_branches = make_obj(wk, obj_array);
 			push_constant(wk, az_branches);
 		}
 
@@ -607,7 +623,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 			end_jmp[1] = wk->vm.code.len;
 			push_constant(wk, 0);
 
-			make_obj(wk, &az_branches, obj_array);
+			az_branches = make_obj(wk, obj_array);
 			push_constant(wk, az_branches);
 
 			obj_array_push(wk,
@@ -652,7 +668,7 @@ vm_comp_node(struct workspace *wk, struct node *n)
 		struct obj_func *func;
 		struct node *arg;
 
-		make_obj(wk, &f, obj_func);
+		f = make_obj(wk, obj_func);
 		func = get_obj_func(wk, f);
 
 		push_code(wk, op_jmp);
@@ -733,11 +749,18 @@ vm_comp_node(struct workspace *wk, struct node *n)
 		}
 		break;
 	}
+	default: UNREACHABLE;
 	}
 }
 
+typedef void((*vm_visit_nodes_cb)(struct workspace *wk, struct node *n));
+
 static void
-vm_compile_expr(struct workspace *wk, struct node *n)
+vm_visit_nodes(struct workspace *wk,
+	struct node *n,
+	const enum node_type privelaged[],
+	uint32_t privelaged_len,
+	vm_visit_nodes_cb cb)
 {
 	struct node *peek, *prev = 0;
 
@@ -745,10 +768,16 @@ vm_compile_expr(struct workspace *wk, struct node *n)
 
 	while (wk->vm.compiler_state.node_stack.len > stack_base || n) {
 		if (n) {
-			if (n->type == node_type_foreach || n->type == node_type_if || n->type == node_type_func_def
-				|| n->type == node_type_ternary || n->type == node_type_and
-				|| n->type == node_type_or) {
-				vm_comp_node(wk, n);
+			bool is_privelaged = false;
+			for (uint32_t i = 0; i < privelaged_len; ++i) {
+				if (n->type == privelaged[i]) {
+					is_privelaged = true;
+					break;
+				}
+			}
+
+			if (is_privelaged) {
+				cb(wk, n);
 				prev = n;
 				n = 0;
 			} else {
@@ -761,13 +790,28 @@ vm_compile_expr(struct workspace *wk, struct node *n)
 				n = peek->r;
 			} else {
 				push_location(wk, peek);
-				vm_comp_node(wk, peek);
+				cb(wk, peek);
 				prev = *(struct node **)arr_pop(&wk->vm.compiler_state.node_stack);
 			}
 		}
 	}
 }
 
+static void
+vm_compile_expr(struct workspace *wk, struct node *n)
+{
+	const enum node_type privelaged[] = {
+		node_type_foreach,
+		node_type_if,
+		node_type_func_def,
+		node_type_ternary,
+		node_type_and,
+		node_type_or,
+	};
+
+	vm_visit_nodes(wk, n, privelaged, ARRAY_LEN(privelaged), vm_comp_node);
+}
+
 static void
 vm_compile_block(struct workspace *wk, struct node *n, enum vm_compile_block_flags flags)
 {
@@ -829,6 +873,84 @@ void
 vm_compile_state_reset(struct workspace *wk)
 {
 	bucket_arr_clear(&wk->vm.compiler_state.nodes);
+	wk->vm.compiler_state.breakpoints = 0;
+}
+
+static bool
+vm_compile_breakpoint_to_off(struct workspace *wk, const struct source *src, obj bp, uint32_t *off)
+{
+	uint32_t t_line, t_col;
+	vm_dbg_unpack_breakpoint(wk, bp, &t_line, &t_col);
+
+	uint32_t i, line = 1, col = 1;
+	for (i = 0; i < src->len; ++i) {
+		if (line == t_line && (col == t_col || !t_col)) {
+			*off = i;
+			return true;
+		}
+
+		if (src->src[i] == '\n') {
+			++line;
+			col = 1;
+		} else {
+			++col;
+		}
+	}
+
+	return false;
+}
+
+static void
+vm_resolve_breakpoint_cb(struct workspace *wk, struct node *n)
+{
+	if (vm_comp_node_skip(n->type)) {
+		return;
+	}
+
+	obj off, node;
+	obj_dict_for(wk, wk->vm.compiler_state.breakpoints, off, node) {
+		struct node *o = node ? (void *)(intptr_t)get_obj_number(wk, node) : 0;
+		bool in_range = n->location.off <= off && off <= n->location.off + n->location.len;
+		if (in_range && (!o || n->location.len < o->location.len)) {
+			obj_dict_seti(wk, wk->vm.compiler_state.breakpoints, off, make_number(wk, (int64_t)n));
+		}
+	}
+}
+
+static void
+vm_resolve_breakpoints(struct workspace *wk, struct node *n)
+{
+	const struct source *src = arr_peek(&wk->vm.src, 1);
+
+	obj file_bp;
+	if (obj_dict_index_str(wk, wk->vm.dbg_state.breakpoints, src->label, &file_bp)) {
+		wk->vm.compiler_state.breakpoints = make_obj(wk, obj_dict);
+
+		obj bp;
+		obj_array_for(wk, file_bp, bp) {
+			uint32_t off;
+			if (vm_compile_breakpoint_to_off(wk, src, bp, &off)) {
+				obj_dict_seti(wk, wk->vm.compiler_state.breakpoints, off, 0);
+			} else {
+				uint32_t t_line, t_col;
+				vm_dbg_unpack_breakpoint(wk, bp, &t_line, &t_col);
+				LOG_W("failed to resolve breakpoint %s:%d:%d", src->label, t_line, t_col);
+			}
+		}
+
+		vm_visit_nodes(wk, n, 0, 0, vm_resolve_breakpoint_cb);
+
+		obj off, node;
+		obj_dict_for(wk, wk->vm.compiler_state.breakpoints, off, node) {
+			struct node *o = node ? (void *)(intptr_t)get_obj_number(wk, node) : 0;
+			if (!o) {
+				LOG_W("failed to resolve breakpoint %s@%d", src->label, off);
+				continue;
+			}
+
+			o->flags |= node_flag_breakpoint;
+		}
+	}
 }
 
 bool
@@ -838,6 +960,10 @@ vm_compile_ast(struct workspace *wk, struct node *n, enum vm_compile_mode mode,
 	wk->vm.compiler_state.err = false;
 	wk->vm.compiler_state.mode = mode;
 
+	if (wk->vm.dbg_state.breakpoints) {
+		vm_resolve_breakpoints(wk, n);
+	}
+
 	*entry = wk->vm.code.len;
 
 	enum vm_compile_block_flags flags = vm_compile_block_final_return | vm_compile_block_start_scope;
@@ -857,7 +983,7 @@ vm_compile_ast(struct workspace *wk, struct node *n, enum vm_compile_mode mode,
 }
 
 bool
-vm_compile(struct workspace *wk, struct source *src, enum vm_compile_mode mode, uint32_t *entry)
+vm_compile(struct workspace *wk, const struct source *src, enum vm_compile_mode mode, uint32_t *entry)
 {
 	struct node *n;
 
