@@ -67,10 +67,10 @@ struct fmt_out_block {
 
 struct fmt_ctx {
 	struct workspace *wk;
-	struct sbuf *out_buf;
+	struct tstr *out_buf;
 	obj raw_blocks;
-	uint32_t indent, enclosing, raw_block_idx, measured_len;
-	bool trailing_comment, fmt_on, measuring, line_has_content;
+	uint32_t indent, raw_block_idx, measured_len;
+	bool trailing_comment, fmt_on, measuring, line_has_content, is_enclosed;
 
 	struct bucket_arr frags;
 
@@ -194,14 +194,14 @@ tree_indent_print(const struct tree_indent *ti)
 	for (i = 0; i < ti->indent; ++i) {
 		if (i < ti->indent - 1) {
 			if (ti->bars & (1 << i)) {
-				log_plain("│   ");
+				log_raw("│   ");
 			} else {
-				log_plain("    ");
+				log_raw("    ");
 			}
 		} else if (!ti->len || ti->i == ti->len - 1) {
-			log_plain("└── ");
+			log_raw("└── ");
 		} else {
-			log_plain("├── ");
+			log_raw("├── ");
 		}
 	}
 }
@@ -231,21 +231,21 @@ fmt_write_frag_set_dbg_ws(struct fmt_ctx *f, const struct fmt_frag *pws, struct
 	for (ws = pws; ws; ws = ws->next) {
 		tree_indent_print(sub_ti);
 
-		log_plain("%s: ", label);
+		log_raw("%s: ", label);
 
 		if (ws->type == fmt_frag_type_ws_newline) {
-			log_plain("newline");
+			log_raw("newline");
 		} else {
-			obj_lprintf(f->wk, "# %o", ws->str);
+			obj_lprintf(f->wk, log_info, "# %o", ws->str);
 			if (ws->type == fmt_frag_type_ws_comment) {
-				log_plain(" comment");
+				log_raw(" comment");
 			} else if (ws->type == fmt_frag_type_ws_comment_trailing) {
-				log_plain(" comment_trailing");
+				log_raw(" comment_trailing");
 			} else {
 				UNREACHABLE;
 			}
 		}
-		log_plain("\n");
+		log_raw("\n");
 		++sub_ti->i;
 	}
 }
@@ -263,26 +263,26 @@ fmt_write_frag_set_dbg(struct fmt_ctx *f, struct fmt_frag *p, const struct tree_
 	struct tree_indent sub_ti;
 
 	if (p == hl) {
-		log_plain("\033[34m");
+		log_raw("\033[34m");
 	}
 
 	tree_indent_print(ti);
 
 	if (p->str) {
-		obj_lprintf(f->wk, "%o", p->str);
+		obj_lprintf(f->wk, log_debug, "%o", p->str);
 	} else if (p->type == fmt_frag_type_block) {
-		obj_lprintf(f->wk, "block");
+		obj_lprintf(f->wk, log_debug, "block");
 	} else if (p->type == fmt_frag_type_line) {
-		obj_lprintf(f->wk, "line");
+		obj_lprintf(f->wk, log_debug, "line");
 	} else if (p->enclosing) {
-		obj_lprintf(f->wk, "%s", p->enclosing);
+		obj_lprintf(f->wk, log_debug, "%s", p->enclosing);
 	} else {
-		obj_lprintf(f->wk, "?");
+		obj_lprintf(f->wk, log_debug, "?");
 	}
 
 	if (p->flags) {
 		obj flags;
-		make_obj(f->wk, &flags, obj_array);
+		flags = make_obj(f->wk, obj_array);
 
 		struct {
 			uint32_t flag;
@@ -315,16 +315,16 @@ fmt_write_frag_set_dbg(struct fmt_ctx *f, struct fmt_frag *p, const struct tree_
 		obj joined;
 		obj_array_join(f->wk, false, flags, make_str(f->wk, ","), &joined);
 
-		log_plain(" <%s>", get_cstr(f->wk, joined));
+		log_raw(" <%s>", get_cstr(f->wk, joined));
 	}
 
 	if (p == hl) {
-		log_plain("\033[0m");
+		log_raw("\033[0m");
 	}
 
-	log_plain(" - %d", fmt_measure_frag(f, p));
+	log_raw(" - %d", fmt_measure_frag(f, p));
 
-	log_plain("\n");
+	log_raw("\n");
 
 	sub_ti = (struct tree_indent){
 		.len = fmt_frag_dbg_len(p),
@@ -356,18 +356,18 @@ fmt_push_out_block(struct fmt_ctx *f)
 {
 	arr_push(&f->out_blocks,
 		&(struct fmt_out_block){
-			.str = sbuf_into_str(f->wk, f->out_buf),
+			.str = tstr_into_str(f->wk, f->out_buf),
 		});
 
-	*f->out_buf = (struct sbuf){ 0 };
-	sbuf_init(f->out_buf, 0, 0, 0);
+	*f->out_buf = (struct tstr){ 0 };
+	tstr_init(f->out_buf, 0, 0, 0);
 }
 
 static void
 fmt_write_nl(struct fmt_ctx *f)
 {
 	if (f->measuring) {
-		if (f->enclosing) {
+		if (f->is_enclosed) {
 			f->measured_len += f->opts.max_line_len + 1;
 		}
 		return;
@@ -376,7 +376,7 @@ fmt_write_nl(struct fmt_ctx *f)
 	}
 
 	f->line_has_content = false;
-	sbuf_push(f->wk, f->out_buf, '\n');
+	tstr_push(f->wk, f->out_buf, '\n');
 }
 
 static void
@@ -399,12 +399,12 @@ fmt_write(struct fmt_ctx *f, const char *s, uint32_t n)
 			switch (f->opts.indent_style) {
 			case fmt_indent_style_space: {
 				for (j = 0; j < f->opts.indent_size; ++j) {
-					sbuf_push(f->wk, f->out_buf, ' ');
+					tstr_push(f->wk, f->out_buf, ' ');
 				}
 				break;
 			}
 			case fmt_indent_style_tab: {
-				sbuf_push(f->wk, f->out_buf, '\t');
+				tstr_push(f->wk, f->out_buf, '\t');
 				break;
 			}
 			}
@@ -413,7 +413,7 @@ fmt_write(struct fmt_ctx *f, const char *s, uint32_t n)
 
 	f->line_has_content = true;
 
-	sbuf_pushn(f->wk, f->out_buf, s, n);
+	tstr_pushn(f->wk, f->out_buf, s, n);
 }
 
 static void
@@ -442,7 +442,7 @@ fmt_write_frag_comment(struct fmt_ctx *f, struct fmt_frag *comment)
 		f->fmt_on = true;
 
 		obj raw_block;
-		obj_array_index(f->wk, f->raw_blocks, f->raw_block_idx, &raw_block);
+		raw_block = obj_array_index(f->wk, f->raw_blocks, f->raw_block_idx);
 		arr_push(&f->out_blocks,
 			&(struct fmt_out_block){
 				.str = raw_block,
@@ -519,7 +519,7 @@ fmt_write_frag_ws(struct fmt_ctx *f, struct fmt_frag *ws, enum fmt_write_frag_ws
 			break;
 		}
 		case fmt_frag_type_ws_newline: {
-			if (f->enclosing) {
+			if (f->is_enclosed) {
 				if (newlines < 1) {
 					// Skip the first newline
 					++newlines;
@@ -592,7 +592,7 @@ fmt_write_frag(struct fmt_ctx *f, struct fmt_frag *p)
 	// Write enclosing characters and indent if necessary
 	if (p->enclosing) {
 		fmt_write(f, &p->enclosing[0], 1);
-		++f->enclosing;
+		stack_push(&f->wk->stack, f->is_enclosed, true);
 
 		if (ml) {
 			++f->indent;
@@ -613,6 +613,7 @@ fmt_write_frag(struct fmt_ctx *f, struct fmt_frag *p)
 			}
 		}
 	} else if (p->type == fmt_frag_type_block) {
+		stack_push(&f->wk->stack, f->is_enclosed, false);
 		++f->indent;
 	}
 
@@ -639,7 +640,7 @@ fmt_write_frag(struct fmt_ctx *f, struct fmt_frag *p)
 				bool stick_line
 					= (fr->flags & fmt_frag_flag_stick_line_right)
 					  || (fr->next->flags & fmt_frag_flag_stick_line_left)
-					  || (!f->enclosing
+					  || (!f->is_enclosed
 						  && (fr->next->flags & fmt_frag_flag_stick_line_left_unless_enclosed
 							  || fr->next->flags
 								     & fmt_frag_flag_stick_left_unless_enclosed));
@@ -679,12 +680,13 @@ fmt_write_frag(struct fmt_ctx *f, struct fmt_frag *p)
 		}
 
 		fmt_write(f, &p->enclosing[1], 1);
-		--f->enclosing;
+		stack_pop(&f->wk->stack, f->is_enclosed);
 
 		if (ml && p->flags & fmt_frag_flag_enclosed_extra_indent) {
 			--f->indent;
 		}
 	} else if (p->type == fmt_frag_type_block) {
+		stack_pop(&f->wk->stack, f->is_enclosed);
 		--f->indent;
 	}
 
@@ -873,8 +875,8 @@ fmt_list(struct fmt_ctx *f, struct node *n, struct fmt_frag *fr, enum fmt_list_f
 				if (prev && (str = fmt_frag_as_simple_str(f, prev->child))) {
 					obj cur = fmt_frag_as_simple_str(f, child->child);
 					const struct str *s = get_str(f->wk, str);
-					bool prev_is_arg_like = str_startswith(s, &WKSTR("-")) && !strchr(s->s, '='),
-					     cur_is_arg_like = cur && str_startswith(get_str(f->wk, cur), &WKSTR("-"));
+					bool prev_is_arg_like = str_startswith(s, &STR("-")) && !strchr(s->s, '='),
+					     cur_is_arg_like = cur && str_startswith(get_str(f->wk, cur), &STR("-"));
 
 					if (prev_is_arg_like && !cur_is_arg_like
 						&& !(child->pre_ws && child->pre_ws->type == fmt_frag_type_ws_comment_trailing)) {
@@ -918,7 +920,7 @@ fmt_list(struct fmt_ctx *f, struct node *n, struct fmt_frag *fr, enum fmt_list_f
 
 		if (is_func_with_single_arg) {
 			if (n->l->type == node_type_string
-				&& str_startswith(get_str(f->wk, n->l->data.str), &WKSTR("'''"))) {
+				&& str_startswith(get_str(f->wk, n->l->data.str), &STR("'''"))) {
 				fr->flags |= fmt_frag_flag_force_single_line;
 			}
 		}
@@ -1014,15 +1016,15 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 	case node_type_string: {
 		obj str;
 		if (f->opts.simplify_string_literals && (str = fmt_obj_as_simple_str(f, n->data.str))
-			&& !str_contains(get_str(f->wk, str), &WKSTR("\n"))
-			&& !str_contains(get_str(f->wk, str), &WKSTR("'"))) {
+			&& !str_contains(get_str(f->wk, str), &STR("\n"))
+			&& !str_contains(get_str(f->wk, str), &STR("'"))) {
 			struct str newstr = *get_str(f->wk, n->data.str);
-			if (str_startswith(&newstr, &WKSTR("f"))) {
+			if (str_startswith(&newstr, &STR("f"))) {
 				++newstr.s;
 				--newstr.len;
 			}
 
-			if (str_startswith(&newstr, &WKSTR("'''"))) {
+			if (str_startswith(&newstr, &STR("'''"))) {
 				newstr.s += 2;
 				newstr.len -= 4;
 			}
@@ -1036,6 +1038,10 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 		fr->str = make_str(f->wk, n->data.num ? "true" : "false");
 		break;
 	}
+	case node_type_null: {
+		fr->str = make_str(f->wk, "null");
+		break;
+	}
 	case node_type_negate: {
 		fr->str = make_str(f->wk, "-");
 		fr->flags |= fmt_frag_flag_stick_right;
@@ -1168,7 +1174,7 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 			enum fmt_special_function function = fmt_special_function_unknown;
 
 			if (n->r->type == node_type_id_lit) {
-				if (str_eql(get_str(f->wk, n->r->data.str), &WKSTR("files"))) {
+				if (str_eql(get_str(f->wk, n->r->data.str), &STR("files"))) {
 					function = fmt_special_function_files;
 				}
 			}
@@ -1184,7 +1190,7 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 
 				// If files() gets a single argument of type
 				// array, un-nest it.
-				if (!args->r && args->l->type == node_type_array) {
+				if (!args->r && args->l && args->l->type == node_type_array) {
 					arr = args->l;
 					args->l = arr->l;
 					args->r = arr->r;
@@ -1196,7 +1202,7 @@ fmt_node(struct fmt_ctx *f, struct node *n)
 					if (args->l) {
 						if (args->l->type != node_type_string
 							|| str_startswith(
-								get_str(f->wk, args->l->data.str), &WKSTR("f"))) {
+								get_str(f->wk, args->l->data.str), &STR("f"))) {
 							all_elements_are_simple_strings = false;
 							break;
 						}
@@ -1636,7 +1642,7 @@ fmt_assemble_out_blocks(struct fmt_ctx *f)
 	for (i = 0; i < f->out_blocks.len; ++i) {
 		s = get_str(f->wk, blocks[i].str);
 		if (blocks[i].raw) {
-			sbuf_pushn(f->wk, f->out_buf, s->s, s->len);
+			tstr_pushn(f->wk, f->out_buf, s->s, s->len);
 			continue;
 		}
 
@@ -1646,7 +1652,7 @@ fmt_assemble_out_blocks(struct fmt_ctx *f)
 			end_of_block = false;
 
 			if (line_end == line) {
-				sbuf_pushs(f->wk, f->out_buf, end_of_line);
+				tstr_pushs(f->wk, f->out_buf, end_of_line);
 				++line;
 				continue;
 			} else if (line_end) {
@@ -1660,16 +1666,16 @@ fmt_assemble_out_blocks(struct fmt_ctx *f)
 			l = str_strip(f->wk, &lstr, 0, str_strip_flag_right_only);
 			s = get_str(f->wk, l);
 
-			sbuf_pushn(f->wk, f->out_buf, s->s, s->len);
+			tstr_pushn(f->wk, f->out_buf, s->s, s->len);
 
 			bool last_line = end_of_block && i == f->out_blocks.len - 1;
 
 			if (last_line) {
 				if (f->opts.insert_final_newline) {
-					sbuf_pushs(f->wk, f->out_buf, end_of_line);
+					tstr_pushs(f->wk, f->out_buf, end_of_line);
 				}
 			} else if (!end_of_block) {
-				sbuf_pushs(f->wk, f->out_buf, end_of_line);
+				tstr_pushs(f->wk, f->out_buf, end_of_line);
 			}
 
 			if (!line_end) {
@@ -1685,7 +1691,7 @@ bool
 fmt(struct source *src, FILE *out, const char *cfg_path, bool check_only, bool editorconfig)
 {
 	bool ret = false;
-	struct sbuf out_buf;
+	struct tstr out_buf;
 	struct workspace wk = { 0 };
 	workspace_init_bare(&wk);
 	struct fmt_ctx f = {
@@ -1730,7 +1736,7 @@ fmt(struct source *src, FILE *out, const char *cfg_path, bool check_only, bool e
 	}
 
 	enum vm_compile_mode compile_mode = vm_compile_mode_fmt;
-	if (str_endswith(&WKSTR(src->label), &WKSTR(".meson"))) {
+	if (str_endswith(&STRL(src->label), &STR(".meson"))) {
 		compile_mode |= vm_compile_mode_language_extended;
 	}
 
@@ -1739,12 +1745,12 @@ fmt(struct source *src, FILE *out, const char *cfg_path, bool check_only, bool e
 		goto ret;
 	}
 
-	sbuf_init(&out_buf, 0, 0, 0);
+	tstr_init(&out_buf, 0, 0, 0);
 	fmt_write_block(&f, fmt_block(&f, n));
 	fmt_push_out_block(&f);
 
 	if (!check_only) {
-		out_buf.flags = sbuf_flag_write;
+		out_buf.flags = tstr_flag_write;
 		out_buf.buf = (void *)out;
 	}
 	fmt_assemble_out_blocks(&f);
