@@ -138,9 +138,9 @@ object_stack_print(struct workspace *wk, struct object_stack *s)
 {
 	for (int32_t i = s->ba.len - 1; i >= 0; --i) {
 		struct obj_stack_entry *e = bucket_arr_get(&s->ba, i);
-		obj_lprintf(wk, "%o%s", e->o, i > 0 ? ", " : "");
+		obj_lprintf(wk, log_debug, "%o%s", e->o, i > 0 ? ", " : "");
 	}
-	log_plain("\n");
+	log_plain(log_debug, "\n");
 }
 
 /******************************************************************************
@@ -203,7 +203,7 @@ vm_inst_location_obj(struct workspace *wk, uint32_t ip)
 	get_detailed_source_location(src, loc, &dloc, (enum get_detailed_source_location_flag)0);
 
 	obj res;
-	make_obj(wk, &res, obj_array);
+	res = make_obj(wk, obj_array);
 	obj_array_push(
 		wk, res, make_strf(wk, "%s%s", src->type == source_type_embedded ? "[embedded] " : "", src->label));
 	obj_array_push(wk, res, make_number(wk, dloc.line));
@@ -211,11 +211,28 @@ vm_inst_location_obj(struct workspace *wk, uint32_t ip)
 	return res;
 }
 
+obj
+vm_inst_location_str(struct workspace *wk, uint32_t ip)
+{
+	obj a, src, line, col;
+
+	a = vm_inst_location_obj(wk, ip);
+	src = obj_array_index(wk, a, 0);
+	line = obj_array_index(wk, a, 1);
+	col = obj_array_index(wk, a, 2);
+
+	return make_strf(wk,
+		"%s:%d:%d",
+		get_str(wk, src)->s,
+		(uint32_t)get_obj_number(wk, line),
+		(uint32_t)get_obj_number(wk, col));
+}
+
 obj
 vm_callstack(struct workspace *wk)
 {
 	obj res;
-	make_obj(wk, &res, obj_array);
+	res = make_obj(wk, obj_array);
 
 	obj_array_push(wk, res, vm_inst_location_obj(wk, wk->vm.ip - 1));
 
@@ -232,38 +249,53 @@ vm_callstack(struct workspace *wk)
 	return res;
 }
 
-void
-vm_diagnostic_v(struct workspace *wk, uint32_t ip, enum log_level lvl, const char *fmt, va_list args)
+static void
+vm_trigger_error(struct workspace *wk)
+{
+	if (wk->vm.in_analyzer) {
+		az_set_error();
+	} else {
+		wk->vm.error = true;
+		wk->vm.run = false;
+	}
+}
+
+static void
+vm_diagnostic_v(struct workspace *wk,
+	uint32_t ip,
+	enum log_level lvl,
+	enum error_message_flag flags,
+	const char *fmt,
+	va_list args)
 {
 	static char buf[1024];
 	obj_vsnprintf(wk, buf, ARRAY_LEN(buf), fmt, args);
 
 	if (!ip) {
 		ip = wk->vm.ip - 1;
+	} else if (ip == UINT32_MAX) {
+		ip = 0;
 	}
 
-	struct source_location loc;
-	struct source *src;
-	vm_lookup_inst_location(&wk->vm, ip, &loc, &src);
+	struct source_location loc = { 0 };
+	struct source *src = 0;
+	if (ip) {
+		vm_lookup_inst_location(&wk->vm, ip, &loc, &src);
+	}
 
-	error_message(src, loc, lvl, buf);
+	error_message(src, loc, lvl, flags, buf);
 
 	if (lvl == log_error) {
-		if (wk->vm.in_analyzer) {
-			az_set_error();
-		} else {
-			wk->vm.error = true;
-			wk->vm.run = false;
-		}
+		vm_trigger_error(wk);
 	}
 }
 
 void
-vm_diagnostic(struct workspace *wk, uint32_t ip, enum log_level lvl, const char *fmt, ...)
+vm_diagnostic(struct workspace *wk, uint32_t ip, enum log_level lvl, enum error_message_flag flags, const char *fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	vm_diagnostic_v(wk, ip, lvl, fmt, args);
+	vm_diagnostic_v(wk, ip, lvl, flags, fmt, args);
 	va_end(args);
 }
 
@@ -272,7 +304,7 @@ vm_error_at(struct workspace *wk, uint32_t ip, const char *fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	vm_diagnostic_v(wk, ip, log_error, fmt, args);
+	vm_diagnostic_v(wk, ip, log_error, 0, fmt, args);
 	va_end(args);
 }
 
@@ -281,7 +313,7 @@ vm_warning(struct workspace *wk, const char *fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	vm_diagnostic_v(wk, 0, log_warn, fmt, args);
+	vm_diagnostic_v(wk, 0, log_warn, 0, fmt, args);
 	va_end(args);
 }
 
@@ -290,7 +322,7 @@ vm_warning_at(struct workspace *wk, uint32_t ip, const char *fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	vm_diagnostic_v(wk, ip, log_warn, fmt, args);
+	vm_diagnostic_v(wk, ip, log_warn, 0, fmt, args);
 	va_end(args);
 }
 
@@ -299,7 +331,7 @@ vm_error(struct workspace *wk, const char *fmt, ...)
 {
 	va_list args;
 	va_start(args, fmt);
-	vm_diagnostic_v(wk, 0, log_error, fmt, args);
+	vm_diagnostic_v(wk, 0, log_error, 0, fmt, args);
 	va_end(args);
 }
 
@@ -357,7 +389,7 @@ typecheck_and_mutate_function_arg(struct workspace *wk, uint32_t ip, obj *val, t
 	if (!listify && (type == obj_file || (type & tc_file) == tc_file)) {
 		if (t == obj_array && get_obj_array(wk, *val)->len == 1) {
 			obj i0;
-			obj_array_index(wk, *val, 0, &i0);
+			i0 = obj_array_index(wk, *val, 0);
 			if (get_obj_type(wk, i0) == obj_file) {
 				*val = i0;
 			}
@@ -368,7 +400,7 @@ typecheck_and_mutate_function_arg(struct workspace *wk, uint32_t ip, obj *val, t
 
 	if (listify) {
 		obj v, arr;
-		make_obj(wk, &arr, obj_array);
+		arr = make_obj(wk, obj_array);
 
 		if (t == obj_array) {
 			obj_array_flat_for_(wk, *val, v, flat_iter) {
@@ -443,7 +475,7 @@ handle_kwarg(struct workspace *wk, struct args_kw akw[], const char *kw, uint32_
 	}
 
 	if (!akw[i].key) {
-		vm_diagnostic(wk, kw_ip, log_error, "unknown kwarg %s", kw);
+		vm_error_at(wk, kw_ip, "unknown kwarg %s", kw);
 		return false;
 	} else if (akw[i].set && !glob) {
 		vm_error_at(wk, kw_ip, "keyword argument '%s' set twice", kw);
@@ -481,7 +513,7 @@ vm_pop_args(struct workspace *wk, struct args_norm an[], struct args_kw akw[])
 			akw[i].val = 0;
 
 			if (akw[i].type & TYPE_TAG_GLOB) {
-				make_obj(wk, &akw[i].val, obj_dict);
+				akw[i].val = make_obj(wk, obj_dict);
 				akw[i].set = true;
 			}
 		}
@@ -550,7 +582,7 @@ vm_pop_args(struct workspace *wk, struct args_norm an[], struct args_kw akw[])
 			type &= ~TYPE_TAG_GLOB;
 			type |= TYPE_TAG_LISTIFY;
 			an[i].set = true;
-			make_obj(wk, &an[i].val, obj_array);
+			an[i].val = make_obj(wk, obj_array);
 			for (j = i; j < wk->vm.nargs; ++j) {
 				entry = object_stack_peek_entry(&wk->vm.stack, wk->vm.nargs - argi);
 				wk->vm.saw_disabler |= entry->o == obj_disabler;
@@ -774,6 +806,8 @@ vm_dis_inst(struct workspace *wk, uint8_t *code, uint32_t base_ip)
 		break;
 	op_case(op_az_merge)
 		break;
+	op_case(op_dbg_break)
+		break;
 	case op_count: UNREACHABLE;
 	}
 	// clang-format on
@@ -796,8 +830,19 @@ vm_dis(struct workspace *wk)
 		struct source_location loc;
 		struct source *src;
 		vm_lookup_inst_location(&wk->vm, i, &loc, &src);
-		snprintf(loc_buf, sizeof(loc_buf), "%s:%3d:%02d", src ? src->label : 0, loc.off, loc.len);
-		printf("%-*s%s\n", w, dis, loc_buf);
+		struct detailed_source_location dloc = { 0 };
+		if (src) {
+			get_detailed_source_location(src, loc, &dloc, (enum get_detailed_source_location_flag)0);
+		}
+		snprintf(loc_buf,
+			sizeof(loc_buf),
+			"%s:%3d:%02d-[%3d:%02d]",
+			src ? src->label : 0,
+			dloc.line,
+			dloc.col,
+			dloc.end_line,
+			dloc.end_col);
+		log_plain(log_info, "%-*s%s\n", w, dis, loc_buf);
 
 		/* if (src) { */
 		/* 	list_line_range(src, loc, 0); */
@@ -816,6 +861,22 @@ vm_push_call_stack_frame(struct workspace *wk, struct call_frame *frame)
 	arr_push(&wk->vm.call_stack, frame);
 }
 
+struct call_frame *
+vm_pop_call_stack_frame(struct workspace *wk)
+{
+	struct call_frame *frame = arr_pop(&wk->vm.call_stack);
+	switch (frame->type) {
+	case call_frame_type_eval: break;
+	case call_frame_type_func:
+		wk->vm.behavior.pop_local_scope(wk);
+		wk->vm.scope_stack = frame->scope_stack;
+		wk->vm.lang_mode = frame->lang_mode;
+		break;
+	}
+
+	return frame;
+}
+
 static void
 vm_push_dummy(struct workspace *wk)
 {
@@ -873,7 +934,7 @@ vm_execute_capture(struct workspace *wk, obj a)
 		if (capture->func->akw[i].set) {
 			val = capture->func->akw[i].val;
 		} else if (capture->defargs) {
-			const struct str s = WKSTR(capture->func->akw[i].key);
+			const struct str s = STRL(capture->func->akw[i].key);
 			obj_dict_index_strn(wk, capture->defargs, s.s, s.len, &val);
 		}
 
@@ -897,7 +958,7 @@ vm_execute_native(struct workspace *wk, uint32_t func_idx, obj self)
 #ifdef TRACY_ENABLE
 		TracyCZoneC(tctx_func, 0xff5000, true);
 		char func_name[1024];
-		snprintf(func_name, sizeof(func_name), "%s", native_funcs[b].name);
+		snprintf(func_name, sizeof(func_name), "%s", native_funcs[func_idx].name);
 		TracyCZoneName(tctx_func, func_name, strlen(func_name));
 #endif
 
@@ -913,7 +974,12 @@ vm_execute_native(struct workspace *wk, uint32_t func_idx, obj self)
 		if (saw_disabler) {
 			res = obj_disabler;
 		} else {
-			vm_error(wk, "in %s", native_funcs[func_idx].name);
+			if (native_funcs[func_idx].flags & func_impl_flag_throws_error) {
+				vm_trigger_error(wk);
+			} else {
+				vm_error(wk, "in function '%s'", native_funcs[func_idx].name);
+			}
+
 			vm_push_dummy(wk);
 			return;
 		}
@@ -950,7 +1016,7 @@ vm_op_constant_list(struct workspace *wk)
 {
 	obj b;
 	uint32_t i, len = vm_get_constant(wk->vm.code.e, &wk->vm.ip);
-	make_obj(wk, &b, obj_array);
+	b = make_obj(wk, obj_array);
 	for (i = 0; i < len; ++i) {
 		obj_array_push(wk, b, object_stack_peek(&wk->vm.stack, len - i));
 	}
@@ -964,7 +1030,7 @@ vm_op_constant_dict(struct workspace *wk)
 {
 	obj b;
 	uint32_t i, len = vm_get_constant(wk->vm.code.e, &wk->vm.ip);
-	make_obj(wk, &b, obj_dict);
+	b = make_obj(wk, obj_dict);
 	for (i = 0; i < len; ++i) {
 		obj key = object_stack_peek(&wk->vm.stack, (len - i) * 2 - 1);
 		if (wk->vm.in_analyzer && get_obj_type(wk, key) == obj_typeinfo) {
@@ -989,14 +1055,14 @@ vm_op_constant_func(struct workspace *wk)
 	defargs = object_stack_pop(&wk->vm.stack);
 	a = vm_get_constant(wk->vm.code.e, &wk->vm.ip);
 
-	make_obj(wk, &c, obj_capture);
+	c = make_obj(wk, obj_capture);
 	capture = get_obj_capture(wk, c);
 
 	capture->func = get_obj_func(wk, a);
 	capture->scope_stack = wk->vm.behavior.scope_stack_dup(wk, wk->vm.scope_stack);
 	capture->defargs = defargs;
 
-	object_stack_push_ip(wk, c, capture->func->entry);
+	object_stack_push(wk, c);
 }
 
 #define typecheck_operand(__o, __o_type, __expect_type, __expect_tc_type, __result_type) \
@@ -1062,7 +1128,7 @@ vm_op_add(struct workspace *wk)
 	case obj_number: {
 		typecheck_operand(b, b_t, obj_number, tc_number, tc_number);
 
-		make_obj(wk, &res, obj_number);
+		res = make_obj(wk, obj_number);
 		set_obj_number(wk, res, get_obj_number(wk, a) + get_obj_number(wk, b));
 		break;
 	}
@@ -1132,7 +1198,7 @@ type_err:
 	case obj_number: {                                                                                  \
 		typecheck_operand(b, b_t, obj_number, tc_number, tc_number);                                \
                                                                                                             \
-		make_obj(wk, &res, obj_number);                                                             \
+		res = make_obj(wk, obj_number);                                                             \
 		set_obj_number(wk, res, get_obj_number(wk, a) __op get_obj_number(wk, b));                  \
 		break;                                                                                      \
 	}                                                                                                   \
@@ -1187,7 +1253,7 @@ vm_op_div(struct workspace *wk)
 	case obj_number:
 		typecheck_operand(b, b_t, obj_number, tc_number, tc_number);
 
-		make_obj(wk, &res, obj_number);
+		res = make_obj(wk, obj_number);
 		set_obj_number(wk, res, get_obj_number(wk, a) / get_obj_number(wk, b));
 		break;
 	case obj_string: {
@@ -1205,9 +1271,9 @@ vm_op_div(struct workspace *wk)
 			return;
 		}
 
-		SBUF(buf);
+		TSTR(buf);
 		path_join(wk, &buf, ss1->s, ss2->s);
-		res = sbuf_into_str(wk, &buf);
+		res = tstr_into_str(wk, &buf);
 		break;
 	}
 	case obj_typeinfo: {
@@ -1293,12 +1359,20 @@ vm_op_in(struct workspace *wk)
 		typecheck_operand(a, a_t, 0, tc_any, tc_bool);
 
 		res = obj_array_in(wk, b, a) ? obj_bool_true : obj_bool_false;
+
+		if (a_t == obj_string) {
+			check_str_enum(wk, a, a_t, b, b_t);
+		}
 		break;
 	}
 	case obj_dict:
 		typecheck_operand(a, a_t, obj_string, tc_string, tc_bool);
 
 		res = obj_dict_in(wk, b, a) ? obj_bool_true : obj_bool_false;
+
+		if (res == obj_bool_false) {
+			check_str_enum(wk, a, a_t, b, b_t);
+		}
 		break;
 	case obj_string:
 		typecheck_operand(a, a_t, obj_string, tc_string, tc_bool);
@@ -1341,9 +1415,19 @@ vm_op_eq(struct workspace *wk)
 	obj res = 0;
 
 	if (a_t == obj_typeinfo || b_t == obj_typeinfo) {
+		if (a_t == obj_string) {
+			check_str_enum(wk, b, b_t, a, a_t);
+		} else if (b_t == obj_string) {
+			check_str_enum(wk, a, a_t, b, b_t);
+		}
+
 		res = make_typeinfo(wk, tc_bool);
 	} else {
 		res = obj_equal(wk, a, b) ? obj_bool_true : obj_bool_false;
+
+		if (a_t == obj_string && b_t == obj_string && res == obj_bool_false) {
+			check_str_enum(wk, a, a_t, b, b_t);
+		}
 	}
 
 	object_stack_push(wk, res);
@@ -1394,7 +1478,7 @@ vm_op_negate(struct workspace *wk)
 
 	switch (a_t) {
 	case obj_number: {
-		make_obj(wk, &res, obj_number);
+		res = make_obj(wk, obj_number);
 		set_obj_number(wk, res, get_obj_number(wk, a) * -1);
 		break;
 	}
@@ -1425,7 +1509,7 @@ vm_op_stringify(struct workspace *wk)
 	obj res = 0;
 
 	if (get_obj_type(wk, a) == obj_typeinfo) {
-		if (!typecheck_typeinfo(wk, a, tc_bool | tc_file | tc_number | tc_string)) {
+		if (!typecheck_typeinfo(wk, a, tc_bool | tc_file | tc_number | tc_string | tc_feature_opt)) {
 			vm_error(wk, "unable to coerce %s to string", obj_typestr(wk, a));
 		}
 
@@ -1566,7 +1650,7 @@ vm_op_store(struct workspace *wk)
 			assign = true;
 			typecheck_operand(val, val_t, obj_number, tc_number, tc_number);
 
-			make_obj(wk, &res, obj_number);
+			res = make_obj(wk, obj_number);
 			set_obj_number(wk, res, get_obj_number(wk, source) + get_obj_number(wk, val));
 			break;
 		}
@@ -1618,7 +1702,7 @@ type_err:
 			if (member_target) {
 				*member_target = res;
 			} else {
-				wk->vm.behavior.assign_variable(wk, id_str->s, res, 0, assign_reassign);
+				wk->vm.behavior.assign_variable(wk, id_str->s, res, wk->vm.ip - 1, assign_reassign);
 			}
 		}
 
@@ -1645,6 +1729,20 @@ type_err:
 		}
 		case obj_array: {
 			val = obj_array_dup_light(wk, val);
+			break;
+		}
+		case obj_typeinfo: {
+			obj dup = make_obj(wk, obj_typeinfo);
+			*get_obj_typeinfo(wk, dup) = *get_obj_typeinfo(wk, val);
+			val = dup;
+			break;
+		}
+		case obj_capture: {
+			struct obj_capture *c = get_obj_capture(wk, val);
+			if (c->func && !c->func->name) {
+				c->func->name = get_str(wk, id)->s;
+			}
+			break;
 		}
 		default: break;
 		}
@@ -1694,6 +1792,9 @@ vm_op_try_load(struct workspace *wk)
 	if (a == obj_disabler) {
 		object_stack_push(wk, obj_disabler);
 		return;
+	} else if (get_obj_type(wk, a) == obj_typeinfo) {
+		vm_push_dummy(wk);
+		return;
 	}
 
 	if (!wk->vm.behavior.get_variable(wk, get_str(wk, a)->s, &res)) {
@@ -1729,7 +1830,7 @@ vm_op_index(struct workspace *wk)
 			break;
 		}
 
-		obj_array_index(wk, a, i, &res);
+		res = obj_array_index(wk, a, i);
 		break;
 	}
 	case obj_dict: {
@@ -1754,7 +1855,7 @@ vm_op_index(struct workspace *wk)
 			break;
 		}
 
-		obj_array_index(wk, tgt->output, i, &res);
+		res = obj_array_index(wk, tgt->output, i);
 		break;
 	}
 	case obj_string: {
@@ -1788,7 +1889,7 @@ vm_op_index(struct workspace *wk)
 			break;
 		}
 
-		make_obj(wk, &res, obj_number);
+		res = make_obj(wk, obj_number);
 		set_obj_number(wk, res, (i * iter->data.range.step) + iter->data.range.start);
 		break;
 	}
@@ -1838,6 +1939,11 @@ vm_op_member(struct workspace *wk)
 		} else if (typecheck_typeinfo(wk, self, tc_dict)) {
 			vm_push_dummy(wk);
 			return;
+		} else if (wk->vm.in_analyzer && get_obj_type(wk, self) == obj_module
+			   && !get_obj_module(wk, self)->found) {
+			// Don't error on missing functions for not-found modules
+			vm_push_dummy(wk);
+			return;
 		}
 
 		vm_error(wk, "member %o not found on %#o", id, obj_type_to_typestr(wk, self));
@@ -1845,11 +1951,15 @@ vm_op_member(struct workspace *wk)
 		return;
 	}
 
-	obj res;
-	make_obj(wk, &res, obj_capture);
+	obj  res = make_obj(wk, obj_capture);
 	struct obj_capture *c = get_obj_capture(wk, res);
 
 	if (f) {
+		if (get_obj_type(wk, f) == obj_typeinfo) {
+			vm_push_dummy(wk);
+			return;
+		}
+
 		*c = *get_obj_capture(wk, f);
 	} else {
 		c->native_func = idx;
@@ -1872,7 +1982,11 @@ vm_op_call(struct workspace *wk)
 
 	obj f = object_stack_pop(&wk->vm.stack);
 
-	unop_disabler_check(f);
+	if (f == obj_disabler) {
+		object_stack_discard(&wk->vm.stack, wk->vm.nargs + wk->vm.nkwargs * 2);
+		object_stack_push(wk, obj_disabler);
+		return;
+	}
 
 	if (wk->vm.in_analyzer && get_obj_type(wk, f) == obj_typeinfo) {
 		object_stack_discard(&wk->vm.stack, wk->vm.nargs + wk->vm.nkwargs * 2);
@@ -1923,7 +2037,7 @@ vm_op_iterator(struct workspace *wk)
 			goto args_to_unpack_mismatch_error;
 		}
 
-		make_obj(wk, &iter, obj_iterator);
+		iter = make_obj(wk, obj_iterator);
 		object_stack_push(wk, iter);
 		iterator = get_obj_iterator(wk, iter);
 
@@ -1939,7 +2053,7 @@ vm_op_iterator(struct workspace *wk)
 			goto args_to_unpack_mismatch_error;
 		}
 
-		make_obj(wk, &iter, obj_iterator);
+		iter = make_obj(wk, obj_iterator);
 		object_stack_push(wk, iter);
 		iterator = get_obj_iterator(wk, iter);
 
@@ -1992,7 +2106,7 @@ vm_op_iterator(struct workspace *wk)
 			goto args_to_unpack_mismatch_error;
 		}
 
-		make_obj(wk, &iter, obj_iterator);
+		iter = make_obj(wk, obj_iterator);
 		object_stack_push(wk, iter);
 		iterator = get_obj_iterator(wk, iter);
 		iterator->type = obj_iterator_type_typeinfo;
@@ -2017,7 +2131,7 @@ args_to_unpack_mismatch_error:
 		obj_typestr(wk, a));
 
 push_dummy_iterator:
-	make_obj(wk, &iter, obj_iterator);
+	iter = make_obj(wk, obj_iterator);
 	object_stack_push(wk, iter);
 	iterator = get_obj_iterator(wk, iter);
 	iterator->type = obj_iterator_type_typeinfo;
@@ -2049,7 +2163,7 @@ vm_op_iterator_next(struct workspace *wk)
 		if (iterator->data.range.i >= iterator->data.range.stop) {
 			should_break = true;
 		} else {
-			make_obj(wk, &val, obj_number);
+			val = make_obj(wk, obj_number);
 			set_obj_number(wk, val, iterator->data.range.i);
 			iterator->data.range.i += iterator->data.range.step;
 		}
@@ -2083,7 +2197,9 @@ vm_op_iterator_next(struct workspace *wk)
 		}
 		break;
 	case obj_iterator_type_typeinfo: {
-		if (iterator->data.typeinfo.i) {
+		// Let it loop twice to catch all variables modified in this impure
+		// loop.
+		if (iterator->data.typeinfo.i > 1) {
 			should_break = true;
 			break;
 		}
@@ -2220,7 +2336,7 @@ vm_op_return(struct workspace *wk)
 	uint32_t a_ip;
 	obj a;
 
-	struct call_frame *frame = arr_pop(&wk->vm.call_stack);
+	struct call_frame *frame = vm_pop_call_stack_frame(wk);
 	wk->vm.ip = frame->return_ip;
 
 	switch (frame->type) {
@@ -2229,9 +2345,6 @@ vm_op_return(struct workspace *wk)
 		break;
 	}
 	case call_frame_type_func:
-		wk->vm.behavior.pop_local_scope(wk);
-		wk->vm.scope_stack = frame->scope_stack;
-		wk->vm.lang_mode = frame->lang_mode;
 		vm_peek(a, 1);
 		typecheck_custom(wk, a_ip, a, frame->expected_return_type, "expected return type %s, got %s");
 		break;
@@ -2247,6 +2360,16 @@ vm_op_typecheck(struct workspace *wk)
 	typecheck(wk, entry->ip, entry->o, vm_get_constant(wk->vm.code.e, &wk->vm.ip));
 }
 
+static void
+vm_op_dbg_break(struct workspace *wk)
+{
+	if (wk->vm.dbg_state.break_cb) {
+		wk->vm.dbg_state.break_cb(wk);
+	} else {
+		repl(wk, true);
+	}
+}
+
 /******************************************************************************
  * vm_execute
  ******************************************************************************/
@@ -2297,7 +2420,7 @@ vm_eval_capture(struct workspace *wk, obj c, const struct args_norm an[], const
 
 	if (wk->vm.error) {
 		object_stack_pop(&wk->vm.stack);
-		arr_pop(&wk->vm.call_stack);
+		vm_pop_call_stack_frame(wk);
 		goto err;
 	}
 
@@ -2317,23 +2440,28 @@ static void
 vm_unwind_call_stack(struct workspace *wk)
 {
 	struct call_frame *frame;
-	uint32_t prev_err_loc = 0;
+
 	while (wk->vm.call_stack.len) {
-		frame = arr_pop(&wk->vm.call_stack);
+		frame = vm_pop_call_stack_frame(wk);
 
 		switch (frame->type) {
 		case call_frame_type_eval: {
+			error_message_flush_coalesced_message();
 			wk->vm.ip = frame->return_ip;
+			// TODO: this is a little hacky?  We need to make sure that
+			// execution can continue even if we emitted errors.
+			wk->vm.run = true;
 			return;
 		}
 		case call_frame_type_func: break;
 		}
 
+		const char *fmt = frame->func->name ? "in function '%s'" : "in %s";
 		const char *fname = frame->func->name ? frame->func->name : "anonymous function";
-		vm_error_at(wk, prev_err_loc, "in %s", fname);
-
-		prev_err_loc = frame->return_ip;
+		vm_diagnostic(wk, frame->return_ip - 1, log_error, error_message_flag_no_source | error_message_flag_coalesce, fmt, fname);
 	}
+
+	error_message_flush_coalesced_message();
 }
 
 obj
@@ -2423,7 +2551,7 @@ static obj
 vm_scope_stack_dup(struct workspace *wk, obj scope_stack)
 {
 	obj r, v;
-	make_obj(wk, &r, obj_array);
+	r = make_obj(wk, obj_array);
 	obj_array_for(wk, scope_stack, v) {
 		obj scope;
 		obj_dict_dup(wk, v, &scope);
@@ -2436,7 +2564,7 @@ static void
 vm_push_local_scope(struct workspace *wk)
 {
 	obj scope;
-	make_obj(wk, &scope, obj_dict);
+	scope = make_obj(wk, obj_dict);
 	obj_array_push(wk, wk->vm.scope_stack, scope);
 }
 
@@ -2486,8 +2614,7 @@ vm_native_func_dispatch(struct workspace *wk, uint32_t func_idx, obj self, obj *
 
 union vm_breakpoint {
 	struct {
-		obj name;
-		uint32_t line;
+		uint32_t line, col;
 	} dat;
 	int64_t i;
 };
@@ -2500,55 +2627,78 @@ vm_dgb_enable(struct workspace *wk)
 	}
 
 	wk->vm.dbg_state.dbg = true;
-	make_obj(wk, &wk->vm.dbg_state.eval_trace, obj_array);
+	wk->vm.dbg_state.eval_trace = make_obj(wk, obj_array);
 	wk->vm.dbg_state.root_eval_trace = wk->vm.dbg_state.eval_trace;
 }
 
 void
-vm_dbg_push_breakpoint(struct workspace *wk, obj file, uint32_t line)
+vm_dbg_push_breakpoint(struct workspace *wk, obj file, uint32_t line, uint32_t col)
 {
 	vm_dgb_enable(wk);
 
-	union vm_breakpoint breakpoint = { .dat = { file, line } };
-
+	union vm_breakpoint breakpoint = { .dat = { line, col } };
 	if (!wk->vm.dbg_state.breakpoints) {
-		make_obj(wk, &wk->vm.dbg_state.breakpoints, obj_array);
+		wk->vm.dbg_state.breakpoints = make_obj(wk, obj_dict);
 	}
 
-	obj_array_push(wk, wk->vm.dbg_state.breakpoints, make_number(wk, breakpoint.i));
+	obj file_bp;
+	if (!obj_dict_index(wk, wk->vm.dbg_state.breakpoints, file, &file_bp)) {
+		file_bp = make_obj(wk, obj_array);
+		obj_dict_set(wk, wk->vm.dbg_state.breakpoints, file, file_bp);
+	}
+
+	L("pushing breakpoint for %s:%d:%d", get_cstr(wk, file), line, col);
+
+	obj_array_push(wk, file_bp, make_number(wk, breakpoint.i));
 }
 
 void
-vm_dbg_get_breakpoint(struct workspace *wk, obj v, obj *file, uint32_t *line)
+vm_dbg_unpack_breakpoint(struct workspace *wk, obj v, uint32_t *line, uint32_t *col)
 {
 	union vm_breakpoint breakpoint = { .i = get_obj_number(wk, v) };
-	*file = breakpoint.dat.name;
 	*line = breakpoint.dat.line;
+	*col = breakpoint.dat.col;
 }
 
 bool
 vm_dbg_push_breakpoint_str(struct workspace *wk, const char *bp)
 {
 	const char *sep = strchr(bp, ':');
-	obj name, line;
+	obj name;
+	uint32_t line, col = 0;
 	if (sep) {
-		int64_t l;
-		if (!str_to_i(&WKSTR(sep + 1), &l, true)) {
-			LOG_E("invalid line number: %s", sep + 1);
-			return false;
+		const char *sep_col = strchr(sep + 1, ':');
+		struct str l = STRL(sep + 1), c = { 0 };
+
+		if (sep_col) {
+			c = STRL(sep_col + 1);
+			l.len -= (c.len + 1);
 		}
 
-		/* SBUF(p); */
-		/* path_make_absolute(wk, &p, get_cstr(wk, make_strn(wk, bp, sep - bp))); */
-		/* name = sbuf_into_str(wk, &p); */
-		name = make_strn(wk, bp, sep - bp);
-		line = l;
+		int64_t i;
+		if (!str_to_i(&l, &i, true)) {
+			LOG_E("invalid line number: %.*s", l.len, l.s);
+			return false;
+		}
+		line = i;
+
+		if (c.s) {
+			if (!str_to_i(&c, &i, true)) {
+				LOG_E("invalid column: %.*s", c.len, c.s);
+				return false;
+			}
+			col = i;
+		}
+
+		TSTR(p);
+		path_make_absolute(wk, &p, get_cstr(wk, make_strn(wk, bp, sep - bp)));
+		name = tstr_into_str(wk, &p);
 	} else {
 		name = make_str(wk, bp);
 		line = 0;
 	}
 
-	vm_dbg_push_breakpoint(wk, name, line);
+	vm_dbg_push_breakpoint(wk, name, line, col);
 	return true;
 }
 
@@ -2558,7 +2708,6 @@ vm_check_break(struct workspace *wk, uint32_t ip)
 	bool should_break = false;
 
 	struct source *src = 0;
-	uint32_t line = 0, col = 0;
 
 	if (wk->vm.dbg_state.stepping) {
 		struct source_location loc = { 0 };
@@ -2566,47 +2715,17 @@ vm_check_break(struct workspace *wk, uint32_t ip)
 
 		if (wk->vm.dbg_state.prev_source_location.off != loc.off) {
 			wk->vm.dbg_state.prev_source_location = loc;
-
-			struct detailed_source_location dloc;
-			get_detailed_source_location(src, loc, &dloc, 0);
-			line = dloc.line;
-			col = dloc.col;
-
 			should_break = true;
 		}
 	}
 
-	if (!should_break && wk->vm.dbg_state.breakpoints) {
-		struct source_location loc = { 0 };
-		vm_lookup_inst_location(&wk->vm, ip, &loc, &src);
-		struct detailed_source_location dloc;
-		get_detailed_source_location(src, loc, &dloc, 0);
-
-		obj v;
-		obj_array_for(wk, wk->vm.dbg_state.breakpoints, v) {
-			union vm_breakpoint breakpoint = { .i = get_obj_number(wk, v) };
-			const struct str *name = get_str(wk, breakpoint.dat.name);
-
-			printf("checking %s:%d / %s:%d\n", name->s, breakpoint.dat.line, src->label, dloc.line);
-
-			if (breakpoint.dat.line == dloc.line && wk->vm.dbg_state.prev_source_location.off != loc.off
-				&& str_contains(&WKSTR(src->label), name)) {
-				wk->vm.dbg_state.prev_source_location = loc;
-
-				line = breakpoint.dat.line;
-				col = dloc.col;
-				should_break = true;
-			}
-		}
-	}
-
 	if (!should_break && wk->vm.dbg_state.break_after) {
 		should_break = wk->vm.dbg_state.icount >= wk->vm.dbg_state.break_after;
 	}
 
 	if (should_break) {
 		if (wk->vm.dbg_state.break_cb) {
-			wk->vm.dbg_state.break_cb(wk, src, line, col);
+			wk->vm.dbg_state.break_cb(wk);
 		} else {
 			repl(wk, true);
 		}
@@ -2618,10 +2737,10 @@ vm_execute_loop(struct workspace *wk)
 {
 	uint32_t cip;
 	while (wk->vm.run) {
-		if (log_should_print(log_debug)) {
-			/* LL("%-50s", vm_dis_inst(wk, wk->vm.code.e, wk->vm.ip)); */
-			/* object_stack_print(wk, &wk->vm.stack); */
-		}
+		// LL("%-50s", vm_dis_inst(wk, wk->vm.code.e, wk->vm.ip));
+		// object_stack_print(wk, &wk->vm.stack);
+
+		log_progress(wk, wk->vm.ip);
 
 		vm_check_break(wk, wk->vm.ip);
 
@@ -2634,50 +2753,221 @@ vm_execute_loop(struct workspace *wk)
 }
 
 /******************************************************************************
- * struct registration
+ * struct/type registration
  ******************************************************************************/
 
+static obj
+vm_struct_type_dict(struct workspace *wk, enum vm_struct_type base_t)
+{
+	switch (base_t) {
+	case vm_struct_type_enum_: return wk->vm.types.enums;
+	case vm_struct_type_struct_: return wk->vm.types.structs;
+	default: UNREACHABLE_RETURN;
+	}
+}
+
+enum vm_struct_type
+vm_make_struct_type(struct workspace *wk, enum vm_struct_type base_t, const char *name)
+{
+	obj def;
+	if (!obj_dict_index_str(wk, vm_struct_type_dict(wk, base_t), name, &def)) {
+		error_unrecoverable("type %s is not registered", name);
+	}
+
+	return base_t | (def << vm_struct_type_shift);
+}
+
+static void
+vm_types_init(struct workspace *wk)
+{
+	if (!wk->vm.types.structs) {
+		wk->vm.types.structs = make_obj(wk, obj_dict);
+		wk->vm.types.enums = make_obj(wk, obj_dict);
+		wk->vm.types.docs = make_obj(wk, obj_dict);
+		wk->vm.types.top_level_docs = make_obj(wk, obj_dict);
+	}
+}
+
+static bool
+vm_register_common(struct workspace *wk, obj dict, const char *name)
+{
+	obj def;
+	if (obj_dict_index_str(wk, dict, name, &def)) {
+		return false;
+	}
+
+	def = make_obj(wk, obj_dict);
+	obj_dict_set(wk, dict, make_str(wk, name), def);
+	return true;
+}
+
+bool
+vm_enum_(struct workspace *wk, const char *name)
+{
+	vm_types_init(wk);
+	return vm_register_common(wk, wk->vm.types.enums, name);
+}
+
+void
+vm_enum_value_(struct workspace *wk, const char *name, const char *member, uint32_t value)
+{
+	obj def;
+	if (!obj_dict_index_str(wk, wk->vm.types.enums, name, &def)) {
+		error_unrecoverable("enums %s is not registered", name);
+	}
+
+	obj_dict_set(wk, def, make_str(wk, member), value);
+}
+
 bool
 vm_struct_(struct workspace *wk, const char *name)
 {
-	if (!wk->vm.registered_structs) {
-		make_obj(wk, &wk->vm.registered_structs, obj_dict);
+	vm_types_init(wk);
+	return vm_register_common(wk, wk->vm.types.structs, name);
+}
+
+const char *
+vm_enum_docs_def(struct workspace *wk, obj def) {
+	obj doc;
+	if (obj_dict_geti(wk, wk->vm.types.docs, def, &doc)) {
+		return get_str(wk, doc)->s;
+	}
+
+	obj arr = make_obj(wk, obj_array);
+	obj k, v;
+	obj_dict_for(wk, def, k, v) {
+		obj_array_push(wk, arr, k);
+	}
+
+	obj_array_join(wk, false, arr, make_str(wk, "|"), &doc);
+
+	obj_dict_seti(wk, wk->vm.types.docs, def, doc);
+	return get_str(wk, doc)->s;
+}
+
+const char *
+vm_struct_docs_def(struct workspace *wk, obj def) {
+	obj doc;
+	if (obj_dict_geti(wk, wk->vm.types.docs, def, &doc)) {
+		return get_str(wk, doc)->s;
+	}
+
+	obj arr = make_obj(wk, obj_array);
+
+	obj k, v;
+	obj_dict_for(wk, def, k, v) {
+		obj t = obj_array_index(wk, v, 1);
+		enum vm_struct_type type = t & vm_struct_type_mask;
+		const char *type_str = 0;
+
+		switch (type) {
+		case vm_struct_type_struct_:
+			type_str = vm_struct_docs_def(wk, t >> vm_struct_type_shift);
+			break;
+		case vm_struct_type_enum_:
+			type_str = vm_enum_docs_def(wk, t >> vm_struct_type_shift);
+			break;
+		case vm_struct_type_bool:
+			type_str = "bool";
+			break;
+		case vm_struct_type_str:
+			type_str = "str";
+			break;
+		case vm_struct_type_obj:
+			type_str = "any";
+			break;
+		}
+
+		assert(type_str);
+
+		obj_array_push(wk, arr, make_strf(wk, "%s?: %s", get_str(wk, k)->s, type_str));
+	}
+
+	obj joined;
+	obj_array_join(wk, false, arr, make_str(wk, ",\n\t"), &joined);
+
+	TSTR(buf);
+	tstr_pushf(wk, &buf, "{\n\t%s\n}", get_str(wk, joined)->s);
+
+	doc = tstr_into_str(wk, &buf);
+	obj_dict_seti(wk, wk->vm.types.docs, def, doc);
+	return get_str(wk, doc)->s;
+}
+
+const char *
+vm_struct_docs_(struct workspace *wk, const char *name, const char *fmt) {
+	vm_types_init(wk);
+
+	obj doc;
+	if (obj_dict_index_str(wk, wk->vm.types.top_level_docs, name, &doc)) {
+		return get_str(wk, doc)->s;
 	}
 
 	obj def;
-	if (obj_dict_index_str(wk, wk->vm.registered_structs, name, &def)) {
-		return false;
+	if (!obj_dict_index_str(wk, wk->vm.types.structs, name, &def)) {
+		error_unrecoverable("struct %s is not registered", name);
 	}
 
-	make_obj(wk, &def, obj_dict);
-	obj_dict_set(wk, wk->vm.registered_structs, make_str(wk, name), def);
-	return true;
+	const char *doc_str = vm_struct_docs_def(wk, def);
+
+	TSTR(buf);
+	tstr_pushf(wk, &buf, fmt, doc_str);
+
+	doc = tstr_into_str(wk, &buf);
+	obj_dict_set(wk, wk->vm.types.top_level_docs, make_str(wk, name), doc);
+	return get_str(wk, doc)->s;
 }
 
 void
-vm_struct_member_(struct workspace *wk, const char *name, const char *member, uint32_t offset, enum vm_struct_type t)
-{
+vm_struct_member_(struct workspace *wk, const char *name, const char *member, uint32_t offset, enum vm_struct_type t) {
 	obj def;
-	if (!obj_dict_index_str(wk, wk->vm.registered_structs, name, &def)) {
+	if (!obj_dict_index_str(wk, wk->vm.types.structs, name, &def)) {
 		error_unrecoverable("struct %s is not registered", name);
 	}
 
 	obj member_def;
-	make_obj(wk, &member_def, obj_array);
+	member_def = make_obj(wk, obj_array);
 	obj_array_push(wk, member_def, offset);
 	obj_array_push(wk, member_def, t);
 
 	obj_dict_set(wk, def, make_str(wk, member), member_def);
 }
 
-bool
-vm_obj_to_struct_(struct workspace *wk, const char *name, obj o, void *s)
+static bool
+vm_obj_to_enum_def(struct workspace *wk, obj def, obj o, void *s)
 {
-	obj def;
-	if (!obj_dict_index_str(wk, wk->vm.registered_structs, name, &def)) {
-		error_unrecoverable("struct %s is not registered", name);
+	if (!typecheck_custom(wk, 0, o, tc_string, 0)) {
+		vm_error(wk,
+			"expected type %s for enum, got %s",
+			typechecking_type_to_s(wk, tc_string),
+			get_cstr(wk, obj_type_to_typestr(wk, o)));
+		return false;
 	}
 
+	obj v;
+	if (!obj_dict_index(wk, def, o, &v)) {
+		vm_error(wk, "unknown enum value %s", get_cstr(wk, o));
+		return false;
+	}
+
+	*(uint32_t *)s = v;
+	return true;
+}
+
+bool
+vm_obj_to_enum_(struct workspace *wk, const char *name, obj o, void *s)
+{
+	obj def;
+	if (!obj_dict_index_str(wk, wk->vm.types.enums, name, &def)) {
+		error_unrecoverable("enums %s is not registered", name);
+	}
+
+	return vm_obj_to_enum_def(wk, def, o, s);
+}
+
+static bool
+vm_obj_to_struct_def(struct workspace *wk, obj def, obj o, void *s)
+{
 	type_tag expected_type;
 	obj k, v, member_def;
 	obj_dict_for(wk, o, k, v) {
@@ -2687,11 +2977,23 @@ vm_obj_to_struct_(struct workspace *wk, const char *name, obj o, void *s)
 		}
 
 		obj offset, t;
-		obj_array_index(wk, member_def, 0, &offset);
-		obj_array_index(wk, member_def, 1, &t);
+		offset = obj_array_index(wk, member_def, 0);
+		t = obj_array_index(wk, member_def, 1);
 		char *dest = (char *)s + offset;
 
-		switch ((enum vm_struct_type)t) {
+		enum vm_struct_type type = t & vm_struct_type_mask;
+
+		switch (type) {
+		case vm_struct_type_struct_:
+			if (!vm_obj_to_struct_def(wk, t >> vm_struct_type_shift, v, dest)) {
+				return false;
+			}
+			break;
+		case vm_struct_type_enum_:
+			if (!vm_obj_to_enum_def(wk, t >> vm_struct_type_shift, v, dest)) {
+				return false;
+			}
+			break;
 		case vm_struct_type_bool:
 			expected_type = tc_bool;
 			if (!typecheck_custom(wk, 0, v, expected_type, 0)) {
@@ -2704,7 +3006,7 @@ vm_obj_to_struct_(struct workspace *wk, const char *name, obj o, void *s)
 			if (!typecheck_custom(wk, 0, v, expected_type, 0)) {
 				goto type_err;
 			}
-			*(const char **)dest = get_cstr(wk, v);
+			*(const struct str **)dest = get_str(wk, v);
 			break;
 		case vm_struct_type_obj: *(obj *)dest = v; break;
 		}
@@ -2714,14 +3016,23 @@ vm_obj_to_struct_(struct workspace *wk, const char *name, obj o, void *s)
 
 type_err:
 	vm_error(wk,
-		"expected type %s for %s member %s, got %s",
+		"expected type %s for member %s, got %s",
 		typechecking_type_to_s(wk, expected_type),
-		name,
 		get_cstr(wk, k),
 		get_cstr(wk, obj_type_to_typestr(wk, v)));
 	return false;
 }
 
+bool
+vm_obj_to_struct_(struct workspace *wk, const char *name, obj o, void *s)
+{
+	obj def;
+	if (!obj_dict_index_str(wk, wk->vm.types.structs, name, &def)) {
+		error_unrecoverable("struct %s is not registered", name);
+	}
+
+	return vm_obj_to_struct_def(wk, def, o, s);
+}
 
 /******************************************************************************
  * init / destroy
@@ -2782,6 +3093,7 @@ vm_init_objects(struct workspace *wk)
 
 	hash_init(&wk->vm.objects.obj_hash, 128, sizeof(obj));
 	hash_init_str(&wk->vm.objects.str_hash, 128);
+	hash_init_str(&wk->vm.objects.dedup_str_hash, 128);
 
 	make_default_objects(wk);
 }
@@ -2857,6 +3169,7 @@ vm_init(struct workspace *wk)
 					      [op_dup] = vm_op_dup,
 					      [op_swap] = vm_op_swap,
 					      [op_typecheck] = vm_op_typecheck,
+					      [op_dbg_break] = vm_op_dbg_break,
 				      } };
 
 	/* objects */
@@ -2866,26 +3179,34 @@ vm_init(struct workspace *wk)
 	build_func_impl_tables();
 
 	/* default scope */
-	make_obj(wk, &wk->vm.default_scope_stack, obj_array);
+	wk->vm.default_scope_stack = make_obj(wk, obj_array);
 	obj scope;
-	make_obj(wk, &scope, obj_dict);
+	scope = make_obj(wk, obj_dict);
 	obj_array_push(wk, wk->vm.default_scope_stack, scope);
 
 	obj_dict_set(wk, scope, make_str(wk, "meson"), obj_meson);
 
 	obj id;
-	make_obj(wk, &id, obj_machine);
+	id = make_obj(wk, obj_machine);
 	set_obj_machine(wk, id, machine_kind_build);
 	obj_dict_set(wk, scope, make_str(wk, "build_machine"), id);
 
-	make_obj(wk, &id, obj_machine);
+	id = make_obj(wk, obj_machine);
 	set_obj_machine(wk, id, machine_kind_host);
 	obj_dict_set(wk, scope, make_str(wk, "host_machine"), id);
 	obj_dict_set(wk, scope, make_str(wk, "target_machine"), id);
 
 	/* module cache */
-	make_obj(wk, &wk->vm.modules, obj_dict);
+	wk->vm.modules = make_obj(wk, obj_dict);
 
+	/* enum types */
+	wk->vm.objects.enums.values = make_obj(wk, obj_dict);
+	wk->vm.objects.enums.types = make_obj(wk, obj_dict);
+
+	/* complex type cache */
+	wk->vm.objects.complex_types = make_obj(wk, obj_dict);
+
+	/* scope stack */
 	wk->vm.scope_stack = wk->vm.behavior.scope_stack_dup(wk, wk->vm.default_scope_stack);
 
 	/* initial code segment */
@@ -2921,6 +3242,7 @@ vm_destroy_objects(struct workspace *wk)
 
 	hash_destroy(&wk->vm.objects.obj_hash);
 	hash_destroy(&wk->vm.objects.str_hash);
+	hash_destroy(&wk->vm.objects.dedup_str_hash);
 }
 
 void
