@@ -20,11 +20,12 @@
 #include "error.h"
 #include "external/libarchive.h"
 #include "external/libcurl.h"
-#include "external/libpkgconf.h"
+#include "external/pkgconfig.h"
 #include "external/samurai.h"
 #include "lang/analyze.h"
 #include "lang/fmt.h"
 #include "lang/func_lookup.h"
+#include "lang/lsp.h"
 #include "lang/object_iterators.h"
 #include "lang/parser.h"
 #include "lang/serial.h"
@@ -79,11 +80,12 @@ static void
 setup_platform_env(const char *build_dir, bool force)
 {
 	if (host_machine.sys == machine_system_windows) {
-		SBUF_manual(cache);
+		TSTR_manual(cache);
 		const char *cache_path = 0;
 		if (build_dir) {
 			path_copy(0, &cache, build_dir);
 			path_push(0, &cache, output_path.private_dir);
+			fs_mkdir_p(cache.buf);
 			if (fs_dir_exists(cache.buf)) {
 				path_push(0, &cache, "vsenv.txt");
 				cache_path = cache.buf;
@@ -212,40 +214,68 @@ ret:
 	return ret;
 }
 
+static bool
+language_mode_from_optarg(const char *arg, enum language_mode *langmode)
+{
+	struct {
+		const char *short_name, *long_name;
+		enum language_mode mode;
+		bool fmt;
+	} modes[] = {
+		{ "n", "normal", language_external },
+		{ "s", "script", language_internal },
+		{ "m", "module", language_extended },
+	};
+
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(modes); ++i) {
+		if (strcmp(arg, modes[i].short_name) == 0 || strcmp(arg, modes[i].long_name) == 0) {
+			*langmode = modes[i].mode;
+			return true;
+		}
+	}
+
+	LOG_E("invalid language mode: %s", arg);
+	LOG_I("supported language modes are:");
+	for (i = 0; i < ARRAY_LEN(modes); ++i) {
+		LOG_I("  - %s | %s", modes[i].short_name, modes[i].long_name);
+	}
+	return false;
+}
+
 static bool
 cmd_check(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct {
 		const char *filename;
+		const char *breakpoint;
 		bool print_ast, print_dis;
 		enum vm_compile_mode compile_mode;
 	} opts = { 0 };
 
-	OPTSTART("pdm:") {
+	OPTSTART("pdm:b:f") {
 	case 'p': opts.print_ast = true; break;
 	case 'd': opts.print_dis = true; break;
+	case 'b': opts.breakpoint = optarg; break;
 	case 'm': {
-		const char *p, *chars = "xf";
-		for (p = optarg; *p; ++p) {
-			if (!strchr(chars, *p)) {
-				LOG_E("invalid mode '%c', must be one of %s", *p, chars);
-				return false;
-			}
+		enum language_mode mode;
+		if (!language_mode_from_optarg(optarg, &mode)) {
+			return false;
+		}
 
-			switch (*p) {
-			case 'x': opts.compile_mode |= vm_compile_mode_language_extended; break;
-			case 'f': opts.compile_mode |= vm_compile_mode_fmt; break;
-			default: UNREACHABLE;
-			}
+		if (mode == language_internal || mode == language_extended) {
+			opts.compile_mode |= vm_compile_mode_language_extended;
 		}
 		break;
 	}
+	case 'f': opts.compile_mode |= vm_compile_mode_fmt; break;
 	}
 	OPTEND(argv[argi],
 		" <filename>",
 		"  -p - print parsed ast\n"
 		"  -d - print dissasembly\n"
-		"  -m <mode> - parse with parse mode <mode>\n",
+		"  -m <mode> - parse with language mode <mode>\n"
+		"  -f - parse in formatting mode\n",
 		NULL,
 		1)
 
@@ -263,6 +293,12 @@ cmd_check(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		goto ret;
 	}
 
+	if (opts.breakpoint) {
+		if (!vm_dbg_push_breakpoint_str(&wk, opts.breakpoint)) {
+			goto ret;
+		}
+	}
+
 	if (opts.print_ast) {
 		struct node *n;
 		if (!(n = parse(&wk, src, opts.compile_mode))) {
@@ -295,26 +331,45 @@ ret:
 static bool
 cmd_analyze(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
-	struct az_opts opts = {
-		.enabled_diagnostics = az_diagnostic_unused_variable | az_diagnostic_dead_code
-				       | az_diagnostic_redirect_script_error,
+	struct {
+		bool subdir_error;
+		enum error_diagnostic_store_replay_opts replay_opts;
+		const char *file_override;
+		uint64_t enabled_diagnostics;
+		enum {
+			action_trace,
+			action_lsp,
+			action_determine_root,
+			action_file,
+			action_default,
+			action_count,
+		} action;
+		enum language_mode lang_mode;
+	} opts = {
+		.enabled_diagnostics = az_diagnostic_unused_variable | az_diagnostic_dead_code,
+		.action = action_default,
+		.lang_mode = language_external,
 	};
 
-	enum {
-		action_trace,
-		action_define,
-		action_default,
-	} action
-		= action_default;
-
 	static const struct command commands[] = {
 		[action_trace] = { "trace", 0, "print a tree of all meson source files that are evaluated" },
-		[action_define] = { "define <var>", 0, "lookup the definition of a variable" },
+		[action_lsp] = { "lsp", 0, "run in lsp mode" },
+		[action_determine_root] = { "root-for", 0, "determine the project root given a meson file" },
+		[action_file] = { "file", 0, "analyze a single file.  implies -m module." },
 		0,
 	};
+	static const int32_t command_args[action_count] = {
+		[action_determine_root] = 1,
+		[action_file] = 1,
+	};
 
-	OPTSTART("luqO:W:i:") {
-	case 'i': opts.internal_file = optarg; break;
+	OPTSTART("luqO:W:m:s") {
+	case 'm': {
+		if (!language_mode_from_optarg(optarg, &opts.lang_mode)) {
+			return false;
+		}
+		break;
+	}
 	case 'l':
 		opts.subdir_error = true;
 		opts.replay_opts |= error_diagnostic_store_replay_dont_include_sources;
@@ -324,7 +379,7 @@ cmd_analyze(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 	case 'W': {
 		bool enable = true;
 		const char *name = optarg;
-		if (str_startswith(&WKSTR(optarg), &WKSTR("no-"))) {
+		if (str_startswith(&STRL(optarg), &STR("no-"))) {
 			enable = false;
 			name += 3;
 		}
@@ -354,63 +409,81 @@ cmd_analyze(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		"",
 		"  -l - optimize output for editor linter plugins\n"
 		"  -q - only report errors\n"
+		"  -m <mode> - analyze with language mode <mode>\n"
 		"  -O <path> - read project file with matching path from stdin\n"
-		"  -i <path> - analyze the single file <path> in internal mode\n"
 		"  -W [no-]<diagnostic> - enable or disable diagnostics\n"
 		"  -W list - list available diagnostics\n"
 		"  -W error - turn all warnings into errors\n",
 		commands,
 		-1);
 
-	uint32_t cmd_i = action_default;
-	if (!find_cmd(commands, &cmd_i, argc, argi, argv, true)) {
-		return false;
-	}
-	if (cmd_i != action_default) {
-		++argi;
-	}
-	action = cmd_i;
-
-	if (action == action_define) {
-		if (!check_operands(argc, argi, 1)) {
+	{
+		// Determine "action".  This is basically a subcommand but is allowed
+		// to be empty
+		uint32_t cmd_i = action_default;
+		if (!find_cmd(commands, &cmd_i, argc, argi, argv, true)) {
 			return false;
 		}
+		if (cmd_i != action_default) {
+			++argi;
+		}
+		opts.action = cmd_i;
+
+		if (!check_operands(argc, argi, command_args[opts.action])) {
+			return false;
+		}
+	}
+
+	if (opts.action == action_lsp) {
+		struct az_opts az_opts = {
+			.enabled_diagnostics = opts.enabled_diagnostics,
+		};
+		return analyze_server(&az_opts);
+	} else if (opts.action == action_determine_root) {
+		struct workspace wk;
+		workspace_init_bare(&wk);
+		const char *root = determine_project_root(&wk, argv[argi]);
+		if (root) {
+			printf("%s\n", root);
+		}
+		workspace_destroy(&wk);
+
+		return root ? true : false;
 	} else {
-		if (!check_operands(argc, argi, 0)) {
-			return false;
+		struct workspace wk;
+		workspace_init_bare(&wk);
+
+		const char *single_file = 0;
+		if (opts.action == action_file) {
+			single_file = argv[argi];
+			if (opts.lang_mode == language_external) {
+				opts.lang_mode = language_extended;
+			}
 		}
-	}
 
-	switch (action) {
-	case action_default: break;
-	case action_trace: {
-		opts.eval_trace = true;
-		break;
-	}
-	case action_define: {
-		opts.get_definition_for = argv[argi];
-		break;
-	}
-	}
+		struct az_opts az_opts;
+		analyze_opts_init(&wk, &az_opts);
+		az_opts.eval_trace = opts.action == action_trace;
+		az_opts.subdir_error = opts.subdir_error;
+		az_opts.replay_opts = opts.replay_opts;
+		az_opts.single_file = single_file;
+		az_opts.enabled_diagnostics = opts.enabled_diagnostics;
+		az_opts.auto_chdir_root = true;
+		az_opts.lang_mode = opts.lang_mode;
 
-	if (opts.internal_file && opts.file_override) {
-		LOG_E("-i and -O are mutually exclusive");
-		return false;
-	}
+		bool res = true;
+		if (opts.file_override) {
+			res = analyze_opts_push_override(&wk, &az_opts, opts.file_override, "-", 0);
+		}
 
-	SBUF_manual(abs);
-	if (opts.file_override) {
-		path_make_absolute(NULL, &abs, opts.file_override);
-		opts.file_override = abs.buf;
-	}
+		if (res) {
+			res = do_analyze(&wk, &az_opts);
+		}
 
-	bool res;
-	res = do_analyze(&opts);
-
-	if (opts.file_override) {
-		sbuf_destroy(&abs);
+		workspace_destroy(&wk);
+		analyze_opts_destroy(&wk, &az_opts);
+		return res;
 	}
-	return res;
 }
 
 static bool
@@ -442,7 +515,7 @@ cmd_summary(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		return false;
 	}
 
-	SBUF_manual(path);
+	TSTR_manual(path);
 	path_join(0, &path, output_path.private_dir, output_path.summary);
 
 	bool ret = false;
@@ -455,7 +528,7 @@ cmd_summary(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 
 	ret = true;
 ret:
-	sbuf_destroy(&path);
+	tstr_destroy(&path);
 	fs_source_destroy(&src);
 	return ret;
 }
@@ -488,11 +561,12 @@ cmd_eval(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct workspace wk;
 	workspace_init_bare(&wk);
+	workspace_setup_paths(&wk, path_cwd(), argv[0], argc, argv);
 
-	const char *filename;
+	const char *string_src = 0;
 	bool embedded = false;
 
-	OPTSTART("esb:") {
+	OPTSTART("esb:c:") {
 	case 'e': embedded = true; break;
 	case 's': {
 		wk.vm.disable_fuzz_unsafe_functions = true;
@@ -502,41 +576,56 @@ cmd_eval(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		vm_dbg_push_breakpoint_str(&wk, optarg);
 		break;
 	}
+	case 'c': {
+		string_src = optarg;
+		break;
+	}
 	}
 	OPTEND(argv[argi],
 		" <filename> [args]",
 		"  -e - lookup <filename> as an embedded script\n"
-		"  -s - disable functions that are unsafe to be called at random\n",
+		"  -s - disable functions that are unsafe to be called at random\n"
+		"  -c cmd - evaluate program passed in as string\n",
 		NULL,
 		-1)
 
-	if (argi >= argc) {
-		LOG_E("missing required filename argument");
-		return false;
-	}
-
-	filename = argv[argi];
-
 	bool ret = false;
 
 	struct source src = { 0 };
 
 	wk.vm.lang_mode = language_internal;
 
-	if (embedded) {
-		if (!(embedded_get(filename, &src))) {
-			LOG_E("failed to find '%s' in embedded sources", filename);
-			goto ret;
+	if (string_src) {
+		if (!check_operands(argc, argi, 0)) {
+			return false;
 		}
+
+		src.label = "commandline";
+		src.src = string_src;
+		src.len = strlen(string_src);
 	} else {
-		if (!fs_read_entire_file(filename, &src)) {
-			goto ret;
+		if (argi >= argc) {
+			LOG_E("missing required filename argument");
+			return false;
+		}
+
+		const char *filename = 0;
+		filename = argv[argi];
+		if (embedded) {
+			if (!(embedded_get(filename, &src))) {
+				LOG_E("failed to find '%s' in embedded sources", filename);
+				goto ret;
+			}
+		} else {
+			if (!fs_read_entire_file(filename, &src)) {
+				goto ret;
+			}
 		}
 	}
 
 	{ // populate argv array
 		obj argv_obj;
-		make_obj(&wk, &argv_obj, obj_array);
+		argv_obj = make_obj(&wk, obj_array);
 		wk.vm.behavior.assign_variable(&wk, "argv", argv_obj, 0, assign_local);
 
 		uint32_t i;
@@ -739,7 +828,7 @@ cmd_dump_toolchains(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[]
 			}
 		} else if (strcmp(optarg, "i1") == 0) {
 			int64_t res;
-			if (!str_to_i(&WKSTR(sep), &res, false)) {
+			if (!str_to_i(&STRL(sep), &res, false)) {
 				LOG_E("invalid value for integer: %s", sep);
 				return false;
 			}
@@ -851,7 +940,11 @@ cmd_test(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct test_options test_opts = { 0 };
 
-	OPTSTART("bs:d:Sfj:lvRe:o:") {
+	OPTSTART("abs:d:Sfj:lvRe:o:") {
+	case 'a': {
+		test_opts.include_subprojects = true;
+		break;
+	}
 	case 'b': {
 		test_opts.cat = test_category_benchmark;
 		test_opts.print_summary = true;
@@ -910,6 +1003,7 @@ cmd_test(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 	}
 	OPTEND(argv[argi],
 		" [test [test [...]]]",
+		"  -a - include all tests from all projects\n"
 		"  -b - run benchmarks instead of tests\n"
 		"  -d <mode> - change progress display mode (auto|dots|bar)\n"
 		"  -o <mode> - set output mode (term|html|json)\n"
@@ -964,12 +1058,21 @@ cmd_install(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 static void
 cmd_setup_help(void)
 {
-	log_plain("\n");
+	log_plain(log_info, "\n");
 
 	struct list_options_opts list_opts = { 0 };
 	list_options(&list_opts);
 
-	log_plain("To see all options, including builtin options, use `muon options -a`.\n");
+	log_plain(log_info, "To see all options, including builtin options, use `muon options -a`.\n");
+}
+
+static void
+make_argv0_absolute(struct workspace *wk, struct tstr *buf, char *const argv[])
+{
+	if (!path_is_basename(argv[0])) {
+		path_make_absolute(wk, buf, argv[0]);
+		((char **)argv)[0] = buf->buf;
+	}
 }
 
 static bool
@@ -983,7 +1086,8 @@ cmd_setup(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 
 	uint32_t original_argi = argi + 1;
 
-	OPTSTART("D:b:") {
+	OPTSTART("D:b:#") {
+	case '#': log_progress_enable(); break;
 	case 'D':
 		if (!parse_and_set_cmdline_option(&wk, optarg)) {
 			goto ret;
@@ -994,9 +1098,45 @@ cmd_setup(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		break;
 	}
 	}
-	OPTEND_CUSTOM(argv[argi], " <build dir>", "  -D <option>=<value> - set options\n", NULL, 1, cmd_setup_help())
+	OPTEND_CUSTOM(argv[argi],
+		" <build dir|source dir>",
+		"  -D <option>=<value> - set options\n"
+		"  -# - enable setup progress bar\n",
+		NULL,
+		1,
+		cmd_setup_help())
 
 	const char *build = argv[argi];
+
+	// The following shenanigans are to support passing the source dir instead
+	// of the build dir.  We decide that the passed dir is a source dir (and
+	// the build dir is the current dir) if the current dir does not contain a
+	// build file.
+	TSTR(argv0);
+	TSTR(new_cwd);
+	TSTR(old_cwd);
+	{
+		path_copy_cwd(&wk, &old_cwd);
+
+		enum build_language _lang;
+		if (!determine_build_file(&wk, path_cwd(), &_lang, true))
+		{
+			// fix argv0 here since if it is a relative path it will be
+			// wrong after chdir
+			make_argv0_absolute(&wk, &argv0, argv);
+
+			if (!path_chdir(build)) {
+				return false;
+			}
+
+			path_copy_cwd(&wk, &new_cwd);
+			wk.source_root = new_cwd.buf;
+			build = old_cwd.buf;
+
+			((const char **)argv)[argi] = build;
+		}
+	}
+
 	++argi;
 
 	// Extract any relevant -D options that need to be handled very early.
@@ -1013,8 +1153,11 @@ cmd_setup(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 				continue;
 			}
 
-			if (str_eql(&WKSTR("vsenv"), get_str(&wk, oo->name))) {
-				opts.vsenv_force = str_eql(&WKSTR("true"), get_str(&wk, oo->val));
+			const struct str *k = get_str(&wk, oo->name);
+			const struct str *v = get_str(&wk, oo->val);
+
+			if (str_eql(&STR("vsenv"), k)) {
+				opts.vsenv_force = str_eql(&STR("true"), v);
 			}
 		}
 	}
@@ -1130,7 +1273,6 @@ cmd_version(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		bool enabled;
 	} feature_names[] = {
 		{ "libcurl", have_libcurl },
-		{ "libpkgconf", have_libpkgconf },
 		{ "libarchive", have_libarchive },
 		{ "samurai", have_samurai },
 #ifdef TRACY_ENABLE
@@ -1145,6 +1287,14 @@ cmd_version(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 		}
 	}
 
+	muon_pkgconfig_init(0);
+
+	for (i = 0; i < ARRAY_LEN(pkgconfig_impls); ++i) {
+		if (pkgconfig_impls[i].get_variable) {
+			printf("  pkgconfig:%s\n", muon_pkgconfig_impl_type_to_s(i));
+		}
+	}
+
 	return true;
 }
 
@@ -1185,11 +1335,44 @@ cmd_ui(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
 	return ui_main();
 }
 
+static bool
+cmd_devenv(void *_ctx, uint32_t argc, uint32_t argi, char *const argv[])
+{
+	OPTSTART("") {
+	}
+	OPTEND(argv[argi], "", "", NULL, -1)
+
+	if (argi >= argc) {
+		LOG_E("missing command");
+		return false;
+	}
+
+	if (!ensure_in_build_dir()) {
+		return false;
+	}
+
+	setup_platform_env(".", true);
+
+	const char *const *cmd = (const char *const *)&argv[argi];
+
+	struct run_cmd_ctx ctx = { 0 };
+	ctx.flags |= run_cmd_ctx_flag_dont_capture;
+
+	if (!run_cmd_argv(&ctx, (char *const *)cmd, 0, 0)) {
+		LOG_E("failed to run command: %s", ctx.err_msg);
+		return false;
+		;
+	}
+
+	exit(ctx.status);
+}
+
 static bool
 cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[])
 {
 	const struct command commands[] = {
 		{ "analyze", cmd_analyze, "run a static analyzer" },
+		{ "devenv", cmd_devenv, "run commands in developer environment" },
 		{ "fmt", cmd_format, "format meson source file" },
 		{ "info", cmd_info, NULL },
 		{ "install", cmd_install, "install files" },
@@ -1206,17 +1389,15 @@ cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[])
 	};
 
 	bool res = false;
-	SBUF_manual(argv0);
+	TSTR_manual(argv0);
 
-	OPTSTART("vC:") {
+	OPTSTART("vqC:") {
 	case 'v': log_set_lvl(log_debug); break;
+	case 'q': log_set_lvl(log_error); break;
 	case 'C': {
 		// fix argv0 here since if it is a relative path it will be
 		// wrong after chdir
-		if (!path_is_basename(argv[0])) {
-			path_make_absolute(NULL, &argv0, argv[0]);
-			argv[0] = argv0.buf;
-		}
+		make_argv0_absolute(0, &argv0, argv);
 
 		if (!path_chdir(optarg)) {
 			return false;
@@ -1227,6 +1408,7 @@ cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[])
 	OPTEND(argv[0],
 		"",
 		"  -v - turn on debug messages\n"
+		"  -q - silence logging except for errors\n"
 		"  -C <path> - chdir to path\n",
 		commands,
 		-1)
@@ -1239,7 +1421,7 @@ cmd_main(void *_ctx, uint32_t argc, uint32_t argi, char *argv[])
 	res = commands[cmd_i].cmd(0, argc, argi, argv);
 
 ret:
-	sbuf_destroy(&argv0);
+	tstr_destroy(&argv0);
 	return res;
 }
 
@@ -1248,7 +1430,7 @@ main(int argc, char *argv[])
 {
 	platform_init();
 
-	log_init();
+	log_set_file(stdout);
 	log_set_lvl(log_info);
 
 	path_init();
@@ -1260,10 +1442,10 @@ main(int argc, char *argv[])
 	bool meson_compat = false;
 
 	{
-		SBUF(basename);
+		TSTR(basename);
 		path_basename(NULL, &basename, argv[0]);
 		meson_compat = strcmp(basename.buf, "meson") == 0 && (argc < 2 || strcmp(argv[1], "internal") != 0);
-		sbuf_destroy(&basename);
+		tstr_destroy(&basename);
 	}
 
 	if (meson_compat) {
