@@ -35,6 +35,7 @@ const char *build_option_type_to_s[build_option_type_count] = {
 	[op_integer] = "integer",
 	[op_array] = "array",
 	[op_feature] = "feature",
+	[op_shell_array] = "shell_array",
 };
 
 static bool
@@ -43,6 +44,9 @@ parse_config_string(struct workspace *wk, const struct str *ss, struct option_ov
 	if (str_has_null(ss)) {
 		LOG_E("option cannot contain NUL");
 		return false;
+	} else if (!key_only && !memchr(ss->s, '=', ss->len)) {
+		LOG_E("option must contain =, got %s", ss->s);
+		return false;
 	}
 
 	struct str subproject = { 0 }, key = { 0 }, val = { 0 }, cur = { 0 };
@@ -93,9 +97,6 @@ parse_config_string(struct workspace *wk, const struct str *ss, struct option_ov
 	} else if (key_only && val.len) {
 		LOG_E("unexpected '=' in option '%s'", ss->s);
 		return false;
-	} else if (!key_only && !val.len) {
-		LOG_E("expected '=' in option '%s'", ss->s);
-		return false;
 	}
 
 	oo->name = make_strn(wk, key.s, key.len);
@@ -120,15 +121,15 @@ subproj_name_matches(struct workspace *wk, const char *name, const char *test)
 }
 
 static void
-log_option_override(struct workspace *wk, struct option_override *oo)
+log_option_override(enum log_level lvl, struct workspace *wk, struct option_override *oo)
 {
-	log_plain("'");
+	log_plain(lvl, "'");
 	if (oo->proj) {
-		log_plain("%s:", get_cstr(wk, oo->proj));
+		log_plain(lvl, "%s:", get_cstr(wk, oo->proj));
 	}
 
-	obj_lprintf(wk, "%s=%#o", get_cstr(wk, oo->name), oo->val);
-	log_plain("'");
+	obj_lprintf(wk, lvl, "%s=%#o", get_cstr(wk, oo->name), oo->val);
+	log_plain(lvl, "'");
 }
 
 bool
@@ -161,8 +162,8 @@ check_invalid_subproject_option(struct workspace *wk)
 
 		if (!found) {
 			LLOG_E("invalid option: ");
-			log_option_override(wk, oo);
-			log_plain(" (no such subproject)\n");
+			log_option_override(log_error, wk, oo);
+			log_plain(log_error, " (no such subproject)\n");
 			ret = false;
 		}
 	}
@@ -175,18 +176,18 @@ coerce_feature_opt(struct workspace *wk, uint32_t ip, const struct str *val, obj
 {
 	enum feature_opt_state f;
 
-	if (str_eql(val, &WKSTR("auto"))) {
+	if (str_eql(val, &STR("auto"))) {
 		f = feature_opt_auto;
-	} else if (str_eql(val, &WKSTR("enabled"))) {
+	} else if (str_eql(val, &STR("enabled"))) {
 		f = feature_opt_enabled;
-	} else if (str_eql(val, &WKSTR("disabled"))) {
+	} else if (str_eql(val, &STR("disabled"))) {
 		f = feature_opt_disabled;
 	} else {
 		vm_error_at(wk, ip, "unable to coerce '%s' into a feature", val->s);
 		return false;
 	}
 
-	make_obj(wk, res, obj_feature_opt);
+	*res = make_obj(wk, obj_feature_opt);
 	set_obj_feature_opt(wk, *res, f);
 	return true;
 }
@@ -203,10 +204,13 @@ check_deprecated_option_iter(struct workspace *wk, void *_ctx, obj old, obj new)
 	struct check_deprecated_option_ctx *ctx = _ctx;
 
 	switch (ctx->opt->type) {
+	case op_shell_array: {
+		break;
+	}
 	case op_array: {
 		uint32_t idx;
 		if (obj_array_index_of(wk, *ctx->val, old, &idx)) {
-			vm_warning(wk, "option value %o is deprecated", old);
+			vm_warning_at(wk, ctx->opt->ip, "option value %o is deprecated", old);
 
 			if (new) {
 				obj_array_set(wk, *ctx->val, idx, new);
@@ -216,7 +220,7 @@ check_deprecated_option_iter(struct workspace *wk, void *_ctx, obj old, obj new)
 	}
 	default:
 		if (str_eql(get_str(wk, ctx->sval), get_str(wk, old))) {
-			vm_warning(wk, "option value %o is deprecated", old);
+			vm_warning_at(wk, ctx->opt->ip, "option value %o is deprecated", old);
 
 			if (new) {
 				*ctx->val = new;
@@ -239,13 +243,13 @@ check_deprecated_option(struct workspace *wk, struct obj_option *opt, obj sval,
 	switch (get_obj_type(wk, opt->deprecated)) {
 	case obj_bool:
 		if (get_obj_bool(wk, opt->deprecated)) {
-			vm_warning(wk, "option %o is deprecated", ctx.opt->name);
+			vm_warning_at(wk, opt->ip, "option %o is deprecated", ctx.opt->name);
 		}
 		break;
 	case obj_string: {
 		struct project *cur_proj = current_project(wk);
 
-		vm_warning(wk, "option %o is deprecated to %o", opt->name, opt->deprecated);
+		vm_warning_at(wk, opt->ip, "option %o is deprecated to %o", opt->name, opt->deprecated);
 
 		obj newopt;
 		if (get_option(wk, cur_proj, get_str(wk, opt->deprecated), &newopt)) {
@@ -275,18 +279,22 @@ coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, o
 {
 	const struct str *val = get_str(wk, sval);
 	*res = 0;
-	if (opt->type == op_array) {
+	if (opt->type == op_array || opt->type == op_shell_array) {
 		// coerce array early so that its elements may be checked for deprecation
 		if (!val->len) {
 			// make -Doption= equivalent to an empty list
-			make_obj(wk, res, obj_array);
+			*res = make_obj(wk, obj_array);
 		} else if (val->s[0] == '[') {
 			if (!eval_str(wk, val->s, eval_mode_repl, res)) {
 				LOG_E("malformed array option value '%s'", val->s);
 				return false;
 			}
 		} else {
-			*res = str_split(wk, val, &WKSTR(","));
+			if (opt->type == op_shell_array) {
+				*res = str_shell_split(wk, val, shell_type_for_host_machine());
+			} else {
+				*res = str_split(wk, val, &STR(","));
+			}
 		}
 	}
 
@@ -310,9 +318,9 @@ coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, o
 	case op_string: *res = sval; break;
 	case op_boolean: {
 		bool b;
-		if (str_eql(val, &WKSTR("true"))) {
+		if (str_eql(val, &STR("true"))) {
 			b = true;
-		} else if (str_eql(val, &WKSTR("false"))) {
+		} else if (str_eql(val, &STR("false"))) {
 			b = false;
 		} else {
 			vm_error_at(wk, opt->ip, "unable to coerce '%s' into a boolean", val->s);
@@ -332,10 +340,11 @@ coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, o
 			return false;
 		}
 
-		make_obj(wk, res, obj_number);
+		*res = make_obj(wk, obj_number);
 		set_obj_number(wk, *res, num);
 		break;
 	}
+	case op_shell_array:
 	case op_array: {
 		// do nothing, array values were already coerced above
 		break;
@@ -367,6 +376,7 @@ typecheck_opt(struct workspace *wk, uint32_t node, obj val, enum build_option_ty
 	case op_combo: expected_type = obj_string; break;
 	case op_integer: expected_type = obj_number; break;
 	case op_array: expected_type = obj_array; break;
+	case op_shell_array: expected_type = obj_array; break;
 	default: UNREACHABLE_RETURN;
 	}
 
@@ -401,14 +411,14 @@ prefix_dir_opts_iter(struct workspace *wk, void *_ctx, obj _k, obj v)
 		return ir_cont;
 	}
 
-	SBUF(new_path);
+	TSTR(new_path);
 	if (path_is_subpath(ctx->prefix->s, path)) {
 		path_relative_to(wk, &new_path, ctx->prefix->s, path);
 	} else {
 		path_join(wk, &new_path, ctx->prefix->s, path);
 	}
 
-	opt->val = sbuf_into_str(wk, &new_path);
+	opt->val = tstr_into_str(wk, &new_path);
 	return ir_cont;
 }
 
@@ -444,38 +454,35 @@ set_option(struct workspace *wk, obj opt, obj new_val, enum option_value_source
 {
 	struct obj_option *o = get_obj_option(wk, opt);
 
-	// Only set options that haven't set from a source with higher
-	// precedence.  This means that option precedence doesn't have to rely
-	// on the order in which options are set.  This means that e.g.
-	// command-line set options can be set before the main meson.build file
-	// has even been parsed.
-	//
-	// This mostly follows meson's behavior, except that deprecated options
-	// cannot override command line options.
+	// Only set options that have not been set from a source with higher
+	// precedence.
 
-	/* { */
-	/* 	const char *sourcenames[] = { */
-	/* 		[option_value_source_unset] = "unset", */
-	/* 		[option_value_source_default] = "default", */
-	/* 		[option_value_source_environment] = "environment", */
-	/* 		[option_value_source_default_options] = "default_options", */
-	/* 		[option_value_source_subproject_default_options] = "subproject_default_options", */
-	/* 		[option_value_source_yield] = "yield", */
-	/* 		[option_value_source_commandline] = "commandline", */
-	/* 		[option_value_source_deprecated_rename] = "deprecated rename", */
-	/* 		[option_value_source_override_options] = "override_options", */
-	/* 	}; */
-	/* 	obj_fprintf(wk, log_file(), */
-	/* 		"%s option %o to %o from %s, last set by %s\n", */
-	/* 		o->source > source */
-	/* 		? "\033[31mnot setting\033[0m" */
-	/* 		: "\033[32msetting\033[0m", */
-	/* 		o->name, new_val, sourcenames[source], sourcenames[o->source]); */
-	/* } */
-
-	if (o->source > source) {
+	if (o->source >= source) {
 		return true;
 	}
+
+	if (false) {
+		const char *sourcenames[] = {
+			[option_value_source_unset] = "unset",
+			[option_value_source_default] = "default",
+			[option_value_source_environment] = "environment",
+			[option_value_source_default_options] = "default_options",
+			[option_value_source_subproject_default_options] = "subproject_default_options",
+			[option_value_source_yield] = "yield",
+			[option_value_source_commandline] = "commandline",
+			[option_value_source_deprecated_rename] = "deprecated rename",
+			[option_value_source_override_options] = "override_options",
+		};
+		obj_lprintf(wk,
+			log_debug,
+			"%s option %o to %o from %s, last set by %s\n",
+			o->source > source ? "\033[31mnot setting\033[0m" : "\033[32msetting\033[0m",
+			o->name,
+			new_val,
+			sourcenames[source],
+			sourcenames[o->source]);
+	}
+
 	o->source = source;
 
 	if (get_obj_type(wk, o->deprecated) == obj_bool && get_obj_bool(wk, o->deprecated)) {
@@ -500,6 +507,9 @@ set_option(struct workspace *wk, obj opt, obj new_val, enum option_value_source
 			vm_error_at(wk, o->ip, "'%o' is not one of %o", new_val, o->choices);
 			return false;
 		}
+
+		const struct str *s = get_str(wk, new_val);
+		new_val = make_strn_enum(wk, s->s, s->len, o->choices);
 		break;
 	}
 	case op_integer: {
@@ -528,6 +538,7 @@ set_option(struct workspace *wk, obj opt, obj new_val, enum option_value_source
 		}
 		break;
 	}
+	case op_shell_array:
 	case op_string:
 	case op_feature:
 	case op_boolean: break;
@@ -594,7 +605,7 @@ void
 get_option_value_overridable(struct workspace *wk, const struct project *proj, obj overrides, const char *name, obj *res)
 {
 	obj opt;
-	if (!get_option_overridable(wk, proj, overrides, &WKSTR(name), &opt)) {
+	if (!get_option_overridable(wk, proj, overrides, &STRL(name), &opt)) {
 		LOG_E("attempted to get unknown option '%s'", name);
 		UNREACHABLE;
 	}
@@ -613,7 +624,7 @@ static void
 set_binary_from_env(struct workspace *wk, const char *envvar, const char *dest)
 {
 	obj opt;
-	if (!get_option(wk, NULL, &WKSTR(dest), &opt)) {
+	if (!get_option(wk, NULL, &STRL(dest), &opt)) {
 		UNREACHABLE;
 	}
 
@@ -622,39 +633,44 @@ set_binary_from_env(struct workspace *wk, const char *envvar, const char *dest)
 		return;
 	}
 
-	// TODO: implement something like shlex.split()
-	obj cmd = str_split(wk, &WKSTR(v), NULL);
+	obj cmd = str_shell_split(wk, &STRL(v), shell_type_for_host_machine());
 	set_option(wk, opt, cmd, option_value_source_environment, false);
 }
 
 static void
-set_compile_opt_from_env(struct workspace *wk, const char *name, const char *flags, const char *extra)
+set_compile_opt_from_env(struct workspace *wk, const char *name, const char *flags)
 {
 	obj opt;
-	if (!get_option(wk, NULL, &WKSTR(name), &opt)) {
+	if (!get_option(wk, NULL, &STRL(name), &opt)) {
 		UNREACHABLE;
 	}
 
 	if ((flags = os_get_env(flags)) && *flags) {
-		extend_array_option(wk, opt, str_split(wk, &WKSTR(flags), NULL), option_value_source_environment);
-	}
-
-	if ((extra = os_get_env(extra)) && *extra) {
-		extend_array_option(wk, opt, str_split(wk, &WKSTR(extra), NULL), option_value_source_environment);
+		extend_array_option(wk,
+			opt,
+			str_shell_split(wk, &STRL(flags), shell_type_for_host_machine()),
+			option_value_source_environment);
 	}
 }
 
 static void
-set_str_opt_from_env(struct workspace *wk, const char *env_name, const char *opt_name)
+set_str_opt_from_env(struct workspace *wk, const char *env_name, const char *opt_name, const char *split)
 {
 	obj opt;
-	if (!get_option(wk, NULL, &WKSTR(opt_name), &opt)) {
+	if (!get_option(wk, NULL, &STRL(opt_name), &opt)) {
 		UNREACHABLE;
 	}
 
 	const char *env_val;
 	if ((env_val = os_get_env(env_name)) && *env_val) {
-		set_option(wk, opt, make_str(wk, env_val), option_value_source_environment, false);
+		obj val;
+		if (split) {
+			val = str_split(wk, &STRL(env_val), &STRL(split));
+		} else {
+			val = make_str(wk, env_val);
+		}
+
+		set_option(wk, opt, val, option_value_source_environment, false);
 	}
 }
 
@@ -714,7 +730,7 @@ set_yielding_project_options_iter(struct workspace *wk, void *_ctx, obj _k, obj
 }
 
 static bool
-determine_option_file(struct workspace *wk, const char *cwd, struct sbuf *out)
+determine_option_file(struct workspace *wk, const char *cwd, struct tstr *out)
 {
 	const char *option_file_names[] = {
 		"meson.options",
@@ -735,6 +751,23 @@ determine_option_file(struct workspace *wk, const char *cwd, struct sbuf *out)
 	return exists;
 }
 
+void
+best_fuzzy_match_in_dict(struct workspace *wk, obj dict, const struct str *input, int32_t *min_d, obj *min)
+{
+	obj k, v;
+	obj_dict_for(wk, dict, k, v) {
+		(void)v;
+		int32_t d;
+		if (!str_fuzzy_match(input, get_str(wk, k), &d)) {
+			continue;
+		}
+		if (d < *min_d) {
+			*min_d = d;
+			*min = k;
+		}
+	}
+}
+
 bool
 setup_project_options(struct workspace *wk, const char *cwd)
 {
@@ -746,13 +779,13 @@ setup_project_options(struct workspace *wk, const char *cwd)
 		return true;
 	}
 
-	SBUF(meson_opts);
+	TSTR(meson_opts);
 	bool exists = determine_option_file(wk, cwd, &meson_opts);
 
 	if (exists) {
 		enum language_mode old_mode = wk->vm.lang_mode;
 		wk->vm.lang_mode = language_opts;
-		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0)) {
+		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0, 0)) {
 			wk->vm.lang_mode = old_mode;
 			return false;
 		}
@@ -790,16 +823,62 @@ setup_project_options(struct workspace *wk, const char *cwd)
 				ret = false;
 			}
 		} else {
+			int32_t min_d = INT32_MAX;
+			obj min_k = 0;
+			best_fuzzy_match_in_dict(wk, wk->global_opts, name, &min_d, &min_k);
+			best_fuzzy_match_in_dict(wk, current_project(wk)->opts,name, &min_d, &min_k);
+
 			LLOG_E("invalid option: ");
-			log_option_override(wk, oo);
-			log_plain("\n");
+			log_option_override(log_error, wk, oo);
+			if (min_k) {
+				log_plain(log_error, ", did you mean '%s'?", get_str(wk, min_k)->s);
+			}
+			log_plain(log_error, "\n");
 			ret = false;
+
 		}
 	}
 
 	return ret;
 }
 
+static void
+make_compiler_option(struct workspace *wk, obj name)
+{
+	obj opt = make_obj(wk, obj_option);
+	struct obj_option *o = get_obj_option(wk, opt);
+	o->name = name;
+	o->type = op_shell_array;
+	o->ip = -1; // ?
+	o->builtin = true;
+
+	if (!create_option(wk, wk->global_opts, opt, make_obj(wk, obj_array))) {
+		UNREACHABLE;
+	}
+}
+
+static void
+make_compiler_env_option(struct workspace *wk, enum compiler_language lang, enum toolchain_component comp)
+{
+	const char *env_opt = toolchain_component_option_name[lang][comp];
+	if (!env_opt) {
+		return;
+	}
+
+	make_compiler_option(wk, make_str(wk, env_opt));
+
+	const char *env_var = strchr(env_opt, '.') + 1;
+	set_binary_from_env(wk, env_var, env_opt);
+}
+
+const char *toolchain_component_option_name[compiler_language_count][toolchain_component_count] = {
+	[compiler_language_c] = { "env.CC", "env.CC_LD" },
+	[compiler_language_cpp] = { "env.CXX", "env.CXX_LD" },
+	[compiler_language_objc] = { "env.OBJC", "env.OBJC_LD" },
+	[compiler_language_objcpp] = { "env.OBJCXX", "env.OBJCXX_LD" },
+	[compiler_language_nasm] = { "env.NASM", "env.NASM_LD" },
+};
+
 bool
 init_global_options(struct workspace *wk)
 {
@@ -807,19 +886,48 @@ init_global_options(struct workspace *wk)
 		return false;
 	}
 
+	static struct {
+		enum compiler_language l;
+		const char *name;
+	} langs[] = {
+#define TOOLCHAIN_ENUM(lang) { compiler_language_##lang, #lang },
+		FOREACH_COMPILER_EXPOSED_LANGUAGE(TOOLCHAIN_ENUM)
+#undef TOOLCHAIN_ENUM
+	};
+
+	static const char *compile_opt_env_var[compiler_language_count][toolchain_component_count] = {
+		[compiler_language_c] = { "CFLAGS", "LDFLAGS" },
+		[compiler_language_cpp] = { "CXXFLAGS", "LDFLAGS" },
+		[compiler_language_objc] = { "OBJCFLAGS", "LDFLAGS" },
+		[compiler_language_objcpp] = { "OBJCXXFLAGS", "LDFLAGS" },
+	};
+
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(langs); ++i) {
+		obj args = make_strf(wk, "%s_args", langs[i].name);
+		make_compiler_option(wk, args);
+		if (compile_opt_env_var[langs[i].l][toolchain_component_compiler]) {
+			set_compile_opt_from_env(wk,
+				get_str(wk, args)->s,
+				compile_opt_env_var[langs[i].l][toolchain_component_compiler]);
+		}
+
+		obj link_args = make_strf(wk, "%s_link_args", langs[i].name);
+		make_compiler_option(wk, link_args);
+		if (compile_opt_env_var[langs[i].l][toolchain_component_linker]) {
+			set_compile_opt_from_env(
+				wk, get_str(wk, link_args)->s, compile_opt_env_var[langs[i].l][toolchain_component_linker]);
+		}
+
+		make_compiler_env_option(wk, langs[i].l, toolchain_component_compiler);
+		make_compiler_env_option(wk, langs[i].l, toolchain_component_linker);
+	}
+
+	set_str_opt_from_env(wk, "PKG_CONFIG_PATH", "pkg_config_path", ENV_PATH_SEP_STR);
+
 	set_binary_from_env(wk, "AR", "env.AR");
-	set_binary_from_env(wk, "CC", "env.CC");
-	set_binary_from_env(wk, "CXX", "env.CXX");
-	set_binary_from_env(wk, "LD", "env.LD");
-	set_binary_from_env(wk, "NASM", "env.NASM");
 	set_binary_from_env(wk, "NINJA", "env.NINJA");
-	set_binary_from_env(wk, "OBJC", "env.OBJC");
-	set_binary_from_env(wk, "OBJCPP", "env.OBJCPP");
-	set_compile_opt_from_env(wk, "c_args", "CFLAGS", "CPPFLAGS");
-	set_compile_opt_from_env(wk, "c_link_args", "CFLAGS", "LDFLAGS");
-	set_compile_opt_from_env(wk, "cpp_args", "CXXFLAGS", "CPPFLAGS");
-	set_compile_opt_from_env(wk, "cpp_link_args", "CXXFLAGS", "LDFLAGS");
-	set_str_opt_from_env(wk, "PKG_CONFIG_PATH", "pkg_config_path");
+	set_binary_from_env(wk, "PKG_CONFIG", "env.PKG_CONFIG");
 
 	return true;
 }
@@ -828,7 +936,7 @@ bool
 parse_and_set_cmdline_option(struct workspace *wk, char *lhs)
 {
 	struct option_override oo = { .source = option_value_source_commandline };
-	if (!parse_config_string(wk, &WKSTR(lhs), &oo, false)) {
+	if (!parse_config_string(wk, &STRL(lhs), &oo, false)) {
 		return false;
 	}
 
@@ -901,7 +1009,7 @@ parse_and_set_option(struct workspace *wk, const struct parse_and_set_option_par
 
 	if (params->flags & parse_and_set_option_flag_override) {
 		obj newopt;
-		make_obj(wk, &newopt, obj_option);
+		newopt = make_obj(wk, obj_option);
 		struct obj_option *o = get_obj_option(wk, newopt);
 		*o = *get_obj_option(wk, opt);
 		opt = newopt;
@@ -987,7 +1095,7 @@ parse_and_set_override_options(struct workspace *wk, uint32_t err_node, obj opts
 		.flags = parse_and_set_option_flag_override,
 	};
 
-	make_obj(wk, &params.opts, obj_dict);
+	params.opts = make_obj(wk, obj_dict);
 	*res = params.opts;
 
 	return parse_and_set_options(wk, &params, opts);
@@ -1027,22 +1135,40 @@ get_option_default_library(struct workspace *wk)
 	obj opt;
 	get_option_value(wk, current_project(wk), "default_library", &opt);
 
-	if (str_eql(get_str(wk, opt), &WKSTR("static"))) {
+	if (str_eql(get_str(wk, opt), &STR("static"))) {
 		return tgt_static_library;
-	} else if (str_eql(get_str(wk, opt), &WKSTR("shared"))) {
+	} else if (str_eql(get_str(wk, opt), &STR("shared"))) {
 		return tgt_dynamic_library;
-	} else if (str_eql(get_str(wk, opt), &WKSTR("both"))) {
+	} else if (str_eql(get_str(wk, opt), &STR("both"))) {
 		return tgt_dynamic_library | tgt_static_library;
 	} else {
 		UNREACHABLE_RETURN;
 	}
 }
 
+enum default_both_libraries
+get_option_default_both_libraries(struct workspace *wk, const struct project *proj, obj overrides)
+{
+	obj opt;
+	get_option_value_overridable(wk, proj, overrides, "default_both_libraries", &opt);
+	const struct str *s = get_str(wk, opt);
+
+	if (str_eql(s, &STR("auto"))) {
+		return default_both_libraries_auto;
+	} else if (str_eql(s, &STR("static"))) {
+		return default_both_libraries_static;
+	} else if (str_eql(s, &STR("shared"))) {
+		return default_both_libraries_shared;
+	}
+
+	UNREACHABLE_RETURN;
+}
+
 bool
 get_option_bool(struct workspace *wk, obj overrides, const char *name, bool fallback)
 {
 	obj opt;
-	if (get_option_overridable(wk, current_project(wk), overrides, &WKSTR(name), &opt)) {
+	if (get_option_overridable(wk, current_project(wk), overrides, &STRL(name), &opt)) {
 		return get_obj_bool(wk, get_obj_option(wk, opt)->val);
 	} else {
 		return fallback;
@@ -1055,9 +1181,9 @@ get_option_backend(struct workspace *wk)
 	obj backend_opt;
 	get_option_value(wk, 0, "backend", &backend_opt);
 	const struct str *backend_str = get_str(wk, backend_opt);
-	if (str_eql(backend_str, &WKSTR("ninja"))) {
+	if (str_eql(backend_str, &STR("ninja"))) {
 		return backend_ninja;
-	} else if (str_eql(backend_str, &WKSTR("xcode"))) {
+	} else if (str_eql(backend_str, &STR("xcode"))) {
 		return backend_xcode;
 	} else {
 		UNREACHABLE;
@@ -1103,16 +1229,27 @@ struct list_options_ctx {
 	const char *subproject_name;
 };
 
+static bool
+is_option_filtered(struct workspace *wk, struct obj_option *opt, bool show_builtin, bool only_modified)
+{
+	if (opt->builtin != show_builtin) {
+		return true;
+	} else if (only_modified
+		   && (opt->source == option_value_source_default || opt->source == option_value_source_yield)) {
+		return true;
+	}
+
+	return false;
+}
+
 static enum iteration_result
 list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 {
 	struct list_options_ctx *ctx = _ctx;
 	struct obj_option *opt = get_obj_option(wk, val);
 
-	if (opt->builtin != ctx->show_builtin) {
-		return ir_cont;
-	} else if (ctx->list_opts->only_modified && opt->source == option_value_source_default) {
-		return ir_cont;
+	if (is_option_filtered(wk, opt, ctx->show_builtin, ctx->list_opts->only_modified)) {
+		return true;
 	}
 
 	/* const char *option_type_names[] = { */
@@ -1127,15 +1264,15 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 	const char *key_clr = "", *val_clr = "", *sel_clr = "", *no_clr = "";
 
 	if (fs_is_a_tty(stdout)) {
-		key_clr = "\033[1;34m";
-		val_clr = "\033[1;37m";
-		sel_clr = "\033[1;32m";
-		no_clr = "\033[0m";
+		key_clr = CLR(c_bold, c_blue);
+		val_clr = CLR(c_bold, c_white);
+		sel_clr = CLR(c_bold, c_green);
+		no_clr = CLR(0);
 	}
 
 	{
 		const char *subp = ctx->subproject_name;
-		obj_lprintf(wk, "  -D %s%s%s%#o%s=", subp ? subp : "", subp ? ":" : "", key_clr, key, no_clr);
+		obj_lprintf(wk, log_info, "  -D %s%s%s%#o%s=", subp ? subp : "", subp ? ":" : "", key_clr, key, no_clr);
 	}
 
 	obj choices = 0;
@@ -1143,26 +1280,26 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 
 	if (opt->type == op_combo) {
 		choices = opt->choices;
-		make_obj(wk, &selected, obj_array);
+		selected = make_obj(wk, obj_array);
 		obj_array_push(wk, selected, opt->val);
 	} else if (opt->type == op_array && opt->choices) {
 		choices = opt->choices;
 		selected = opt->val;
 	} else {
-		make_obj(wk, &choices, obj_array);
+		choices = make_obj(wk, obj_array);
 		switch (opt->type) {
 		case op_string: obj_array_push(wk, choices, make_str(wk, "string")); break;
 		case op_boolean:
 			obj_array_push(wk, choices, make_str(wk, "true"));
 			obj_array_push(wk, choices, make_str(wk, "false"));
-			make_obj(wk, &selected, obj_array);
+			selected = make_obj(wk, obj_array);
 			obj_array_push(wk, selected, make_str(wk, get_obj_bool(wk, opt->val) ? "true" : "false"));
 			break;
 		case op_feature:
 			obj_array_push(wk, choices, make_str(wk, "enabled"));
 			obj_array_push(wk, choices, make_str(wk, "disabled"));
 			obj_array_push(wk, choices, make_str(wk, "auto"));
-			make_obj(wk, &selected, obj_array);
+			selected = make_obj(wk, obj_array);
 			obj_array_push(wk,
 				selected,
 				make_str(wk,
@@ -1174,6 +1311,7 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 			break;
 		case op_combo:
 		case op_array:
+		case op_shell_array:
 		case op_integer: break;
 		default: UNREACHABLE;
 		}
@@ -1196,71 +1334,114 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 	switch (opt->type) {
 	case op_boolean:
 	case op_combo:
-	case op_feature: obj_lprintf(wk, "<%s>", get_cstr(wk, choices)); break;
+	case op_feature: {
+		obj_lprintf(wk, log_info, "<%s>", get_cstr(wk, choices));
+		break;
+	}
 	case op_string: {
-		const struct str *def = get_str(wk, opt->val);
-		obj_lprintf(
-			wk, "<%s>, default: %s%s%s", get_cstr(wk, choices), sel_clr, def->len ? def->s : "''", no_clr);
+		if (opt->source == option_value_source_default) {
+			const struct str *def = get_str(wk, opt->val);
+			obj_lprintf(wk,
+				log_info,
+				"<%s>, default: %s%s%s",
+				get_cstr(wk, choices),
+				sel_clr,
+				def->len ? def->s : "''",
+				no_clr);
+		} else {
+			obj_lprintf(wk, log_info, "%s%o%s", sel_clr, opt->val, no_clr);
+		}
 		break;
 	}
 	case op_integer:
-		log_plain("<%sN%s>", val_clr, no_clr);
-		if (opt->min || opt->max) {
-			log_plain(" where ");
-			if (opt->min) {
-				obj_lprintf(wk, "%o <= ", opt->min);
-			}
-			log_plain("%sN%s", val_clr, no_clr);
-			if (opt->max) {
-				obj_lprintf(wk, " <= %o", opt->max);
+		if (opt->source == option_value_source_default) {
+			log_plain(log_info, "<%sN%s>", val_clr, no_clr);
+			if (opt->min || opt->max) {
+				log_plain(log_info, " where ");
+				if (opt->min) {
+					obj_lprintf(wk, log_info, "%o <= ", opt->min);
+				}
+				log_plain(log_info, "%sN%s", val_clr, no_clr);
+				if (opt->max) {
+					obj_lprintf(wk, log_info, " <= %o", opt->max);
+				}
 			}
+			obj_lprintf(wk, log_info, ", default: %s%o%s", sel_clr, opt->val, no_clr);
+		} else {
+			obj_lprintf(wk, log_info, "%s%o%s", sel_clr, opt->val, no_clr);
 		}
-		obj_lprintf(wk, ", default: %s%o%s", sel_clr, opt->val, no_clr);
-
 		break;
+	case op_shell_array:
 	case op_array:
-		log_plain("<%svalue%s[,%svalue%s[...]]>", val_clr, no_clr, val_clr, no_clr);
-		if (opt->choices) {
-			obj_lprintf(wk, " where value in %s", get_cstr(wk, choices));
+		if (opt->source == option_value_source_default) {
+			if (opt->type == op_shell_array) {
+				log_plain(log_info, "<shell array>");
+			} else {
+				log_plain(log_info, "<array>");
+			}
+			if (opt->choices) {
+				obj_lprintf(wk, log_info, " where value in %s", get_cstr(wk, choices));
+			}
+
+			obj_lprintf(wk, log_info, ", default: %s%o%s", sel_clr, opt->val, no_clr);
+		} else {
+			obj_lprintf(wk, log_info, "%s%o%s", sel_clr, opt->val, no_clr);
 		}
 		break;
 	default: UNREACHABLE;
 	}
 
-	if (opt->source != option_value_source_default) {
-		obj_lprintf(wk, "*");
+	const char *source_name = 0;
+	switch (opt->source) {
+	case option_value_source_unset:
+	case option_value_source_default: break;
+	case option_value_source_environment: source_name = "environment"; break;
+	case option_value_source_yield: break;
+	case option_value_source_default_options: source_name = "default_options"; break;
+	case option_value_source_subproject_default_options: source_name = "subproject default_options"; break;
+	case option_value_source_override_options: source_name = "override option"; break;
+	case option_value_source_deprecated_rename: source_name = "deprecated rename"; break;
+	case option_value_source_commandline: source_name = "commandline"; break;
+	}
+
+	if (source_name) {
+		obj_lprintf(wk, log_info, "*");
 	}
 
 	if (opt->description) {
-		obj_lprintf(wk, " - %#o", opt->description);
+		obj_lprintf(wk, log_info, " - %#o", opt->description);
 	}
 
-	log_plain("\n");
+	if (source_name && ctx->list_opts->list_all) {
+		obj_lprintf(wk, log_info, "\n     using value from: %s", source_name);
+	}
+
+	log_plain(log_info, "\n");
 	return ir_cont;
 }
 
 static enum iteration_result
 list_options_for_subproject(struct workspace *wk, struct subprojects_common_ctx *ctx, const char *path)
 {
-	SBUF(cwd);
+	TSTR(cwd);
 
 	struct wrap wrap = { 0 };
-	if (!wrap_parse(path, &wrap)) {
+	if (!wrap_parse(wk, get_str(wk, current_project(wk)->subprojects_dir)->s, path, &wrap)) {
 		return ir_err;
 	}
 
-	SBUF(meson_opts);
+	TSTR(meson_opts);
 	bool exists = determine_option_file(wk, wrap.dest_dir.buf, &meson_opts);
 
 	make_project(wk, &wk->cur_project, wrap.name.buf, wrap.dest_dir.buf, "");
 	current_project(wk)->cfg.name = make_str(wk, wrap.name.buf);
 
 	if (exists) {
-		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0)) {
+		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, build_language_meson, 0, 0)) {
 			goto cont;
 		}
 	} else {
-		make_obj(wk, &current_project(wk)->opts, obj_dict);
+		current_project(wk)->opts = make_obj(wk, obj_dict);
 	}
 
 cont:
@@ -1275,36 +1456,59 @@ list_options(const struct list_options_opts *list_opts)
 {
 	bool ret = false;
 	struct workspace wk = { 0 };
-	workspace_init(&wk);
+	workspace_init_bare(&wk);
+	workspace_init_runtime(&wk);
 	wk.vm.lang_mode = language_opts;
 
+	bool load_from_build_dir = false;
+
+	if (fs_file_exists("meson.build")) {
+		load_from_build_dir = false;
+	} else {
+		TSTR(option_info);
+		path_join(&wk, &option_info, output_path.private_dir, output_path.option_info);
+		if (!fs_file_exists(option_info.buf)) {
+			LOG_I("this command must be run from a build directory or the project root");
+			goto ret;
+		}
+
+		load_from_build_dir = true;
+	}
+
+	if (!load_from_build_dir && list_opts->list_all) {
+		if (!init_global_options(&wk)) {
+			UNREACHABLE;
+		}
+	}
+
 	uint32_t idx;
 	make_project(&wk, &idx, "dummy", path_cwd(), "");
 
-	if (fs_file_exists("meson.build")) {
-		SBUF(meson_opts);
+	if (!load_from_build_dir) {
+		TSTR(meson_opts);
 		bool exists = determine_option_file(&wk, ".", &meson_opts);
 
 		if (exists) {
-			if (!wk.vm.behavior.eval_project_file(&wk, meson_opts.buf, build_language_meson, 0)) {
+			if (!wk.vm.behavior.eval_project_file(&wk, meson_opts.buf, build_language_meson, 0, 0)) {
 				goto ret;
 			}
 		} else {
-			make_obj(&wk, &current_project(&wk)->opts, obj_dict);
+			current_project(&wk)->opts = make_obj(&wk, obj_dict);
 		}
 
 		{
-			SBUF(subprojects_dir);
+			TSTR(subprojects_dir);
 			struct workspace az_wk = { 0 };
 			analyze_project_call(&az_wk);
-			path_make_absolute(&wk, &subprojects_dir, get_cstr(&az_wk, current_project(&az_wk)->subprojects_dir));
+			path_make_absolute(
+				&wk, &subprojects_dir, get_cstr(&az_wk, current_project(&az_wk)->subprojects_dir));
 
 			obj name = current_project(&az_wk)->cfg.name;
 			if (name) {
 				current_project(&wk)->cfg.name = make_str(&wk, get_cstr(&az_wk, name));
 			}
 
-			current_project(&wk)->subprojects_dir = sbuf_into_str(&wk, &subprojects_dir);
+			current_project(&wk)->subprojects_dir = tstr_into_str(&wk, &subprojects_dir);
 
 			workspace_destroy(&az_wk);
 
@@ -1317,20 +1521,34 @@ list_options(const struct list_options_opts *list_opts)
 			}
 		}
 	} else {
-		SBUF(option_info);
-		path_join(&wk, &option_info, output_path.private_dir, output_path.option_info);
-		if (!fs_file_exists(option_info.buf)) {
-			LOG_I("run this command must be run from a build directory or the project root");
-			goto ret;
-		}
-
 		obj arr;
 		if (!serial_load_from_private_dir(&wk, &arr, output_path.option_info)) {
 			goto ret;
 		}
 
-		obj_array_index(&wk, arr, 0, &wk.global_opts);
-		obj_array_index(&wk, arr, 1, &current_project(&wk)->opts);
+		uint32_t i = 0;
+		obj v;
+		obj_array_for(&wk, arr, v) {
+			if (i == 0) {
+				wk.global_opts = v;
+			} else {
+				uint32_t proj_i = (i - 1) / 2;
+				if (proj_i >= wk.projects.len) {
+					make_project(&wk, &proj_i, 0, "", "");
+				}
+
+				struct project *proj = arr_get(&wk.projects, proj_i);
+				if ((i - 1) & 1) {
+					proj->cfg.name = v;
+				} else {
+					proj->opts = v;
+				}
+			}
+
+			++i;
+		}
+		wk.global_opts = obj_array_index(&wk, arr, 0);
+		current_project(&wk)->opts = obj_array_index(&wk, arr, 1);
 	}
 
 	struct list_options_ctx ctx = { .list_opts = list_opts };
@@ -1340,36 +1558,69 @@ list_options(const struct list_options_opts *list_opts)
 	uint32_t i;
 	for (i = 0; i < wk.projects.len; ++i) {
 		struct project *proj = arr_get(&wk.projects, i);
-		if (get_obj_dict(&wk, proj->opts)->len) {
+
+		uint32_t builtin_count = 0, non_builtin_count = 0;
+
+		obj k, v;
+		obj_dict_for(&wk, proj->opts, k, v) {
+			(void)k;
+			struct obj_option *opt = get_obj_option(&wk, v);
+
+			if (opt->builtin) {
+				if (is_option_filtered(&wk, opt, true, list_opts->only_modified)) {
+					continue;
+				}
+
+				if (list_opts->list_all) {
+					++builtin_count;
+				}
+			} else {
+				if (is_option_filtered(&wk, opt, false, list_opts->only_modified)) {
+					continue;
+				}
+
+				++non_builtin_count;
+			}
+		}
+
+		if (non_builtin_count || builtin_count) {
 			had_project_options = true;
 
 			const char *name = get_cstr(&wk, proj->cfg.name);
-			if (!name) {
-				name = "project";
+			ctx.subproject_name = i == 0 ? 0 : name;
+
+			if (non_builtin_count || builtin_count) {
+				log_plain(log_info, "%s options:\n", name);
 			}
 
-			log_plain("%s options:\n", name);
+			if (non_builtin_count) {
+				obj_dict_foreach(&wk, proj->opts, &ctx, list_options_iter);
+
+				if (!builtin_count) {
+					log_plain(log_info, "\n");
+				}
+			}
+
+			if (builtin_count) {
+				ctx.show_builtin = true;
+				log_plain(log_info, "  builtin options:\n");
+				obj_dict_foreach(&wk, proj->opts, &ctx, list_options_iter);
+				ctx.show_builtin = false;
+
+				log_plain(log_info, "\n");
+			}
 
-			ctx.subproject_name = i == 0 ? 0 : name;
-			obj_dict_foreach(&wk, proj->opts, &ctx, list_options_iter);
 			ctx.subproject_name = 0;
-
-			log_plain("\n");
 		}
 	}
 
 	if (!had_project_options && !list_opts->list_all) {
-		log_plain("no project options defined\n");
+		log_plain(log_info, "no project options defined\n");
 	}
 
 	if (list_opts->list_all) {
 		ctx.show_builtin = true;
-
-		log_plain("project builtin-options:\n");
-		obj_dict_foreach(&wk, current_project(&wk)->opts, &ctx, list_options_iter);
-		log_plain("\n");
-
-		log_plain("global options:\n");
+		log_plain(log_info, "builtin global options:\n");
 		obj_dict_foreach(&wk, wk.global_opts, &ctx, list_options_iter);
 	}
 
