@@ -22,15 +22,19 @@
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
-#ifndef _WIN32
+#ifdef _WIN32
+#include <io.h>
+#else
 #include <unistd.h>
 #endif
 
 #include "buf_size.h"
 #include "lang/string.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 
 bool
@@ -49,32 +53,36 @@ fs_mkdir_p(const char *path)
 {
 	bool res = false;
 	uint32_t i, len = strlen(path);
-	SBUF_manual(buf);
+	TSTR_manual(buf);
 	path_copy(NULL, &buf, path);
 
-	assert(len > 1);
+	assert(len >= 1);
 
-	for (i = 1; i < len; ++i) {
+	i = 0;
+	if (path_is_absolute(buf.buf)) {
+		char *p;
+		if ((p = strchr(buf.buf, PATH_SEP))) {
+			i = (p - buf.buf) + 1;
+		}
+	}
+
+	for (; i < len; ++i) {
 		if (buf.buf[i] == PATH_SEP) {
 			buf.buf[i] = 0;
-			if (!fs_dir_exists(buf.buf)) {
-				if (!fs_mkdir(buf.buf)) {
-					goto ret;
-				}
+			if (!fs_mkdir(buf.buf, true)) {
+				goto ret;
 			}
 			buf.buf[i] = PATH_SEP;
 		}
 	}
 
-	if (!fs_dir_exists(path)) {
-		if (!fs_mkdir(path)) {
-			goto ret;
-		}
+	if (!fs_mkdir(path, true)) {
+		goto ret;
 	}
 
 	res = true;
 ret:
-	sbuf_destroy(&buf);
+	tstr_destroy(&buf);
 	return res;
 }
 
@@ -169,7 +177,7 @@ fs_read_entire_file(const char *path, struct source *src)
 	size_t read;
 	char *buf = NULL;
 
-	*src = (struct source) { .label = path };
+	*src = (struct source){ .label = path, .type = source_type_file };
 
 	if (strcmp(path, "-") == 0) {
 		f = stdin;
@@ -204,14 +212,16 @@ fs_read_entire_file(const char *path, struct source *src)
 		read = fread(buf, 1, src->len, f);
 
 		if (read != src->len) {
-			LOG_E("failed to read entire file, only read %" PRIu64 "/%" PRId64 "bytes", (uint64_t)read, src->len);
+			LOG_E("failed to read entire file, only read %" PRIu64 "/%" PRId64 "bytes",
+				(uint64_t)read,
+				src->len);
 			goto err;
 		}
 	} else {
 		uint32_t buf_size = BUF_SIZE_4k;
 		buf = z_calloc(buf_size + 1, 1);
 
-		while ((read = fread(&buf[src->len], 1, BUF_SIZE_4k, f))) {
+		while ((read = fread(&buf[src->len], 1, buf_size - src->len, f))) {
 			src->len += read;
 
 			if (src->len >= buf_size) {
@@ -219,7 +229,6 @@ fs_read_entire_file(const char *path, struct source *src)
 				buf = z_realloc(buf, buf_size);
 				memset(&buf[src->len], 0, buf_size - src->len);
 			}
-
 		}
 
 		assert(src->len < buf_size && buf[src->len] == 0);
@@ -257,6 +266,7 @@ fs_source_dup(const struct source *src, struct source *dup)
 	dup->label = &buf[src->len];
 	dup->src = buf;
 	dup->len = src->len;
+	dup->type = src->type;
 
 	memcpy(buf, src->src, src->len);
 	memcpy(&buf[src->len], src->label, label_len);
@@ -265,9 +275,13 @@ fs_source_dup(const struct source *src, struct source *dup)
 void
 fs_source_destroy(struct source *src)
 {
-	if (src->src) {
-		z_free((char *)src->src);
+	if (!src->is_weak_reference) {
+		if (src->src) {
+			z_free((char *)src->src);
+		}
 	}
+	src->src = 0;
+	src->len = 0;
 }
 
 bool
@@ -311,7 +325,9 @@ fs_fread(void *ptr, size_t size, FILE *f)
 	if (r == size) {
 		return true;
 	} else {
-		if ((err = ferror(f))) {
+		if (feof(f)) {
+			LOG_E("fread got EOF");
+		} else if ((err = ferror(f))) {
 			LOG_E("fread failed: %s", strerror(err));
 		} else {
 			LOG_E("fread failed: unknown");
@@ -320,6 +336,17 @@ fs_fread(void *ptr, size_t size, FILE *f)
 	}
 }
 
+int32_t
+fs_read(int fd, void *buf, uint32_t buf_len)
+{
+	int32_t res = read(fd, buf, buf_len);
+	if (res < 0) {
+		LOG_E("read: %s", strerror(errno));
+	}
+
+	return res;
+}
+
 bool
 fs_write(const char *path, const uint8_t *buf, uint64_t buf_len)
 {
@@ -344,23 +371,12 @@ fs_write(const char *path, const uint8_t *buf, uint64_t buf_len)
 bool
 fs_has_cmd(const char *cmd)
 {
-	SBUF_manual(buf);
+	TSTR_manual(buf);
 	bool res = fs_find_cmd(NULL, &buf, cmd);
-	sbuf_destroy(&buf);
+	tstr_destroy(&buf);
 	return res;
 }
 
-static bool
-fs_dup2(int oldfd, int newfd)
-{
-	if ((dup2(oldfd, newfd)) == -1) {
-		LOG_E("failed dup2(%d, %d): %s", oldfd, newfd, strerror(errno));
-		return false;
-	}
-
-	return true;
-}
-
 bool
 fs_fileno(FILE *f, int *ret)
 {
@@ -376,51 +392,28 @@ fs_fileno(FILE *f, int *ret)
 }
 
 bool
-fs_redirect(const char *path, const char *mode, int fd, int *old_fd)
+fs_make_writeable_if_exists(const char *path)
 {
-	FILE *out;
-	int out_fd;
-
-	if ((*old_fd = dup(fd)) == -1) {
-		LOG_E("failed dup(%d): %s", fd, strerror(errno));
-		return false;
-	} else if (!(out = fs_fopen(path, mode))) {
-		return false;
-	} else if (!fs_fileno(out, &out_fd)) {
-		return false;
-	} else if (!fs_dup2(out_fd, fd)) {
-		return false;
-	} else if (!fs_fclose(out)) {
-		return false;
+	struct stat sb;
+	if (stat(path, &sb) == 0) {
+		if (!(sb.st_mode & S_IWUSR)) {
+			if (!fs_chmod(path, sb.st_mode | S_IWUSR)) {
+				return false;
+			}
+		}
 	}
 
 	return true;
 }
 
-bool
-fs_redirect_restore(int fd, int old_fd)
-{
-	if (!fs_dup2(old_fd, fd)) {
-		return false;
-	} else if (close(old_fd) == -1) {
-		LOG_E("failed close(%d): %s", old_fd, strerror(errno));
-		return false;
-	}
-	return true;
-}
-
-struct fs_copy_dir_ctx {
-	const char *src_base, *dest_base;
-};
-
-static enum iteration_result
+enum iteration_result
 fs_copy_dir_iter(void *_ctx, const char *path)
 {
 	enum iteration_result res = ir_err;
 	struct fs_copy_dir_ctx *ctx = _ctx;
 	struct stat sb;
-	SBUF_manual(src);
-	SBUF_manual(dest);
+	TSTR_manual(src);
+	TSTR_manual(dest);
 
 	path_join(NULL, &src, ctx->src_base, path);
 	path_join(NULL, &dest, ctx->dest_base, path);
@@ -430,17 +423,22 @@ fs_copy_dir_iter(void *_ctx, const char *path)
 	}
 
 	if (S_ISDIR(sb.st_mode)) {
-		if (!fs_dir_exists(dest.buf)) {
-			if (!fs_mkdir(dest.buf)) {
-				goto ret;
-			}
+		if (!fs_mkdir(dest.buf, ctx->force)) {
+			goto ret;
 		}
 
-		if (!fs_copy_dir(src.buf, dest.buf)) {
+		struct fs_copy_dir_ctx sub_ctx = *ctx;
+		sub_ctx.src_base = src.buf;
+		sub_ctx.dest_base = dest.buf;
+		if (!fs_copy_dir_ctx(&sub_ctx)) {
 			goto ret;
 		}
 	} else if (S_ISREG(sb.st_mode)) {
-		if (!fs_copy_file(src.buf, dest.buf)) {
+		if (ctx->file_cb) {
+			ctx->file_cb(ctx->usr_ctx, src.buf, dest.buf);
+		}
+
+		if (!fs_copy_file(src.buf, dest.buf, ctx->force)) {
 			goto ret;
 		}
 	} else {
@@ -450,26 +448,100 @@ fs_copy_dir_iter(void *_ctx, const char *path)
 
 	res = ir_cont;
 ret:
-	sbuf_destroy(&src);
-	sbuf_destroy(&dest);
+	tstr_destroy(&src);
+	tstr_destroy(&dest);
 	return res;
 }
 
 bool
-fs_copy_dir(const char *src_base, const char *dest_base)
+fs_copy_dir_ctx(struct fs_copy_dir_ctx *ctx)
+{
+	if (!fs_mkdir(ctx->dest_base, true)) {
+		return ir_err;
+	}
+
+	return fs_dir_foreach(ctx->src_base, ctx, fs_copy_dir_iter);
+}
+
+bool
+fs_copy_dir(const char *src_base, const char *dest_base, bool force)
 {
 	struct fs_copy_dir_ctx ctx = {
 		.src_base = src_base,
 		.dest_base = dest_base,
+		.force = force,
 	};
 
-	if (!fs_dir_exists(dest_base)) {
-		if (!fs_mkdir(dest_base)) {
-			return ir_err;
+	return fs_copy_dir_ctx(&ctx);
+}
+
+struct fs_rmdir_ctx {
+	const char *base_dir;
+	bool force;
+};
+
+static enum iteration_result
+fs_rmdir_iter(void *_ctx, const char *path)
+{
+	enum iteration_result ir_res = ir_err;
+	struct fs_rmdir_ctx *ctx = _ctx;
+	struct stat sb;
+
+	TSTR(name);
+
+	path_join(NULL, &name, ctx->base_dir, path);
+
+	if (fs_symlink_exists(name.buf)) {
+		if (!fs_remove(name.buf)) {
+			goto ret;
 		}
+
+		ir_res = ir_cont;
+		goto ret;
+	} else if (stat(name.buf, &sb) != 0) {
+		if (ctx->force) {
+			ir_res = ir_cont;
+		} else {
+			LOG_E("failed stat(%s): %s", path, strerror(errno));
+		}
+		goto ret;
 	}
 
-	return fs_dir_foreach(src_base, &ctx, fs_copy_dir_iter);
+	if (S_ISDIR(sb.st_mode)) {
+		if (!fs_rmdir_recursive(name.buf, ctx->force)) {
+			goto ret;
+		}
+
+		if (!fs_rmdir(name.buf, ctx->force)) {
+			goto ret;
+		}
+
+	} else if (S_ISREG(sb.st_mode)) {
+		if (!fs_remove(name.buf)) {
+			goto ret;
+		}
+
+	} else {
+		LOG_E("unhandled file type: %s", name.buf);
+		goto ret;
+	}
+
+	ir_res = ir_cont;
+
+ret:
+	tstr_destroy(&name);
+	return ir_res;
+}
+
+bool
+fs_rmdir_recursive(const char *path, bool force)
+{
+	struct fs_rmdir_ctx ctx = {
+		.base_dir = path,
+		.force = force,
+	};
+
+	return fs_dir_foreach(path, &ctx, fs_rmdir_iter);
 }
 
 bool
