@@ -23,6 +23,8 @@
 
 extern char **environ;
 
+static const int32_t fail_to_exec_exit_code = 163;
+
 enum copy_pipe_result {
 	copy_pipe_result_finished,
 	copy_pipe_result_waiting,
@@ -30,7 +32,7 @@ enum copy_pipe_result {
 };
 
 static enum copy_pipe_result
-copy_pipe(int pipe, struct sbuf *sbuf, FILE *tee_out)
+copy_pipe(int pipe, struct tstr *tstr, FILE *tee_out)
 {
 	ssize_t b;
 	char buf[4096];
@@ -48,7 +50,7 @@ copy_pipe(int pipe, struct sbuf *sbuf, FILE *tee_out)
 			return copy_pipe_result_finished;
 		}
 
-		sbuf_pushn(0, sbuf, buf, b);
+		tstr_pushn(0, tstr, buf, b);
 
 		if (tee_out) {
 			fwrite(buf, b, 1, tee_out);
@@ -162,6 +164,11 @@ run_cmd_collect(struct run_cmd_ctx *ctx)
 		return run_cmd_error;
 	}
 
+	if (ctx->status == fail_to_exec_exit_code) {
+		ctx->err_msg = "command failed to execute";
+		return run_cmd_error;
+	}
+
 	return run_cmd_finished;
 }
 
@@ -192,21 +199,22 @@ static bool
 run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, const char *envstr, uint32_t envc)
 {
 	const char *p;
-	SBUF_manual(cmd);
+	TSTR_manual(cmd);
 
 	if (!fs_find_cmd(NULL, &cmd, _cmd)) {
+		L("failed to run command '%s': not found", _cmd);
 		ctx->err_msg = "command not found";
 		return false;
 	}
 
-	if (log_should_print(log_debug)) {
+	{
 		LL("executing %s:", cmd.buf);
 		char *const *ap;
 
 		for (ap = argv; *ap; ++ap) {
-			log_plain(" '%s'", *ap);
+			log_plain(log_debug, " '%s'", *ap);
 		}
-		log_plain("\n");
+		log_plain(log_debug, "\n");
 
 		if (envstr) {
 			const char *k;
@@ -218,7 +226,7 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 					if (!k) {
 						k = p + 1;
 					} else {
-						log_plain(" %s='%s'", k, p + 1);
+						log_plain(log_debug, " %s='%s'", k, p + 1);
 						k = NULL;
 
 						if (++i >= envc) {
@@ -228,7 +236,7 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 				}
 			}
 
-			log_plain("\n");
+			log_plain(log_debug, "\n");
 		}
 	}
 
@@ -244,8 +252,8 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 	ctx->input_fd_open = true;
 
 	if (!(ctx->flags & run_cmd_ctx_flag_dont_capture)) {
-		sbuf_init(&ctx->out, 0, 0, sbuf_flag_overflow_alloc);
-		sbuf_init(&ctx->err, 0, 0, sbuf_flag_overflow_alloc);
+		tstr_init(&ctx->out, 0, 0, tstr_flag_overflow_alloc);
+		tstr_init(&ctx->err, 0, 0, tstr_flag_overflow_alloc);
 
 		if (!open_run_cmd_pipe(ctx->pipefd_out, ctx->pipefd_out_open)) {
 			goto err;
@@ -260,25 +268,25 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 		if (ctx->chdir) {
 			if (chdir(ctx->chdir) == -1) {
 				LOG_E("failed to chdir to %s: %s", ctx->chdir, strerror(errno));
-				exit(1);
+				exit(fail_to_exec_exit_code);
 			}
 		}
 
 		if (ctx->stdin_path) {
 			if (dup2(ctx->input_fd, 0) == -1) {
 				LOG_E("failed to dup stdin: %s", strerror(errno));
-				exit(1);
+				exit(fail_to_exec_exit_code);
 			}
 		}
 
 		if (!(ctx->flags & run_cmd_ctx_flag_dont_capture)) {
 			if (dup2(ctx->pipefd_out[1], 1) == -1) {
 				LOG_E("failed to dup stdout: %s", strerror(errno));
-				exit(1);
+				exit(fail_to_exec_exit_code);
 			}
 			if (dup2(ctx->pipefd_err[1], 2) == -1) {
 				LOG_E("failed to dup stderr: %s", strerror(errno));
-				exit(1);
+				exit(fail_to_exec_exit_code);
 			}
 		}
 
@@ -297,7 +305,7 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 								k,
 								p + 1,
 								strerror(err));
-							exit(1);
+							exit(fail_to_exec_exit_code);
 						}
 						k = NULL;
 
@@ -311,14 +319,14 @@ run_cmd_internal(struct run_cmd_ctx *ctx, const char *_cmd, char *const *argv, c
 
 		if (execve(cmd.buf, (char *const *)argv, environ) == -1) {
 			LOG_E("%s: %s", cmd.buf, strerror(errno));
-			exit(1);
+			exit(fail_to_exec_exit_code);
 		}
 
 		abort();
 	}
 
 	/* parent */
-	sbuf_destroy(&cmd);
+	tstr_destroy(&cmd);
 
 	if (ctx->pipefd_err_open[1] && close(ctx->pipefd_err[1]) == -1) {
 		LOG_E("failed to close: %s", strerror(errno));
@@ -345,7 +353,7 @@ build_argv(struct run_cmd_ctx *ctx,
 	const char *argstr,
 	uint32_t argstr_argc,
 	char *const *old_argv,
-	struct sbuf *cmd,
+	struct tstr *cmd,
 	const char ***argv)
 {
 	const char *argv0, *new_argv0 = NULL, *new_argv1 = NULL;
@@ -365,8 +373,8 @@ build_argv(struct run_cmd_ctx *ctx,
 
 	assert(*argv0 && "argv0 cannot be empty");
 
-	sbuf_clear(cmd);
-	sbuf_pushs(NULL, cmd, argv0);
+	tstr_clear(cmd);
+	tstr_pushs(NULL, cmd, argv0);
 
 	if (!path_is_basename(cmd->buf)) {
 		path_make_absolute(NULL, cmd, argv0);
@@ -418,7 +426,7 @@ run_cmd_argv(struct run_cmd_ctx *ctx, char *const *argv, const char *envstr, uin
 	struct source src = { 0 };
 	const char **new_argv = NULL;
 
-	SBUF_manual(cmd);
+	TSTR_manual(cmd);
 	if (!build_argv(ctx, &src, NULL, 0, argv, &cmd, &new_argv)) {
 		goto err;
 	}
@@ -433,7 +441,7 @@ err:
 	if (new_argv) {
 		z_free((void *)new_argv);
 	}
-	sbuf_destroy(&cmd);
+	tstr_destroy(&cmd);
 	return ret;
 }
 
@@ -444,7 +452,7 @@ run_cmd(struct run_cmd_ctx *ctx, const char *argstr, uint32_t argc, const char *
 	struct source src = { 0 };
 	const char **argv = NULL;
 
-	SBUF_manual(cmd);
+	TSTR_manual(cmd);
 	if (!build_argv(ctx, &src, argstr, argc, NULL, &cmd, &argv)) {
 		goto err;
 	}
@@ -455,7 +463,7 @@ err:
 	if (argv) {
 		z_free((void *)argv);
 	}
-	sbuf_destroy(&cmd);
+	tstr_destroy(&cmd);
 	return ret;
 }
 
@@ -464,8 +472,8 @@ run_cmd_ctx_destroy(struct run_cmd_ctx *ctx)
 {
 	run_cmd_ctx_close_fds(ctx);
 
-	sbuf_destroy(&ctx->out);
-	sbuf_destroy(&ctx->err);
+	tstr_destroy(&ctx->out);
+	tstr_destroy(&ctx->err);
 }
 
 bool
