@@ -14,6 +14,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "log.h"
 #include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
@@ -149,3 +150,43 @@ run_cmd_determine_interpreter(struct source *src,
 
 	return true;
 }
+
+void
+run_cmd_print_error(struct run_cmd_ctx *ctx, enum log_level lvl)
+{
+	if (ctx->err_msg) {
+		log_print(true, lvl, "%s", ctx->err_msg);
+	}
+
+	if (ctx->out.len) {
+		log_print(false, lvl, "stdout:\n%s", ctx->out.buf);
+	}
+
+	if (ctx->err.len) {
+		log_print(false, lvl, "stderr:\n%s", ctx->err.buf);
+	}
+}
+
+bool
+run_cmd_checked(struct run_cmd_ctx *ctx, const char *argstr, uint32_t argc, const char *envstr, uint32_t envc)
+{
+	if (!run_cmd(ctx, argstr, argc, envstr, envc) || ctx->status != 0) {
+		run_cmd_print_error(ctx, log_error);
+		run_cmd_ctx_destroy(ctx);
+		return false;
+	}
+
+	return true;
+}
+
+bool
+run_cmd_argv_checked(struct run_cmd_ctx *ctx, char *const *argv, const char *envstr, uint32_t envc)
+{
+	if (!run_cmd_argv(ctx, argv, envstr, envc) || ctx->status != 0) {
+		run_cmd_print_error(ctx, log_error);
+		run_cmd_ctx_destroy(ctx);
+		return false;
+	}
+
+	return true;
+}
