@@ -19,15 +19,15 @@ build() {
 
 	git clone https://git.sr.ht/~lattis/muon "$tmp"
 	cd "$tmp"
-	git checkout $1
+	git checkout "$1"
 
 	export CC=gcc
 	export CFLAGS="-D_POSIX_C_SOURCE=200112L -D__EXTENSIONS__"
 
 	./bootstrap.sh build
-	build/muon setup build
-	build/muon -C build samu
-	build/muon -C build test -d dots -s lang
+	build/muon-bootstrap setup build
+	build/muon-bootstrap -C build samu
+	build/muon-bootstrap -C build test -d dots -s lang
 }
 
 submit() {
@@ -45,7 +45,6 @@ _receive() {
 	exec 2>&1
 	send_status pending
 
-	dir=ci/muon
 	tmp="ci/muon/$(date +%s)"
 	mkdir -p "$tmp"
 
